/*
 * Copyright (c) 2005-2006 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: IntSet.java,v 1.2 2007/11/27 18:41:05 almo Exp $
 */

package jatva.examples.intset;

public class IntSet
{
  protected IntSetNode _node = null;
  protected int _size = 0;
  
  public IntSet()
  {
  }
  
  public boolean checkInvariants()
  {
    if(_node != null)
      return _node.check() && _node.checkColors() && _node.checkPaths();
    
    return true;
  }
  
  public synchronized String toString()
  {
    StringBuffer st = new StringBuffer("{");
    
    if(_node != null)
      _node.appendStringBuffer(st);
    
    st.append("}");
    
    return st.toString();
  }
  
  protected void rotate(boolean left, IntSetNode x)
  {
    IntSetNode y;
    
    if(left)
    {
      y = x._right;
      x._right = y._left;
      if(y._left != null)
        y._left._parent = x;
      
      y._parent = x._parent;
      if(x._parent != null)
      {
        if(x == x._parent._left)
          x._parent._left = y;
        else
          x._parent._right = y;
      }
      else
        _node = y;
      
      y._left = x;
      x._parent = y;
    }
    else
    {
      y = x._left;
      x._left = y._right;
      if(y._right != null)
        y._right._parent = x;
      
      y._parent = x._parent;
      if(x._parent != null)
      {
        if(x == x._parent._left)
          x._parent._left = y;
        else
          x._parent._right = y;
      }
      else
        _node = y;
      
      y._right = x;
      x._parent = y;
    }
  }
  
  public synchronized boolean contains(int value)
  {
    if(_node == null)
      return false;
    
    IntSetNode p = _node;
    
    do
    {
      if(p._value == value)
        return true;
      else if(p._value > value)
        p = p._left;
      else
        p = p._right;
    }
    while(p != null);
    
    return false;
  }
  
  protected IntSetNode minimum()
  {
    if(_node == null)
      return null;
    else
      return _node.minimum();
  }
  
  protected IntSetNode maximum()
  {
    if(_node == null)
      return null;
    else
      return _node.maximum();
  }
  
  protected IntSetNode next(IntSetNode p)
  {
    if(p._right != null)
      return p._right.minimum();
    
    while(p._parent != null && p == p._parent._right)
      p = p._parent;

    return p._parent;
  }
  
  protected IntSetNode prev(IntSetNode p)
  {
    if(p._left != null)
      return p._left.maximum();
    
    while(p._parent != null && p == p._parent._left)
      p = p._parent;

    return p._parent;
  }
  
  public synchronized boolean add(int value)
  {
    if(_node == null)
    {
      _node = new IntSetNode(value);
      _node._red = false;
      _size++;
      
      return true;
    }
    
    IntSetNode p = _node, q;
    
    do
    {
      q = p;
      
      if(p._value == value)
        return false;
      else if(p._value > value)
        p = p._left;
      else
        p = p._right;
    }
    while(p != null);
    
    if(q._value > value)
      q._left = ( p =  new IntSetNode(value) );
    else
      q._right = ( p = new IntSetNode(value) );
    p._parent = q;
    
    _size++;
    
    IntSetNode r;
    
//    System.out.println("After add : " + toString());
    
    while(q != null && q._red == true)
    {
      if(q == q._parent._left)
      {
//        System.out.print("q is on the left, ");
        
        r = q._parent._right;
        if(r == null || !r._red )
        {
          if(p == q._right)
          {
//            System.out.println("case 2");
                        
            q._right = p._left;
            if(p._left != null) p._left._parent = q;
            p._left = q;
            p._parent = q._parent;
            q._parent._left = p;
            q._parent = p;
            p = q;
            q = q._parent;
            
//            System.out.println("After rotation : " + toString());
          }
          
//          System.out.println("case 3");
          
          if(q._parent._parent != null)
          {
            if(q._parent._parent._left == q._parent)
              q._parent._parent._left = q;
            else
              q._parent._parent._right = q;
          }
          else
            _node = q;
          
          q._red = false;
          q._parent._red = true;
          q._parent._left = q._right;
          if(q._right != null) q._right._parent = q._parent;
          q._right = q._parent;
          q._parent = q._parent._parent;
          q._right._parent = q;
          
          return true;
        }
        else
        {
//          System.out.println("case 1");
          
          q._red = false;
          r._red = false;
          q._parent._red = true;
          p = q._parent;
          q = p._parent;
        }
      }
      else
      {
//        System.out.print("q is on the right, ");

        r = q._parent._left;
        if(r == null || !r._red)
        {
          if(p == q._left)
          {
//            System.out.println("case 2");

            q._left = p._right;
            if(p._right != null) p._right._parent = q;
            p._right = q;
            p._parent = q._parent;
            q._parent._right = p;
            q._parent = p;
            p = q;
            q = q._parent;

//            System.out.println("After rotation : " + toString());
          }
          
//         System.out.println("case 3");
         
         if(q._parent._parent != null)
          {
            if(q._parent._parent._left == q._parent)
              q._parent._parent._left = q;
            else
              q._parent._parent._right = q;
          }
          else
            _node = q;
          
          q._red = false;
          q._parent._red = true;
          q._parent._right = q._left;
          if(q._left != null) q._left._parent = q._parent;
          q._left = q._parent;
          q._parent = q._parent._parent;
          q._left._parent = q;
          
          return true;
        }
        else
        {
          q._red = false;
          r._red = false;
          q._parent._red = true;
          p = q._parent;
          q = p._parent;
        }
      }
    }
    
    _node._red = false;
    
    return true;
    
  }
  
  public synchronized boolean remove(int value)
  {
    if(_node == null)
      return false;
    
    IntSetNode p = _node;
    
    do
    {
      if(p._value > value)
        p = p._left;
      else if(p._value < value)
        p = p._right;
      else
        break;
    }
    while(p != null);
    
    if(p == null)
      return false;
    
    boolean isLeft;
    
    if(p._left == null)
    {
      if(p._parent == null)
      {
        _node = p._right;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._right;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._right;
          isLeft = false;
        }
      }
      
      if(p._right != null)
        p._right._parent = p._parent;
      
      if(!p._red)
        fixup(p._parent, p._right, isLeft);
    }
    else if(p._right == null)
    {
      if(p._parent == null)
      {
        _node = p._left;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._left;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._left;
          isLeft = false;
        }
      }
      
      p._left._parent = p._parent;
      if(!p._red)
        fixup(p._parent, p._left, isLeft);
    }
    else
    {
      IntSetNode q = next(p);
      p._value = q._value;
      
      if(q._parent == null)
      {
        _node = q._right;
        isLeft = true;
      }
      else
      {
        if(q == q._parent._left)
        {
          q._parent._left = q._right;
          isLeft = true;
        }
        else
        {
          q._parent._right = q._right;
          isLeft = false;
        }
      }
      
      if(q._right != null)
        q._right._parent = q._parent;
      
      if(!q._red)
        fixup(q._parent, q._right, isLeft); 
    }
    
    _size--;
    return true;
  }
  
  protected void fixup(IntSetNode p, IntSetNode q, boolean isLeft)
  {
    while(p != null && ( q == null || q._red == false ) )
    {
      IntSetNode w, v;
      
      if(isLeft)
      {
        w = p._right;
        if(w._red)
        {
          w._red = false;
          p._red = true;
          rotate(true, p);
          w = p._right;
        }
        
        v = w._left;
        
        if(w._right == null || !w._right._red)
        {
          if(v == null || !v._red)
          {
            w._red = true;
            
            q = p;
            p = p._parent;
            isLeft = (p == null)?true:(q == p._left);
            
            continue;
          }
          else
          {
            w._red = true;
            v._red = false;
            rotate(false, w);
            
            w = v;
            v = v._left;
          }
        }
        
        w._red = p._red;
        p._red = false;
        if(w._right != null) w._right._red = false;
        rotate(true, p);
        
        q = _node;
        p = null;
      }
      else
      {
        w = p._left;
        if(w._red)
        {
          w._red = false;
          p._red = true;
          rotate(false, p);
          w = p._left;
        }
        
        v = w._right;
        
        if(w._left == null || !w._left._red)
        {
          if(v == null || !v._red)
          {
            w._red = true;
            
            q = p;
            p = p._parent;
            isLeft = (p == null)?true:(q == p._left);
            
            continue;
          }
          else
          {
            w._red = true;
            v._red = false;
            rotate(true, w);
            
            w = v;
            v = v._right;
          }
        }
        
        w._red = p._red;
        p._red = false;
        if(w._left != null) w._left._red = false;
        rotate(false, p);
        
        q = _node;
        p = null;
      }
    }
    
    if(q != null)
      q._red = false;
  }
  
  public synchronized void clear()
  {
    _node = null;
    _size = 0;
  }
  
  public synchronized boolean isEmpty()
  {
    return (_node == null);
  }
  
  public synchronized boolean equals(Object o)
  {
    if(o == null)
      return false;
    
    if( !(o instanceof IntSet) )
      return false;
    
    IntSet r = (IntSet)o;
    
    if(_size != r._size)
      return false;
    
    if(r == this)
      return true;
            
    IntSetNode p = minimum(), q = r.minimum();
    
    while(p != null)
    {
      if(p._value != q._value)
        return false;
      
      p = next(p);
      q = r.next(q);
    }
    
    return true;
  }
  
  public synchronized int size()
  {
    return _size;
  }
  
  public synchronized int[] toArray()
  {
    int result[] = new int[_size];
    
    if(_node != null)
      _node.fill(result, 0);
    
    return result;
  }
  
  public IntSet(IntSet r)
  {
    _size = r._size;
    
    if(r._node != null)
    {
      _node = new IntSetNode(0);
      r._node.nodeClone(_node);
    }
  }
  
  public synchronized IntSet clone()
  {
    return new IntSet(this);
  }

  static class CheckException
	extends Throwable
  {
	CheckException(int value)
	{
	  super();
	  System.out.println(  "CHECK: not all the paths from "
						 + value
						 + " have the same black length"
						);
	}
  }

  static class IntSetNode
  {
	IntSetNode _left = null;
	IntSetNode _right = null;
	IntSetNode _parent = null;
	int _value = 0;
	boolean _red = true;
  
	IntSetNode(int value)
	{
	  _value = value;
	}
  
	IntSetNode minimum()
	{
	  IntSetNode p = this;
    
	  while(p._left != null)
		p = p._left;
    
	  return p;
	}
  
	IntSetNode maximum()
	{
	  IntSetNode p = this;
    
	  while(p._right != null)
		p = p._right;
    
	  return p;
	}
  
	int fill(int array[], int ind)
	{
	  int res = ind+1;
    
	  array[ind] = _value;
    
	  if(_left != null)
	  {
		res = _left.fill(array, ind+1);
      
		if(_right != null)
		  return _right.fill(array, res);
	  }
	  else if(_right != null)
		return _right.fill(array, ind+1);
    
	  return res;
	}
  
	void appendStringBuffer(StringBuffer st)
	{
	  st.append(_value);
    
	  if(_left != null)
	  {
		st.append(", ");
		_left.appendStringBuffer(st);
      
		if(_right != null)
		{
		  st.append(", ");
		  _right.appendStringBuffer(st);
		}
	  }
	  else if(_right != null)
	  {
		st.append(", ");
		_right.appendStringBuffer(st);
	  }
	}
  
	void nodeClone(IntSetNode n)
	{
	  n._value = _value;
	  n._red = _red;
    
	  if(_left != null)
	  {
		n._left = new IntSetNode(0);
		n._left._parent = n;
		_left.nodeClone(n._left);
	  }
    
	  if(_right != null)
	  {
		n._right = new IntSetNode(0);
		n._right._parent = n;
		_right.nodeClone(n._right);
	  }
	}
  
	boolean check()
	{
	  boolean result = true;
    
	  if(_left != null && _left._parent != this)
	  {
		result = false;
		System.out.println(  "CHECK: Inconsistent parent-left links: left is "
						   + _left._value
						   + ", parent must be "
						   + _value
						   + ", but is "
						   + (_left._parent == null?"null":Integer.toString(_left._parent._value))
						  );
      
		_left.check();  
	  }
	  else if(_left != null)
		result = _left.check();
      
	  if(_right != null && _right._parent != this)
	  {
		result = false;
		System.out.println(  "CHECK: Inconsistent parent-right links: right is "
						   + _right._value
						   + ", parent must be "
						   + _value
						   + ", but is "
						   + (_right._parent == null?"null":Integer.toString(_right._parent._value))
						  );
      
		_right.check();
	  }
	  else if(_right != null)
		result = result && _right.check();
    
	  return result;
	}
  
	boolean checkColors()
	{
	  boolean result = true;
    
	  if(_red && _left != null && _left._red)
	  {
		result = false;
		System.out.println(  "CHECK: Inconsistent colors: "
						   + _value
						   + " is red, while left "
						   + _left._value
						   + " is red"
						  );

		_left.checkColors();
	  }
	  else if(_left != null)
		result = _left.checkColors();
    
	  if(_red && _right != null && _right._red)
	  {
		result = false;
		System.out.println(  "CHECK: Inconsistent colors: "
						   + _value
						   + " is red, while right "
						   + _right._value
						   + " is red"
						  );
      
		result = _right.checkColors();
	  }
	  else if(_right != null)
		result = result && _right.checkColors();
    
	  return result;  
	}
  
	boolean checkPaths()
	{
	  try
	  {
		System.out.println("Black height: " + blackHeight());
    
		return true;
	  }
	  catch(CheckException err)
	  {
		return false;
	  }
	}
  
	int blackHeight()
	  throws CheckException
	{
	  if(_left == null && _right == null)
		return 1 + (_red?0:1);
	  else if(_left == null)
	  {
		if(_right.blackHeight() != 1)
		  throw new CheckException(_value);
		else
		  return 1 + (_red?0:1);
	  }
	  else if(_right == null)
	  {
		if(_left.blackHeight() != 1)
		  throw new CheckException(_value);
		else
		  return 1 + (_red?0:1);
	  }
	  else
	  {
		if(_left.blackHeight() != _right.blackHeight())
		  throw new CheckException(_value);
		else
		  return _left.blackHeight() + (_red?0:1);
	  }
	}
  }
}

