/*
 * Copyright (c) 2005-2006 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: List.java,v 1.2 2007/11/26 18:53:54 almo Exp $
 * Modified on   06.03.2006
 * First version 01.04.2001
 * 
 */

package jatva.examples.pqueue;

public class List<T>
{
  int _size = 0;
  ListElement<T> _head = null;
  
  public List()
  {
  }
  
  public List(List<T> other)
  {
    ListElement<T> p = other._head;
    
    if(p == null)
      return;
    
    _size = other._size;
    
    ListElement<T> s = other._head.clone(), q = null;
    _head = s;
    
    p = p._next;
    
    while(p != other._head)
    {
      q = p.clone();
      s._next = q;
      q._prev = s;
      p = p._next;
      s = q;
    }
    
    _head._prev = s;
    s._next = _head;
  }
  
  public synchronized int size()
  {
    return _size;
  }
  
  public synchronized boolean isEmpty()
  {
    return (_size == 0);
  }
  
  public synchronized T firstElement()
  {
    if(_head != null)
      return _head._elem;
    else
      return null;
  }
  
  public synchronized T lastElement()
  {
    if(_head != null)
      return _head._prev._elem;
    else
      return null;
  }
  
  public synchronized void addLastElement(T obj)
  {
    if(_head == null)
    {
      _head = new ListElement<T>(obj);
      _head._next = _head;
      _head._prev = _head;
    }
    else
    {
      ListElement<T> p = new ListElement<T>(obj);
      p._prev = _head._prev;
      p._next = _head;
      _head._prev._next = p;
      _head._prev = p;
    }

    _size++;
  }
  
  public synchronized void addFirstElement(T obj)
  {
    if(_head == null)
    {
      _head = new ListElement<T>(obj);
      _head._next = _head;
      _head._prev = _head;
    }
    else
    {
      ListElement<T> p = new ListElement<T>(obj);
      p._prev = _head._prev;
      p._next = _head;
      _head._prev._next = p;
      _head._prev = p;
      _head = p;
    }
    _size++;
  }
  
  public synchronized void clear()
  {
    _head = null;
    _size = 0;
  }
  
  public synchronized void removeLastElement()
  {
    if(_head == null)
      return;
    else if(_size == 1)
    {
      _head = null;
    }
    else
    {
      _head._prev._prev._next = _head;
      _head._prev = _head._prev._prev;
    }

    _size--;
  }
  
  public synchronized void removeFirstElement()
  {
    if(_head == null)
      return;
    else if(_size == 1)
    {
      _head = null;
    }
    else
    {
      _head._prev._next = _head._next;
      _head._next._prev = _head._prev;
      _head = _head._next;
    }

    _size--;
  }
  
  public synchronized void removeAllElements()
  {
    _size = 0;
    _head = null;
  }
    
  public synchronized void insertElementAt(T obj, int i)
    throws InvalidListIndex
  {
    if(i < 0 || i > _size)
      throw new InvalidListIndex(i, _size);
    
    if(i == _size)
    {
      addLastElement(obj);
      return;
    }
    
    ListElement<T> p = new ListElement<T>(obj);
    ListElement<T> s = _head;
    
    do
    {
      if(i == 0)
      {
        p._prev = s._prev;
        p._next = s;
        s._prev._next = p;
        s._prev = p;
        
        if(s == _head)
          _head = p;
      }

      s = s._next;
      i--;
    }
    while(i >= 0);
    
    _size++;
  }

  public synchronized void setElementAt(T obj, int i)
    throws InvalidListIndex
  {
    if(i < 0 || i >= _size)
      throw new InvalidListIndex(i, _size);
    
    ListElement<T> s = _head;
    
    do
    {
      if(i == 0)
        s._elem = obj;

      s = s._next;
      i--;
    }
    while(i >= 0);
  }

  public synchronized T elementAt(int i)
    throws InvalidListIndex
  {
    if(i < 0 || i >= _size)
      throw new InvalidListIndex(i, _size);
    
    ListElement<T> s = _head;
    
    do
    {
      if(i == 0)
        return s._elem;

      s = s._next;
      i--;
    }
    while(i >= 0);
    
    return s._elem;
  }

  public synchronized void removeElementAt(int i)
    throws InvalidListIndex
  {
    if(i < 0 || i >= _size)
      throw new InvalidListIndex(i, _size);
    
    ListElement<T> s = _head;
    
    if(_size == 1)
    {
      _head = null;
      _size--;
      return;
    }
    
    do
    {
      if(i == 0)
      {
        s._prev._next = s._next;
        s._next._prev = s._prev;
        
        if(s == _head)
          _head = s._next;
      }

      s = s._next;
      i--;
    }
    while(i >= 0);
    
    _size--;
  }
  
  public synchronized boolean equals(Object obj)
  {
    if( !(obj instanceof List) )
      return false;
    
    List<T> other = (List<T>)obj;
    
    if(other.size() != _size)
      return false;
    
    if(_size == 0)
      return true;
    
    ListElement<T> p = _head;
    
    for(int i = 0; i < _size; i++)
    {
      if(p._elem != other.elementAt(i))
        return false;
      
      p = p._next;
    }
    
    return true;
  }
  
  public synchronized boolean equalsByElement(Object obj)
  {
    if( !(obj instanceof List) )
      return false;
    
    List<T> other = (List<T>)obj;
    
    if(other.size() != _size)
      return false;
    
    if(_size == 0)
      return true;
    
    ListElement<T> p = _head;
    
    for(int i = 0; i < _size; i++)
    {
      if(!p._elem.equals(other.elementAt(i)))
        return false;
      
      p = p._next;
    }
    
    return true;
  }
  
  public synchronized List<T> clone()
  {
    List<T> result = new List<T>();
    
    ListElement<T> p = _head;
    
    if(p == null)
      return result;
    
    result._size = _size;
    
    ListElement<T> s = _head.clone(), q = null;
    result._head = s;
    
    p = p._next;
    
    while(p != _head)
    {
      q = p.clone();
      s._next = q;
      q._prev = s;
      p = p._next;
      s = q;
    }
    
    result._head._prev = s;
    s._next = result._head;
    
    return result;
  }
  
  public synchronized boolean equalsByFields(T t)
  {
    return equals(t);
  }
  
  public synchronized List<T> cloneByFields()
  {
    return clone();
  }
  
  public synchronized boolean contains(T t)
  {
    ListElement<T> p = _head;
    
    if(p == null)
      return false;
    
    do
    {
      if(p._elem == t)
        return true;
      
      p = p._next;
    }
    while(p != _head);
    
    return false;
  }
  
  public synchronized boolean containsEqual(T t)
  {
    ListElement<T> p = _head;
    
    if(p == null)
      return false;
    
    do
    {
      if(p._elem.equals(t))
        return true;
      
      p = p._next;
    }
    while(p != _head);
    
    return false;
  }
  
  static class ListElement<T> implements Cloneable
  {
	ListElement<T> _next = null;
	ListElement<T> _prev = null;
	T _elem = null;
    
	ListElement(T elem)
	{
	  _elem = elem;
	}
  
	public ListElement<T> clone()
	{
	  return new ListElement<T>(_elem);
	}
  }
}
