/*
 * Copyright (c) 2005-2006 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: PQueue.java,v 1.1 2007/11/22 14:49:30 almo Exp $
 * Modified on   06.03.2006
 * First version 01.04.2001
 * 
 */

package jatva.examples.pqueue;

import java.util.ArrayList;

/**
 *  The implementation of the queue with priorities.
 *  Priorities are represented as integers between MIN_Priority and MAX_Priority
 */

public class PQueue<T>
{
  public final static int MIN_Priority = 0;
  public final static int MAX_Priority = 255;
  protected final static int ClusterInitialCapacity = 10;
  protected final static int ClusterIncrement = 10;
  
  public ArrayList<ArrayList<T>> _clusters 
              = new ArrayList<ArrayList<T>>(MAX_Priority - MIN_Priority + 1);
  
  public PQueue()
  {
    for(int i = 0; i <= MAX_Priority - MIN_Priority; i++)
      _clusters.add(i, new ArrayList<T>(ClusterInitialCapacity));
  }
  
  public final synchronized boolean isEmpty()
  {
    int i;
    
    for(i = MAX_Priority - MIN_Priority;
        i >= 0 && (_clusters.get(i)).size() == 0;
        i--
       );  
    
    if(i < 0) return true;
    else      return false;
  }
  
  private void impl_enq(T t, int prty)
  {
    _clusters.get(prty - MIN_Priority).add(t);
  }
  
  static int counter = 0;
  
  public final synchronized void enq(T t)
  {
    //if(counter++ == 56) return;
    impl_enq(t, MIN_Priority);
  }
  
  public final synchronized void enq(T t, int prty)
    throws IllegalPriority
  {
    if(prty < MIN_Priority || prty > MAX_Priority)
      throw new IllegalPriority("The specified priority isn't in priorities interval: " + prty);
    
    impl_enq(t, prty);
  }
  
  public synchronized T deq()
  {
    int i;
    
    for(i = MAX_Priority - MIN_Priority;
        i >= 0 && _clusters.get(i).size() == 0;
        i--
       );  
    
    if(i < 0) return null;
    else
    {
      T result = _clusters.get(i).get(0);
      _clusters.get(i).remove(0);
      
      return result;
    }
  }
}