package jatva.examples.set;

import java.util.Arrays;

public class SetOnArray<T extends Object> implements Cloneable
{
    public T[] array;

    public SetOnArray( int max_size )
    {
        this.array = (T[])(new Object[ max_size ]);
        Arrays.fill( this.array, null );
    }

    /**
     * Used only in 'clone' opration
     */
    private SetOnArray()
    {}

    public int countNonNullElements()
    {
        return countNonNullElements( array );
    }

    public static <T> int countNonNullElements( T[] array )
    {
        int res = 0;
        for( T o : array )
        {
            if( o != null ){ res ++; }
        }

        return res;
    }

    public void addElement( T obj )
    {
        int size = countNonNullElements();

        if( size == array.length )
            throw new IndexOutOfBoundsException();

        if( hasElement( obj ) )
            return;

        for( int i = 0; i < array.length; i ++ )
        {
            if( array[ i ] == null )
            {
                array[ i ] = obj;
                return;
            }
        }
    }

    public void removeElement( T obj )
    {
        for( int i = 0; i < array.length; i ++ )
        {
            if( obj.equals( array[ i ] ) )
            {
                array[ i ] = null;
                return;
            }
        }
    }

    public boolean hasElement( T obj )
    {
        for( T o : array )
        {
            if( o == null )
                continue;
            else
            if( o.equals( obj ) )
                return true;
        }
        return false;
    }

    public boolean contentEquals( SetOnArray<T> set2 )
    {
        return contentEquals( set2.array );
    }

    public boolean contentEquals( T[] array2 )
    {
        Object[] arrCopy = Arrays.copyOf( array, array.length );

        if( array.length != array2.length )
            return false;

        externalLoop:
        for( Object o : array2 )
        {
            if( o == null )
                continue;

            for( int i = 0; i < arrCopy.length; i ++ )
            {
                if( arrCopy[ i ] == null )
                    continue;
                else
                if( arrCopy[ i ].equals( o ) )
                {
                    arrCopy[ i ] = null;
                    // we found this element in arrCopy
                    continue externalLoop;
                }
            }
        }

        return countNonNullElements( arrCopy ) == 0;
    }

    @Override
    public SetOnArray<T> clone()
    {
        SetOnArray<T> _clone = new SetOnArray<T>();
        _clone.array = Arrays.copyOf( this.array, this.array.length );
        return _clone;
    }
}
