/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class ByteIterator
  extends    AbstractByteIterator
  implements IByteIterator
{
  public static byte[] defaultValues = {
     Byte.MIN_VALUE
   , Byte.MAX_VALUE
   , (byte)(Byte.MIN_VALUE + 1)
   , (byte)(Byte.MAX_VALUE - 1)
   , 0
   , 1
   , -1
   , (byte)(Byte.MIN_VALUE + 32)
   , (byte)(Byte.MAX_VALUE - 32)
   , (byte)(Byte.MIN_VALUE + 73)
  };
  
  protected IByteIterator[] items = new IByteIterator[] { new ArrayByteIterator() };
  protected int index = 0;
  
  public ByteIterator()
  {
  }
  
  public ByteIterator(byte[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = new IByteIterator[] { new ArrayByteIterator(values) };
  }
  
  public ByteIterator(byte start, byte end, int offset, int divisor)
  {
    items = new IByteIterator[] { new SegmentByteIterator(start, end, offset, divisor) };
  }

  public ByteIterator(byte point, int offset)
  {
    items = new IByteIterator[] { new NeighbourhoodByteIterator(point, offset) };
  }

  
  public ByteIterator(IByteIterator[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = values;
  }
  
  public void addSegment(byte start, byte end, int offset, int divisor)
  {
    IByteIterator[] newItems = new IByteIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new SegmentByteIterator(start, end, offset, divisor);
    items = newItems;
  }
  
  public void addSegment(byte start, byte end, int offset)
  {
    addSegment(start, end, offset, 1);
  }
  
  public void addSegment(byte start, byte end)
  {
    addSegment(start, end, 0, 1);
  }
  
  public void addPoint(byte point, int offset)
  {
    IByteIterator[] newItems = new IByteIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new NeighbourhoodByteIterator(point, offset);
    items = newItems;
  }
  
  public void addArray(byte[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    IByteIterator[] newItems = new IByteIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new ArrayByteIterator(values);
    items = newItems;
  }
  
  public void addIterator(IByteIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    IByteIterator[] newItems = new IByteIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = (IByteIterator)it.clone();
    items = newItems;
  }
  
  public void removeItem(int index)
  {
    if(index >= this.index && index < items.length)
    {
      IByteIterator[] newItems = new IByteIterator[items.length-1];
      System.arraycopy(items, 0, newItems, 0, index);
      System.arraycopy(items, index+1, newItems, index, items.length-index-1);
      items = newItems;
    }
  }
  
  public ByteIterator(ByteIterator r)
  {
    index = r.index;
    items = new IByteIterator[r.items.length];
    
    for(int i = 0; i < items.length; i++)
    {
      if(r.items[i] != null)
      items[i] = (IByteIterator)r.items[i].clone();
    }
  }
  
  public synchronized Object clone()
  {
    return new ByteIterator(this);
  }
  
  public synchronized void shift()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    if(index < items.length) items[index].next();

    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
  }
  
  public byte value()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    return items[index].value();
  }
  
  public boolean stop()
  {
    int i = index;
    while(i < items.length && (items[i] == null || items[i].stop()))
    {
      i++;
    }

    if(i >= items.length) return true;
    else                  return false;
  }

  public boolean equals(Object o) 
  {
    if(!(o instanceof ByteIterator))
      return false;

    ByteIterator obj = (ByteIterator)o;

    if(   index != obj.index
       || items.length != obj.items.length
      )
      return false;
      
    for(int i = 0; i < items.length; i++)
      if(items[i] == null && obj.items[i] != null || items[i] != null && !items[i].equals(obj.items[i]))
        return false;
        
    return true;
  }

  public int hashCode ()
  {
    int r = 0;
    for(int i = 0; i < items.length; i++)
      r += items[i].hashCode();
    
    return r;
  }
}
