/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

import jatva.iterators.AbstractIterator;

public class ByteObjectIterator
  extends AbstractIterator
  implements jatva.iterators.Iterator
{
  public static Byte[] defaultValues = {
     null
   , new Byte((byte)(Byte.MIN_VALUE + 1))
   , new Byte((byte)0)
   , null
  };
  static
  {
    defaultValues[3] = defaultValues[1];
  }

  protected Byte[] values = defaultValues;
  protected IByteIterator it = new ArrayByteIterator();
  protected int index = 0;
  
  public ByteObjectIterator()
  {
  }

  public ByteObjectIterator(Byte[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Byte[values.length];
    it = null;
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:(new Byte(values[i].byteValue()));
  }

  public ByteObjectIterator(IByteIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    values = null;
    this.it = (IByteIterator)it.clone();
  }
  
  public ByteObjectIterator(Byte[] values, IByteIterator it)
  {
    if(values == null && it == null) throw new IllegalArgumentException("Both arguments should not be null");
    if(values != null)
    {
      this.values = new Byte[values.length];
      
      for(int i = 0; i < values.length; i++)
        this.values[i] = (values[i] == null)?null:(new Byte(values[i].byteValue()));
    }
    if(it != null) this.it = (IByteIterator)it.clone();
  }
  
  public void setValueIterator(IByteIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    this.it = (IByteIterator)it.clone();
  }
  
  public void setArray(Byte[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Byte[values.length];
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:(new Byte(values[i].byteValue()));
  }
  
  public ByteObjectIterator(ByteObjectIterator r)
  {
    index = r.index;
    it = (r.it == null)?null:(IByteIterator)r.it.clone();
    if(r.values == null) { values = null; return; }
    
    values = new Byte[r.values.length];
    
    for(int i = 0; i < values.length; i++)
      values[i] = (r.values[i] == null)?null:(new Byte(r.values[i].byteValue()));
  }
  
  public synchronized Object clone()
  {
    return new ByteObjectIterator(this);
  }
  
  public boolean equals(Object o) 
  {
    if(!(o instanceof ByteObjectIterator))
      return false;

    ByteObjectIterator obj = (ByteObjectIterator)o;

    if(index != obj.index) return false;
    if(   it == null && obj.it != null
       || it != null && !it.equals(obj.it)
      ) return false;
    if(values == null) return (obj.values == null);
    if(values != null && obj.values == null) return false;
    if(values.length != obj.values.length) return false;
    
    for(int i = 0; i < values.length; i++)
      if(   values[i] == null && obj.values[i] != null
         || values[i] != null && !values[i].equals(obj.values[i])
        ) return false;
    
    return true;   
  }

  public Byte value()
  {
    if(values != null && index < values.length) return values[index];
    else if(it != null)       return new Byte(it.value());
    else                      return null;
  }
  
  public boolean stop()
  {
    if(it == null)          return (index >= values.length);
    else if(values == null) return it.stop();
    else                    return index >= values.length && it.stop();
  }
  
  public void shift()
  {
    if(values != null && index < values.length) index++;
    else if(it != null)       it.next();
  }
  
  public String toString()
  {
    if(stop()) return jatva.iterators.Iterator.EXHAUSTED;
    return "" + value();
  }

  public Object next()
  {
    Object r = value();
    shift();
    return r;    
  }

  public int hashCode ()
  {
    int r = index;
    if(values != null)
      for(int i = 0; i < values.length; i++)
        r += values[i].byteValue()*i;
    
    if(it != null)
      r += it.hashCode();
    
    return r;
  }
}
