/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class CharacterIterator
  extends AbstractCharacterIterator
  implements ICharacterIterator
{
  public static char[] defaultValues = {
     0x0000
   , 0xFFFF
   , 0xFFFe
   , 0x0001
   , 0x00FF
   , 0x0100
   , ' '
   , '\n'
   , '\r'
   , '0'
   , 'Z'
   , 'z'
   , 0xABCD
   , '#'
   , ''
   , 0x0180
   , 0x0FFF
   , 0x0250
   , 0x02b0
  };
  
  protected ICharacterIterator[] items = new ICharacterIterator[] { new ArrayCharacterIterator() };
  protected int index = 0;
  
  public CharacterIterator()
  {
  }
  
  public CharacterIterator(char[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = new ICharacterIterator[] { new ArrayCharacterIterator(values) };
  }
  
  public CharacterIterator(char start, char end, int offset, int divisor)
  {
    items = new ICharacterIterator[] { new SegmentCharacterIterator(start, end, offset, divisor) };
  }

  public CharacterIterator(char point, int offset)
  {
    items = new ICharacterIterator[] { new NeighbourhoodCharacterIterator(point, offset) };
  }

  
  public CharacterIterator(ICharacterIterator[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = values;
  }
  
  public void addSegment(char start, char end, int offset, int divisor)
  {
    ICharacterIterator[] newItems = new ICharacterIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new SegmentCharacterIterator(start, end, offset, divisor);
    items = newItems;
  }
  
  public void addSegment(char start, char end, int offset)
  {
    addSegment(start, end, offset, 1);
  }
  
  public void addSegment(char start, char end)
  {
    addSegment(start, end, 0, 1);
  }
  
  public void addPoint(char point, int offset)
  {
    ICharacterIterator[] newItems = new ICharacterIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new NeighbourhoodCharacterIterator(point, offset);
    items = newItems;
  }
  
  public void addArray(char[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    ICharacterIterator[] newItems = new ICharacterIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new ArrayCharacterIterator(values);
    items = newItems;
  }
  
  public void addIterator(ICharacterIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    ICharacterIterator[] newItems = new ICharacterIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = (ICharacterIterator)it.clone();
    items = newItems;
  }
  
  public void removeItem(int index)
  {
    if(index >= this.index && index < items.length)
    {
      ICharacterIterator[] newItems = new ICharacterIterator[items.length-1];
      System.arraycopy(items, 0, newItems, 0, index);
      System.arraycopy(items, index+1, newItems, index, items.length-index-1);
      items = newItems;
    }
  }
  
  public CharacterIterator(CharacterIterator r)
  {
    index = r.index;
    items = new ICharacterIterator[r.items.length];
    
    for(int i = 0; i < items.length; i++)
      if(r.items[i] != null)
        items[i] = (ICharacterIterator)r.items[i].clone();
  }
  
  public Object clone()
  {
    return new CharacterIterator(this);
  }
  
  public void shift()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    if(index < items.length) items[index].next();

    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
  }
  
  public char value()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    return items[index].value();
  }
  
  public boolean stop()
  {
    int i = index;
    while(i < items.length && (items[i] == null || items[i].stop()))
    {
      i++;
    }

    if(i >= items.length) return true;
    else                  return false;
  }

  public boolean equals(Object o) 
  {
    if(!(o instanceof CharacterIterator))
      return false;

    CharacterIterator obj = (CharacterIterator)o;

    if(   index != obj.index
       || items.length != obj.items.length
      )
      return false;
      
    for(int i = 0; i < items.length; i++)
      if(items[i] == null && obj.items[i] != null || items[i] != null && !items[i].equals(obj.items[i]))
        return false;
        
    return true;
  }

  public int hashCode ()
  {
    int r = 0;
    for(int i = 0; i < items.length; i++)
      r += items[i].hashCode();
    
    return r;
  }
}
