/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

import jatva.iterators.AbstractIterator;

public class ClassIterator
  extends AbstractIterator
  implements jatva.iterators.Iterator
{
  public static Class[] defaultValues = new Class[8];
  static
  {
    defaultValues[0] = defaultValues[7] = null;
    try
    {
      defaultValues[1] = Class.forName("java.lang.Object");
      defaultValues[2] = defaultValues[5] = Class.forName("java.lang.String");
      defaultValues[3] = Class.forName("java.lang.Object");
      defaultValues[4] = Class.forName("java.util.Vector");
      defaultValues[6] = Class.forName("java.lang.Class");
    }
    catch(ClassNotFoundException ex)
    {
      ex.printStackTrace();
    }
  };

  
  protected Class[] values = defaultValues;
  protected int index = 0;
  
  public ClassIterator()
  {
  }
  
  public ClassIterator(Class[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = values;
  }
  
  public ClassIterator(ClassIterator right)
  {
    index = right.index;
    values = right.values;
  }
  
  public void addClass(Class n)
  {
    Class[] newValues = new Class[values.length+1];
    System.arraycopy(values, 0, newValues, 0, values.length);
    newValues[values.length] = n;
    values = newValues;
  }

  public Object clone()
  {
    return new ClassIterator(this);
  }
  
  public void shift()
  {
    index++;
  }
  
  public Class value()
  {
    return values[index];
  }
  
  public boolean stop()
  {
    return index >= values.length;
  }

  public boolean equals(Object o) 
  {
    if(!(o instanceof ClassIterator))
      return false;

    ClassIterator obj = (ClassIterator)o;

    if(index != obj.index) return false;
    if(values.length != obj.values.length) return false;
    
    for(int i = 0; i < values.length; i++)
      if(   values[i] == null && obj.values[i] != null
         || values[i] != null && !values[i].equals(obj.values[i])
        ) return false;
    
    return true;   
  }

  public String toString()
  {
    if(stop()) return jatva.iterators.Iterator.EXHAUSTED;
    return "" + value();
  }

  public Object next()
  {
    Object r = value();
    shift();
    return r;    
  }

  public int hashCode ()
  {
    int r = index;
    for(int i = 0; i < values.length; i++)
      r += values[i].hashCode()*i;
    
    return r;
  }
}
