/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class DoubleIterator
  extends AbstractDoubleIterator
  implements IDoubleIterator
{
  public static double shiftDouble(double v, long i)
  {
    long l = Double.doubleToLongBits(v);
    
    if(l < 0)
    {
      if(i > 0 && l - i > 0)
        return Double.longBitsToDouble(0x7fffffffffffffffL - (l - i));
      else
        return Double.longBitsToDouble(l-i);
    }
    else
    {
      if(i < 0 && l + i < 0)
        return Double.longBitsToDouble(0x8000000000000000L - (l + i) - 1L);
      else
        return Double.longBitsToDouble(l+i);
    }
  }

  public static double nextDouble(double v)
  {
    return shiftDouble(v, 1);
  }
  
  public static double prevDouble(double v)
  {
    return shiftDouble(v, -1);
  }
  
  public static long numberOfDoubles(double v, double w)
  {
    if(v == w)
    {
      if(v != 0) return 0;
      else if(Double.doubleToLongBits(w) != Double.doubleToLongBits(v)) return 1;
      else return 0;
    }
      
    if(Double.isNaN(v) && Double.isNaN(w)) return 0;
    else if(Double.isNaN(v)) return numberOfDoubles(Double.NEGATIVE_INFINITY, w)+1;
    else if(Double.isNaN(w)) return numberOfDoubles(v, Double.POSITIVE_INFINITY)+1;

    if(v < 0 && w == 0)
    {
      if(Double.doubleToLongBits(w) == 0)
        return Math.abs(Double.doubleToLongBits(v)-Double.doubleToLongBits(-0.0) + 1);
      else
        return Math.abs(Double.doubleToLongBits(v)-Double.doubleToLongBits(-0.0));
    }
    else if(w < 0 && v == 0)
    {
      if(Double.doubleToLongBits(v) == 0)
        return Math.abs(Double.doubleToLongBits(w)-Double.doubleToLongBits(-0.0) + 1);
      else
        return Math.abs(Double.doubleToLongBits(w)-Double.doubleToLongBits(-0.0));
    }
    else if(v > 0 && w == 0)
    {
      if(Double.doubleToLongBits(w) == 0)
        return Double.doubleToLongBits(v);
      else
        return Double.doubleToLongBits(v)+1;
    }
    else if(w > 0 && v == 0)
    {
      if(Double.doubleToLongBits(v) == 0)
        return Double.doubleToLongBits(w);
      else
        return Double.doubleToLongBits(w)+1;
    }
    else if(v > w && w > 0 || v < w && w < 0)
      return Math.abs(Double.doubleToLongBits(v)-Double.doubleToLongBits(w));
    else if(w > v && v > 0 || w < v && v < 0)
      return Math.abs(Double.doubleToLongBits(w)-Double.doubleToLongBits(v));
    else if(v > 0 && 0 > w)
      return numberOfDoubles(v, 0.0) + numberOfDoubles(0.0, w);
    else
      return numberOfDoubles(w, 0.0) + numberOfDoubles(0.0, v);
  }

  public static boolean compareDoubles(double v, double w)
  {
    return Double.isNaN(v) && Double.isNaN(w) || !Double.isNaN(v) && v == w;
  }

  public static double[] defaultValues =
  {
     0
   , Double.MIN_VALUE
   , Double.MAX_VALUE
   , -Double.MIN_VALUE
   , -Double.MAX_VALUE
   , Double.POSITIVE_INFINITY
   , Double.NEGATIVE_INFINITY
   , Double.NaN
   , 1.0
   , -1.0
   , Double.longBitsToDouble(0x3ff0000000000001L)
   , Double.longBitsToDouble(0x3fefffffffffffffL)
   , -Double.longBitsToDouble(0x3ff0000000000001L)
   , -Double.longBitsToDouble(0x3fefffffffffffffL)
   , prevDouble(Float.MIN_VALUE)
   , nextDouble(Float.MAX_VALUE)
   , Double.longBitsToDouble(0x5ac3a8f30ed179b4L)
   , -Double.longBitsToDouble(0x31fca7b029e4f8d6L)
  };
  
  protected IDoubleIterator[] items = new IDoubleIterator[] { new ArrayDoubleIterator() };
  protected int index = 0;
  
  public DoubleIterator()
  {
  }
  
  public DoubleIterator(double[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = new IDoubleIterator[] { new ArrayDoubleIterator(values) };
  }
  
  public DoubleIterator(double start, double end, int offset, int divisor)
  {
    items = new IDoubleIterator[] { new SegmentDoubleIterator(start, end, offset, divisor) };
  }

  public DoubleIterator(double point, int offset)
  {
    items = new IDoubleIterator[] { new NeighbourhoodDoubleIterator(point, offset) };
  }

  
  public DoubleIterator(IDoubleIterator[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = values;
  }
  
  public void addSegment(double start, double end, int offset, int divisor)
  {
    IDoubleIterator[] newItems = new IDoubleIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new SegmentDoubleIterator(start, end, offset, divisor);
    items = newItems;
  }
  
  public void addSegment(double start, double end, int offset)
  {
    addSegment(start, end, offset, 1);
  }
  
  public void addSegment(double start, double end)
  {
    addSegment(start, end, 0, 1);
  }
  
  public void addPoint(double point, int offset)
  {
    IDoubleIterator[] newItems = new IDoubleIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new NeighbourhoodDoubleIterator(point, offset);
    items = newItems;
  }
  
  public void addArray(double[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    IDoubleIterator[] newItems = new IDoubleIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new ArrayDoubleIterator(values);
    items = newItems;
  }
  
  public void addIterator(IDoubleIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    IDoubleIterator[] newItems = new IDoubleIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = (IDoubleIterator)it.clone();
    items = newItems;
  }
  
  public void removeItem(int index)
  {
    if(index >= this.index && index < items.length)
    {
      IDoubleIterator[] newItems = new IDoubleIterator[items.length-1];
      System.arraycopy(items, 0, newItems, 0, index);
      System.arraycopy(items, index+1, newItems, index, items.length-index-1);
      items = newItems;
    }
  }
  
  public DoubleIterator(DoubleIterator r)
  {
    index = r.index;
    items = new IDoubleIterator[r.items.length];
    
    for(int i = 0; i < items.length; i++)
      if(r.items[i] != null)
        items[i] = (IDoubleIterator)r.items[i].clone();
  }
  
  public Object clone()
  {
    return new DoubleIterator(this);
  }
  
  public void shift()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    if(index < items.length) items[index].next();

    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
  }
  
  public double value()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    return items[index].value();
  }
  
  public boolean stop()
  {
    int i = index;
    while(i < items.length && (items[i] == null || items[i].stop()))
    {
      i++;
    }

    if(i >= items.length) return true;
    else                  return false;
  }

  public boolean equals(Object o) 
  {
    if(!(o instanceof DoubleIterator))
      return false;

    DoubleIterator obj = (DoubleIterator)o;

    if(   index != obj.index
       || items.length != obj.items.length
      )
      return false;
      
    for(int i = 0; i < items.length; i++)
      if(items[i] == null && obj.items[i] != null || items[i] != null && !items[i].equals(obj.items[i]))
        return false;
        
    return true;
  }

  public int hashCode ()
  {
    int r = 0;
    for(int i = 0; i < items.length; i++)
      r += items[i].hashCode();
    
    return r;
  }
}
