/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

import jatva.iterators.AbstractIterator;

public class DoubleObjectIterator
  extends AbstractIterator
  implements jatva.iterators.Iterator
{
  public static Double[] defaultValues = {
     null
   , new Double((double)(Double.MIN_VALUE + 1))
   , new Double((double)0)
   , null
  };
  static
  {
    defaultValues[3] = defaultValues[1];
  }

  protected Double[] values = defaultValues;
  protected IDoubleIterator it = new ArrayDoubleIterator();
  protected int index = 0;
  
  public DoubleObjectIterator()
  {
  }

  public DoubleObjectIterator(Double[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Double[values.length];
    it = null;
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:(new Double(values[i].doubleValue()));
  }

  public DoubleObjectIterator(IDoubleIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    values = null;
    this.it = (IDoubleIterator)it.clone();
  }
  
  public DoubleObjectIterator(Double[] values, IDoubleIterator it)
  {
    if(values == null && it == null) throw new IllegalArgumentException("Both arguments should not be null");
    if(values != null)
    {
      this.values = new Double[values.length];
      
      for(int i = 0; i < values.length; i++)
        this.values[i] = (values[i] == null)?null:(new Double(values[i].doubleValue()));
    }
    if(it != null) this.it = (IDoubleIterator)it.clone();
  }
  
  public void setValueIterator(IDoubleIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    this.it = (IDoubleIterator)it.clone();
  }
  
  public void setArray(Double[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Double[values.length];
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:new Double(values[i].doubleValue());
  }

  public DoubleObjectIterator(DoubleObjectIterator r)
  {
    index = r.index;
    it = (r.it == null)?null:(IDoubleIterator)r.it.clone();
    if(r.values == null) { values = null; return; }
    values = new Double[r.values.length];
    
    for(int i = 0; i < values.length; i++)
      values[i] = (r.values[i] == null)?null:(new Double(r.values[i].doubleValue()));
  }
  
  public Object clone()
  {
    return new DoubleObjectIterator(this);
  }
  
  public boolean equals(Object o) 
  {
    if(!(o instanceof DoubleObjectIterator))
      return false;

    DoubleObjectIterator obj = (DoubleObjectIterator)o;

    if(index != obj.index) return false;
    if(   it == null && obj.it != null
       || it != null && !it.equals(obj.it)
      ) return false;
    if(values == null) return (obj.values == null);
    if(values != null && obj.values == null) return false;
    if(values.length != obj.values.length) return false;
    
    for(int i = 0; i < values.length; i++)
      if(   values[i] == null && obj.values[i] != null
         || values[i] != null && !values[i].equals(obj.values[i])
        ) return false;
    
    return true;   
  }

  public Double value()
  {
    if(values != null && index < values.length) return values[index];
    else if(it != null)       return new Double(it.value());
    else                      return null;
  }
  
  public boolean stop()
  {
    if(it == null)          return (index >= values.length);
    else if(values == null) return it.stop();
    else                    return index >= values.length && it.stop();
  }
  
  public void shift()
  {
    if(values != null && index < values.length) index++;
    else if(it != null)       it.next();
  }
  
  public String toString()
  {
    if(stop()) return jatva.iterators.Iterator.EXHAUSTED;
    return "" + value();
  }

  public Object next()
  {
    Object r = value();
    shift();
    return r;    
  }

  public int hashCode ()
  {
    int r = index;
    if(values != null)
      for(int i = 0; i < values.length; i++)
        r += (int)Double.doubleToLongBits(values[i].doubleValue())*i;
    
    if(it != null)
      r += it.hashCode();
    
    return r;
  }
}
