/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class FloatIterator
  extends AbstractFloatIterator
  implements IFloatIterator
{
  public static float shiftFloat(float v, int i)
  {
    int l = Float.floatToRawIntBits(v);
    
    if(l < 0)
    {
      if(i > 0 && l - i > 0)
        return Float.intBitsToFloat(0x7fffffff - (l - i));
      else
        return Float.intBitsToFloat(l-i);
    }
    else
    {
      if(i < 0 && l + i < 0)
        return Float.intBitsToFloat(0x80000000 - (l + i) - 1);
      else
        return Float.intBitsToFloat(l+i);
    }
  }

  public static float nextFloat(float v)
  {
    return shiftFloat(v, 1);
  }
  
  public static float prevFloat(float v)
  {
    return shiftFloat(v, -1);
  }
  
  public static long numberOfFloats(float v, float w)
  {
    if(v == w)
    {
      if(v != 0) return 0;
      else if(Float.floatToIntBits(w) != Float.floatToIntBits(v)) return 1;
      else return 0;
    }
    if(Float.isNaN(v) && Float.isNaN(w)) return 0;
    else if(Float.isNaN(v)) return numberOfFloats(Float.NEGATIVE_INFINITY, w)+1;
    else if(Float.isNaN(w)) return numberOfFloats(v, Float.POSITIVE_INFINITY)+1;
    
    if(v < 0 && w == 0)
    {
      if(Float.floatToIntBits(w) == 0)
        return Math.abs((long)Float.floatToIntBits(v)-(long)Float.floatToIntBits(-0.0f) + 1);
      else
        return Math.abs((long)Float.floatToIntBits(v)-(long)Float.floatToIntBits(-0.0f));
    }
    else if(w < 0 && v == 0)
    {
      if(Float.floatToIntBits(v) == 0)
        return Math.abs((long)Float.floatToIntBits(w)-(long)Float.floatToIntBits(-0.0f) + 1);
      else
        return Math.abs((long)Float.floatToIntBits(w)-(long)Float.floatToIntBits(-0.0f));
    }
    else if(v > 0 && w == 0)
    {
      if(Float.floatToIntBits(w) == 0)
        return (long)Float.floatToIntBits(v);
      else
        return (long)Float.floatToIntBits(v)+1;
    }
    else if(w > 0 && v == 0)
    {
      if(Float.floatToIntBits(v) == 0)
        return (long)Float.floatToIntBits(w);
      else
        return (long)Float.floatToIntBits(w)+1;
    }
    else if(v > w && w > 0 || v < w && w < 0)
      return Math.abs((long)Float.floatToIntBits(v)-(long)Float.floatToIntBits(w));
    else if(w > v && v > 0 || w < v && v < 0)
      return Math.abs((long)Float.floatToIntBits(w)-(long)Float.floatToIntBits(v));
    else if(v > 0 && 0 > w)
      return numberOfFloats(v, 0.0f) + numberOfFloats(0.0f, w);
    else
      return numberOfFloats(w, 0.0f) + numberOfFloats(0.0f, v);
  }
  
  public static boolean compareFloats(float v, float w)
  {
    return Float.isNaN(v) && Float.isNaN(w) || !Float.isNaN(v) && v == w;
  }

  public static float[] defaultValues =
  {
     0
   , Float.MIN_VALUE
   , Float.MAX_VALUE
   , -Float.MIN_VALUE
   , -Float.MAX_VALUE
   , Float.POSITIVE_INFINITY
   , Float.NEGATIVE_INFINITY
   , Float.NaN
   , (float)(1.0)
   , (float)(-1.0)
   , nextFloat((float)(1.0))
   , prevFloat((float)(1.0))
   , -Float.intBitsToFloat(0x3f800001)
   , -Float.intBitsToFloat(0x3f7fffff)
   , Float.intBitsToFloat(0x5ac3a8f3)
   , -Float.intBitsToFloat(0x31fca7b0)
  };
  
  protected IFloatIterator[] items = new IFloatIterator[] { new ArrayFloatIterator() };
  protected int index = 0;
  
  public FloatIterator()
  {
  }
  
  public FloatIterator(float[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = new IFloatIterator[] { new ArrayFloatIterator(values) };
  }
  
  public FloatIterator(float start, float end, int offset, int divisor)
  {
    items = new IFloatIterator[] { new SegmentFloatIterator(start, end, offset, divisor) };
  }

  public FloatIterator(float point, int offset)
  {
    items = new IFloatIterator[] { new NeighbourhoodFloatIterator(point, offset) };
  }

  
  public FloatIterator(IFloatIterator[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = values;
  }
  
  public void addSegment(float start, float end, int offset, int divisor)
  {
    IFloatIterator[] newItems = new IFloatIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new SegmentFloatIterator(start, end, offset, divisor);
    items = newItems;
  }
  
  public void addSegment(float start, float end, int offset)
  {
    addSegment(start, end, offset, 1);
  }
  
  public void addSegment(float start, float end)
  {
    addSegment(start, end, 0, 1);
  }
  
  public void addPoint(float point, int offset)
  {
    IFloatIterator[] newItems = new IFloatIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new NeighbourhoodFloatIterator(point, offset);
    items = newItems;
  }
  
  public void addArray(float[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    IFloatIterator[] newItems = new IFloatIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new ArrayFloatIterator(values);
    items = newItems;
  }
  
  public void addIterator(IFloatIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    IFloatIterator[] newItems = new IFloatIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = (IFloatIterator)it.clone();
    items = newItems;
  }
  
  public void removeItem(int index)
  {
    if(index >= this.index && index < items.length)
    {
      IFloatIterator[] newItems = new IFloatIterator[items.length-1];
      System.arraycopy(items, 0, newItems, 0, index);
      System.arraycopy(items, index+1, newItems, index, items.length-index-1);
      items = newItems;
    }
  }
  
  public FloatIterator(FloatIterator r)
  {
    index = r.index;
    items = new IFloatIterator[r.items.length];
    
    for(int i = 0; i < items.length; i++)
      if(r.items[i] != null)
        items[i] = (IFloatIterator)r.items[i].clone();
  }
  
  public Object clone()
  {
    return new FloatIterator(this);
  }
  
  public void shift()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    if(index < items.length) items[index].next();

    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
  }
  
  public float value()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    return items[index].value();
  }
  
  public boolean stop()
  {
    int i = index;
    while(i < items.length && (items[i] == null || items[i].stop()))
    {
      i++;
    }

    if(i >= items.length) return true;
    else                  return false;
  }

  public boolean equals(Object o) 
  {
    if(!(o instanceof FloatIterator))
      return false;

    FloatIterator obj = (FloatIterator)o;

    if(   index != obj.index
       || items.length != obj.items.length
      )
      return false;
      
    for(int i = 0; i < items.length; i++)
      if(items[i] == null && obj.items[i] != null || items[i] != null && !items[i].equals(obj.items[i]))
        return false;
        
    return true;
  }

  public int hashCode ()
  {
    int r = 0;
    for(int i = 0; i < items.length; i++)
      r += items[i].hashCode();
    
    return r;
  }
}
