/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

import jatva.iterators.AbstractIterator;

public class FloatObjectIterator
  extends AbstractIterator
  implements jatva.iterators.Iterator
{
  public static Float[] defaultValues = {
     null
   , new Float((float)(Float.MIN_VALUE + 1))
   , new Float((float)0)
   , null
  };
  static
  {
    defaultValues[3] = defaultValues[1];
  }

  protected Float[] values = defaultValues;
  protected IFloatIterator it = new ArrayFloatIterator();
  protected int index = 0;
  
  public FloatObjectIterator()
  {
  }

  public FloatObjectIterator(Float[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Float[values.length];
    it = null;
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:(new Float(values[i].floatValue()));
  }

  public FloatObjectIterator(IFloatIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    values = null;
    this.it = (IFloatIterator)it.clone();
  }
  
  public FloatObjectIterator(Float[] values, IFloatIterator it)
  {
    if(values == null && it == null) throw new IllegalArgumentException("Both arguments should not be null");
    if(values != null)
    {
      this.values = new Float[values.length];
      
      for(int i = 0; i < values.length; i++)
        this.values[i] = (values[i] == null)?null:(new Float(values[i].floatValue()));
    }
    if(it != null) this.it = (IFloatIterator)it.clone();
  }
  
  public void setValueIterator(IFloatIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    this.it = (IFloatIterator)it.clone();
  }
  
  public void setArray(Float[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Float[values.length];
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:new Float(values[i].floatValue());
  }
  
  public FloatObjectIterator(FloatObjectIterator r)
  {
    index = r.index;
    it = (r.it == null)?null:(IFloatIterator)r.it.clone();
    if(r.values == null) { values = null; return; }
    values = new Float[r.values.length];
    
    for(int i = 0; i < values.length; i++)
      values[i] = (r.values[i] == null)?null:(new Float(r.values[i].floatValue()));
  }
  
  public Object clone()
  {
    return new FloatObjectIterator(this);
  }
  
  public boolean equals(Object o) 
  {
    if(!(o instanceof FloatObjectIterator))
      return false;

    FloatObjectIterator obj = (FloatObjectIterator)o;

    if(index != obj.index) return false;
    if(   it == null && obj.it != null
       || it != null && !it.equals(obj.it)
      ) return false;
    if(values == null) return (obj.values == null);
    if(values != null && obj.values == null) return false;
    if(values.length != obj.values.length) return false;
    
    for(int i = 0; i < values.length; i++)
      if(   values[i] == null && obj.values[i] != null
         || values[i] != null && !values[i].equals(obj.values[i])
        ) return false;
    
    return true;   
  }

  public Float value()
  {
    if(values != null && index < values.length) return values[index];
    else if(it != null)       return new Float(it.value());
    else                      return null;
  }
  
  public boolean stop()
  {
    if(it == null)          return (index >= values.length);
    else if(values == null) return it.stop();
    else                    return index >= values.length && it.stop();
  }
  
  public void shift()
  {
    if(values != null && index < values.length) index++;
    else if(it != null)       it.next();
  }
  
  public String toString()
  {
    if(stop()) return jatva.iterators.Iterator.EXHAUSTED;
    return "" + value();
  }

  public Object next()
  {
    Object r = value();
    shift();
    return r;    
  }

  public int hashCode ()
  {
    int r = index;
    if(values != null)
      for(int i = 0; i < values.length; i++)
        r += Float.floatToIntBits(values[i].floatValue())*i;
    
    if(it != null)
      r += it.hashCode();
    
    return r;
  }
}
