/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class IntegerIterator
  extends AbstractIntegerIterator
  implements IIntegerIterator
{
  public static int[] defaultValues = {
     Integer.MIN_VALUE
   , Integer.MAX_VALUE
   , Integer.MIN_VALUE + 1
   , Integer.MAX_VALUE - 1
   , 0
   , 1
   , -1
   , Byte.MIN_VALUE - 1
   , Byte.MAX_VALUE + 1  
   , Short.MIN_VALUE - 1
   , Short.MAX_VALUE + 1
   , Integer.MIN_VALUE + 572
   , Integer.MAX_VALUE - 328947
   , Integer.MIN_VALUE + 73901635
   , Integer.MAX_VALUE - 106724983
  };
  
  protected IIntegerIterator[] items = new IIntegerIterator[] { new ArrayIntegerIterator() };
  protected int index = 0;
  
  public IntegerIterator()
  {
  }
  
  public IntegerIterator(int[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = new IIntegerIterator[] { new ArrayIntegerIterator(values) };
  }
  
  public IntegerIterator(int start, int end, int offset, int divisor)
  {
    items = new IIntegerIterator[] { new SegmentIntegerIterator(start, end, offset, divisor) };
  }

  public IntegerIterator(int point, int offset)
  {
    items = new IIntegerIterator[] { new NeighbourhoodIntegerIterator(point, offset) };
  }

  
  public IntegerIterator(IIntegerIterator[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = values;
  }
  
  public void addSegment(int start, int end, int offset, int divisor)
  {
    IIntegerIterator[] newItems = new IIntegerIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new SegmentIntegerIterator(start, end, offset, divisor);
    items = newItems;
  }
  
  public void addSegment(int start, int end, int offset)
  {
    addSegment(start, end, offset, 1);
  }
  
  public void addSegment(int start, int end)
  {
    addSegment(start, end, 0, 1);
  }
  
  public void addPoint(int point, int offset)
  {
    IIntegerIterator[] newItems = new IIntegerIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new NeighbourhoodIntegerIterator(point, offset);
    items = newItems;
  }
  
  public void addArray(int[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    IIntegerIterator[] newItems = new IIntegerIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new ArrayIntegerIterator(values);
    items = newItems;
  }
  
  public void addIterator(IIntegerIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    IIntegerIterator[] newItems = new IIntegerIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = (IIntegerIterator)it.clone();
    items = newItems;
  }
  
  public void removeItem(int index)
  {
    if(index >= this.index && index < items.length)
    {
      IIntegerIterator[] newItems = new IIntegerIterator[items.length-1];
      System.arraycopy(items, 0, newItems, 0, index);
      System.arraycopy(items, index+1, newItems, index, items.length-index-1);
      items = newItems;
    }
  }
  
  public IntegerIterator(IntegerIterator r)
  {
    index = r.index;
    items = new IIntegerIterator[r.items.length];
    
    for(int i = 0; i < items.length; i++)
      if(r.items[i] != null)
        items[i] = (IIntegerIterator)r.items[i].clone();
  }
  
  public Object clone()
  {
    return new IntegerIterator(this);
  }
  
  public void shift()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    if(index < items.length) items[index].next();

    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
  }
  
  public int value()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    return items[index].value();
  }
  
  public boolean stop()
  {
    int i = index;
    while(i < items.length && (items[i] == null || items[i].stop()))
    {
      i++;
    }

    if(i >= items.length) return true;
    else                  return false;
  }

  public boolean equals(Object o) 
  {
    if(!(o instanceof IntegerIterator))
      return false;

    IntegerIterator obj = (IntegerIterator)o;

    if(   index != obj.index
       || items.length != obj.items.length
      )
      return false;
      
    for(int i = 0; i < items.length; i++)
      if(items[i] == null && obj.items[i] != null || items[i] != null && !items[i].equals(obj.items[i]))
        return false;
        
    return true;
  }

  public int hashCode ()
  {
    int r = 0;
    for(int i = 0; i < items.length; i++)
      r += items[i].hashCode();
    
    return r;
  }
}
