/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

import jatva.iterators.AbstractIterator;

public class IntegerObjectIterator
  extends AbstractIterator
  implements jatva.iterators.Iterator
{
  public static Integer[] defaultValues = {
     null
   , new Integer((int)(Integer.MIN_VALUE + 1))
   , new Integer((int)0)
   , null
  };
  static
  {
    defaultValues[3] = defaultValues[1];
  }

  protected Integer[] values = defaultValues;
  protected IIntegerIterator it = new ArrayIntegerIterator();
  protected int index = 0;
  
  public IntegerObjectIterator()
  {
  }

  public IntegerObjectIterator(Integer[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Integer[values.length];
    it = null;
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:(new Integer(values[i].intValue()));
  }

  public IntegerObjectIterator(IIntegerIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    values = null;
    this.it = (IIntegerIterator)it.clone();
  }
  
  public IntegerObjectIterator(Integer[] values, IIntegerIterator it)
  {
    if(values == null && it == null) throw new IllegalArgumentException("Both arguments should not be null");
    if(values != null)
    {
      this.values = new Integer[values.length];
      
      for(int i = 0; i < values.length; i++)
        this.values[i] = (values[i] == null)?null:(new Integer(values[i].intValue()));
    }
    if(it != null) this.it = (IIntegerIterator)it.clone();
  }
  
  public void setValueIterator(IIntegerIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    this.it = (IIntegerIterator)it.clone();
  }
  
  public void setArray(Integer[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Integer[values.length];
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:(new Integer(values[i].intValue()));
  }
  
  public IntegerObjectIterator(IntegerObjectIterator r)
  {
    index = r.index;
    it = (r.it == null)?null:(IIntegerIterator)r.it.clone();
    if(r.values == null) { values = null; return; }
    values = new Integer[r.values.length];
    
    for(int i = 0; i < values.length; i++)
      values[i] = (r.values[i] == null)?null:(new Integer(r.values[i].intValue()));
  }
  
  public Object clone()
  {
    return new IntegerObjectIterator(this);
  }
  
  public boolean equals(Object o) 
  {
    if(!(o instanceof IntegerObjectIterator))
      return false;

    IntegerObjectIterator obj = (IntegerObjectIterator)o;

    if(index != obj.index) return false;
    if(   it == null && obj.it != null
       || it != null && !it.equals(obj.it)
      ) return false;
    if(values == null) return obj.values == null;
    if(values != null && obj.values == null) return false;
    if(values.length != obj.values.length) return false;
    
    for(int i = 0; i < values.length; i++)
      if(   values[i] == null && obj.values[i] != null
         || values[i] != null && !values[i].equals(obj.values[i])
        ) return false;
    
    return true;   
  }

  public Integer value()
  {
    if(values != null && index < values.length) return values[index];
    else if(it != null)       return new Integer(it.value());
    else                      return null;
  }
  
  public boolean stop()
  {
    if(it == null)          return (index >= values.length);
    else if(values == null) return it.stop();
    else                    return index >= values.length && it.stop();
  }
  
  public void shift()
  {
    if(values != null && index < values.length) index++;
    else if(it != null)       it.next();
  }
  
  public String toString()
  {
    if(stop()) return jatva.iterators.Iterator.EXHAUSTED;
    return "" + value();
  }

  public Object next()
  {
    Object r = value();
    shift();
    return r;    
  }

  public int hashCode ()
  {
    int r = index;
    if(values != null)
      for(int i = 0; i < values.length; i++)
        r += values[i].intValue()*i;
    
    if(it != null)
      r += it.hashCode();
    
    return r;
  }
}
