/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class LongIterator
  extends AbstractLongIterator
  implements ILongIterator
{
  public static long[] defaultValues = {
     Long.MIN_VALUE
   , Long.MAX_VALUE
   , Long.MIN_VALUE + 1
   , Long.MAX_VALUE - 1
   , 0
   , 1
   , -1
   , Byte.MIN_VALUE - 1L
   , Byte.MAX_VALUE + 1L  
   , Short.MIN_VALUE - 1L
   , Short.MAX_VALUE + 1L
   , Integer.MIN_VALUE - 1L
   , Integer.MAX_VALUE + 1L
   , Long.MIN_VALUE + 334234L
   , Long.MAX_VALUE - 7129825903218947L
   , Long.MIN_VALUE + 239674803709521659L
  };
  
  protected ILongIterator[] items = new ILongIterator[] { new ArrayLongIterator() };
  protected int index = 0;
  
  public LongIterator()
  {
  }
  
  public LongIterator(long[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = new ILongIterator[] { new ArrayLongIterator(values) };
  }
  
  public LongIterator(long start, long end, int offset, int divisor)
  {
    items = new ILongIterator[] { new SegmentLongIterator(start, end, offset, divisor) };
  }

  public LongIterator(long point, int offset)
  {
    items = new ILongIterator[] { new NeighbourhoodLongIterator(point, offset) };
  }

  
  public LongIterator(ILongIterator[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = values;
  }
  
  public void addSegment(long start, long end, int offset, int divisor)
  {
    ILongIterator[] newItems = new ILongIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new SegmentLongIterator(start, end, offset, divisor);
    items = newItems;
  }
  
  public void addSegment(long start, long end, int offset)
  {
    addSegment(start, end, offset, 1);
  }
  
  public void addSegment(long start, long end)
  {
    addSegment(start, end, 0, 1);
  }
  
  public void addPoint(long point, int offset)
  {
    ILongIterator[] newItems = new ILongIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new NeighbourhoodLongIterator(point, offset);
    items = newItems;
  }
  
  public void addArray(long[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    ILongIterator[] newItems = new ILongIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new ArrayLongIterator(values);
    items = newItems;
  }
  
  public void addIterator(ILongIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    ILongIterator[] newItems = new ILongIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = (ILongIterator)it.clone();
    items = newItems;
  }
  
  public void removeItem(int index)
  {
    if(index >= this.index && index < items.length)
    {
      ILongIterator[] newItems = new ILongIterator[items.length-1];
      System.arraycopy(items, 0, newItems, 0, index);
      System.arraycopy(items, index+1, newItems, index, items.length-index-1);
      items = newItems;
    }
  }
  
  public LongIterator(LongIterator r)
  {
    index = r.index;
    items = new ILongIterator[r.items.length];
    
    for(int i = 0; i < items.length; i++)
      if(r.items[i] != null)
        items[i] = (ILongIterator)r.items[i].clone();
  }
  
  public Object clone()
  {
    return new LongIterator(this);
  }
  
  public void shift()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    if(index < items.length) items[index].next();

    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
  }
  
  public long value()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    return items[index].value();
  }
  
  public boolean stop()
  {
    int i = index;
    while(i < items.length && (items[i] == null || items[i].stop()))
    {
      i++;
    }

    if(i >= items.length) return true;
    else                  return false;
  }

  public boolean equals(Object o) 
  {
    if(!(o instanceof LongIterator))
      return false;

    LongIterator obj = (LongIterator)o;

    if(   index != obj.index
       || items.length != obj.items.length
      )
      return false;
      
    for(int i = 0; i < items.length; i++)
      if(items[i] == null && obj.items[i] != null || items[i] != null && !items[i].equals(obj.items[i]))
        return false;
        
    return true;
  }
 
  public int hashCode ()
  {
    int r = 0;
    for(int i = 0; i < items.length; i++)
      r += items[i].hashCode();
    
    return r;
  }
}
