/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

import jatva.iterators.AbstractIterator;

public class LongObjectIterator
  extends AbstractIterator
  implements jatva.iterators.Iterator
{
  public static Long[] defaultValues = {
     null
   , new Long((long)(Long.MIN_VALUE + 1))
   , new Long((long)0)
   , null
  };
  static
  {
    defaultValues[3] = defaultValues[1];
  }

  protected Long[] values = defaultValues;
  protected ILongIterator it = new ArrayLongIterator();
  protected int index = 0;
  
  public LongObjectIterator()
  {
  }

  public LongObjectIterator(Long[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Long[values.length];
    it = null;
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:(new Long(values[i].longValue()));
  }

  public LongObjectIterator(ILongIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    values = null;
    this.it = (ILongIterator)it.clone();
  }
  
  public LongObjectIterator(Long[] values, ILongIterator it)
  {
    if(values == null && it == null) throw new IllegalArgumentException("Both arguments should not be null");
    if(values != null)
    {
      this.values = new Long[values.length];
      
      for(int i = 0; i < values.length; i++)
        this.values[i] = (values[i] == null)?null:(new Long(values[i].longValue()));
    }
    if(it != null) this.it = (ILongIterator)it.clone();
  }
  
  public void setValueIterator(ILongIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    this.it = (ILongIterator)it.clone();
  }
  
  public void setArray(Long[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Long[values.length];
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:(new Long(values[i].longValue()));
  }
  
  public LongObjectIterator(LongObjectIterator r)
  {
    index = r.index;
    it = (r.it == null)?null:(ILongIterator)r.it.clone();
    if(r.values == null) { values = null; return; }
    values = new Long[r.values.length];
    
    for(int i = 0; i < values.length; i++)
      values[i] = (r.values[i] == null)?null:(new Long(r.values[i].longValue()));
  }
  
  public Object clone()
  {
    return new LongObjectIterator(this);
  }
  
  public boolean equals(Object o) 
  {
    if(!(o instanceof LongObjectIterator))
      return false;

    LongObjectIterator obj = (LongObjectIterator)o;

    if(index != obj.index) return false;
    if(   it == null && obj.it != null
       || it != null && !it.equals(obj.it)
      ) return false;
    if(values == null) return obj.values == null;
    if(values != null && obj.values == null) return false;
    if(values.length != obj.values.length) return false;
    
    for(int i = 0; i < values.length; i++)
      if(   values[i] == null && obj.values[i] != null
         || values[i] != null && !values[i].equals(obj.values[i])
        ) return false;
    
    return true;   
  }

  public Long value()
  {
    if(values != null && index < values.length) return values[index];
    else if(it != null)       return new Long(it.value());
    else                      return null;
  }
  
  public boolean stop()
  {
    if(it == null)          return (index >= values.length);
    else if(values == null) return it.stop();
    else                    return index >= values.length && it.stop();
  }
  
  public void shift()
  {
    if(values != null && index < values.length) index++;
    else if(it != null)       it.next();
  }
  
  public String toString()
  {
    if(stop()) return jatva.iterators.Iterator.EXHAUSTED;
    return "" + value();
  }

  public Object next()
  {
    Object r = value();
    shift();
    return r;    
  }

  public int hashCode ()
  {
    int r = index;
    if(values != null)
      for(int i = 0; i < values.length; i++)
        r += (int)values[i].longValue()*i;
    
    if(it != null)
      r += it.hashCode();
    
    return r;
  }
}
