/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class NeighbourhoodFloatIterator
  extends    AbstractFloatIterator
  implements IFloatIterator
{
  protected float point = 0;
  protected int offset = 0;
  protected int index  = 0;
  
  public NeighbourhoodFloatIterator(float point, int offset)
  {
    if(offset < 0) offset = 0;
    if(offset > Integer.MAX_VALUE/2) offset = Integer.MAX_VALUE/2;
     
    this.point  = point;
    this.offset = offset;
  }

  public NeighbourhoodFloatIterator(float point)
  {
    this(point, 0);
  }
  
  public NeighbourhoodFloatIterator(NeighbourhoodFloatIterator r)
  {
    this.point  = r.point;
    this.offset = r.offset;
    this.index  = r.index;
  }

  public Object clone()
  {
    return new NeighbourhoodFloatIterator(this);
  }
  
  public boolean equals(Object o)
  {
    if(!(o instanceof NeighbourhoodFloatIterator))
      return false;

    NeighbourhoodFloatIterator obj = (NeighbourhoodFloatIterator)o;

    return    index  == obj.index 
           && (point  == obj.point || Float.isNaN(point) && Float.isNaN(obj.point))
           && offset == obj.offset;
  }
  
  public float value()
  {
    if(Float.isNaN(point))
      switch(index)
      {
        case 0:  return Float.NaN;
        case 1:  return Float.POSITIVE_INFINITY;
        case 2:  return Float.NEGATIVE_INFINITY;
        default: return ((index&1) == 0)?
                          FloatIterator.shiftFloat(Float.NEGATIVE_INFINITY, index/2-1)
                        : FloatIterator.shiftFloat(Float.POSITIVE_INFINITY, -index/2);
      }
    else return FloatIterator.shiftFloat(point, (index%2 == 0)?(index/2):(-(index/2+1)));
  }
  
  public boolean stop()
  {
    return index > 2*offset;
  }
  
  public void shift()
  {
    index++;
    
    if(index == Integer.MAX_VALUE || Float.isNaN(point)) return;
    else if(point == Float.POSITIVE_INFINITY)
    {
      if((index&1) == 0 && index > 2) { index++; return; }
      else                                       return;
    }
    else if(point == Float.NEGATIVE_INFINITY)
    {
      if((index&1) != 0 && index > 1) { index++; return; } 
      else                                       return;
    }
    else
    {
      if(      index%2 == 0
            && point > FloatIterator.shiftFloat(Float.MAX_VALUE, - index/2)
         ||    index%2 == 1
            && point < FloatIterator.shiftFloat(-Float.MAX_VALUE, (index+1)/2)
        )
        index++;
    }
  }

  public int hashCode ()
  {
    return index + 3*Float.floatToIntBits(point) + 11*offset;
  }
}
