/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

import jatva.iterators.AbstractIterator;

public class NumberIterator
  extends AbstractIterator
  implements jatva.iterators.Iterator
{
  public static Number[] defaultValues = {
     null
   , new Byte((byte)(Byte.MIN_VALUE + 1))
   , new Short((short)(0))
   , new Integer(Integer.MIN_VALUE)
   , new Double(Double.MIN_VALUE)
   , new Long(Long.MAX_VALUE - 1)
   , null
   , new Float(Float.POSITIVE_INFINITY)
   , new Double(Double.NaN)
   , new Integer(739016735)
   , new Double(1.286492e-85)
   , null
   , null
  };
  static
  {
    try
    {
      defaultValues[6] = defaultValues[3];
      defaultValues[11] = new java.math.BigInteger("829431735066921579304918267384910539");
      defaultValues[12] = new java.math.BigDecimal("-3095812250763.0047623943602193863754018347e-1872");
    }
    catch(NumberFormatException ex) {}
  }
  
  public static Number cloneNumber(Number n)
  {
    if     (n ==         null   )              return null;
    if     (n instanceof Byte   )              return new Byte(n.byteValue());
    else if(n instanceof Short  )              return new Short(n.shortValue());
    else if(n instanceof Integer)              return new Integer(n.intValue()); 
    else if(n instanceof Long   )              return new Long(n.longValue()); 
    else if(n instanceof Float  )              return new Float(n.floatValue()); 
    else if(n instanceof Double )              return new Double(n.doubleValue()); 
    else if(n instanceof java.math.BigInteger) return new java.math.BigInteger(
                                                 ((java.math.BigInteger)n).toByteArray()
                                               ); 
    else if(n instanceof java.math.BigDecimal) return new java.math.BigDecimal(
                                                  ((java.math.BigDecimal)n).unscaledValue()
                                                , ((java.math.BigDecimal)n).scale()
                                               );
    else                                       return null; 
  }

  protected Number[] values = defaultValues;
  protected int index = 0;
  
  public NumberIterator()
  {
  }
  
  public NumberIterator(Number[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new Number[values.length];
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:cloneNumber(values[i]);
  }
  
  public NumberIterator(NumberIterator r)
  {
    index = r.index;
    values = new Number[r.values.length];
    
    for(int i = 0; i < values.length; i++)
      values[i] = (r.values[i] == null)?null:cloneNumber(r.values[i]);
  }
  
  public void addNumber(Number n)
  {
    Number[] newValues = new Number[values.length+1];
    System.arraycopy(values, 0, newValues, 0, values.length);
    newValues[values.length] = cloneNumber(n);
    values = newValues;
  }
  
  public Object clone()
  {
    return new NumberIterator(this);
  }
  
  public void shift()
  {
    index++;
  }
  
  public Number value()
  {
    return values[index];
  }
  
  public boolean stop()
  {
    return index >= values.length;
  }

  public boolean equals(Object o) 
  {
    if(!(o instanceof NumberIterator))
      return false;

    NumberIterator obj = (NumberIterator)o;

    if(index != obj.index) return false;
    if(values.length != obj.values.length) return false;
    
    for(int i = 0; i < values.length; i++)
      if(   values[i] == null && obj.values[i] != null
         || values[i] != null && !values[i].equals(obj.values[i])
        ) return false;
    
    return true;   
  }

  public String toString()
  {
    if(stop()) return jatva.iterators.Iterator.EXHAUSTED;
    return "" + value();
  }

  public Object next ()
  {
    Object r = value();
    shift();
    return r;    
  }

  public int hashCode ()
  {
    int r = index;
    for(int i = 0; i < values.length; i++)
      r += values[i].hashCode()*i;
    return r;
  }
}
