/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class SegmentCharacterIterator
  extends AbstractCharacterIterator
  implements ICharacterIterator
{
  protected char start = 0;
  protected char end = 0;
  protected int offset = 0;
  protected int divisor = 1;
  protected int index = 0;
  protected int subindex = 0;
  
  public SegmentCharacterIterator(char start, char end, int offset, int divisor)
  {
    if(divisor <= 0) divisor = 1;
    if(offset < 0) offset = 0;
    if(offset > Short.MAX_VALUE) offset = Short.MAX_VALUE;

    if((2*offset+1)*divisor >= Math.abs(end - start))
    {
      divisor = Math.abs(end - start);
      offset = 0;
    }
    
    this.start   = start;
    this.end     = end;
    this.offset  = offset;
    this.divisor = divisor;
  }

  public SegmentCharacterIterator(char start, char end, int offset)
  {
    this(start, end, offset, 1);
  }
  
  public SegmentCharacterIterator(char start, char end)
  {
    this(start, end, 0, 1);
  }
  
  public SegmentCharacterIterator(SegmentCharacterIterator r)
  {
    this.start   = r.start;
    this.end     = r.end;
    this.offset  = r.offset;
    this.divisor = r.divisor;
    this.index    = r.index;
    this.subindex = r.subindex;
  }

  public Object clone()
  {
    return new SegmentCharacterIterator(this);
  }
  
  public boolean equals(Object o)
  {
    if(!(o instanceof SegmentCharacterIterator))
      return false;

    SegmentCharacterIterator obj = (SegmentCharacterIterator)o;

    return    index    == obj.index 
           && subindex == obj.subindex 
           && start    == obj.start 
           && end      == obj.end
           && offset   == obj.offset
           && divisor  == obj.divisor;
  }
  
  public char value()
  {
    if(index == 0)
      return (char)(start + subindex*((start <= end)?(1):(-1)));
    else if(index == divisor)
      return (char)(end - subindex*((start <= end)?(1):(-1)));
    else
      return (char)(start + (end-start)*index/divisor + ((subindex%2 == 0)?(subindex/2):(-(subindex+1)/2)));
  }
  
  public boolean stop()
  {
    return index > divisor;
  }
  
  public void shift()
  {
    subindex++;

    if(   index == 0
       || index == divisor
      )
    {
      if(subindex > offset)
      {
        index++;
        subindex = 0;
      }
    }
    else if(subindex > 2*offset)
    {
      index++;
      subindex = 0;
    }
  }

  public int hashCode ()
  {
    return index + 3*subindex + 5*start + 7*end + 11*offset + 13*divisor;
  }
}
