/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class SegmentDoubleIterator
  extends AbstractDoubleIterator
  implements IDoubleIterator
{
  protected double start  = 0;
  protected double end    = 0;
  protected int offset   = 0;
  protected int divisor  = 1;
  
  /**
   * index field provides enumeration of iteration base points - a set of
   * points partitioning the segment [start, end] into divisor parts
   */
  protected int index    = 0;
  
  /**
   * subindex field in different opeartion modes means different things.
   * In jumping mode it provides enumeration of points in a neighbourhood of
   * one base point.
   * In forward mode it stores the current value of iteration - 
   * beware that this value is double and should be interpreted adequeately!  
   */
  protected long subindex = 0; 
  
  /**
   * auxiliary field to store flags designating type of segment ends and
   * modes of operation
   */
  protected int exAttribute = NORMAL;
  
  protected final static int NORMAL                             = 0;
  protected final static int FROM_POSITIVE_INFINITY             = 1;
  protected final static int FROM_NEGATIVE_INFINITY             = 2;
  protected final static int FROM_NAN                           = 4;
  protected final static int TO_POSITIVE_INFINITY               = 8;
  protected final static int TO_NEGATIVE_INFINITY               = 16;
  protected final static int TO_NAN                             = 32;
  protected final static int POSITIVE_INFINITY                  = 9;
  protected final static int FROM_NEGATIVE_TO_POSITIVE_INFINITY = 10;
  protected final static int FROM_NAN_TO_POSITIVE_INFINITY      = 12;
  protected final static int FROM_POSITIVE_TO_NEGATIVE_INFINITY = 17;
  protected final static int NEGATIVE_INFINITY                  = 18;
  protected final static int FROM_NAN_TO_NEGATIVE_INFINITY      = 20;
  protected final static int FROM_POSITIVE_INFINITY_TO_NAN      = 33;
  protected final static int FROM_NEGATIVE_INFINITY_TO_NAN      = 34;
  protected final static int NAN                                = 36;
  protected final static int TO_EXTENSION                       = 56;
  
  /*
   * Iterator can operate in jumping mode and forward mode.
   * In jumping mode (JUMPING bit is set) it provides values
   * in neighbourhood of base points by jumping around a base point.
   * In forward mode (JUMPING bit is removed) it provides consequent
   * double values in start-end direction.  
   */
  protected final static int JUMPING                            = 64;
  
  public SegmentDoubleIterator(double start, double end, int offset, int divisor)
  {
    if(Double.isNaN(start))
    {
      start = -Double.MAX_VALUE;
      exAttribute |= FROM_NAN;
    }
    else if(start == Double.NEGATIVE_INFINITY)
    {
      start = -Double.MAX_VALUE;
      exAttribute |= FROM_NEGATIVE_INFINITY;
    }
    else if(start == Double.POSITIVE_INFINITY)
    {
      start = Double.MAX_VALUE;
      exAttribute |= FROM_POSITIVE_INFINITY;
    }
    
    if(Double.isNaN(end))
    {
      end = Double.MAX_VALUE;
      exAttribute |= TO_NAN;
    }
    else if(end == Double.NEGATIVE_INFINITY)
    {
      end = -Double.MAX_VALUE;
      exAttribute |= TO_NEGATIVE_INFINITY;
    }
    else if(end == Double.POSITIVE_INFINITY)
    {
      end = Double.MAX_VALUE;
      exAttribute |= TO_POSITIVE_INFINITY;
    }
    
    if(divisor <= 0) divisor = 1;
    if(offset < 0) offset = 0;
    if(offset > Integer.MAX_VALUE/2) offset = Integer.MAX_VALUE/2;
    
    if((2*((long)offset)+1)*divisor > Integer.MAX_VALUE)
    {
      if(divisor < Short.MAX_VALUE/2)
        offset = (int)(offset/((2*((double)offset)+1)*divisor/Integer.MAX_VALUE));
      else if(offset < Short.MAX_VALUE/2)
        divisor = (int)(divisor/((2*((double)offset)+1)*divisor/Integer.MAX_VALUE));
      else
      {
        double x = (divisor + Math.sqrt(((double)divisor)*divisor + 8*((double)Integer.MAX_VALUE)*offset*divisor))/2/Integer.MAX_VALUE;
        offset = (int)(offset/x);
        divisor = (int)(divisor/x);
      }
    }

    if(   !(end == start) 
       &&
          (   exAttribute == FROM_POSITIVE_INFINITY
           || exAttribute == TO_POSITIVE_INFINITY
           || exAttribute == FROM_NEGATIVE_INFINITY
           || exAttribute == TO_NEGATIVE_INFINITY
          )
       && DoubleIterator.numberOfDoubles(end, start) >=0
       && DoubleIterator.numberOfDoubles(end, start) <= Integer.MAX_VALUE
       && (2*(long)offset+1)*divisor >= DoubleIterator.numberOfDoubles(end, start)+1
      )
    {
      //System.out.println("        ::cons:: " + "one-point exception");
      //System.out.println("        ::cons:: " + "number of points is great");
      //System.out.println("        ::cons:: " + "number = " + DoubleIterator.numberOfDoubles(end, start));
      //System.out.println("        ::cons:: " + "offset = " + offset);
      //System.out.println("        ::cons:: " + "divisor = " + divisor);

      divisor = 0;
      offset = 0;
    }
    else if(   !(end == start) 
            &&
               (   exAttribute == FROM_NAN
                || exAttribute == TO_NAN
                || exAttribute == FROM_NEGATIVE_TO_POSITIVE_INFINITY
                || exAttribute == FROM_POSITIVE_TO_NEGATIVE_INFINITY
               )
            && DoubleIterator.numberOfDoubles(end, start) >=0
            && DoubleIterator.numberOfDoubles(end, start) <= Integer.MAX_VALUE
            && (2*(long)offset+1)*divisor >= DoubleIterator.numberOfDoubles(end, start)+2
           )
    {
      //System.out.println("        ::cons:: " + "two-point extension");
      //System.out.println("        ::cons:: " + "number of points is great");
      //System.out.println("        ::cons:: " + "number = " + DoubleIterator.numberOfDoubles(end, start));
      //System.out.println("        ::cons:: " + "offset = " + offset);
      //System.out.println("        ::cons:: " + "divisor = " + divisor);

      divisor = 0;
      offset = 0;
    }
    else if(   exAttribute == NORMAL
            && (   !(end == start)
                && DoubleIterator.numberOfDoubles(end, start) >=0
                && DoubleIterator.numberOfDoubles(end, start) <= Integer.MAX_VALUE
                && (2*(long)offset+1)*divisor >= DoubleIterator.numberOfDoubles(end, start)
               
            || end == start )
           )
    {
      //System.out.println("        ::cons:: " + "no extension");
      //System.out.println("        ::cons:: " + "number of points is great");
      //System.out.println("        ::cons:: " + "number = " + DoubleIterator.numberOfDoubles(end, start));
      //System.out.println("        ::cons:: " + "offset = " + offset);
      //System.out.println("        ::cons:: " + "divisor = " + divisor);
      
      divisor = 0;
      offset = 0;
    }
    else if(end == start || exAttribute == NAN)
    {
      switch(exAttribute)
      {
        case FROM_NAN:
        case TO_NAN:
          if((2*offset+1)*divisor >= 2) { divisor = 0; offset = 0; }
          break;
        case NAN:
          start = end;
        case NEGATIVE_INFINITY:
        case POSITIVE_INFINITY:
          divisor = 0;
          offset  = 0;
          break;
        default:
          divisor = 1;
          offset  = 0;
          break;
      }
    }
    
    this.start   = start;
    this.end     = end;
    this.offset  = offset;
    this.divisor = divisor;
    
    if((exAttribute & FROM_NAN) != 0)
      subindex = 0xfff0000000000001L;
    else if((exAttribute & FROM_NEGATIVE_INFINITY) != 0)
      subindex = 0xfff0000000000000L;
    else if((exAttribute & FROM_POSITIVE_INFINITY) != 0)
      subindex = 0x7ff0000000000000L;
    else
      subindex = Double.doubleToLongBits(start);
    
    //System.out.println("        ::cons:: " + "start   = "  + start);
    //System.out.println("        ::cons:: " + "end     = "  + end);
    //System.out.println("        ::cons:: " + "offset  = "  + offset);
    //System.out.println("        ::cons:: " + "divisor = "  + divisor);
  }

  public SegmentDoubleIterator(double start, double end, int offset)
  {
    this(start, end, offset, 1);
  }
  
  public SegmentDoubleIterator(double start, double end)
  {
    this(start, end, 0, 1);
  }
  
  public SegmentDoubleIterator(SegmentDoubleIterator r)
  {
    this.start       = r.start;
    this.end         = r.end;
    this.offset      = r.offset;
    this.divisor     = r.divisor;
    this.index       = r.index;
    this.subindex    = r.subindex;
    this.exAttribute = r.exAttribute;
  }

  public Object clone()
  {
    return new SegmentDoubleIterator(this);
  }
  
  public boolean equals(Object o)
  {
    if(!(o instanceof SegmentDoubleIterator))
      return false;

    SegmentDoubleIterator obj = (SegmentDoubleIterator)o;

    return    index       == obj.index 
           && subindex    == obj.subindex 
           && ( start == obj.start || Double.isNaN(start) && Double.isNaN(obj.start) )
           && ( end   == obj.end   || Double.isNaN(end)   && Double.isNaN(obj.end) )
           && offset      == obj.offset
           && divisor     == obj.divisor
           && exAttribute == obj.exAttribute;
  }
  
  public double value()
  {
    if((exAttribute & JUMPING) == 0)
    {
      return Double.longBitsToDouble(subindex);
    }
    else
    {
      if(Math.abs(end/2-start/2) < Double.MAX_VALUE/divisor/2)
        return DoubleIterator.shiftDouble((start + (end-start)*index/divisor), (subindex%2 == 0)?(subindex/2):(-(subindex+1)/2));
      else if(index < ((double)divisor)/2 || index == ((double)divisor)/2 && (start != -end))
        return DoubleIterator.shiftDouble((start + (end/2-start/2)/divisor*index*2), (subindex%2 == 0)?(subindex/2):(-(subindex+1)/2));
      else if(index > ((double)divisor)/2)
        return DoubleIterator.shiftDouble((end - (end/2-start/2)/divisor*(divisor - index)*2), (subindex%2 == 0)?(subindex/2):(-(subindex+1)/2));
      else
        return DoubleIterator.shiftDouble(0.0, (subindex%2 == 0)?(subindex/2):(-(subindex+1)/2));
    }
  }
  
  public boolean stop()
  {
/*    if(divisor == 0)
      return checkLimitWithZeroDivisor();
    else*/
      return index > divisor || index < 0;
  }
  
  private int checkCurrentRightLimit()
  {
    long rl;
    switch(exAttribute)
    {
      case NORMAL:
      {
        if(start == end) return -1;
        
        long iend = Double.doubleToLongBits(end);
        if(subindex == iend) return -1;
        
        if(divisor != 0)
        {
          if(Math.abs(end/2-start/2) < Double.MAX_VALUE/divisor/2)
            rl = shiftLongAsDouble(
                    Double.doubleToLongBits((start + (end-start)*index/divisor))
                  , offset*((start <= end)?(1):(-1))
                 );
          else if(index < ((double)divisor)/2 || index == ((double)divisor)/2 && (start != -end))
            rl = shiftLongAsDouble(
                    Double.doubleToLongBits((start + (end/2-start/2)/divisor*index*2))
                  , offset*((start <= end)?(1):(-1))
                 );
          else if(index > ((double)divisor)/2)
            rl = shiftLongAsDouble(
                    Double.doubleToLongBits((end - (end/2-start/2)/divisor*(divisor-index)*2))
                  , offset*((start <= end)?(1):(-1))
                 );
          else
            rl = shiftLongAsDouble(
                    0L
                  , offset*((start <= end)?(1):(-1))
                 );
          
          if(subindex == rl) return 1;
        }
        
        return 0;
      }
      case NAN: case POSITIVE_INFINITY: case NEGATIVE_INFINITY:
        return -1;
      default:
      {
        if(index == divisor)
        {
          if(   (exAttribute & TO_POSITIVE_INFINITY) != 0 && subindex == 0x7ff0000000000000L
             || (exAttribute & TO_NEGATIVE_INFINITY) != 0 && subindex == 0xfff0000000000000L
             || ((exAttribute & TO_NAN) != 0 && (subindex > 0x7ff0000000000000L || subindex < 0 && subindex > 0xfff0000000000000L))
             || (exAttribute & TO_EXTENSION) == 0 && subindex == Double.doubleToLongBits(end)
            )
            return -1;
          else
            return 0;
        }
        else if(index == 0)
        {
          if((exAttribute & FROM_POSITIVE_INFINITY) != 0)
          {
            rl = shiftLongAsDouble( 0x7ff0000000000000L, -offset );
          }
          else if((exAttribute & FROM_NEGATIVE_INFINITY) != 0)
          {
            rl = shiftLongAsDouble( 0xfff0000000000000L, offset );
          }
          else if((exAttribute & FROM_NAN) != 0)
          {
            rl = shiftLongAsDouble( 0xfff0000000000001L, offset );
          }
          else
            rl = shiftLongAsDouble(
                   Double.doubleToLongBits(start), offset*((start <= end)?(1):(-1))
                 );
          
          if(subindex == rl) return 1;
          else               return 0;
        }
        else
        {
          if((exAttribute & TO_EXTENSION) == 0 && subindex == Double.doubleToLongBits(end))
          {
            index++;
            return -1;
          }
          
          if(Math.abs(end/2-start/2) < Double.MAX_VALUE/divisor/2)
            rl = shiftLongAsDouble(
                    Double.doubleToLongBits((start + (end-start)*index/divisor))
                  , offset*((start <= end)?(1):(-1))
                 );
          else if(index < ((double)divisor)/2 || index == ((double)divisor)/2 && (start != -end))
            rl = shiftLongAsDouble(
                    Double.doubleToLongBits((start + (end/2-start/2)/divisor*index*2))
                  , offset*((start <= end)?(1):(-1))
                 );
          else if(index > ((double)divisor)/2)
            rl = shiftLongAsDouble(
                    Double.doubleToLongBits((end - (end/2-start/2)/divisor*(divisor-index)*2))
                  , offset*((start <= end)?(1):(-1))
                 );
          else
            rl = shiftLongAsDouble(
                    0L
                  , offset*((start <= end)?(1):(-1))
                 );
      
          if(subindex == rl) return 1;
          else               return 0;
        }
      }
    }
  }
  
  protected static boolean geLongsAsDoubles(long x, long y)
  {
    if(x >= 0 && y >= 0)     return (x >= y);
    else if(x < 0 && y < 0)  return (x <= y);
    else if(x < 0 && y >= 0) return false;
    else                     return true; // x >= 0 && y <= 0
  }
  
  private boolean checkLeftLimit()
  {
    int shift = 1;
    if(index == divisor)
    {
      if(   (exAttribute & TO_POSITIVE_INFINITY) != 0
         || (exAttribute & TO_NEGATIVE_INFINITY) != 0
        )
        shift = 0;
      else if((exAttribute & TO_NAN) != 0)
        shift = -1;
    }
    
    long ll;
    if(Math.abs(end/2-start/2) < Double.MAX_VALUE/divisor/2)
      ll = shiftLongAsDouble(
              Double.doubleToLongBits((start + (end-start)*index/divisor))
            , (offset + shift)*((start <= end)?(-1):(1))
           );
    else if(index < ((double)divisor)/2 || index == ((double)divisor)/2 && (start != -end))
      ll = shiftLongAsDouble(
              Double.doubleToLongBits((start + (end/2-start/2)/divisor*index*2))
            , (offset + shift)*((start <= end)?(-1):(1))
           );
    else if(index > ((double)divisor)/2)
      ll = shiftLongAsDouble(
              Double.doubleToLongBits((end - (end/2-start/2)/divisor*(divisor - index)*2))
            , (offset + shift)*((start <= end)?(-1):(1))
           );
    else
      ll = shiftLongAsDouble(
              0L
            , (offset + shift)*((start <= end)?(-1):(1))
           );
      
    return (start <= end)?geLongsAsDoubles(subindex, ll):geLongsAsDoubles(ll, subindex);
  }
  
  private boolean checkNextIntersection()
  {
    //if(index == divisor) return true;
    int shift = 1;
    if(index + 1 == divisor)
    {
      if(   (exAttribute & TO_POSITIVE_INFINITY) != 0
         || (exAttribute & TO_NEGATIVE_INFINITY) != 0
        )
        shift = 0;
      else if((exAttribute & TO_NAN) != 0)
        shift = -1;
    }
    
    long rl, ll;
    if(Math.abs(end/2-start/2) < Double.MAX_VALUE/divisor/2)
    {
      rl = shiftLongAsDouble(
              Double.doubleToLongBits((start + (end-start)*index/divisor))
            , offset*((start <= end)?(1):(-1))
           );
      ll = shiftLongAsDouble(
              Double.doubleToLongBits((start + (end-start)*(index+1)/divisor))
            , (offset + shift)*((start <= end)?(-1):(1))
           );
    }
    else if(index < ((double)divisor)/2 || index == ((double)divisor)/2 && (start != -end))
    {
      rl = shiftLongAsDouble(
              Double.doubleToLongBits((start + (end/2-start/2)/divisor*index*2))
            , offset*((start <= end)?(1):(-1))
           );
      if(index+1 < ((double)divisor)/2 || index+1 == ((double)divisor)/2 && (start != -end))
        ll = shiftLongAsDouble(
                Double.doubleToLongBits((start + (end/2-start/2)/divisor*(index+1)*2))
              , (offset + shift)*((start <= end)?(-1):(1))
             );
      else if(index+1 > ((double)divisor)/2)
        ll = shiftLongAsDouble(
                Double.doubleToLongBits((end - (end/2-start/2)/divisor*(divisor-index-1)*2))
              , (offset + shift)*((start <= end)?(-1):(1))
             );
      else
        ll = shiftLongAsDouble(
                0L
              , (offset + shift)*((start <= end)?(-1):(1))
             );
    }
    else if(index > ((double)divisor)/2)
    {
      rl = shiftLongAsDouble(
              Double.doubleToLongBits((end - (end/2-start/2)/divisor*(divisor-index)*2))
            , offset*((start <= end)?(1):(-1))
           );
      ll = shiftLongAsDouble(
              Double.doubleToLongBits((end - (end/2-start/2)/divisor*(divisor-index-1)*2))
            , (offset + shift)*((start <= end)?(-1):(1))
           );
    }
    else
    {
      rl = shiftLongAsDouble(
              0L
            , offset*((start <= end)?(1):(-1))
           );
      ll = shiftLongAsDouble(
              Double.doubleToLongBits((end - (end/2-start/2)/divisor*(divisor-index-1)*2))
            , (offset + shift)*((start <= end)?(-1):(1))
           );
    }
    
    return (start <= end)?geLongsAsDoubles(rl, ll):geLongsAsDoubles(ll, rl);
  }
  
  protected static long shiftLongAsDouble(long x, int i)
  {
    if(x < 0)
    {
      if(i > 0 && x - i > 0)
        return 0x7fffffffffffffffL - (x - i);
      else
        return x-i;
    }
    else
    {
      if(i < 0 && x + i < 0)
        return 0x8000000000000000L - (x + i) - 1;
      else
        return x+i;
    }
  }
  
  public void shift()
  {
    if((exAttribute & JUMPING) == 0)
    {
      //System.out.println("        ::next:: " + "no jumping");
      
      switch(checkCurrentRightLimit())
      {
        case -1:
        {
          //System.out.println("        ::next:: " + "complete"); 
          
          index++; 
          return; 
        }
        case 1:
        {
          //System.out.println("        ::next:: " + "partial complete"); 
          
          index++; // check mode, set subindex
          if(checkLeftLimit())
          {
            //System.out.println("        ::next:: " + "overmove"); 
            
            switch(exAttribute)
            {
              case FROM_POSITIVE_INFINITY:
                subindex = shiftLongAsDouble(subindex, -1); return;
              default:
                subindex = shiftLongAsDouble(subindex, (start <= end)?(1):(-1)); return;                    
            }
          }
          else if(index == divisor)
          {
            //System.out.println("        ::next:: " + "no overmove, last subsegment"); 
            
            if((exAttribute & TO_POSITIVE_INFINITY) != 0)
            {
              subindex = Double.doubleToLongBits(
                           DoubleIterator.shiftDouble(
                              end
                            , -offset+1
                           )
                         );
            }
            else if((exAttribute & TO_NEGATIVE_INFINITY) != 0)
            {
              subindex = Double.doubleToLongBits(
                  DoubleIterator.shiftDouble(
                     end
                   , offset-1
                  )
                );
            }
            else if((exAttribute & TO_NAN) != 0)
            {
              subindex = Double.doubleToLongBits(
                  DoubleIterator.shiftDouble(
                     end
                   , -offset+2
                  )
                );
            }
            else
            {
              subindex = Double.doubleToLongBits(
                  DoubleIterator.shiftDouble(
                     end
                   , offset*((start <= end)?(-1):(1))
                  )
                );
            }
            return;
          }
          else if(checkNextIntersection())
          {
            //System.out.println("        ::next:: " + "no overmove, next intersection"); 
            
            if(Math.abs(end/2-start/2) < Double.MAX_VALUE/divisor/2)
              subindex = Double.doubleToLongBits(
                           DoubleIterator.shiftDouble(
                              (start + (end-start)*index/divisor)
                            , offset*((start <= end)?(-1):(1))
                           )
                         );
            else if(index < ((double)divisor)/2 || index == ((double)divisor)/2 && (start != -end))
              subindex = Double.doubleToLongBits(
                            DoubleIterator.shiftDouble(
                               (start + (end/2-start/2)/divisor*index*2)
                             , offset*((start <= end)?(-1):(1))
                            )
                          );
            else if(index > ((double)divisor)/2)
              subindex = Double.doubleToLongBits(
                            DoubleIterator.shiftDouble(
                               (end - (end/2-start/2)/divisor*(divisor-index)*2)
                             , offset*((start <= end)?(-1):(1))
                            )
                          );
            else
              subindex = Double.doubleToLongBits(
                            DoubleIterator.shiftDouble(
                               0.0
                             , offset*((start <= end)?(-1):(1))
                            )
                          );
              
            return;
          }
          else
          {
            //System.out.println("        ::next:: " + "no overmove, turn to jumping"); 
            
            exAttribute |= JUMPING;
            subindex = 0;
            return;
          }
        }
        default:
        {
          //System.out.println("        ::next:: " + "continue"); 
          
          subindex = shiftLongAsDouble(subindex, (start <= end)?(1):(-1));
          return;
        }
      }
    }
    else
    {
      //System.out.println("        ::next:: " + "jumping"); 
      
      subindex++;
  
      if(subindex > 2*offset)
      {
        index++;
        if(index == divisor)
        {
          //System.out.println("        ::next:: " + "jumping, last subsegment"); 
          
          exAttribute &= ~JUMPING;
          if((exAttribute & TO_POSITIVE_INFINITY) != 0)
          {
            subindex = Double.doubleToLongBits(
                         DoubleIterator.shiftDouble(
                            end
                          , -offset+1
                         )
                       );
          }
          else if((exAttribute & TO_NEGATIVE_INFINITY) != 0)
          {
            subindex = Double.doubleToLongBits(
                DoubleIterator.shiftDouble(
                   end
                 , offset-1
                )
              );
          }
          else if((exAttribute & TO_NAN) != 0)
          {
            subindex = Double.doubleToLongBits(
                DoubleIterator.shiftDouble(
                   end
                 , -offset+2
                )
              );
          }
          else
          {
            subindex = Double.doubleToLongBits(
                DoubleIterator.shiftDouble(
                   end
                 , offset*((start <= end)?(-1):(1))
                )
              );
          }
          return;
        }
        else if(checkNextIntersection())
        {
          //System.out.println("        ::next:: " + "jumping, turn to forward"); 
          
          exAttribute &= ~JUMPING;
          if(Math.abs(end/2-start/2) < Double.MAX_VALUE/divisor/2)
            subindex = Double.doubleToLongBits(
                         DoubleIterator.shiftDouble(
                            (start + (end-start)*index/divisor)
                          , offset*((start <= end)?(-1):(1))
                         )
                       );
          else if(index < ((double)divisor)/2 || index == ((double)divisor)/2 && (start != -end))
            subindex = Double.doubleToLongBits(
                          DoubleIterator.shiftDouble(
                             (start + (end/2-start/2)/divisor*index*2)
                           , offset*((start <= end)?(-1):(1))
                          )
                        );
          else if(index > ((double)divisor)/2)
            subindex = Double.doubleToLongBits(
                          DoubleIterator.shiftDouble(
                             (end - (end/2-start/2)/divisor*(divisor-index)*2)
                           , offset*((start <= end)?(-1):(1))
                          )
                        );
          else
            subindex = Double.doubleToLongBits(
                          DoubleIterator.shiftDouble(
                             0.0
                           , offset*((start <= end)?(-1):(1))
                          )
                        );
            
          return;
        }
        else
        {
          //System.out.println("        ::next:: " + "jumping, keeping"); 
          
          subindex = 0;
        }
      }
    }
  }

  public int hashCode ()
  {
    return index + 3*(int)subindex + 5*(int)Double.doubleToLongBits(start) + 7*(int)Double.doubleToLongBits(end) + 11*offset + 13*divisor;
  }
}
