/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class SegmentFloatIterator
  extends AbstractFloatIterator
  implements IFloatIterator
{
  protected float start  = 0;
  protected float end    = 0;
  protected int offset   = 0;
  protected int divisor  = 1;
  
  /**
   * index field provides enumeration of iteration base points - a set of
   * points partitioning the segment [start, end] into divisor parts
   */
  protected int index    = 0;
  
  /**
   * subindex field in different opeartion modes means different things.
   * In jumping mode it provides enumeration of points in a neighbourhood of
   * one base point.
   * In forward mode it stores the current value of iteration - 
   * beware that this value is float and should be interpreted adequeately!  
   */
  protected int subindex = 0; 
  
  /**
   * auxiliary field to store flags designating type of segment ends and
   * modes of operationInfo
   */
  protected int exAttribute = NORMAL;
  
  protected final static int NORMAL                             = 0;
  protected final static int FROM_POSITIVE_INFINITY             = 1;
  protected final static int FROM_NEGATIVE_INFINITY             = 2;
  protected final static int FROM_NAN                           = 4;
  protected final static int TO_POSITIVE_INFINITY               = 8;
  protected final static int TO_NEGATIVE_INFINITY               = 16;
  protected final static int TO_NAN                             = 32;
  protected final static int POSITIVE_INFINITY                  = 9;
  protected final static int FROM_NEGATIVE_TO_POSITIVE_INFINITY = 10;
  protected final static int FROM_NAN_TO_POSITIVE_INFINITY      = 12;
  protected final static int FROM_POSITIVE_TO_NEGATIVE_INFINITY = 17;
  protected final static int NEGATIVE_INFINITY                  = 18;
  protected final static int FROM_NAN_TO_NEGATIVE_INFINITY      = 20;
  protected final static int FROM_POSITIVE_INFINITY_TO_NAN      = 33;
  protected final static int FROM_NEGATIVE_INFINITY_TO_NAN      = 34;
  protected final static int NAN                                = 36;
  protected final static int TO_EXTENSION                       = 56;
  
  /*
   * Iterator can operate in jumping mode and forward mode.
   * In jumping mode (JUMPING bit is set) it provides values
   * in neighbourhood of base points by jumping around a base point.
   * In forward mode (JUMPING bit is removed) it provides consequent
   * float values in start-end direction.  
   */
  protected final static int JUMPING                            = 64;
  
  public SegmentFloatIterator(float start, float end, int offset, int divisor)
  {
    if(Float.isNaN(start))
    {
      start = -Float.MAX_VALUE;
      exAttribute |= FROM_NAN;
    }
    else if(start == Float.NEGATIVE_INFINITY)
    {
      start = -Float.MAX_VALUE;
      exAttribute |= FROM_NEGATIVE_INFINITY;
    }
    else if(start == Float.POSITIVE_INFINITY)
    {
      start = Float.MAX_VALUE;
      exAttribute |= FROM_POSITIVE_INFINITY;
    }
    
    if(Float.isNaN(end))
    {
      end = Float.MAX_VALUE;
      exAttribute |= TO_NAN;
    }
    else if(end == Float.NEGATIVE_INFINITY)
    {
      end = -Float.MAX_VALUE;
      exAttribute |= TO_NEGATIVE_INFINITY;
    }
    else if(end == Float.POSITIVE_INFINITY)
    {
      end = Float.MAX_VALUE;
      exAttribute |= TO_POSITIVE_INFINITY;
    }
    
    if(divisor <= 0) divisor = 1;
    if(offset < 0) offset = 0;
    if(offset > Integer.MAX_VALUE/2) offset = Integer.MAX_VALUE/2;
    
    if((2*((long)offset)+1)*divisor > Integer.MAX_VALUE)
    {
      if(divisor < Short.MAX_VALUE/2)
        offset = (int)(offset/((2*((double)offset)+1)*divisor/Integer.MAX_VALUE));
      else if(offset < Short.MAX_VALUE/2)
        divisor = (int)(divisor/((2*((double)offset)+1)*divisor/Integer.MAX_VALUE));
      else
      {
        double x = (divisor + Math.sqrt(((double)divisor)*divisor + 8*((double)Integer.MAX_VALUE)*offset*divisor))/2/Integer.MAX_VALUE;
        offset = (int)(offset/x);
        divisor = (int)(divisor/x);
      }
    }

    if(   !(end == start) 
       &&
          (   exAttribute == FROM_POSITIVE_INFINITY
           || exAttribute == TO_POSITIVE_INFINITY
           || exAttribute == FROM_NEGATIVE_INFINITY
           || exAttribute == TO_NEGATIVE_INFINITY
          )
       && (2*(long)offset+1)*divisor >= FloatIterator.numberOfFloats(end, start)+1
      )
    {
      //System.out.println("        ::cons:: " + "one-point exception");
      //System.out.println("        ::cons:: " + "number of points is great");
      //System.out.println("        ::cons:: " + "number = " + FloatIterator.numberOfFloats(end, start));
      //System.out.println("        ::cons:: " + "offset = " + offset);
      //System.out.println("        ::cons:: " + "divisor = " + divisor);

      divisor = 0;
      offset = 0;
    }
    else if(   !(end == start) 
            &&
               (   exAttribute == FROM_NAN
                || exAttribute == TO_NAN
                || exAttribute == FROM_NEGATIVE_TO_POSITIVE_INFINITY
                || exAttribute == FROM_POSITIVE_TO_NEGATIVE_INFINITY
               )
            && (2*(long)offset+1)*divisor >= FloatIterator.numberOfFloats(end, start)+2
           )
    {
      //System.out.println("        ::cons:: " + "two-point extension");
      //System.out.println("        ::cons:: " + "number of points is great");
      //System.out.println("        ::cons:: " + "number = " + FloatIterator.numberOfFloats(end, start));
      //System.out.println("        ::cons:: " + "offset = " + offset);
      //System.out.println("        ::cons:: " + "divisor = " + divisor);

      divisor = 0;
      offset = 0;
    }
    else if(   exAttribute == NORMAL
            && (   !(end == start)
  //            && FloatIterator.numberOfFloats(end, start) <= Integer.MAX_VALUE
                && (2*(long)offset+1)*divisor >= FloatIterator.numberOfFloats(end, start)
               
            || end == start )
           )
    {
      //System.out.println("        ::cons:: " + "no extension");
      //System.out.println("        ::cons:: " + "number of points is great");
      //System.out.println("        ::cons:: " + "number = " + FloatIterator.numberOfFloats(end, start));
      //System.out.println("        ::cons:: " + "offset = " + offset);
      //System.out.println("        ::cons:: " + "divisor = " + divisor);
      
      divisor = 0;
      offset = 0;
    }
    else if(end == start || exAttribute == NAN)
    {
      switch(exAttribute)
      {
        case FROM_NAN:
        case TO_NAN:
          if((2*offset+1)*divisor >= 2) { divisor = 0; offset = 0; }
          break;
        case NAN:
          start = end;
        case NEGATIVE_INFINITY:
        case POSITIVE_INFINITY:
          divisor = 0;
          offset  = 0;
          break;
        default:
          divisor = 1;
          offset  = 0;
          break;
      }
    }
    
    this.start   = start;
    this.end     = end;
    this.offset  = offset;
    this.divisor = divisor;
    
    if((exAttribute & FROM_NAN) != 0)
      subindex = 0xff800001;
    else if((exAttribute & FROM_NEGATIVE_INFINITY) != 0)
      subindex = 0xff800000;
    else if((exAttribute & FROM_POSITIVE_INFINITY) != 0)
      subindex = 0x7f800000;
    else
      subindex = Float.floatToIntBits(start);
    
    //System.out.println("        ::cons:: " + "start   = "  + start);
    //System.out.println("        ::cons:: " + "end     = "  + end);
    //System.out.println("        ::cons:: " + "offset  = "  + offset);
    //System.out.println("        ::cons:: " + "divisor = "  + divisor);
  }

  public SegmentFloatIterator(float start, float end, int offset)
  {
    this(start, end, offset, 1);
  }
  
  public SegmentFloatIterator(float start, float end)
  {
    this(start, end, 0, 1);
  }
  
  public SegmentFloatIterator(SegmentFloatIterator r)
  {
    this.start       = r.start;
    this.end         = r.end;
    this.offset      = r.offset;
    this.divisor     = r.divisor;
    this.index       = r.index;
    this.subindex    = r.subindex;
    this.exAttribute = r.exAttribute;
  }

  public Object clone()
  {
    return new SegmentFloatIterator(this);
  }
  
  public boolean equals(Object o)
  {
    if(!(o instanceof SegmentFloatIterator))
      return false;

    SegmentFloatIterator obj = (SegmentFloatIterator)o;

    return    index       == obj.index 
           && subindex    == obj.subindex 
           && ( start == obj.start || Float.isNaN(start) && Float.isNaN(obj.start) )
           && ( end   == obj.end   || Float.isNaN(end)   && Float.isNaN(obj.end) )
           && offset      == obj.offset
           && divisor     == obj.divisor
           && exAttribute == obj.exAttribute;
  }
  
  public float value()
  {
    if((exAttribute & JUMPING) == 0)
    {
      return Float.intBitsToFloat(subindex);
    }
    else
    {
      return FloatIterator.shiftFloat((float)(start + ((double)end-start)*index/divisor), (subindex%2 == 0)?(subindex/2):(-(subindex+1)/2));
    }
  }
  
  public boolean stop()
  {
/*    if(divisor == 0)
      return checkLimitWithZeroDivisor();
    else*/
      return index > divisor || index < 0;
  }
  
  private int checkCurrentRightLimit()
  {
    int rl;
    switch(exAttribute)
    {
      case NORMAL:
      {
        if(start == end) return -1;
        
        int iend = Float.floatToIntBits(end);
        if(subindex == iend) return -1;
        
        if(divisor != 0)
        {
          rl = shiftIntAsFloat(
                  Float.floatToIntBits((float)(start + ((double)end-start)*index/divisor))
                , offset*((start <= end)?(1):(-1))
               ); 
          
          if(subindex == rl) return 1;
        }
        
        return 0;
      }
      case NAN: case POSITIVE_INFINITY: case NEGATIVE_INFINITY:
        return -1;
      default:
      {
        if(index == divisor)
        {
          if(   (exAttribute & TO_POSITIVE_INFINITY) != 0 && subindex == 0x7f800000
             || (exAttribute & TO_NEGATIVE_INFINITY) != 0 && subindex == 0xff800000
             || ((exAttribute & TO_NAN) != 0 && (subindex > 0x7f800000 || subindex < 0 && subindex > 0xff800000))
             || (exAttribute & TO_EXTENSION) == 0 && subindex == Float.floatToIntBits(end)
            )
            return -1;
          else
            return 0;
        }
        else if(index == 0)
        {
          if((exAttribute & FROM_POSITIVE_INFINITY) != 0)
          {
            rl = shiftIntAsFloat( 0x7f800000, -offset );
          }
          else if((exAttribute & FROM_NEGATIVE_INFINITY) != 0)
          {
            rl = shiftIntAsFloat( 0xff800000, offset );
          }
          else if((exAttribute & FROM_NAN) != 0)
          {
            rl = shiftIntAsFloat( 0xff800001, offset );
          }
          else
            rl = shiftIntAsFloat(
                   Float.floatToIntBits((float)start), offset*((start <= end)?(1):(-1))
                 );
          
          if(subindex == rl) return 1;
          else               return 0;
        }
        else
        {
          if((exAttribute & TO_EXTENSION) == 0 && subindex == Float.floatToIntBits(end))
          {
            index++;
            return -1;
          }
          
          rl = shiftIntAsFloat(
                  Float.floatToIntBits((float)(start + ((double)end-start)*index/divisor))
                , offset*((start <= end)?(1):(-1))
               ); 
      
          if(subindex == rl) return 1;
          else               return 0;
        }
      }
    }
  }
  
  protected static boolean geIntsAsFloats(int x, int y)
  {
    if(x >= 0 && y >= 0)     return (x >= y);
    else if(x < 0 && y < 0)  return (x <= y);
    else if(x < 0 && y >= 0) return false;
    else                     return true; // x >= 0 && y <= 0
  }
  
  private boolean checkLeftLimit()
  {
    int shift = 1;
    if(index == divisor)
    {
      if(   (exAttribute & TO_POSITIVE_INFINITY) != 0
         || (exAttribute & TO_NEGATIVE_INFINITY) != 0
        )
        shift = 0;
      else if((exAttribute & TO_NAN) != 0)
        shift = -1;
    }
    
    int ll  = shiftIntAsFloat(
                 Float.floatToIntBits((float)(start + ((double)end-start)*index/divisor))
               , (offset + shift)*((start <= end)?(-1):(1))
              );
    return (start <= end)?geIntsAsFloats(subindex, ll):geIntsAsFloats(ll, subindex);
  }
  
  private boolean checkNextIntersection()
  {
    //if(index == divisor) return true;
    int shift = 1;
    if(index + 1 == divisor)
    {
      if(   (exAttribute & TO_POSITIVE_INFINITY) != 0
         || (exAttribute & TO_NEGATIVE_INFINITY) != 0
        )
        shift = 0;
      else if((exAttribute & TO_NAN) != 0)
        shift = -1;
    }
    
    int rl = shiftIntAsFloat(
                Float.floatToIntBits((float)(start + ((double)end-start)*index/divisor))
              , offset*((start <= end)?(1):(-1))
             )
      , ll = shiftIntAsFloat(
                Float.floatToIntBits((float)(start + ((double)end-start)*(index+1)/divisor))
              , (offset + shift)*((start <= end)?(-1):(1))
             );
    return (start <= end)?geIntsAsFloats(rl, ll):geIntsAsFloats(ll, rl);
  }
  
  protected static int shiftIntAsFloat(int x, int i)
  {
    if(x < 0)
    {
      if(i > 0 && x - i > 0)
        return 0x7fffffff - (x - i);
      else
        return x-i;
    }
    else
    {
      if(i < 0 && x + i < 0)
        return 0x80000000 - (x + i) - 1;
      else
        return x+i;
    }
  }
  
  public void shift()
  {
    if((exAttribute & JUMPING) == 0)
    {
      //System.out.println("        ::next:: " + "no jumping");
      
      switch(checkCurrentRightLimit())
      {
        case -1:
        {
          //System.out.println("        ::next:: " + "complete"); 
          
          index++; 
          return; 
        }
        case 1:
        {
          //System.out.println("        ::next:: " + "partial complete"); 
          
          index++; // check mode, set subindex
          if(checkLeftLimit())
          {
            //System.out.println("        ::next:: " + "overmove"); 
            
            switch(exAttribute)
            {
              case FROM_POSITIVE_INFINITY:
                subindex = shiftIntAsFloat(subindex, -1); return;
              default:
                subindex = shiftIntAsFloat(subindex, (start <= end)?(1):(-1)); return;                    
            }
          }
          else if(index == divisor)
          {
            //System.out.println("        ::next:: " + "no overmove, last subsegment"); 
            
            if((exAttribute & TO_POSITIVE_INFINITY) != 0)
            {
              subindex = Float.floatToIntBits(
                           FloatIterator.shiftFloat(
                              end
                            , -offset+1
                           )
                         );
            }
            else if((exAttribute & TO_NEGATIVE_INFINITY) != 0)
            {
              subindex = Float.floatToIntBits(
                  FloatIterator.shiftFloat(
                     end
                   , offset-1
                  )
                );
            }
            else if((exAttribute & TO_NAN) != 0)
            {
              subindex = Float.floatToIntBits(
                  FloatIterator.shiftFloat(
                     end
                   , -offset+2
                  )
                );
            }
            else
            {
              subindex = Float.floatToIntBits(
                  FloatIterator.shiftFloat(
                     end
                   , offset*((start <= end)?(-1):(1))
                  )
                );
            }
            return;
          }
          else if(checkNextIntersection())
          {
            //System.out.println("        ::next:: " + "no overmove, next intersection"); 
            
            subindex = Float.floatToIntBits(
                         FloatIterator.shiftFloat(
                            (float)(start + ((double)end-start)*index/divisor)
                          , offset*((start <= end)?(-1):(1))
                         )
                       );
            return;
          }
          else
          {
            //System.out.println("        ::next:: " + "no overmove, turn to jumping"); 
            
            exAttribute |= JUMPING;
            subindex = 0;
            return;
          }
        }
        default:
        {
          //System.out.println("        ::next:: " + "continue"); 
          
          subindex = shiftIntAsFloat(subindex, (start <= end)?(1):(-1));
          return;
        }
      }
    }
    else
    {
      //System.out.println("        ::next:: " + "jumping"); 
      
      subindex++;
  
      if(subindex > 2*offset)
      {
        index++;
        if(index == divisor)
        {
          //System.out.println("        ::next:: " + "jumping, last subsegment"); 
          
          exAttribute &= ~JUMPING;
          if((exAttribute & TO_POSITIVE_INFINITY) != 0)
          {
            subindex = Float.floatToIntBits(
                         FloatIterator.shiftFloat(
                            end
                          , -offset+1
                         )
                       );
          }
          else if((exAttribute & TO_NEGATIVE_INFINITY) != 0)
          {
            subindex = Float.floatToIntBits(
                FloatIterator.shiftFloat(
                   end
                 , offset-1
                )
              );
          }
          else if((exAttribute & TO_NAN) != 0)
          {
            subindex = Float.floatToIntBits(
                FloatIterator.shiftFloat(
                   end
                 , -offset+2
                )
              );
          }
          else
          {
            subindex = Float.floatToIntBits(
                FloatIterator.shiftFloat(
                   end
                 , offset*((start <= end)?(-1):(1))
                )
              );
          }
          return;
        }
        else if(checkNextIntersection())
        {
          //System.out.println("        ::next:: " + "jumping, turn to forward"); 
          
          exAttribute &= ~JUMPING;
          subindex = Float.floatToIntBits(
                       FloatIterator.shiftFloat(
                          (float)(start + ((double)end-start)*index/divisor)
                        , offset*((start <= end)?(-1):(1))
                       )
                     );
          return;
        }
        else
        {
          //System.out.println("        ::next:: " + "jumping, keeping"); 
          
          subindex = 0;
        }
      }
    }
  }

  public int hashCode ()
  {
    return index + 3*subindex + 5*Float.floatToIntBits(start) + 7*Float.floatToIntBits(end) + 11*offset + 13*divisor;
  }
}
