/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class SegmentLongIterator
  extends AbstractLongIterator
  implements ILongIterator
{
  protected long start = 0;
  protected long end = 0;
  protected int offset = 0;
  protected int divisor = 1;
  protected int index = 0;
  protected int subindex = 0;
  
  public SegmentLongIterator(long start, long end, int offset, int divisor)
  {
    if(divisor <= 0) divisor = 1;
    if(offset < 0) offset = 0;
    if(offset > Integer.MAX_VALUE/2) offset = Integer.MAX_VALUE/2;

    if((2*((long)offset)+1)*divisor > Integer.MAX_VALUE)
    {
      if(divisor < Short.MAX_VALUE/2)
        offset = (int)(offset/((2*((double)offset)+1)*divisor/Integer.MAX_VALUE));
      else if(offset < Short.MAX_VALUE/2)
        divisor = (int)(divisor/((2*((double)offset)+1)*divisor/Integer.MAX_VALUE));
      else
      {
        double x = (divisor + Math.sqrt(((double)divisor)*divisor + 8*((double)Integer.MAX_VALUE)*offset*divisor))/2/Integer.MAX_VALUE;
        offset = (int)(offset/x);
        divisor = (int)(divisor/x);
      }
    }
    
    if(   (   start - end != Long.MIN_VALUE
           && (   start <= end && start - end <= 0
               || start >= end && start - end >= 0
              )
          )
       && Math.abs(end - start) <= Integer.MAX_VALUE
       && (2*(long)offset+1)*divisor >= Math.abs(end - start)
      )
    {
      divisor = (int)Math.abs(end - start);
      offset = 0;
    }
    
    this.start   = start;
    this.end     = end;
    this.offset  = offset;
    this.divisor = divisor;
  }

  public SegmentLongIterator(long start, long end, int offset)
  {
    this(start, end, offset, 1);
  }
  
  public SegmentLongIterator(long start, long end)
  {
    this(start, end, 0, 1);
  }
  
  public SegmentLongIterator(SegmentLongIterator r)
  {
    this.start   = r.start;
    this.end     = r.end;
    this.offset  = r.offset;
    this.divisor = r.divisor;
    this.index    = r.index;
    this.subindex = r.subindex;
  }

  public Object clone()
  {
    return new SegmentLongIterator(this);
  }
  
  public boolean equals(Object o)
  {
    if(!(o instanceof SegmentLongIterator))
      return false;

    SegmentLongIterator obj = (SegmentLongIterator)o;

    return    index    == obj.index 
           && subindex == obj.subindex 
           && start    == obj.start 
           && end      == obj.end
           && offset   == obj.offset
           && divisor  == obj.divisor;
  }
  
  public long value()
  {
    if(index == 0)
      return (long)(start + subindex*((start <= end)?(1):(-1)));
    else if(index == divisor)
      return (long)(end - subindex*((start <= end)?(1):(-1)));
    else if(    (   start - end != Long.MIN_VALUE
                 && (   start <= end && start - end <= 0
                     || start >= end && start - end >= 0
                    )
                )
             && Math.abs(end - start) < Integer.MAX_VALUE
           )
      return (long)(start + (end-start)*index/divisor + ((subindex%2 == 0)?(subindex/2):(-(subindex+1)/2)));
    else if(    (   start - end != Long.MIN_VALUE
                 && (   start <= end && start - end <= 0
                     || start >= end && start - end >= 0
                    )
                )
           )
      return (long)(start + (end-start)/divisor*index + ((subindex%2 == 0)?(subindex/2):(-(subindex+1)/2)));
    else if(start - end == Long.MIN_VALUE)
      return (long)(start + ((start <= end)?Long.MAX_VALUE:Long.MIN_VALUE)/divisor*index + ((subindex%2 == 0)?(subindex/2):(-(subindex+1)/2)));
    else
      return (long)(start + (end/2-start/2)/divisor*index*2 + index/3 + ((subindex%2 == 0)?(subindex/2):(-(subindex+1)/2)));
  }
  
  public boolean stop()
  {
    return index > divisor || index < 0;
  }
  
  public void shift()
  {
    subindex++;

    if(   index == 0
       || index == divisor
      )
    {
      if(subindex > offset)
      {
        index++;
        subindex = 0;
      }
    }
    else if(subindex > 2*offset)
    {
      index++;
      subindex = 0;
    }
  }

  public int hashCode ()
  {
    return index + 3*subindex + 5*(int)start + 7*(int)end + 11*offset + 13*divisor;
  }
}
