/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

public class ShortIterator
  extends AbstractShortIterator
  implements IShortIterator
{
  public static short[] defaultValues = {
     Short.MIN_VALUE
   , Short.MAX_VALUE
   , Short.MIN_VALUE + 1
   , Short.MAX_VALUE - 1
   , 0
   , 1
   , -1
   , Byte.MIN_VALUE - 1
   , Byte.MAX_VALUE + 1  
   , Short.MIN_VALUE + 8572
   , Short.MAX_VALUE - 43185
   , Short.MIN_VALUE + 52603
  };
  
  protected IShortIterator[] items = new IShortIterator[] { new ArrayShortIterator() };
  protected int index = 0;
  
  public ShortIterator()
  {
  }
  
  public ShortIterator(short[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = new IShortIterator[] { new ArrayShortIterator(values) };
  }
  
  public ShortIterator(short start, short end, int offset, int divisor)
  {
    items = new IShortIterator[] { new SegmentShortIterator(start, end, offset, divisor) };
  }

  public ShortIterator(short point, int offset)
  {
    items = new IShortIterator[] { new NeighbourhoodShortIterator(point, offset) };
  }

  
  public ShortIterator(IShortIterator[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    items = values;
  }
  
  public void addSegment(short start, short end, int offset, int divisor)
  {
    IShortIterator[] newItems = new IShortIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new SegmentShortIterator(start, end, offset, divisor);
    items = newItems;
  }
  
  public void addSegment(short start, short end, int offset)
  {
    addSegment(start, end, offset, 1);
  }
  
  public void addSegment(short start, short end)
  {
    addSegment(start, end, 0, 1);
  }
  
  public void addPoint(short point, int offset)
  {
    IShortIterator[] newItems = new IShortIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new NeighbourhoodShortIterator(point, offset);
    items = newItems;
  }
  
  public void addArray(short[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    IShortIterator[] newItems = new IShortIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = new ArrayShortIterator(values);
    items = newItems;
  }
  
  public void addIterator(IShortIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    IShortIterator[] newItems = new IShortIterator[items.length+1];
    System.arraycopy(items, 0, newItems, 0, items.length);
    newItems[items.length] = (IShortIterator)it.clone();
    items = newItems;
  }
  
  public void removeItem(int index)
  {
    if(index >= this.index && index < items.length)
    {
      IShortIterator[] newItems = new IShortIterator[items.length-1];
      System.arraycopy(items, 0, newItems, 0, index);
      System.arraycopy(items, index+1, newItems, index, items.length-index-1);
      items = newItems;
    }
  }
  
  public ShortIterator(ShortIterator r)
  {
    index = r.index;
    items = new IShortIterator[r.items.length];
    
    for(int i = 0; i < items.length; i++)
      if(r.items[i] != null)
        items[i] = (IShortIterator)r.items[i].clone();
  }
  
  public Object clone()
  {
    return new ShortIterator(this);
  }
  
  public void shift()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    if(index < items.length) items[index].next();

    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
  }
  
  public short value()
  {
    while(index < items.length && (items[index] == null || items[index].stop()))
    {
      index++;
    }
    
    return items[index].value();
  }
  
  public boolean stop()
  {
    int i = index;
    while(i < items.length && (items[i] == null || items[i].stop()))
    {
      i++;
    }

    if(i >= items.length) return true;
    else                  return false;
  }

  public boolean equals(Object o) 
  {
    if(!(o instanceof ShortIterator))
      return false;

    ShortIterator obj = (ShortIterator)o;

    if(   index != obj.index
       || items.length != obj.items.length
      )
      return false;
      
    for(int i = 0; i < items.length; i++)
      if(items[i] == null && obj.items[i] != null || items[i] != null && !items[i].equals(obj.items[i]))
        return false;
        
    return true;
  }

  public int hashCode ()
  {
    int r = 0;
    for(int i = 0; i < items.length; i++)
      r += items[i].hashCode();
    
    return r;
  }
}
