/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.iterators.lang;

import jatva.iterators.AbstractIterator;

public class StringBufferIterator
  extends AbstractIterator
  implements jatva.iterators.Iterator
{
  public static StringBuffer[] defaultValues = {
     null
   , new StringBuffer("")
   , new StringBuffer("\u0000")
   , null
  };
  static
  {
    defaultValues[3] = defaultValues[1];
  }

  protected StringBuffer[] values = defaultValues;
  protected IStringIterator it = new StringIterator();
  protected int index = 0;
  
  public StringBufferIterator()
  {
  }
  
  public StringBufferIterator(StringBuffer[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new StringBuffer[values.length];
    it = null;
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:new StringBuffer(values[i].toString());
  }

  public StringBufferIterator(IStringIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    values = null;
    this.it = (IStringIterator)it.clone();
  }
  
  public StringBufferIterator(StringBuffer[] values, IStringIterator it)
  {
    if(values == null && it == null)
      throw new IllegalArgumentException("Both arguments should not be null");
    if(values != null)
    {
      this.values = new StringBuffer[values.length];
      for(int i = 0; i < values.length; i++)
        this.values[i] = (values[i] == null)?null:new StringBuffer(values[i].toString());
    }
    if(it != null) this.it = (IStringIterator)it.clone();
  }
  
  public void setValueIterator(IStringIterator it)
  {
    if(it == null) throw new IllegalArgumentException("Argument should not be null");
    this.it = (IStringIterator)it.clone();
  }
  
  public void setArray(StringBuffer[] values)
  {
    if(values == null) throw new IllegalArgumentException("Argument should not be null");
    this.values = new StringBuffer[values.length];
    
    for(int i = 0; i < values.length; i++)
      this.values[i] = (values[i] == null)?null:new StringBuffer(values[i].toString());
  }
  
  public StringBufferIterator(StringBufferIterator r)
  {
    index = r.index;
    it = (r.it == null)?null:(IStringIterator)r.it.clone();
    if(r.values == null) { values = null; return; }
    values = new StringBuffer[r.values.length];
    
    for(int i = 0; i < values.length; i++)
      values[i] = (r.values[i] == null)?null:new StringBuffer(r.values[i].toString());
  }
  
  public Object clone()
  {
    return new StringBufferIterator(this);
  }
  
  public boolean equals(Object o) 
  {
    if(!(o instanceof StringBufferIterator))
      return false;

    StringBufferIterator obj = (StringBufferIterator)o;

    if(index != obj.index) return false;
    if(   it == null && obj.it != null
       || it != null && !it.equals(obj.it)
      ) return false;
    if(values == null) return (obj.values == null);
    if(values != null && obj.values == null) return false;
    if(values.length != obj.values.length) return false;
    
    for(int i = 0; i < values.length; i++)
      if(   values[i] == null && obj.values[i] != null
         || values[i] != null && !values[i].equals(obj.values[i])
        ) return false;
    
    return true;   
  }

  public StringBuffer value()
  {
    if(values != null && index < values.length) return values[index];
    else if(it != null)            return new StringBuffer(it.value());
    else                           return null;
  }
  
  public boolean stop()
  {
    if(it == null)          return (index >= values.length);
    else if(values == null) return it.stop();
    else                    return index >= values.length && it.stop();
  }

  public void shift()
  {
    if(values != null && index < values.length) index++;
    else if(it != null)       it.next();
  }
  
  public String toString()
  {
    if(stop()) return jatva.iterators.Iterator.EXHAUSTED;
    return "" + value();
  }

  public Object next ()
  {
    Object r = value();
    shift();
    return r;    
  }

  public int hashCode ()
  {
    int r = index;
    if(values != null)
      for(int i = 0; i < values.length; i++)
        r += values[i].hashCode()*i;
    
    if(it != null)
      r += it.hashCode();
    
    return r;
  }
}
