/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: MediatorSet.java,v 1.5 2006/09/05 15:33:17 almo Exp $
 * Created on 01.08.2005
 *
 */
package jatva.lang;

import jatva.iterators.lang.IObjectIterator;
import jatva.util.IdentityList;
import jatva.util.WeakIdentitySet;

public class MediatorSet
{
  WeakIdentitySet set      = new WeakIdentitySet();
  MediatorSet     parent   = null;
  IdentityList    children = new IdentityList();
  
  ModelClass      mclass   = null;

  public MediatorSet(ModelClass c, MediatorSet t)
  {
    mclass = c;
    parent = t;
    if(t != null)
    {
      t.addChild(this);
    }
  }
  
  public ModelObject add(ModelObject o, Object[] impls)
  {
    ModelObject res = find(impls, null, true);
    
    if(res == null) set.add(o);
      
    return res;
  }
  
  public void add(ModelObject o)
  {
    if(o == null) return; 
      
    set.add(o);
  }

  public void remove(ModelObject o)
  {
    if(o != null) set.remove(o);
  }
  
  public void addChild(MediatorSet s)
  {
    children.addLast(s);
  }

  public ModelObject find(Object[] implFields)
  {
    IObjectIterator i = set.iterator();
    for(; !i.stop(); i.shift())
    {
      if(((ModelObject)i.value())._ts_compareImplementationFields(implFields))
        return (ModelObject)i.value();
    }

    return null;
  }
  
  public ModelObject find(Object[] impls, MediatorSet ignore, boolean goUp)
  {
    ModelObject r1 = null;
    
    IObjectIterator i = set.iterator();
    for(; !i.stop(); i.shift())
    {
      ModelObject o = (ModelObject)i.value();
      if(o._ts_compareImplementationFields(impls))
      {
        r1 = o;
        break;
      }
    }

    ModelObject r2 = null;
    
    for(int j = 0; j < children.size(); j++)
    {
      MediatorSet c = (MediatorSet)(children.get(j));
      
      if( ignore != c )
      {
        r2 = c.find( impls, null, false );
           
        if( r1 != null && r2 != null )
        {
          throw new ConflictingMediatorObjectsException(r1, r2, impls);
        }
        else if(r2 != null)
          r1 = r2;
       }
    }

    if( goUp && parent != null )
    {
       r2 = parent.find( impls, this, true );
       if( r2 != null )
       {
         throw new ConflictingMediatorObjectsException(r2, mclass._ts_name, impls);
       }
    }

    return r1;
  }
  
}
