/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.lang;

import jatva.lang.services.DeepSubject;

/**
 * The <code>ModelClass</code> class is the base class of classes modeling classes.
 * <p>
 * This class can be used to represent objects of Java <code>Class</code> class.
 * <p>
 * 
 * @author  Victor Kuliamin
 * @version 0.1, 12/03/2002
 * @see     jatva.lang.ModelObject
 * @see     java.lang.Class
 * @since   JavaTESK 0.2
 */
public class ModelClass
  extends ModelObject
{
  protected String _ts_name                = null;
  protected String _ts_staticUpdateManager = null;

  public ModelClass() {}
  
  public ModelClass(String name, String staticUpdateManagerName)
  {
    _ts_name                = name;
    _ts_staticUpdateManager = staticUpdateManagerName;
  }
  
  public ModelClass(ModelClass r)
  {
    _ts_name = r._ts_name;
    _ts_staticUpdateManager = r._ts_staticUpdateManager;
  }
  
  public void _ts_setStaticUpdateManager(String str)
  {
    _ts_staticUpdateManager = str;
  }
  
  public String _ts_getStaticUpdateManager()
  {
    return _ts_staticUpdateManager;
  }

  //---------------------------------------------------------------------------
  // Methods overrding the ones of DeepSubject class
  //
  /**
   * Method overrides the corresponding method in {@link DeepSubject} class.
   * 
   * @return  the clone created.
   * @see     jatva.lang.services.DeepSubject
   */
  public DeepSubject _ts_shallowClone ()
  {
    return new ModelClass(this);
  }

  /**
   * Overrides the corresponding method in {@link DeepSubject}
   * class.
   * 
   * @param obj  The object to compare with
   * @return     <code>true</code> if this object and <code>obj</code>
   *             have identical fields 
   */
  public boolean _ts_shallowEquals (Object obj)
  {
    if(!(obj instanceof ModelClass)) return false;
    
    ModelClass r = (ModelClass)obj;
    
    return    _ts_name == r._ts_name
           && _ts_staticUpdateManager == r._ts_staticUpdateManager;
  }
  
  /**
   * Overrides the corresponding method in {@link DeepSubject}
   * class.
   * 
   * @param obj  The object, which fields values are assigned into fields of
   *             this object 
   */
  public void _ts_shallowAssign (Object obj)
  {
    if(!(obj instanceof ModelClass)) return;
    
    ModelClass r = (ModelClass)obj;
    
    _ts_name = r._ts_name;
    _ts_staticUpdateManager = r._ts_staticUpdateManager;
  }

  //---------------------------------------------------------------------------
  // State synchronization methods
  //
  /**
   * Synchronizes fields of class object with
   * the corresponding static fields.
   */
  public void _ts_updateClassFields()
  {
  }
  
  public void mapStateDown ()
  {
    throw new MapStateDownNotSupportedException(_ts_name);
  }
  
  //------------------------------------------------------
  // miscellaneous general-purpose methods
  //
  public Object clone()
  {
    return new ModelClass(this);
  }
}
