/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: TestingModel.java,v 1.10 2008/02/12 17:00:46 almo Exp $
 * Created on 25.02.2005
 *
 */
package jatva.lang;

import jatva.lang.services.DeepSubject;
import jatva.lang.synchronization.TestManager;
import jatva.util.Utils;

/**
 * Base class for scenarios. 
 *
 * @author Victor Kuliamin
 *
 */
public class TestingModel
  extends DeepSubject
  implements Runnable
{
  public TestingModel ( )
  {
  }

  public TestingModel(TestingModel right)
  {
  }

  public synchronized Object clone()
  {
    return new TestingModel(this);
  }
  
  private GraphExplorer _ts_engine = new GraphExplorer();
  
  public GraphExplorer testEngine()
  {
    return _ts_engine;
  }
  
  public void setTestEngine(GraphExplorer te)
  {
    _ts_engine = te;
  }
  
  public void run ( )
  {
    TestProperty.loadProperties();
    TestManager _lv_registrar = TestManager.getRegistrar();
    _lv_registrar.scenarioStart(this);
      
    try
    {
      _ts_engine.run ( this );
    }
    catch(TestException err)
    {
    }
        
    _lv_registrar.scenarioEnd(_ts_engine.getFinishReason());
  }
  
  public void reduce()
  {
  }
  
  public void reset()
  {
    throw new ResetNotSupportedException(getClass().getName());
  }
  
  /**
   * Informs test system that scenario method is starting some
   * meaningfull activity its operation should be properly traced. 
   */
  public void confirmTransition()
  {
      TestManager.getRegistrar().confirmTransition();
  }
  
  // A bunch of methods for writing JUnit-like scenarios that use SingleStateWithReset_TestEngine.
  // Unlike JUnit with its lots and lots of assert... methods we use simple and clean approach of SuiteRunner.

  /**
   * The setup method for JUnit-like testing. This method is equivalent to
   * setUp method while reset method is equivalent to tearDown method.
   *
   */
  public void setup()
  {
    throw new SetupNotSupportedException(getClass().getName());
  }

  /**
   * Verify that a boolean condition is true.
   * If verification succeeds, returns normally.
   * If verification fails, throws <code>jatva.lang.TestException</code>.
   * 
   * @param condition The boolean condition to verify.
   */
  protected void verify(boolean condition)
  {
    verify(condition, null);
  }

  /**
   * Verify that a boolean condition, described in String message, is true.
   * If verification succeeds, returns normally.
   * If verification fails, throws <code>jatva.lang.TestException</code>.
   * 
   * @param condition The boolean condition to verify.
   */
  protected void verify(boolean condition, String message)
  {
    if (!condition)
    {
      throw new ScenarioConstraintViolationException( getClass().getName(), "verify()", message );
    }
  }

  /**
   * Throws <code>jatva.lang.TestException</code> to indicate a test failed.
   */
  protected void fail() 
  {
    throw new ScenarioConstraintViolationException( getClass().getName(), "fail()", null);
  }

  /**
   * Throws <code>jatva.lang.TestException</code>, with the passed
   * <code>String</code> <code>message</code> as the exception's detail
   * message, to indicate a test failed.
   *
   * @param message A message describing the failure.
   */
  protected void fail(String message) 
  {
    throw new ScenarioConstraintViolationException( getClass().getName(), "fail()", message);
  }

  /**
   * Throws <code>jatva.lang.TestException</code>, with the passed
   * <code>Throwable</code> cause, to indicate a test failed.
   *
   * @param cause A <code>Throwable</code> that indicates the cause of the failure.
   */
  protected void fail(Throwable cause) 
  {
    fail(cause.getMessage(), cause);
  }

  /**
   * Throws <code>jatva.lang.TestException</code>, with the passed
   * <code>String</code> <code>message</code> as the exception's detail
   * message and <code>Throwable</code> cause, to indicate a test failed.
   *
   * @param message A message describing the failure.
   * @param cause A <code>Throwable</code> that indicates the cause of the failure.
   */
  protected void fail(String message, Throwable cause) 
  {
    throw new ScenarioConstraintViolationException( getClass().getName(), "fail()", message, cause);
  }

  public AbstractAction _ts_initialize ()
  {
    return null;
  }

  protected int _ts_getMaxMethodNumber()
  { return -1; }

  private void throwNullAbstractAction( String method )
  {
    throw new TestSystemInternalError
        ( "Null abstract action argument"
        , new String[]{ "object", "method" }
        , new String[]{ Utils.getSimpleString( this ), method }
        );
  }

  private void throwIllegalAbstractActionNumber( String method, int numberFound )
  {
      throw new TestSystemInternalError
          ( "Illegal abstract action method number"
          , new String[]{ "object", "method", "maximum number", "number found" }
          , new String[]{ Utils.getSimpleString( this )
                        , method
                        , "" + _ts_getMaxMethodNumber()
                        , "" + numberFound
                        }
          );
  }

  public AbstractAction _ts_next( AbstractAction a )
  {
    if( a == null )
    {
      throwNullAbstractAction( "next()" );
      return null;
    }

    if( a.next() )
      return a;
    else
      return _ts_nextDeclared( a.method() );
  }

  protected AbstractAction _ts_nextDeclared( int currentMethodNumber )
  {
    int maxNumber = _ts_getMaxMethodNumber();

    if( currentMethodNumber == maxNumber )
      // last defined action
      return null;
    else
    {
      throwIllegalAbstractActionNumber( "next()", currentMethodNumber );
      return null;
    }
  }

  public void _ts_execute( AbstractAction a )
  {
    if( a == null )
    {
      throwNullAbstractAction( "execute()" );
      return;
    }
    
    _ts_executeDeclared( a );
  }

  protected void _ts_executeDeclared( AbstractAction a )
  {
      throwIllegalAbstractActionNumber( "execute()", a.method() );
  }

  public void _ts_traceAction( AbstractAction a )
  {
    if( a == null )
    {
      throwNullAbstractAction( "traceAction()" );
      return;
    }

    String a_name = _ts_getScenMethodName( a.method() );

    jatva.lang.synchronization
            .TestManager.getRegistrar().registerScenarioMethod( this, a_name, a );
  }

  protected String _ts_getScenMethodName( int method_number )
  {
      throwIllegalAbstractActionNumber( "traceAction()", method_number );
      return null;
  }

  public Object _ts_state()
  {
    return "Single";
  }
}
