/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.lang.coverage;

import jatva.lang.ModelObject;

public class CoverageInfo
{
  public final static String BRANCH_COVERAGE = "branch_coverage"; 
  public final static String MARK_COVERAGE = "mark_coverage"; 
  public final static String PREDICATE_COVERAGE = "predicate_coverage"; 
  public final static String DISJUNCT_COVERAGE = "disjunct_coverage"; 

  protected String _identifier = null;
  protected jatva.util.IntSet _elements_achieved = new jatva.util.IntSet();
  //protected java.util.HashSet _elements_achieved = new java.util.HashSet(); 
  
  //protected String _class_name = null;
  protected String _method_signature = null;
  protected ModelObject _object_under_test = null;

  protected int _number_of_elements = 0;

  public CoverageInfo ()
  {
  } 

  public CoverageInfo(ModelObject p_object_under_test, String p_method_signature, String p_identifier)
  {
    _object_under_test = p_object_under_test;
    _method_signature = p_method_signature;
    _identifier = p_identifier;
  }
  
  public CoverageInfo(ModelObject p_object_under_test, String p_method_signature)
  {
    this(p_object_under_test, p_method_signature, BRANCH_COVERAGE);
  }
  
  public CoverageInfo ( CoverageInfo right )
  {
    _identifier = right._identifier;
    _elements_achieved = new jatva.util.IntSet(right._elements_achieved);
    _number_of_elements = right._number_of_elements;
    
    _object_under_test = right._object_under_test;
    _method_signature = right._method_signature;
  } 

  public synchronized Object clone()
  {
    return new CoverageInfo ( this );
  } 
  
  public synchronized ModelObject trackedObject()
  {
    return _object_under_test;
  }
  
  public synchronized void setIdentifier(String ident)
  {
    _identifier = ident;
  }
  
  public synchronized String identifier()
  {
    return _identifier;
  }
  
  public synchronized void setMethodSignature(String p_method_signature)
  {
    _method_signature = p_method_signature;
  }
  
  public synchronized String methodSignature()
  {
    return _method_signature;
  }
  
  public synchronized void setSize( int number )
  {
    if(_identifier != null) _number_of_elements = number;
  }
  
  public synchronized int size()
  {
    if(_identifier == null) return Integer.MAX_VALUE;
    else                    return _number_of_elements;
  }
  
  public synchronized boolean isAchieved(int id)
  {
    if(_identifier == null) return false;
    else                    return _elements_achieved.contains(id);
  }
  
  public synchronized void addToAchieved ( int id )
  {
    if(_identifier != null) _elements_achieved.add( id );
  } 

  public synchronized boolean isComplete()
  {
    if(_identifier == null) return false;
    else return    (_elements_achieved.size() == _number_of_elements)
           || (_number_of_elements == 0);
  } 
  
  public synchronized String toString()
  {
    return "" + _method_signature + "/" + _identifier + " : " + _elements_achieved.toString();
  }
} 
