/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: DisjunctInfo.java,v 1.3 2006/03/30 12:47:09 barancev Exp $
 * Created on 14.02.2005
 *
 */
package jatva.lang.coverage;

import jatva.util.TwoBitSet;

/**
 * The class storing disjunct data. 
 * 
 * @author Victor Kuliamin
 */
public class DisjunctInfo
{
  // Formula values constants
  public final static byte Undefined = 0;
  public final static byte True      = 1;
  public final static byte False     = 2;
  
  private int       _predicateId  = -1;
  private int       _markedPathId = -1;
  private int       _branchId     = -1;
  private TwoBitSet _values  = null;

  public int getPredicateId()
  {
    return _predicateId;
  }

  public int getMarkedPathId()
  {
    return _markedPathId; 
  }

  public int getBranchId()
  {
    return _branchId;
  }

  public TwoBitSet getValues()
  {
    return _values;
  }

  public static boolean compareValues(TwoBitSet v1, TwoBitSet v2)
  {
    int n = (v1.size() > v2.size())?v2.size():v1.size();
    for(int i = 0; i < n; i++)
      if(v1.get(i) != v2.get(i)) return false;

    if(v1.size() > n)
      for(int i = n; i < v1.size(); i++)
        if(v1.get(i) != 0) return false;

    if(v2.size() > n)
      for(int i = n; i < v2.size(); i++)
        if(v2.get(i) != 0) return false;

    return true;
  }

  public DisjunctInfo
  (   int branchId
    , int markedPathId
    , int predicateId
    , int formulasNumber
  )
  {
    _predicateId  = predicateId;
    _markedPathId = markedPathId;
    _branchId     = branchId;
    _values       = new TwoBitSet(formulasNumber);
  }

  public DisjunctInfo
  (   int branchId
    , int markedPathId
    , int predicateId
    , String values
  )
  {
    _predicateId  = predicateId;
    _markedPathId = markedPathId;
    _branchId     = branchId;
    _values       = new TwoBitSet(values.length());

    for(int i = 0; i < values.length(); i++)
    {
      char c = values.charAt(i);
      if     (c == '+') _values.set(i, True);
      else if(c == '-') _values.set(i, False);
    }
  }

  public byte getFormulaValue(int formulaId)
  {
    return _values.get(formulaId);
  }
  
  public void setFormulaValue(int formulaId, byte value)
  {
    _values.set(formulaId, value);
  }

  public String getFormulasValues()
  {
    StringBuffer str = new StringBuffer();
    for(int i = 0; i < _values.size(); i++)
    {
      byte d = _values.get(i);
      switch (d)
      {
        case True:
          str.append('+');
          break;
        case False:
          str.append('-');
          break;
        default:
          str.append('*');
          break;
      }
    }
    return str.toString();
  }
  
  public static String getString(TwoBitSet values)
  {
    StringBuffer str = new StringBuffer();
    for(int i = 0; i < values.size(); i++)
    {
      byte d = values.get(i);
      switch (d)
      {
        case True:
          str.append('+');
          break;
        case False:
          str.append('-');
          break;
        default:
          str.append('*');
          break;
      }
    }
    return str.toString();
  }
}

