/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: SEOperationInfo.java,v 1.1 2008/07/25 09:48:16 almo Exp $
 * Created on 15.02.2005
 *
 */
package jatva.lang.coverage;

import jatva.lang.InsufficientAccessibilityError;
import jatva.lang.TranslationError;
import jatva.util.IntToStringMap;
import jatva.util.MethodInfo;
import jatva.util.Utils;

import java.lang.reflect.Method;

/**
 * The class storing data of all disjuncts.
 *
 * @author Victor Kuliamin
 */
public class SEOperationInfo extends MethodInfo
{
  /*public final static String Branch_Coverage_ID     = "branch_coverage";
  public final static String MarkedPath_Coverage_ID = "mark_coverage";
  public final static String Predicate_Coverage_ID  = "predicate_coverage";
  public final static String Disjunct_Coverage_ID   = "disjunct_coverage";

  protected String translateCoverageName(String str)
  {
    if(str == null)                   return null;
    else if(   str.equals(CoverageInfo.BRANCH_COVERAGE)
            || str.equals(CoverageInfo.MARK_COVERAGE)
            || str.equals(CoverageInfo.PREDICATE_COVERAGE)
            || str.equals(CoverageInfo.DISJUNCT_COVERAGE)
           )                          return str;
    else if(str.equals("branch"))     return CoverageInfo.BRANCH_COVERAGE;
    else if(str.equals("mark"))       return CoverageInfo.MARK_COVERAGE;
    else if(str.equals("predicate"))  return CoverageInfo.PREDICATE_COVERAGE;
    else if(str.equals("disjunct"))   return CoverageInfo.DISJUNCT_COVERAGE;
    else                              return null;
  }
  */
  public final static String BRANCHES   = "branch";
  public final static String MARKS      = "mark";
  public final static String PREDICATES = "predicate";
  public final static String DISJUNCTS  = "disjunct";


//  Class   _class  = null;

  int _classLevel  = 0;

  SEOperationInfo directlyRefinedOperation = null;

  DisjunctsMap   _disjuncts   = new DisjunctsMap();
  IntToStringMap _formulas    = new IntToStringMap();
  IntToStringMap _branches    = new IntToStringMap();
  IntToStringMap _markedPaths = new IntToStringMap();
  IntToStringMap _predicates  = new IntToStringMap();

  private Method _oracleMethod = null;
  private Method _implementationMethod = null;

//  private int _getClassLevel()
//  {
//    Field fld = null;
//
//    try
//    {
//      fld = _class.getDeclaredField("_ts_classLevel");
//      fld.setAccessible(true);
//    }
//    catch (SecurityException e)
//    {
//      throw new InsufficientAccessibilityError(  "Insufficient rights to get field value through reflection"
//                                               , e
//                                               , new String[]{"class", "field"}
//                                               , new String[]{_class.getName(), "_ts_classLevel"}
//                                              );
//    }
//    catch (NoSuchFieldException e)
//    {
//      return 0;
//    }
//
//    if((fld.getModifiers() & Modifier.STATIC) == 0 || fld.getType() != Integer.TYPE)
//    {
//      return 0;
//    }
//    else try
//    {
//      return fld.getInt(null);
//    }
//    catch (IllegalArgumentException e)
//    {
//      throw new TestSystemInternalError(  "Impossible situation"
//                                        , e
//                                        , new String[]{"class", "field"}
//                                        , new String[]{_class.getName(), "_ts_classLevel"}
//                                       );
//    }
//    catch (IllegalAccessException e)
//    {
//      throw new InsufficientAccessibilityError(  "Field is inaccessible through reflection"
//                                               , e
//                                               , new String[]{"class", "field"}
//                                               , new String[]{_class.getName(), "_ts_classLevel"}
//                                              );
//    }
//  }

  public SEOperationInfo( int classLevel, String extendedReflectSignature, SEOperationInfo refinedOperation )
  {
    super( extendedReflectSignature );
    this._classLevel = classLevel;

//    _classLevel = _getClassLevel();
    this.directlyRefinedOperation = refinedOperation;

    Class<?> cls = method.getDeclaringClass();

    if(reflectSignature.indexOf("static ") != -1)
    {
      try
      {
        Class<?>[] nested_classes = cls.getDeclaredClasses();
        cls = null;
        for( Class<?> nested_class : nested_classes )
        {
          if( nested_class.getName().endsWith("$ClassModel"))
          {
            cls = nested_class;
            break;
          }
        }
        if(cls == null)
        {
          throw new TranslationError( "Incorrect translation of specification class - class model class not found"
                                    , new String[]{"class"}
                                    , new String[]{declaringClass}
                                    );
        }
      }
      catch(SecurityException e)
      {
        throw new InsufficientAccessibilityError( "No rights to call Class.getDeclaredClasses() method"
                                                , e
                                                , new String[]{"class"}
                                                , new String[]{declaringClass}
                                                );
      }
    }

    _oracleMethod         = getOracleMethodFor(cls, reflectSignature);
    _implementationMethod = getImplMethodFor(cls, reflectSignature);

    if(_oracleMethod == null)
    {
      //System.out.println(" CLASS: " + cls);
      //System.out.println(" METHOD: " + method);

      throw new TranslationError(  "Incorrect translation of specification class - oracle method not found"
                                 , new String[]{"class", "method"}
                                 , new String[]{ declaringClass, prettySignature}
                                );
    }

    if(_implementationMethod == null)
    {
      //System.out.println(" CLASS: " + cls);
      //System.out.println(" METHOD: " + method);

      throw new TranslationError(  "Incorrect translation of specification class - implementation method not found"
                                 , new String[]{"class", "method"}
                                 , new String[]{declaringClass, prettySignature}
                                );
    }
  }

  public int getClassLevel()
  {
    return _classLevel;
  }

  public void addBranch(int id, String identifier)
  {
    _branches.put(id, identifier);
  }

  public void addMarkedPath(int id, String description)
  {
    _markedPaths.put(id, description);
  }

  public void addPredicate(int id, String description)
  {
    _predicates.put(id, description);
  }

  public void addFormula(int id, String description)
  {
    _formulas.put(id, description);
  }

  public void addDisjunct(int id, DisjunctInfo info)
  {
    _disjuncts.put(id, info);
  }

/*
  protected ModelCall lastCall = null;

  public ModelCall LastCall
  {
    get { return lastCall; }
  }*/

//----- Getters
  public IntToStringMap getBranchesMap()
  {
    return _branches;
  }

  public IntToStringMap getMarkedPathsMap()
  {
    return _markedPaths;
  }

  public IntToStringMap getPredicatesMap()
  {
    return _predicates;
  }

  public DisjunctsMap getDisjunctsMap()
  {
    return _disjuncts;
  }

  public IntToStringMap getFormulasMap()
  {
    return _formulas;
  }

  public Method getImplementationMethod()
  {
    return _implementationMethod;
  }

  public Method getOracleMethod()
  {
    return _oracleMethod;
  }
  public int size( String coverage )
  {
    if     (coverage.equals(BRANCHES))   return _branches.size();
    else if(coverage.equals(MARKS))      return _markedPaths.size();
    else if(coverage.equals(PREDICATES)) return _predicates.size();
    else if(coverage.equals(DISJUNCTS))  return _disjuncts.size();
    else                                 return 0;
  }

  /**
   * @param p_class
   * @param method
   * @param parameterTypes
   * @return
   */
  public static Method getImplMethodFor (Class<?> cls, String method)
  {
    int opening_brace = method.indexOf( '(' );
    int closing_brace = method.indexOf( ')' );
    assert opening_brace < closing_brace;

    String pars = method.substring( opening_brace, closing_brace + 1 );

    int method_name_start = method.lastIndexOf( '.', opening_brace ) + 1;

    String impl_name = "_ts_implementation_" + method.substring( method_name_start
                                                               , opening_brace
                                                               );

    StringBuffer res = new StringBuffer();
    // see comments to 'getMethodByStringPrefix'. return type is not significant here
    res.append( "void " );
    res.append( cls.getName() );
    res.append( "." ).append( impl_name ).append( pars );
    String s = res.toString();
    return Utils.getMethodByStringPrefix(cls, s.toString());
  }

  /**
   * @param p_class
   * @param method
   * @param parameterTypes
   * @return
   */
  public static Method getOracleMethodFor( Class<?> cls, String method )
  {
    int opening_brace = method.indexOf( '(' );
    int closing_brace = method.indexOf( ')' );
    assert opening_brace < closing_brace;

    String pars = method.substring( opening_brace, closing_brace + 1 );

    int method_name_start = method.lastIndexOf( '.', opening_brace ) + 1;

    String oracle_name = "_ts_oracle_" + method.substring( method_name_start
                                                         , opening_brace
                                                         );

    StringBuffer res = new StringBuffer();
    res.append( "boolean " );
    res.append( cls.getName() );
    res.append( "." ).append( oracle_name ).append( pars );

    return Utils.getMethodByStringPrefix(cls, res.toString() );
  }

  public SEOperationInfo getDirectlyRefinedOperation()
  { return directlyRefinedOperation; }

  public SEOperationInfo getRefinedOperation( int targetClassLevel )
  {
      SEOperationInfo op = this;
      while( op.getClassLevel() != targetClassLevel )
      {
          op = op.getDirectlyRefinedOperation();
      }
      return op;
  }
}
