/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: DeepSubject.java,v 1.5 2007/10/19 06:51:49 kuliamin Exp $
 * Created on 02.06.2004
 */
package jatva.lang.services;

import jatva.util.IdentityMap;
import jatva.util.Utils;


/**
 * The base class of classes, which objects can 
 * be copied and compared for equality on the full depth along references on 
 * other such objects. 
 * 
 * @author Victor Kuliamin
 */
public class DeepSubject 
{
  /**
   * Default constructor.
   */
  public DeepSubject()
  {
  }

  /**
   * Copy constructor. Constructs a shallow copy of the parameter.
   * To be overriden in descendants.
   * 
   * @param r  The object to be copied into this one.
   */
  public DeepSubject(DeepSubject r)
  {
  }

  /**
   * Provides deep comparison between this and other object.
   * Returns <code>true</code> if this object and <code>obj</code> are
   * roots of isomorphic object graphs, which corresponding nodes are 
   * shallow equal or coincide.  
   * Should not be overriden. Uses template method pattern.
   * 
   * @param obj  The object to compare with 
   * @return     <code>true</code> if this object and <code>obj</code> are
   *             roots of isomorphic object graphs (having references as edges),
   *             which corresponding nodes are shallow equal or coincide.
   *             <code>false</code> otherwise.
   */
  public final boolean _ts_deepEquals(Object obj)
  {
    return DeepOperationManager.areEqual((Object)this, obj);
  }

  /**
   * Provides temporary result of deep comparison between this and 
   * other object.
   * It is used in deep comparisons for this and other objects in
   * {@link #_ts_deepEquals(Object)}.
   * Should not be overriden and used by classes other than {@link DeepOperationManager}.
   * Uses template method pattern.
   *  
   * @param obj  The object to compare with
   * @return     <code>true</code> if this object and <code>obj</code> are
   *             roots of isomorphic object graphs (having references as edges),
   *             which corresponding nodes are shallow equal or coincide.
   *             <code>false</code> otherwise.
   
  final boolean _ts_implDeepEquals(DeepSubject obj, IdentityMap m)
  {
    if(!DeepOperationManager.areEqual(this, obj)) return false;
    if(!_ts_fieldsDeepEquals(obj))                return false;

    return true;
  }*/

  /**
   * Returns a deep copy of this object.
   * Should not be overriden. Uses template method pattern.
   * 
   * @return  The deep copy of this object
   */
  public final DeepSubject _ts_deepClone()
  {
    return (DeepSubject)DeepOperationManager.getDeepCloneIndependent(this);
  }

  /**
   * Returns a temporary deep copy of this object.
   * Copying is not considered to be finished after this method execution.
   * It is used in deep copy construction for this and other objects in
   * {@link #_ts_deepClone()}.
   * Should not be overriden and used by classes other than {@link DeepOperationManager}.
   * Uses template method pattern.
   * 
   * @return  The deep copy of this object
   */
  final DeepSubject _ts_implDeepClone()
  {
    DeepSubject res = DeepOperationManager.getCloneIndependent(this);

    _ts_fieldsDeepCloneIndependent(res);

    return res;
  }
  
  /**
   * Returns deep string rperesentation of this object.
   * Should not be overriden.
   * 
   * @return    The deep string represetantion of this object
   */
  public final String _ts_deepToString()
  {
    return DeepOperationManager.getDeepString(this);
  }
  
  /**
   * Appends the given <code>StringBuffer</code> with string
   * representation of this object.
   * 
   * @param buf    The <code>StringBuffer</code> to append
   */
  final void _ts_implDeepAppend(StringBuffer buf, IdentityMap m)
  {
    buf.append("{ ");
    _ts_fieldsDeepAppend(buf, m);
    buf.append(" }");
  }

  /*
   * Returns shallow string representation of the object.
   */
  public final String toShallowString()
  {
    StringBuffer buf = new StringBuffer();
    
    buf.append(Utils.getSimpleString(this));
    buf.append("{ ");
    _ts_appendFields(buf);
    buf.append(" }");

    return buf.toString();
  }
  
  /**
   * Returns a shallow copy of this object.
   * In {@link DeepSubject} class return new object.
   * To be overriden in descendant classes.
   * 
   * @return A shallow copy of this object
   */
  public DeepSubject _ts_shallowClone ()
  {
    return new DeepSubject(this);
  }

  /**
   * Compares this and other object for identity of their
   * fields.
   * In {@link DeepSubject} class compares objects for identity.
   * To be overriden in descendant classes.
   * 
   * @param obj  The object to compare with
   * @return     <code>true</code> if this object and <code>obj</code>
   *             have identical fields 
   */
  public boolean _ts_shallowEquals (Object obj)
  {
    return this == obj;
  }

  /**
   * Performs assignment of <code>obj</code> fields into fields of
   * this object.  
   * In {@link DeepSubject} class does nothing.
   * To be overriden in descendant classes.
   * 
   * @param obj  The object, which fields values are assigned into fields of
   *             this object 
   */
  public void _ts_shallowAssign (Object obj)
  {
  }
  
  protected void _ts_fieldsDeepAssign(DeepSubject r, IdentityMap m)
  {
    
  }

  /**
   * Compares fields of this object and of <code>obj</code> of the full
   * depth.
   * It is used in deep comparison for this and other objects in
   * {@link #_ts_deepEquals(Object)}.
   * To be overriden in descendant classes.
   * 
   * @param obj  The object which fields are compared with the fields of 
   *             this one
   * @return     <code>true</code> if all the fields are equals on the full
   *             depth along the references on the deeply cloneable objects.
   * @see        #_ts_deepEquals(Object)
   */
  protected boolean _ts_fieldsDeepEquals(DeepSubject obj, IdentityMap m)
  {
    return true;
  }
  
  protected boolean _ts_fieldsDeepEquals(DeepSubject obj, IdentityMap m, IdentityMap tmp)
  {
    return true;
  }

  /**
   * Constructs deep copies of this object's fields storing references
   * and assigns them into the corresponding fields of <code>obj</code>.
   * It is used in deep copy construction for this and other objects in
   * {@link #_ts_deepClone()}.
   * To be overriden in descendant classes.
   * 
   * @param obj  The object to assign fields of 
   */
  protected void _ts_fieldsDeepCloneIndependent(DeepSubject obj)
  {
  }
  
  /**
   * Appends the specified <code>StringBuffer</code> object
   * with short representation of this object's fields.
   * 
   * @param buf    The <code>StringBuffer</code> to append
   */
  protected void _ts_appendFields(StringBuffer buf)
  {
  }
  
  /**
   * Appends the specified <code>StringBuffer</code> object
   * with deep representation of this object's fields.
   * 
   * @param buf    The <code>StringBuffer</code> to append
   */
  protected void _ts_fieldsDeepAppend(StringBuffer buf, IdentityMap m)
  {
  }

  
  public void _ts_store(IdentityMap m)
  {
    DeepOperationManager.getDeepClone(this, m);
  }
  
  public void _ts_restore(IdentityMap m)
  {
    DeepOperationManager.restore(this, m);
  }
  
  public boolean _ts_compare(IdentityMap m)
  {
    return DeepOperationManager.compare(this, m);
  }
  
  protected void _ts_fieldsDeepClone(DeepSubject o, IdentityMap m)
  {
    
  }
}
