package jatva.lang.synchronization;

import jatva.util.IdentityMap;
import jatva.util.IdentityToIntMap;

import java.util.Iterator;

public class ActorContext
{
  public static final int ACTOR_FINISHED = -1;
  //public static final int ACTOR_INTERRUPTED = -2;
  
  private IdentityToIntMap actorPositions = new IdentityToIntMap();
  private IdentityToIntMap snapshot;
  private IdentityMap /* ActorThread X Queue{ Object } */ queues = new IdentityMap();
  
  private int newActors = 0;
  private int actorsInLatch = 0;
  
  private Object counterLatch;
  
  
  public ActorContext( ActorContext previous )
  {
//    super();
    if(previous == null)
    {
      counterLatch = new Object();
    }
    else
    {
      if(!previous.actorPositions.equals(previous.snapshot))
      {
        throw new Error("ACTORS ARE NOT QUIESCENT IN THE PREVIOUS FRAME");
      }
      
      Iterator i = previous.actorPositions.keySet().iterator();
      
      while(i.hasNext())
      {
        Object o = i.next();
        int k = previous.actorPositions.get(o);
        if(k != ACTOR_FINISHED)
        {
          actorPositions.put(o, k);
          queues.put(o, previous.queues.get(o));
        }
      }
      
      counterLatch = previous.counterLatch;
    }
  }

  public void registerActor( ActorThread actor )
  {
    actorPositions.put( actor, 0 );
    queues.put(actor, new LinkedQueue());
    newActors++;
  }
/*
  private boolean isReadyToRelease()
  {
    return actorsInLatch == newActors;
  }
  */
  
  public void _ts_startActors()
  {
    Iterator i = actorPositions.keySet().iterator();
    
    while( i.hasNext() )
    {
      ((ActorThread)i.next()).start();
    }
    
    synchronized( counterLatch )
    {
      while( !(actorsInLatch == newActors) )
      {
        try
        {
          counterLatch.wait();
        }
        catch( InterruptedException e )
        {
          throw new Error("MAIN TEST THREAD IS INTERRUPTED WHILE WAITING FOR ACTORS START-UP");
        }
      }
    }
    
    synchronized( this )  // brought from _ts_releaseActors()
    {
      //initPerform();
      snapshot = (IdentityToIntMap)actorPositions.clone();
      //queues = new IdentityMap();
      
      //i = actorPositions.keySet().iterator();
      //while( i.hasNext() )
      //{
      //  queues.put( i.next(), new ArrayList() );
      //}
      
      notifyAll();
    }
  }
  
  public synchronized void reachCommonStart()
  {
    synchronized( counterLatch )
    {
      actorsInLatch++;
      counterLatch.notify();
    }
        
    try
    {
      wait();
    }
    catch( InterruptedException e )
    {
      //actorInterrupted();
      throw new Error("ACTOR THREAD IS INTERRUPTED WHILE WAITING IN START-UP LATCH");
    }
  }
  /*
  private void initPerform()
  {
    snapshot = (IdentityToIntMap)actorPositions.clone();
    queues = new HashMap();
    
    Iterator i = actorPositions.keySet().iterator();
    while( i.hasNext() )
    {
      queues.put( i.next(), new ArrayList() );
    }
  }

  public synchronized void _ts_releaseActors()
  {
    initPerform();
    
    notifyAll();
  }
  
  public void releaseAll( int timeOut )
  {
    initPerform();
    
    for( int i = 0; i < actorPositions.size(); i ++ )
    {
      try
      {
        Thread.sleep( timeOut );
      }
      catch( InterruptedException ie )
      {
        System.exit( 1 );
      }
      
      synchronized( this )
      {
        notify();
      }
    }
  }*/

  public void _ts_waitForFrameEnd( int timeout )
  {
    /*
    if( snapshot == null )
        System.exit( 1 );
    */
    
    //int i = 0;
    do
    {
      //i++;
      //TestManager.getRegistrar().traceMessage("waiting for timeout: " + i + "-th cycle");
      try
      {
        Thread.sleep( timeout );
      }
      catch( InterruptedException e )
      {
        throw new Error("MAIN TEST THREAD IS INTERRUPTED WHILE WAITING FOR QUIESCENCE");
      }
      
      synchronized( this )
      {
        if( actorPositions.equals( snapshot ) )
        {
          //TestManager.getRegistrar().traceMessage("result: " + i + " cycles");
          break;
        } 
        else
        {
          snapshot = (IdentityToIntMap)actorPositions.clone();
        }
      }
    }
    while( true );
  }

  // ----------------------------------------------------
  //   a c t o r   s t a t e s
  // ----------------------------------------------------
  
  public void _ts_shift()
  {
    Thread t = Thread.currentThread();
    
    synchronized( actorPositions )
    {
      actorPositions.put( t, actorPositions.get( t ) + 1 );
    }
  }
  
  public void finish()
  {
    synchronized( actorPositions )
    {
      actorPositions.put( Thread.currentThread(), ACTOR_FINISHED );
    }
  }
  /*
  public void actorInterrupted()
  {
    synchronized( actorPositions )
    {
      actorPositions.put( Thread.currentThread()
                              , new Integer( ACTOR_INTERRUPTED )
                              );
    }
  }*/
  
  // ----------------------------------------------------
  //   s e n d  /  r e c e i v e
  // ----------------------------------------------------
  
  public void send( Object dest, Object o )
  {
    LinkedQueue q = (LinkedQueue)queues.get( dest );
    
    try
    {
      q.put(o);
    }
    catch (InterruptedException e)
    {
      throw new Error("ACTOR THREAD INTERRUPTED WHILE SENDING AN EVENT");
    }
    /*
    synchronized( queues )
    {
        ArrayList dest_queue = getActorQueue( dest );
        
        synchronized( dest_queue )
        {
            dest_queue.add( m );
            dest_queue.notify();
        }
    }*/
  }
  
  public void sendAll( Object source, Object o )
  {
    Iterator i = queues.keys().iterator();
    
    while(i.hasNext())
    {
      Object t = i.next();
      
      if(t != source)
      {
        try
        {
          ((LinkedQueue)queues.get(t)).put(o);
        }
        catch (InterruptedException e)
        {
          throw new Error("ACTOR THREAD INTERRUPTED WHILE SENDING AN EVENT");
        }
      }
    }
    
    /*
    synchronized( queues )
    {
      Iterator ai = queues.values().iterator();
      
      while( ai.hasNext() )
      {
        ArrayList queue = (ArrayList)ai.next();
    
        synchronized( queue )
        {
          queue.add( o );
          queue.notify();
        }
      }
    }
    */
  }

  public Object receive( Object recipient )
  {
    LinkedQueue q = (LinkedQueue)queues.get( recipient );
    
    try
    {
      return q.take();
    }
    catch (InterruptedException e)
    {
      throw new Error("ACTOR THREAD INTERRUPTED WHILE RECEIVING AN EVENT");
    }
    
    /*
      ArrayList queue = getActorQueue( recipient );
      
      synchronized( queue )
      {
          while( queue.size() == 0 )
              try
              {
                  queue.wait();
              }
              catch( InterruptedException ie )
              {
                  System.exit( 1 );
              }
              
          return queue.remove( 0 );
      }
      */
  }
  /*
  private ArrayList getActorQueue( Object actor )
  {*/
      /*ArrayList queue = null;
      
      synchronized( queues )
      {
          if( !(queues.containsKey( actor )) )
          {
              queue = new ArrayList();
              queues.put( actor, queue );
          } else
              queue = (ArrayList)queues.get( actor );
      }
      
      return queue;
      */
   /*
      return (ArrayList)queues.get( actor );
  }*/
}