/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: EventPool.java,v 1.4 2008/08/20 12:56:26 almo Exp $
 * Created on 11.02.2005
 *
 */
package jatva.lang.synchronization;

import jatva.lang.ModelEvent;
import jatva.util.Utils;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;

/**
 * @author Victor Kuliamin
 *
 */
public class EventPool
{
  Hashtable<Object, LinkedList<ModelEvent>> items;
  ArrayList<Object> channelsList;

  public EventPool()
  {
    this(1);
  }

  public EventPool(int channels)
  {
    items = new Hashtable<Object, LinkedList<ModelEvent>>(channels);
    channelsList= new ArrayList<Object>(channels);
  }

  public void clear()
  {
      items.clear();
      channelsList.clear();
      size = 0;
  }

  public int size()
  {
    return size;
  }

  public int channels()
  {
    return channelsList.size();
  }

  public void addEvent(ModelEvent e)
  {
    Object channelId = e.getChannelId();
    synchronized( items )
    {
        if( !(items.containsKey( channelId )) )
        {
            channelsList.add( channelId );
            items.put( channelId, new LinkedList<ModelEvent>() );
        }

        items.get( channelId ).addLast( e );
        size ++;
    }
    /*//TestManager m = TestManager.getRegistrar();
    //m.traceMessage("!!!! addEvent: ( <" + e + "> , " + ch + " )");
    //m.traceMessage("!!!! addEvent: channels: " + channels);
    if(ch < 0) return;
    if(ch < channels)
    {
      //m.traceMessage("!!!! addEvent: adding immediately");
        items[ch].addLast(e);
    }
    else
    {
      //m.traceMessage("!!!! addEvent: extending channels");

      EventList[] newItems = new EventList[ch+1];
      System.arraycopy(items, 0, newItems, 0, channels);
      for(int i = channels; i < ch+1; i++)
        newItems[i] = new EventList();
      channels = ch+1;
      items = newItems;
      items[ch].addLast(e);
    }
    size++;*/
  }

  /*public void addEvent(ModelEvent e)
  {
    //if(TestManager.getDefaultEventDistribution())
    {
      EventList[] newItems = new EventList[channels+1];
      System.arraycopy(items, 0, newItems, 0, channels);
      newItems[channels] = new EventList();
      channels++;
      items = newItems;
      items[channels-1].addLast(e);
    }
    //else
    {
      //items[0].addLast(e);
    }
    size++;
  }*/

  /*public void removeEvents(int ch, int num)
  {
    if(ch < channels && ch >= 0)
    {
      EventList l = items[ch];
      if(l.size() > num)
      {
        for(int i = 0; i < num; i++) l.removeLast();
      }
      if(l.size() <= num) l.clear();
    }
  }*/

  public LinkedList<ModelEvent> getChannel(int i)
  {
    return items.get( channelsList.get( i ) );
  }

  @Override
public String toString()
  {
    StringBuffer r = new StringBuffer("{");
    for( Object channelId : items.keySet() )
    {
        r.append( Utils.endl );
        r.append(channelId);
        r.append( ":" );
        List<ModelEvent> events_in_channel = items.get( channelId );
        for( ModelEvent me : events_in_channel )
        {
            r.append( Utils.endl );
            r.append( "    " );
            r.append( me.toString( false ) );
        }
    }
    r.append( Utils.endl );
    r.append("}");
    return r.toString();
  }

  /*
  boolean containsKey(Object key);
  boolean containsValue(Object value);
  boolean contains(Object key, Object val);

  Object  get(Object key);

  Object  removeKey(Object key);
  boolean removeValue(Object val);
  boolean remove(Object key, Object val);

  Object  put(Object key, Object value);

  Object clone();
  String toString();
  boolean equals(Object o);
  boolean equalsByRefs(Object o);*/

  private int size = 0;
  // private EventList[] items;

//  public void unboundAllEvents()
//  {
//      for( LinkedList<ModelEvent> events :items.values() )
//      {
//          for( ModelEvent me : events )
//              me.unbound();
//      }
//  }
}
