/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: Serializer.java,v 1.13 2008/06/17 17:06:10 almo Exp $
 * Created on 15.02.2005
 *
 */
package jatva.lang.synchronization;

import jatva.lang.ModelEvent;
import jatva.lang.MultipleCorrectStatesException;
import jatva.lang.NoCorrectSeriesException;
import jatva.lang.services.DeepOperationManager;
import jatva.util.IdentityMap;
import jatva.util.IntList;
import jatva.util.Utils;

import java.util.Arrays;
import java.util.Iterator;

/**
 * @author Victor Kuliamin
 *
 */
public class Serializer
{
  //private int       timeDelta = 100;
  
  private boolean seekAll  = true;//false;//
  private boolean printAll = false;

  private ModelEvent[] sequence = null;
  // the index of the leftmost null element in sequence array.
  private int seq_length = 0;

  private EventPool events   = null;
  private IntList   stack    = new IntList();
  private int[]     used     = null;

  private IdentityMap storedState = null;      
  private IdentityMap endState    = null;      
  
  public Serializer() {}
  
  public void setSeekAllSeries(boolean s)
  {
    seekAll = s;
  }
  
  public boolean performAnalysis(EventPool p, IdentityMap s)
  {
    events = p;
    sequence = new ModelEvent[ events.size() ];
    Arrays.fill( sequence, null );
    seq_length = 0;
    used = new int[events.channels()]; 
    TestManager man = TestManager.getRegistrar();
    
    int num = 0;
    
//    man.traceModelMessage("serializer: previous event at start: " + previous);
    //man.traceMessage("serializer: number of channels: " + events.channels());
    //man.traceMessage("serializer: number of events: " + events.size());
    man.traceModelMessage("serializer: events: " + Utils.endl + events.toString());
    
    storeState(s);
    
    if(!seekAll)
    {
      if(makeNextSequence(s))
      {
        man.traceModelMessage(printStackWithPrefix("serializer: found series:"));
        previous = null;
        stack.clear();
        return true;
      }
      else
      {
        man.traceModelMessage("serializer: no correct series found");
        previous = null;
        stack.clear();
        
        throw new NoCorrectSeriesException();
        //return false;  
      }
    }
    else
    {
      if(makeNextSequence(s))
      {
        man.traceModelMessage(printStackWithPrefix("serializer: found " + (++num) + " series:"));
        storeEndState(s);
        
        previous = null;
        restoreState(s);
        while(makeNextSequence(s))
        {
          man.traceModelMessage(printStackWithPrefix("serializer: found " + (++num) + " series:"));
          if(!compareStates(s))
          {
            man.traceModelMessage("serializer: found several correct series with different end states");
            
            throw new MultipleCorrectStatesException();
          }
          restoreState(s);
          previous = null;
        }
        
        restoreEndState(s);

        man.traceModelMessage("serializer: no more series");
        previous = null;
        
        return true;
      }
      else
      {
        man.traceModelMessage("serializer: no correct series found");
        previous = null;
        
        throw new NoCorrectSeriesException();
        //return false;
      }
    }
  }
  
  public void storeState(IdentityMap s)
  {
    storedState = new IdentityMap();
    for(Iterator i = s.keys().iterator(); i.hasNext(); )
    {
      DeepOperationManager.getDeepClone(i.next(), storedState);
    }
    
    //TestManager.getRegistrar().traceTestMessage("stored state : " + DeepOperationManager.getDeepString(storedState));
  }
  
  public void restoreState(IdentityMap s)
  {
    //TestManager.getRegistrar().traceTestMessage("current state : " + DeepOperationManager.getDeepString(s));

    for(Iterator i = s.keys().iterator(); i.hasNext(); )
    {
      DeepOperationManager.restoreFromClone(i.next(), storedState);
    }

    //TestManager.getRegistrar().traceTestMessage("restored state : " + DeepOperationManager.getDeepString(s));
  }
  
  public void storeEndState(IdentityMap s)
  {
    endState = new IdentityMap();
    for(Iterator i = s.keys().iterator(); i.hasNext(); )
    {
      DeepOperationManager.getDeepClone(i.next(), endState);
    }
    
    //TestManager.getRegistrar().traceTestMessage("end state : " + DeepOperationManager.getDeepString(endState));
  }
  
  public void restoreEndState(IdentityMap s)
  {
    //TestManager.getRegistrar().traceTestMessage("current state : " + DeepOperationManager.getDeepString(s));

    for(Iterator i = s.keys().iterator(); i.hasNext(); )
    {
      DeepOperationManager.restoreFromClone(i.next(), endState);
    }

    //TestManager.getRegistrar().traceTestMessage("restored state : " + DeepOperationManager.getDeepString(s));
  }
  
  public boolean compareStates(IdentityMap s)
  {
    return DeepOperationManager.areStatesDeeplyEqual(s, endState);
  }
  
  public boolean constructSequence(IdentityMap st)
  {
    boolean res = false;
    IntList stack = new IntList();

    // can process < 1024 events in one channel
    
    int position = 0;
    int t        = 0;
    int i = 0, num = 0;
    boolean tmp = false;
    
    for(i = 0; i < events.channels(); i++) used[i] = -1;
    
    if(!seekAll)
    {
      while(position < events.size())
      {
        ModelEvent e = events.getChannel(channel(t)).get(index(t));
        sequence[ seq_length ++ ] = e;
        tmp = (checkSequence(st) == -1);
        
        if(tmp)
        {
          stack.addLast(t);
          position++;
          used[channel(t)]++;
          
          i = getNext(-1, events.channels(), events, used);
          
          if(i < 0 && position == events.size()) break;

          s2:while(i < 0 && position > 0)
          {
            sequence[ seq_length -- ] = null;
            stack.removeLast();
            position--;
            used[channel(t)]--;
            
            i = getNext(t, events.channels(), events, used);

            if(i >= 0) break s2;
            t = stack.last();
          }
          
          if(position <= 0) return false;
          else              t = i;
        }
        else
        {
          sequence[ seq_length -- ] = null;
          i = getNext(t, events.channels(), events, used);
          
          while(i < 0 && position > 0)
          {
            sequence[ seq_length -- ] = null;
            stack.removeLast();
            position--;
            used[channel(t)]--;
            t = stack.last();
            
            i = getNext(t, events.channels(), events, used);
          }
          
          if(position <= 0) return false;
          else              t = i;
        }
      }
    }
    else // seekAll == true
    {
      while(true)
      {
        while(position < events.size())
        {
          ModelEvent e = events.getChannel(channel(t)).get(index(t));
          sequence[ seq_length ++ ] = e;
          tmp = (checkSequence(st) == -1);
          
          if(tmp)
          {
            stack.addLast(t);
            position++;
            used[channel(t)]++;
            i = getNext(-1, events.channels(), events, used);
            
            if(i < 0/*i == channels*/ && position == events.size())
            {
              break;
            }
            
            s2:while(i < 0/*i >= channels*/ && position > 0)
            {
              sequence[ seq_length -- ] = null;
              stack.removeLast();
              position--;
              used[channel(t)]--;
              
              i = getNext(t, events.channels(), events, used);
              if(i >= 0) break s2;
              t = stack.last();
            }
            
            if(position <= 0 && i < 0/*i >= channels*/)
            {
              return false;
            }
            else
            {
              t = i;//(i << 10) + j;
            }
          }
          else
          {
            sequence[ seq_length -- ] = null;
            i = getNext(t, events.channels(), events, used);

            while(i < 0/*i >= channels*/ && position > 0)
            {
              t = stack.last();
              sequence[ seq_length -- ] = null;
              stack.removeLast();
              position--;
              used[channel(t)]--;

              i = getNext(t, events.channels(), events, used);
            }
            
            if(position <= 0 && i < 0/*i >= channels*/)
            {
              return false;
            }
            else
            {
              t = i;//(i << 10) + j;
            }
          }
        }
      
        System.out.println("" + (++num) + " sequence: " + sequence);
        i = -1;//channels;
        
        while(i < 0/*i >= channels*/ && position > 0)
        {
          t = stack.last();
          sequence[ seq_length -- ] = null;
          stack.removeLast();
          position--;
          used[channel(t)]--;
          i = getNext(t, events.channels(), events, used);
        }
        
        if(position <= 0 && i < 0/*i >= channels*/)
        {
          return false;
        }
        else
        {
          t = i;//(i << 10) + j;
        }
      }
    }
    
    return res;
  }
  
  
  
  public boolean makeNextSequence(IdentityMap st)
  {
    // can process <= 65536 channels and events in one channel
    int i = 0, j = 0, e = -1;
    
    
    if(stack.isEmpty())
    {
      //TestManager man = TestManager.getRegistrar();
      //man.traceMessage("makeNextSequence: found " + events.channels() + " channels");
      for(i = 0; i < events.channels(); i++)
      {
        used[i] = events.getChannel(i).size()-1;
        //man.traceMessage("makeNextSequence: in " + i + "-th channel found " + events.getChannel(i).size() + " events");
        for(j = 0; j < events.getChannel(i).size(); j++)
          stack.addLast((i << 16) + j);
      }
      
      if((e = checkSequence(st)) == -1)
      {
        return true;
      }
      
      
      e++;
    }
    else e = events.size(); 
    
    while(true)
    {
      do
      {
        e--;
        for(i = stack.size()-1; i >= e; i--)
        { 
          j = stack.last() >> 16;
          used[j]--; 
          stack.removeLast();
        }
        
        for(i = j + 1; i < events.channels(); i++)
          if(used[i] + 1 < events.getChannel(i).size()) break;
      }
      while(e != -1 && i == events.channels());
      
      if(e == -1)
      {
        return false;
      }
  
      while(true)
      {
        
        stack.addLast((i << 16) + (++used[i]));
          
        e++;
        if(e >= events.size())
        {
          if((e = checkSequence(st)) == -1)
          {
            
            return true;
          }
          else 
          { 
            e++; 
            break; 
          }
        }
        else
        {
          for(i = 0; i < events.channels(); i++)
            if(used[i] + 1 < events.getChannel(i).size()) break;
          
          if(i == events.channels()) { e--; break; }
        }
      }
    }
  }
  
  public final static int channel(int t) { return t >> 16; }
  public final static int index  (int t) { return t & 65535; }
  public final static int composite(int t, int m) { return (t << 16) + m; }
  
  private final int getNext(int t, int channels, EventPool events, int[] used)
  {
    for(int i = (t < 0)?0:(channel(t)+1); i < channels; i++)
      if(used[i] + 1 < events.getChannel(i).size()) return (i << 10) + used[i] + 1;
    
    return -1;
  }
  
  public int checkSequence(IdentityMap s)
  {
    int t;
    ModelEvent e;
    TestManager mgr = TestManager.getRegistrar();

    mgr.serializationStart();

    //sequence.clear();
    /*
    StringBuffer str = new StringBuffer("checking sequence: ");
    for(int i = 0; i < stack.size(); i++)
    {
      t = stack.get(i);
      str.append(" [" + channel(t) + ", " + index(t) + "] : <" + events.getChannel(channel(t)).get(index(t)) + ">");
    }*/
//    str.append(System.getProperty("line.separator"));
    mgr.traceModelMessage(printStackWithPrefix("serializer: checking sequence:"));
        
    for(int i = 0; i < stack.size(); i++)
    {
      t = stack.get(i);
      e = events.getChannel(channel(t)).get(index(t));
      //sequence.addLast(e);
      
      if(    stack.size() == events.size()
          && i == ( stack.size() - 1 )
        )
      {
          mgr.eventIsTheLastInSequence();
      }
      mgr.setCurrentEvent(e);
      
//      mgr.traceModelMessage("previous: " + e);
      
      if(previous != null && !checkImmediateResult(e))
      {
        previous = null;
        restoreState(s);
        mgr.removeCurrentEvent();
        
        mgr.serializationEnd();
        return i;
      }
      else if(previous != null)
      {
        previous = null;
      }
      else if(!e.check()) 
      {
        restoreState(s);
        mgr.removeCurrentEvent();

        mgr.serializationEnd();
        return i;
      }

//      mgr.traceModelMessage("previous: " + e);
      
      mgr.removeCurrentEvent();
    }
    
    if(previous != null) 
    { 
      previous = null; 
      
      mgr.serializationEnd();
      return stack.size()-1; 
    }
    
    mgr.serializationEnd();
    return -1;

    /*
    boolean flag31 = false;
    boolean flag1_1 = false;
    boolean flag21_01 = false;
    
    for(int i = 0; i < sequence.size(); i++)
    {
      if(!flag31 && sequence.get(i).kind().startsWith("1"))
        return i;
      else if(sequence.get(i).kind().startsWith("3"))
        flag31 = true;
      
      if(!flag1_1 && sequence.get(i).kind().endsWith("1"))
        return i;
      else if(sequence.get(i).kind().startsWith("1"))
        flag1_1 = true;
      
      if(!flag21_01 && sequence.get(i).kind().startsWith("01"))
        return i;
      else if(sequence.get(i).kind().startsWith("21"))
        flag21_01 = true;
    }
    
    if(sequence.size() > 2 && !sequence.get(2).kind().startsWith("21")) return 2;
    
    if(sequence.size() > 5 && !sequence.get(5).kind().startsWith("10")) return 5;
    */
  }
  
  private ModelEvent previous = null;
  
  public void markImmediateResult(ModelEvent e)
  {
    previous = e;
  }
  
  private boolean checkImmediateResult(ModelEvent e)
  {
    if(e.getManager() != previous.getManager())
    {
      TestManager.getRegistrar().traceModelMessage("NO!!! Next event is not the resulting event of the previous call");
      return false;
    }
  
    TestManager.getRegistrar().traceModelMessage("YES!!! Next event is the resulting event of the previous call");
    return true;
  }
  
  private String printStackWithPrefix(String prefix)
  {
    StringBuffer s = new StringBuffer(30);
    
    s.append(prefix);
    s.append( Utils.endl );
    s.append( "{" );
    int t;
    for(int i = 0; i < stack.size(); i++)
    {
      t = stack.get(i);
      s.append( Utils.endl );
      s.append("[");
      s.append(channel(t));
      s.append(", ");
      s.append(index(t));
      s.append( "] :" );
      s.append( Utils.endl );
      s.append( "    " );
      s.append( events.getChannel(channel(t)).get(index(t)).toString( true ) );
    }
    s.append( Utils.endl );
    s.append( "}" );
    return s.toString();
  }
    /*
  public static void xmain(String[] args)
  {
    Serializer s = new Serializer();
    EventPool     m = new EventPool(4);
    
    m.addEvent(new ModelEvent("00"), 0);
    m.addEvent(new ModelEvent("01"), 0);
    m.addEvent(new ModelEvent("02"), 0);
    m.addEvent(new ModelEvent("10"), 1);
    m.addEvent(new ModelEvent("11"), 1);
    m.addEvent(new ModelEvent("20"), 2);
    m.addEvent(new ModelEvent("21"), 2);
    m.addEvent(new ModelEvent("30"), 3);
    
    System.out.println(m.toString());
    
    s.init(m);
    
    s.constructSequence();
    
  }*/

  /*
  public static void ymain(String[] args)
  {
    Serializer s = new Serializer();
    EventPool     m = new EventPool(4);
    
    m.addEvent(new ModelEvent("00"), 0);
    m.addEvent(new ModelEvent("01"), 0);
    m.addEvent(new ModelEvent("02"), 0);
    m.addEvent(new ModelEvent("10"), 1);
    m.addEvent(new ModelEvent("11"), 1);
    m.addEvent(new ModelEvent("20"), 2);
    m.addEvent(new ModelEvent("21"), 2);
    m.addEvent(new ModelEvent("30"), 3);
    
    System.out.println(m.toString());
    
    s.init(m);
    
    int num = 0;
    while(s.makeNextSequence())
    {
      System.out.println("" + (++num) + " sequence: " + s.sequence);
    }
    
    System.out.println("no more sequences");
  }
  */
}