/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: Test.java,v 1.20 2008/08/07 19:17:36 almo Exp $
 * Created on 25.05.2005
 *
 */
package jatva.lang.synchronization;

import jatva.iterators.lang.IObjectPairIterator;
import jatva.lang.ModelEvent;
import jatva.lang.coverage.SEOperationInfo;
import jatva.lang.services.DeepOperationManager;
import jatva.tracer.Tracer;
import jatva.util.IdentityMap;
import jatva.util.IdentitySet;
import jatva.util.IdentityToIntMap;
import jatva.util.WeakIdentityMap;

/**
 *
 * @author Victor Kuliamin
 *
 */
public class Test
{
  private static WeakIdentityMap map = new WeakIdentityMap(); // ThreadGroup -> Test
  private static int             counter = 1;                 // Test counter

  //------------------------------------------------------
  // Test modes
  public static final int SYNCHRONOUS      = 0; // events are analysed in the sequence of their detection
  public static final int WEAK_SYNCHRONOUS = 1; // synchronous until events come from different channels or two calls mix with event, then it becomes asynchronous
  public static final int ASYNCHRONOUS     = 2; // only first call preconditions are checked
                                                // currently this mode is equivalent to being inside some test frame
  public static final int CONCURRENT       = 3; // no analysis of events until frame end

  public static final int EXECUTING        = 0; // executing a test frame
  public static final int CHECKING         = 1; // checking results of a test frame

  public static final long DEFAULT_FRAME_TIMEOUT = 100;

  private int mode                         = SYNCHRONOUS;
  private int frameExecutionMode           = SYNCHRONOUS;

  private int  status       = EXECUTING;
  private long frameTimeout = DEFAULT_FRAME_TIMEOUT;

  private int              id         = counter++;              // test id
  private Serializer       serializer = new Serializer();       // serializer, or composite oracle
  private EventPool        eventPool  = new EventPool();        // event pool
//  private IdentityToIntMap frames     = new IdentityToIntMap(); // registered enterings in test frame
  private IdentityMap      state      = new IdentityMap();      // model state to store
  private IdentityToIntMap managers   = new IdentityToIntMap(); // test managers with their ids
  private TestManager      mainThreadManager;
  private Tracer           tracer;
  private int              manCounter = 1;                      // test managers' counter to create their ids
  //private Latch            frameLatch = null;                   // Latch to lock exit from test frames

  private IdentitySet      objects = new IdentitySet();

  private SynchronizationManager smanager = new SynchronizationManager();

  //private StringToObjectMap calledOperations = new StringToObjectMap();  // Map of class names to sets of called operations in all test threads

  private IdentitySet      calledOperations = new IdentitySet();  // Set of model called operations in all test threads

  public final static int CERTAINTY_BASED_ON_MODEL_CALLS = 1;
  public final static int CERTAINTY_BASED_ON_FRAME_STARTS    = 2;

  // Flag of actual transition (transitions that are not actual are faked)
  boolean      transitionPerformed = true;

  private int certaintyMode =   CERTAINTY_BASED_ON_MODEL_CALLS
                              | CERTAINTY_BASED_ON_FRAME_STARTS
                            ;

  /**
   * Stores messages generated during single transition.
   * Outside transitions this buffer must be empty.
   */
  private TraceMessagesBuffer traceMessageBuffer = new TraceMessagesBuffer();

  private Test()
  {
      tracer = Tracer.createTracerObject();
      tracer.setTestId( id );
      traceMessageBuffer.switchToImmediateTracing();
  }

  // ------------------------------------------------------------------
  //   A C C E S S O R S
  public synchronized boolean isTransitionPerformed()
  { return transitionPerformed; }

  public synchronized void setTransitionPerformed( boolean transitionPerformed )
  { this.transitionPerformed = transitionPerformed; }

  public void setCertaintyBasedOnModelCalls()
  { certaintyMode = CERTAINTY_BASED_ON_MODEL_CALLS; }

  public void setCertaintyBasedOnFrameStarts()
  { certaintyMode = CERTAINTY_BASED_ON_FRAME_STARTS; }

  public synchronized int getCertaintyMode()
  { return certaintyMode; }

  public Tracer getTracer()
  { return tracer; }

  public synchronized boolean isCalledOperation(SEOperationInfo op)
  {
    if( calledOperations.contains( op ) )
    {
        return true;
    } else {
        calledOperations.add( op );
        return false;
    }
//    IdentitySet is;
//
//    if(o == null)
//    {
//      is = new IdentitySet();
//      is.add(op);
//      calledOperations.put(cls, is);
//
//      return false;
//    }
//    else
//    {
//      is = (IdentitySet)o;
//      if(is.contains(op))
//        return true;
//      else
//      {
//        is.add(op);
//        return false;
//      }
//    }
  }

  /*
   * TO DO: Finish model event trace buffering
   *
  private TestManager modelEventTracingManager = null;  // Test manager currently tracing model event
  private IdentityList waitingModelEventTracingManagers = new IdentityList(); // Tests managers waiting to trace model events

  public synchronized boolean tryTracingModelEvent(TestManager m)
  {
    if(modelEventTracingManager == null)
    {
      modelEventTracingManager = m;
      return true;
    }
    else
    {
      waitingModelEventTracingManagers.addLast(m);
      return false;
    }
  }


  public synchronized void removeTracingModelEvent(TestManager m)
  {
    if(modelEventTracingManager == m)
    {
      modelEventTracingManager = null;
    }
  }
  */

  static Test getInitTest(TestManager m)
  {
    ThreadGroup g = Thread.currentThread().getThreadGroup();
    Test res;

    synchronized(Test.class)
    {
      res = (Test)map.get(g);

      if(res == null)
      {
        res = new Test();
        res.mainThreadManager = m;
        map.cput(g, res);
      }
    }

    res.managers.put(m, res.manCounter++);

    //System.out.println("" + Integer.toHexString(System.identityHashCode(res)) + " test is assigned");

    return res;
  }

  public static Test getTest()
  {
    ThreadGroup g = Thread.currentThread().getThreadGroup();

    synchronized(Test.class)
    {
      return (Test)map.get(g);
    }
  }

  /**
   *
   * @pre   m - subordinate manager of the test
   * @param m
   * @return
   */
  int getManagerId(TestManager m)
  {
    return Serializer.composite(id, managers.get(m));
  }

  int getLocalId(TestManager m)
  {
    return managers.get(m);
  }

  public int getId()
  {
    return id;
  }

  /**
   * @return Returns the mode.
   */
  public int getMode ()
  {
    return mode;
  }
  /**
   * @param mode The mode to set.
   */
  public void setMode (int mode)
  {
    this.mode = mode;
  }

  public int getStatus()
  {
    return status;
  }

  int getManagerDefaultChannel(TestManager m)
  {
    return managers.get(m);
  }
  /*
  Latch getFrameLatch()
  {
    return frameLatch;
  }*/

  SynchronizationManager getSynchronizationManager()
  {
    return smanager;
  }

  void registerEvent(ModelEvent e)
  {
    synchronized(this)
    {

      // TODO: ?????????
      if(   !transitionPerformed
          && ( certaintyMode & CERTAINTY_BASED_ON_MODEL_CALLS ) != 0
          && !traceMessageBuffer.isEmpty()
        )
      {
        // here non-emptiness of message buffer is equivalent to
        // being inside some transition.
        // not inside a 'state'-block for example.

        transitionPerformed = true;
        if( mode == SYNCHRONOUS )
        {
          traceMessageBuffer.switchToImmediateTracing();
        }
      }

      //m.getTracer().traceMessage(m, new StringBuffer("test: registering event <").append(e).append("> in ").append(managers.get(m)).append("-th channel"));
      if(mode == ASYNCHRONOUS)
      {
        eventPool.addEvent(e);
      }
      //m.getTracer().traceMessage(m, new StringBuffer("test: events: ").append(eventPool));
    }
  }

  void registerFrameStorage(Object r)
  {
    if( frameExecutionMode == ASYNCHRONOUS )
    {
        // frames are not "physically" opened if frame mode is synchronous
        // see 'TestManager.registerFrameStorage'
        if(    ( ( certaintyMode & CERTAINTY_BASED_ON_FRAME_STARTS ) != 0 )
            && !transitionPerformed
          )
        {
          transitionPerformed = true;
//          traceTransitionDetails();
        }

        mode = ASYNCHRONOUS;

        // assume that only main scenario thread works here !!!
        DeepOperationManager.getDeepClone(r, state);
    }
  }
  /*
  void registerFrame(TestManager m)
  {
    /*
    int c = 0;
    synchronized(this)
    {
      //if(r != null) r.store(state);
      m.getTracer().traceTestFrameEnter(m);

     /*
      if(frames.isEmpty())
      {
//        frameLatch = new Latch();       // init new latch for this frame
        frames.put(m, 1);
      }
      else if((c = frames.get(m)) != 0)
      {
        frames.put(m, c+1);
      }
      else
      {
        frames.put(m, 1);
      }
    }
  }*/
  /*
  boolean unregisterFrame(TestManager m)  // returns true if the thread exits the frame last time or not entered
  {
    int c = 0;
    synchronized(this)
    {
      if((c = frames.get(m)) > 1)
      {
        frames.put(m, c-1);
        return false;
      }
      else if(c == 1)
      {
        frames.remove(m);

        if(frames.isEmpty())
        {
          m.getTracer().traceSerializationStart(this);

          status = CHECKING;

          serializer.performAnalysis(eventPool, state);
          eventPool.clear();

          status = EXECUTING;
          m.getTracer().traceTestFrameEnd(this);
          //frameLatch.release();                  // release all the threads wating for the test frame end
        }
      }
    }
    return true;
  }*/

  void closeFrame()
  {
    if(mode == ASYNCHRONOUS)
    {
      try
      {
        Thread.sleep(frameTimeout);
      }
      catch(InterruptedException e)
      {}

      mainThreadManager.prepareMessageBuffers();

      synchronized(this)
      {
        //m.getTracer().traceSerializationStart(m);

        status = CHECKING;

        serializer.performAnalysis(eventPool, state);
        eventPool.clear();

        status = EXECUTING;

        mode = SYNCHRONOUS;
      }
    }
    else
    {
      //m.getTracer().traceTestFrameEnd(m);
    }

    // outside frame bounds everything in test goes synchronously
  }

  void markImmediateResult(ModelEvent e)
  {
    serializer.markImmediateResult(e);
  }

  private class TracerFinalizer extends Thread
  {
    public TracerFinalizer () {}

    @Override
    public void run ()
    {
      tracer.finishWriter();
//      IdentitySet s = managers.keySet();
//      for (IObjectIterator i = s.iterator(); !i.stop(); i.shift())
//      {
//        TestManager m = (TestManager)i.value();
//        m.traceTestMessage("finishing writer");
//        m.getTracer().finishWriter();
//      }
    }
  }

  {
    Runtime.getRuntime().addShutdownHook(new TracerFinalizer());
  }

  //int quiescenceTimeout = 1000; // milliseconds

  public long getFrameTimeout ()
  {
    return frameTimeout;
  }

  public void setFrameTimeout (long millis)
  {
    frameTimeout = millis;
  }

  public boolean registerObject (Object o)
  {
    IObjectPairIterator i = map.iterator();
    for(; !i.stop(); i.shift())
    {
      Test t = (Test)i.value2();
      if(t.objects.contains(o)) return (t == this);

    }

    objects.add(o);
    return true;
  }

  public int getFrameExecutionMode()
  {
    return frameExecutionMode;
  }

  public void setFrameExecutionMode( int frameExecutionMode )
  {
    this.frameExecutionMode = frameExecutionMode;
  }

  public void setSeekAllSeries( boolean s )
  {
      serializer.setSeekAllSeries( s );
  }

  public TraceMessagesBuffer getTraceMessageBuffer()
  {
    return traceMessageBuffer;
  }

  public boolean isManagerTheMain( TestManager tm )
  { return mainThreadManager == tm; }

  public TestManager getMainThreadManager()
  {
    return mainThreadManager;
  }

  public IdentityToIntMap getManagers()
  {
    return managers;
  }

  public Serializer getSerializer()
  {
      return serializer;
  }
}
