/*
 * Copyright (c) 2000-2003 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 * TraceConstants
 * $Id: TraceConstants.java,v 1.5 2006/04/21 09:56:03 kuliamin Exp $
 *
 * Author: Alexei Barantsev <barancev@ispras.ru>
 * Created on June 20, 2003
 */
//=============================================================================

package jatva.tracer;

//=============================================================================

//=============================================================================

/**
 * This class contains constants related to trace format. They can (and should)
 * be used not only by tracer but by trace analyzers too.
 * 
 * @author <a href="mailto:barancev@ispras.ru">Alexei Barantsev</a>
 */
public final class TraceConstants {

    /**
     * This class is not instantiable. It contains only static constants.
     */
    private TraceConstants() {}

    public final static String NO_SCENARIO = "<empty>";
    public final static String UNKNOWN_HOST = "[unknown host]";

    public final static String TRACE = "trace";
    public final static String THREAD_ID_ATTR = "trace";

    public final static String EXCEPTION = "exception";
    public final static String EXCEPTION_INTERIM_ATTR = "interim";
    public final static String EXCEPTION_INTERNAL_ATTR = "internal";
    public final static String EXCEPTION_WHERE = "where";
    public final static String EXCEPTION_INFO = "info";
    public final static String EXCEPTION_BACKTRACE = "backtrace";

    public final static String INFO = "info";
    public final static String USER_INFO = "user_info";
    public final static String USER_INFO_MAP = "user_info_map";
    public final static String PROPERTY = "property";

    public final static String KIND_ATTR = "kind";
    public final static String TYPE_ATTR = "type";
    public final static String NAME_ATTR = "name";
    public final static String VALUE_ATTR = "value";

    public final static String SIGNATURE_ATTR = "signature";

    public final static String SCENARIO_START = "scenario_start";
    public final static String SCENARIO_NAME_ATTR = "name";
    public final static String SCENARIO_START_TIME_ATTR = "time";
    public final static String SCENARIO_ID_ATTR = "id";
    public final static String HOST_ATTR = "host";
    public final static String OS_ATTR = "os";
    public final static String JVM_ATTR = "jvm";

    public final static String TRACE_ENVIRONMENT = "environment";
    public final static String ENV_NAME_ATTR = "name";
    public final static String ENV_VALUE_ATTR = "value";

    public final static String SCENARIO_END = "scenario_end";
    public final static String SCENARIO_END_TIME_ATTR = "time";

    public final static String SCENARIO_VALUE = "scenario_value";

    public final static String STATE = "state";
    public final static String STATE_ID_ATTR = "id";

    public final static String TRANSITION_START = "transition_start";
    public final static String TRANSITION_ID_ATTR = "id";
    public final static String TRANSITION_END = "transition_end";
    public final static String TRANSITION_CERTAIN_ATTR = "certain";

    public final static String JUMP_START = "jump_start";
    public final static String JUMP_KIND_ATTR = "kind";
    public final static String JUMP_END = "jump_end";

    public final static String SERIALIZATION_START = "serialization_start";
    public final static String SERIALIZATION_END = "serialization_end";

    public final static String MODEL_OBJECT = "model_object";
    public final static String MODEL_OBJECT_ID_ATTR = "id";
    public final static String MODEL_CLASS = "model_class";
    public final static String MODEL_CLASS_NAME_ATTR = "name";
    public final static String MODEL_CALL_START = "model_operation_start";
    public final static String MODEL_KIND_ATTR = "kind";
    public final static String MODEL_SIGNATURE_ATTR = "signature";
    public final static String MODEL_REFID_ATTR = "refid";
    public final static String MODEL_TEST_AGENT_ID_ATTR = "test_agent_id";
    public final static String MODEL_TEST_AGENT_EVENT_ID_ATTR = "test_agent_event_id";
    public final static String MODEL_CHANNEL_ID_ATTR = "channel_id";
    public final static String MODEL_CALL_END = "model_operation_end";

    public final static String MODEL_VALUE = "model_value";

    public final static String ORACLE_START = "oracle_start";
    public final static String ORACLE_CLASS_ATTR = "class_name";
    public final static String ORACLE_SIGNATURE_ATTR = "signature";
    public final static String ORACLE_REF_ATTR = "ref";
    public final static String ORACLE_END = "oracle_end";
    public final static String PRECONDITION_END = "precondition_end";
    public final static String POSTCONDITION_START = "postcondition_start";

    public final static String INVARIANT_START = "invariant_start";
    public final static String INVARIANT_SIGNATURE_ATTR = "signature";
    public final static String INVARIANT_END = "invariant_end";

    public final static String COVERAGE_STRUCTURE = "coverage_structure";
    public final static String COVERAGE_FORMULAE = "formulae";
    public final static String COVERAGE_FORMULA = "formula";
    public final static String COVERAGE_FORMULA_ID_ATTR = "id";
    public final static String COVERAGE_BRANCHES = "branches";
    public final static String COVERAGE_BRANCH = "branch";
    public final static String COVERAGE_BRANCH_ID_ATTR = "id";
    public final static String COVERAGE_MPATHS = "marked_paths";
    public final static String COVERAGE_MPATH = "marked_path";
    public final static String COVERAGE_MPATH_ID_ATTR = "id";
    public final static String COVERAGE_PREDICATES = "predicates";
    public final static String COVERAGE_PREDICATE = "predicate";
    public final static String COVERAGE_PREDICATE_ID_ATTR = "id";
    public final static String COVERAGE_DISJUNCTS = "disjuncts";
    public final static String COVERAGE_DISJUNCT = "disjunct";
    public final static String COVERAGE_DISJUNCT_ID_ATTR = "id";
    public final static String COVERAGE_USER = "coverage";
    public final static String COVERAGE_USER_ID_ATTR = "id";
    public final static String COVERAGE_USER_EL = "element";
    public final static String COVERAGE_USER_EL_ID_ATTR = "id";
    public final static String COVERAGE_USER_EL_NAME_ATTR = "name";

    public final static String BRANCH = "branch";
    public final static String DEFERRED_BRANCH = "deferred_branch";
    public final static String BRANCH_NAME_ATTR = "name";
    public final static String MARK = "mark";
    public final static String MARK_NAME_ATTR = "name";
    public final static String PRIME_FORMULA = "prime_formula";
    public final static String FORMULA_ID_ATTR = "id";
    public final static String FORMULA_VALUE_ATTR = "value";
    public final static String COVERAGE_ELEMENT = "coverage_element";
    public final static String COVERAGE_NAME_ATTR = "coverage";
    public final static String COVERAGE_EL_ID_ATTR = "id";

    public final static String IMPL_OBJECT = "implementation_object";
    public final static String IMPL_OBJECT_ID_ATTR = "id";
    public final static String IMPL_CLASS = "implementation_class";
    public final static String IMPL_CLASS_NAME_ATTR = "name";
    public final static String IMPL_CALL_START = "implementation_operation_start";
    public final static String IMPL_SIGNATURE_ATTR = "signature";
    public final static String IMPL_CALL_END = "implementation_operation_end";
    public final static String IMPL_WORKING_TIME = "implementation_working_time";
    public final static String IMPL_START_TIME_ATTR = "start";
    public final static String IMPL_END_TIME_ATTR = "end";
    public final static String IMPL_VALUE = "implementation_value";

}

//=============================================================================
// End of TraceConstants.java