/*
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 * AbstractFormatter
 * $Id: AbstractFormatter.java,v 1.8 2008/07/25 09:58:04 almo Exp $
 *
 * Author: Alexei Barantsev <barancev@ispras.ru>
 * Created on March 27, 2003
 */
//=============================================================================

package jatva.tracer.impl;

//=============================================================================

import jatva.lang.services.DeepOperationManager;
import jatva.tracer.TraceFormatter;
import jatva.tracer.TracerInternalError;
import jatva.util.Utils;

//=============================================================================

/** This class formats each message as an xml tag that contains message data
 * in attributes and/or in the body.
 *
 * @author  <a href="mailto:barancev@ispras.ru">Alexei Barantsev</a>
 * @version 1.0
 */
public abstract class AbstractFormatter
  implements TraceFormatter
{
  @Override
public AbstractFormatter clone()
  {
    try
    {
      return (AbstractFormatter)super.clone();
    }
    catch (CloneNotSupportedException e)
    {
      // Never should happen
      throw new TracerInternalError("Class " + this.getClass().getName() + " cannot clone formatter", e);
    }
  }

  public void setModelObjectsFormatDepth(int l)
  {
    modelObjectsTracingDepth = l;
  }

  public void setImplementationObjectsFormatDepth(int l)
  {
    implementationObjectsTracingDepth = l;
  }

  public void setTestId( int testId )
  {
    this.testId = testId;
  }

  protected int modelObjectsTracingDepth = 0;
  protected int implementationObjectsTracingDepth = 1;
  protected int testId = 0;

  protected String printObjectValue( boolean isPrimitive, boolean isModelObject, Object value )
  {
      if( isPrimitive )
          return "" + value;
      else
          return printObjectValue( isModelObject, value );
  }

  protected String printObjectValue( boolean isModelObject, Object value )
  {
      int tracingDepth =   isModelObject
                         ? modelObjectsTracingDepth
                         : implementationObjectsTracingDepth
                       ;

      if( tracingDepth == 2 )
      {
          return Utils.getFieldsString(value);
      }
      else if( tracingDepth == 1 )
      {
          return Utils.getSimpleString(value);
      } else {
          return DeepOperationManager.getDeepString(value);
      }
  }

  protected String[] printArgsValues( boolean areModelObjects, String[] argTypes, Object[] args )
  {
      int tracingDepth =   areModelObjects
                         ? modelObjectsTracingDepth
                         : implementationObjectsTracingDepth
                       ;

      String argsDeepStrings[] = null;

      if( tracingDepth > 2 || tracingDepth < 1 )
          argsDeepStrings = DeepOperationManager.deepToString( args );

      String[] res = new String[ args.length ];

      for( int i = 0; i < args.length; i ++ )
      {
          if( Utils.isPrimitiveType( argTypes[ i ] ) )
          {
              res[ i ] = args[ i ].toString();
          }
          else if( tracingDepth == 2 )
          {
              res[ i ] = Utils.getFieldsString( args[ i ] );
          }
          else if( tracingDepth == 1 )
          {
              res[ i ] = Utils.getSimpleString( args[ i ] );
          } else {
              res[ i ] = argsDeepStrings[ i ];
          }
      }
      return res;
  }

  protected String getRealValueType( String static_type, Object o )
  {
      if( o == null )
          return static_type;
      if( Utils.isPrimitiveType( static_type ) )
          return static_type;
      else
          return Utils.prettifyClass( o.getClass() );
  }
}
//=============================================================================
// End of file AbstractFormatter
