/*
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 * FormatAsText
 * $Id: FormatAsText.java,v 1.32 2008/07/25 10:05:31 almo Exp $
 *
 * Author: Alexei Barantsev <barancev@ispras.ru>
 * Created on December 14, 2002
 */
//=============================================================================

package jatva.tracer.impl;

//=============================================================================

import jatva.lang.ModelCall;
import jatva.lang.ModelClass;
import jatva.lang.ModelEvent;
import jatva.lang.ModelObject;
import jatva.lang.TestingModel;
import jatva.lang.coverage.SEOperationInfo;
import jatva.lang.synchronization.Serializer;
import jatva.lang.synchronization.TestManager;
import jatva.tracer.TraceConstants;
import jatva.util.IntSet;
import jatva.util.IntToIntMap;
import jatva.util.OperationInfo;
import jatva.util.Utils;

import java.io.CharArrayWriter;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.util.Date;
import java.util.Map;
import java.util.Properties;

//=============================================================================

/** This class formats trace in JavaTESK 1.0 compatible style.
 *
 * @author  <a href="mailto:barancev@ispras.ru">Alexei Barantsev</a>
 * @version 1.0
 *
 * @author Victor Kuliamin
 * @version 2.0
 */
public class FormatAsText
  extends AbstractFormatter
{
  private static int threadIdentCounter = 0;

  public  static int INDENT_SIZE        = 2;

  private int indent      = 0;

  private IntSet visitedStates      = new IntSet();
  private IntSet visitedTransitions = new IntSet();

  private IntToIntMap threadIndents = new IntToIntMap();

  /** Creates a new instance of FormatAsText */
  public FormatAsText()
  {
  }

  @Override
  public FormatAsText clone()
  {
    FormatAsText newObj       = (FormatAsText) super.clone();
    newObj.visitedStates      = new IntSet();
    newObj.visitedTransitions = new IntSet();
    newObj.threadIndents      = new IntToIntMap();
    newObj.indent = 0;
    return newObj;
  }

  public void configure(Properties formatterProperties)
  {
  }

  private StringBuffer getPrefix(int threadId)
  {
    if( !(threadIndents.containsKey( threadId )) )
    {
        threadIndents.put( threadId, threadIdentCounter ++ );
    }

    int threadIndent = threadIndents.get( threadId );

    int m = threadIndent*INDENT_SIZE*2;
    //System.out.println("!!! ti: " + threadIndent);
    StringBuffer buf = new StringBuffer(m);

    for (int i = 0; i < m; i++)
    {
      buf.append(' ');
    }

    buf.append(Serializer.channel(threadId))
       .append('.')
       .append(Serializer.index(threadId))
       .append('(')
       .append(Thread.currentThread().getName())
       .append(')')
       .append(": ");
    //System.out.println("!!! ii: " + indent);

    m = indent*INDENT_SIZE;
    for (int i = 0; i < m; i++)
    {
      buf.append(' ');
    }

    return buf;
  }

  /** Replaces all occurences of one string to another string. */
  private static String replaceAll(String str, String regex, String replacement)
  {
    if (str == null)
    {
      return null;
    }

    StringBuffer buffer = new StringBuffer(str.length());
    int start = 0, end = 0;
    while ((end = str.indexOf(regex, start)) != -1) {
        buffer.append(str.substring(start, end)).append(replacement);
        start = end + regex.length();
    }

    buffer.append(str.substring(start));
    return buffer.toString();
  }

  /** Replaces newlines with \\n's */
  private static StringBuffer normalizeString(String source)
  {
    StringBuffer s = new StringBuffer(source);

    return s.insert(0,"[<[ ").append(" ]>]");
  }

  private static StringBuffer normalizeStrings(String[] ss)
  {
    StringBuffer s = new StringBuffer();
    if(ss == null) return s;

    s.append("[<[ ");

    for(int i = 0; i < ss.length; i++)
    {
      s.append(ss[i]);
      if(i != ss.length-1)
      {
        s.append(Utils.endl);
      }
    }

    s.append(" ]>]");

    return s;
  }


  public StringBuffer formatTestThreadStart(TestManager manager)
  {
    Date d = new Date();
    return getPrefix(manager.getId())
             .append(" start: ")
             .append(d.getTime())
             .append(" - ")
             .append(normalizeString(DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.FULL).format(d)))
             .append(Utils.endl);
  }

  public StringBuffer formatScenarioStart(TestManager manager, TestingModel scenario)
  {
    Date d;
    return getPrefix(manager.getId())
             .append(" scenario_start: ")
             .append(scenario == null ? TraceConstants.NO_SCENARIO : scenario.getClass().getName())
             .append(" : ")
             .append((d = new Date()).getTime())
             .append(" - ")
             .append(normalizeString(DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.FULL).format(d)))
             .append(Utils.endl);
  }

  public StringBuffer formatEnvironmentProperties(TestManager manager, Map<String,String> props)
  {
    StringBuffer str = new StringBuffer();

    for( Map.Entry<String,String> e : props.entrySet() )
    {
    	str.append(getPrefix(manager.getId()))
         .append(" environment: ")
         .append(normalizeString(e.getKey()))
         .append(" = ")
         .append(normalizeString(e.getValue()))
         .append(Utils.endl);
    }

    return str;
  }

  public StringBuffer formatScenarioEnd(TestManager manager, String msg)
  {
    Date d;
    return getPrefix(manager.getId())
             .append(" scenario_end: ")
             .append((d = new Date()).getTime())
             .append(" - ")
             .append(normalizeString(DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.FULL).format(d)))
             .append(Utils.endl)
             .append(getPrefix(manager.getId()))
             .append(" reason: ")
             .append(msg)
             .append(Utils.endl);
  }

  public StringBuffer formatException( TestManager manager
                                     , boolean isInterim
                                     , Throwable cause
                                     , String info
                                     , String[] tags
                                     , String[] values
                                     )
  {
    StringBuffer s;

    //if(manager.getStatus() == Test.EXECUTING)
    //{
      s = getPrefix(manager.getId());
    //}
    //else
    //{
    //  s = getCheckingPrefix(wrid, manager.getTestId());
    //}

    s.append(" exception: ")
     .append(info)
     .append(Utils.endl);

    if(tags != null && values != null && tags.length > 0)
    {
      indent += 2;
      for(int i = 0; i < tags.length; i++)
      {
        //if(manager.getStatus() == Test.EXECUTING)
        //{
          s.append(getPrefix(manager.getId()));
        //}
        //else
        //{
        //  s.append(getCheckingPrefix(wrid, manager.getTestId()));
        //}
        s.append(tags[i])
         .append(" = ")
         .append((i < values.length)?values[i]:"null")
         .append(Utils.endl);
      }
      indent -= 2;
    }

    if(cause != null)
    {
      CharArrayWriter buf = new CharArrayWriter();
      cause.printStackTrace(new PrintWriter(buf));
      indent += 2;
      //if(manager.getStatus() == Test.EXECUTING)
      //{
        s.append(getPrefix(manager.getId()));
      //}
      //else
      //{
      //  s.append(getCheckingPrefix(wrid, manager.getTestId()));
      //}
      s.append(normalizeString(buf.toString()))
       .append(Utils.endl);
      indent -= 2;
    }

    return s;
  }

  public StringBuffer formatValidException(TestManager manager, Throwable ex)
  {
    StringBuffer s = getPrefix(manager.getId());

    s.append(" valid exception: ")
     .append(Utils.getSimpleString(ex))
     .append(" : ")
     .append(ex.toString())
     .append(Utils.endl);

    CharArrayWriter buf = new CharArrayWriter();
    ex.printStackTrace(new PrintWriter(buf));
    indent += 2;
    s.append(getPrefix(manager.getId()));
    s.append(normalizeString(buf.toString()))
     .append(Utils.endl);
    indent -= 2;

    return s;
  }

  public StringBuffer formatInternalError(TestManager manager, Throwable cause, String info, String[] tags, String[] values)
  {
    StringBuffer s;

      //if(manager.getStatus() == Test.EXECUTING)
      //{
        s = getPrefix(manager.getId());
      //}
      //else
      //{
      //  s = getCheckingPrefix(wrid, manager.getTestId());
      //}

      s.append(" BUG: ")
       .append(info)
       .append(Utils.endl);

      if(tags != null && values != null && tags.length > 0)
      {
        indent += 2;
        for(int i = 0; i < tags.length; i++)
        {
          //if(manager.getStatus() == Test.EXECUTING)
          //{
            s.append(getPrefix(manager.getId()));
          //}
          //else
          //{
          //  s.append(getCheckingPrefix(wrid, manager.getTestId()));
          //}
          s.append(tags[i])
           .append(" = ")
           .append((i < values.length)?values[i]:"null")
           .append(Utils.endl);
        }
        indent -= 2;
      }

      if(cause != null)
      {
        CharArrayWriter buf = new CharArrayWriter();
        cause.printStackTrace(new PrintWriter(buf));
        indent += 2;
        //if(manager.getStatus() == Test.EXECUTING)
        //{
          s.append(getPrefix(manager.getId()));
        //}
        //else
        //{
        //  s.append(getCheckingPrefix(wrid, manager.getTestId()));
        //}
        s.append(normalizeString(buf.toString()))
         .append(Utils.endl);
        indent -= 2;
      }

      return s;
  }

  public StringBuffer formatState(TestManager manager, int stateId, Object state)
  {
    StringBuffer s = getPrefix(manager.getId());

    if (visitedStates.contains(stateId))
    {
      s.append(" old_state: ");
    }
    else
    {
      visitedStates.add(stateId);
      s.append(" new_state: ");
    }

    return s.append(stateId)
            .append(" = ")
            .append("" + state)
            .append(Utils.endl);
  }

  public StringBuffer formatTransitionStart(TestManager manager, int transitionId)
  {
    StringBuffer s = getPrefix(manager.getId());

    if (visitedTransitions.contains(transitionId))
    {
      s.append(" old_transition: ");
    }
    else
    {
      visitedTransitions.add(transitionId);
      s.append(" new_transition: ");
    }

    s.append(transitionId)
     .append(Utils.endl)
     .append(getPrefix(manager.getId()))
     .append(" [")
     .append(Utils.endl);

    indent++;

    return s;
  }

  public StringBuffer formatTransitionEnd(TestManager manager)
  {
    indent--;
    return getPrefix(manager.getId())
             .append(" ]")
             .append(Utils.endl);
  }

  public StringBuffer formatJumpStart(TestManager manager, String jumpKind)
  {
    return getPrefix(manager.getId())
             .append(" jump_start: ")
             .append(jumpKind)
             .append(Utils.endl);
  }

  public StringBuffer formatJumpEnd(TestManager manager)
  {
    return getPrefix(manager.getId())
             .append(" jump_end: ")
             .append(Utils.endl);
  }


  public StringBuffer formatTestFrameStart(TestManager manager)
  {
    StringBuffer s = getPrefix(manager.getId())
                       .append('[');

//    s.append(getCheckingPrefix(wrid, test.getId()))
//     .append('[');

    for(int i = 0; i < 80; i++)
    {
      s.append('=');
    }

    return s.append(Utils.endl);
  }

  public StringBuffer formatTestFrameEnter(TestManager manager)
  {
    return getPrefix(manager.getId())
             .append(" enter the frame")
             .append(Utils.endl);
  }

  public StringBuffer formatTestFrameExit(TestManager manager)
  {
    return getPrefix(manager.getId())
             .append(" exit the frame")
             .append(Utils.endl);
  }

  public StringBuffer formatSerializationStart(TestManager manager)
  {
    StringBuffer s = getPrefix(manager.getId());

    for(int i = 0; i < 80; i++)
    {
      s.append('-');
    }

    return s.append(Utils.endl);
  }

  public StringBuffer formatSerializationEnd(TestManager manager)
  {
    StringBuffer s = getPrefix(manager.getId());

    for(int i = 0; i < 80; i++)
    {
      s.append('-');
    }

    return s.append(Utils.endl);
  }


  public StringBuffer formatTestFrameEnd(TestManager manager)
  {
    StringBuffer s = getPrefix(manager.getId())
                       .append(']');

    for(int i = 0; i < 80; i++)
    {
      s.append('=');
    }

    return s.append(Utils.endl);
  }

  public StringBuffer formatScenarioMethod(TestManager manager, TestingModel obj, String mth)
  {
    return getPrefix(manager.getId())
            .append(" scenario: method ")
            .append(obj.getClass().getName())
            .append(".")
            .append(mth)
            .append("()")
            .append(Utils.endl);
  }

  public StringBuffer formatScenarioValue(TestManager manager, String kind, String type, String name, boolean isPrimitive, Object value)
  {
    return getPrefix(manager.getId())
             .append(" scenario: ")
             .append(kind)
             .append(" ")
             .append((type == null || type.equals(""))? "" : type + " ")
             .append((name == null || name.equals(""))? "" : name + " ")
             .append("= ")
             .append( printObjectValue( isPrimitive, true, value ) )
             .append(Utils.endl);
  }

  public StringBuffer formatScenarioMessage(TestManager manager, String info)
  {
    return getPrefix(manager.getId())
             .append(" info: scenario: ")
             .append(normalizeString(info))
             .append(Utils.endl);
  }

  public StringBuffer formatTestMessage(TestManager manager, String info)
  {
    return getPrefix(manager.getId())
             .append(" info: test: ")
             .append(normalizeString(info))
             .append(Utils.endl);
  }

  public StringBuffer formatModelEvent(TestManager manager, ModelEvent e)
  {
    ///System.out.println("UUUUUU");
    StringBuffer s =  getPrefix(manager.getId())
                        .append(" model: ");

    if(e.getObject() instanceof ModelClass)
    {
      s.append("class ")
       .append(e.getOperation().getDeclaringClass());
    }
    else
    {
      s.append("object ")
       .append(normalizeString( printObjectValue( true, e.getObject() ) ) );
    }
    s.append(Utils.endl)
     .append(getPrefix(manager.getId()))
     .append((e instanceof ModelCall)?("        operation "):("        reaction "))
     .append(e.getOperation().getPrettySignature())
     .append(Utils.endl)
     .append(getPrefix(manager.getId()))
     .append("        arguments: ")
     .append(Utils.endl);

    String[] argNames = e.getOperation().getParNames();
    String[] argTypes = e.getOperation().getParTypes();
    String[] argStrings = printArgsValues( true, argTypes, e.getArguments() );

    //String sg = e.getOperation().getSignature();

    ///System.out.println("SIGNATURE: " + sg);

    //int k = sg.indexOf('('), l, m;

    ///System.out.println("k: " + k);

    for(int i = 0; i < argNames.length; i++)
    {
      //l = sg.indexOf(',', k);
      //m = sg.indexOf(')', k);
      ///System.out.println("l: " + l);
      ///System.out.println("m: " + m);

      //if(l == -1 || l > m) l = m;
      //System.out.println("l: " + l);

      s.append(getPrefix(manager.getId()))
       .append("          ")
       //.append(sg.substring(k+1, l))
       .append(argTypes[i])
       .append(" ")
       .append(argNames[i])
       .append(" = ")
       .append(argStrings[i] );

//      if(Utils.isPrimitiveType(argTypes[i]/*sg.substring(k+1, l)*/))
//      {
//        //System.out.println("PRIME");
//        s.append(e.getArguments()[i].toString());
//      }
//      else
//      {
//        //System.out.println("NON-PRIME");
//        switch(modelObjectsTracingDepth)
//        {
//          case 2:
//            s.append(normalizeString( Utils.getFieldsString(e.getArguments()[i]) ));
//            break;
//          case 1:
//            s.append(normalizeString( Utils.getSimpleString(e.getArguments()[i]) ));
//            break;
//          default:
//            s.append(normalizeString(args[i]));
//        }
//      }

      s.append(Utils.endl);
      //k = l;
    }

    s.append(getPrefix(manager.getId()))
     .append(" {")
     .append(Utils.endl);

    indent++;

    ///System.out.println("WWWWWW");
    return s;
  }

  public StringBuffer formatModelEventEnd(TestManager manager, ModelEvent e)
  {
    StringBuffer s = new StringBuffer();

    if(e.getResult() != null)
    {
      String rt = e.getOperation().getReturnType();

      if(!Utils.isVoidType(rt))
      {
        s.append(getPrefix(manager.getId()))
         .append(" model: ")
         .append("result: ")
         .append( getRealValueType( rt, e.getResult() ) )
         .append(" ")
         .append(normalizeString( printObjectValue( Utils.isPrimitiveType( rt )
                                                  , true
                                                  , e.getResult()
                                                  )
                                )
                )
        .append(Utils.endl);
      }
    }
    else if(e.getException() != null)
    {
      s.append(getPrefix(manager.getId()))
       .append(" model: ")
       .append("exception: ")
       .append( normalizeString( printObjectValue( true, e.getException() ) ) )
       .append(Utils.endl);
    }

    indent--;

    s.append(getPrefix(manager.getId()))
     .append(" }")
     .append(Utils.endl);

    return s;
  }

  public StringBuffer formatModelValue(TestManager manager, String kind, String type, String name, Object value)
  {
    return getPrefix(manager.getId())
             .append(" model: ")
             .append(kind)
             .append("".equals(type) ? "" : " " + type)
             .append("".equals(name) ? "" : " " + name)
             .append(" = ")
             .append( normalizeString( printObjectValue( true, value ) ) )
             .append(Utils.endl);
  }

  public StringBuffer formatModelMessage(TestManager manager, String info)
  {
    //if(manager.getStatus() == Test.EXECUTING)
    //{
      return getPrefix(manager.getId())
               .append(" info: model: ")
               .append(normalizeString(info))
               .append(Utils.endl);
    /*}
    else
    {
      return getCheckingPrefix(wrid, manager.getTestId())
              .append(" info: model: ")
              .append(normalizeString(info))
              .append(Utils.endl);
    }*/
  }

  public StringBuffer formatOraclePreconditionEnter(TestManager manager, String signature, int id)
  {
    StringBuffer s;

    //if(manager.getStatus() == Test.EXECUTING)
    //{
      s = getPrefix(manager.getId());
      s.append(" oracle: ")
       .append(signature)
       .append(Utils.endl)
       .append(getPrefix(manager.getId()))
       .append(" {")
       .append(Utils.endl);
    /*}
    else
    {
      s = getCheckingPrefix(wrid, manager.getTestId());
      s.append(" oracle: ")
      .append(signature)
      .append(Utils.endl)
      .append(getCheckingPrefix(wrid, manager.getTestId()))
      .append(" {")
      .append(Utils.endl);
    }*/

    indent++;

    return s;
  }

  public StringBuffer formatOraclePostconditionEnter(TestManager manager, String signature)
  {
    StringBuffer s;

    //if(manager.getStatus() == Test.EXECUTING)
    //{
      s = getPrefix(manager.getId());
      s.append(" postcondition: ")
       .append(signature)
       .append(Utils.endl);
    /*}
    else
    {
      s = getCheckingPrefix(wrid, manager.getTestId());
      s.append(" oracle: ")
      .append(signature)
      .append(Utils.endl)
      .append(getCheckingPrefix(wrid, manager.getTestId()))
      .append(" {")
      .append(Utils.endl);
    }*/

    return s;
  }

  public StringBuffer formatOracleExit(TestManager manager)
  {
    indent--;
    //if(manager.getStatus() == Test.EXECUTING)
    //{
      return getPrefix(manager.getId())
               .append(" }")
               .append(Utils.endl);
    /*}
    else
    {
      return getCheckingPrefix(wrid, manager.getTestId())
              .append(" }")
              .append(Utils.endl);
    }*/
  }

  public StringBuffer formatPreconditionExit(TestManager manager)
  {
    //if(manager.getStatus() == Test.EXECUTING)
    //{
      return getPrefix(manager.getId())
               .append(" model: precondition end")
               .append(Utils.endl);
    /*}
    else
    {
      return getCheckingPrefix(wrid, manager.getTestId())
              .append(" model: precondition end")
              .append(Utils.endl);
    }*/
  }

  /* (non-Javadoc)
   * @see jatva.tracer.TraceFormatter#formatCoverageStructure(long, java.lang.String, jatva.lang.coverage.OperationInfo)
   */
  public StringBuffer formatCoverageStructure (TestManager manager, SEOperationInfo op)
  {
    StringBuffer s = getPrefix(manager.getId());

    s.append(" model: coverage structure: ")
     .append(Utils.endl);

    if(op != null)
    {
      s.append(getPrefix(manager.getId()))
       .append("        branches: ")
       .append(Utils.endl);

      int i = 0;
      for(;i < op.getBranchesMap().size(); i++)
      {
        s.append(getPrefix(manager.getId()))
         .append("          ")
         .append(i)
         .append(" : ")
         .append(op.getBranchesMap().get(i))
         .append(Utils.endl);
      }

      s.append(getPrefix(manager.getId()))
       .append("        formulas: ")
       .append(Utils.endl);
      for(i = 0;i < op.getFormulasMap().size(); i++)
      {
        s.append(getPrefix(manager.getId()))
         .append("          ")
         .append(i)
         .append(" : ")
         .append(op.getFormulasMap().get(i))
         .append(Utils.endl);
      }

      if( op.getMarkedPathsMap().size() > 0 )
      {
          s.append(getPrefix(manager.getId()))
         .append("        marked paths: ")
         .append(Utils.endl);
        for(i = 0;i < op.getMarkedPathsMap().size(); i++)
        {
          s.append(getPrefix(manager.getId()))
           .append("          ")
           .append(i)
           .append(" : ")
           .append(op.getMarkedPathsMap().get(i))
           .append(Utils.endl);
        }
      }

      if( op.getPredicatesMap().size() > 0 )
      {
        s.append(getPrefix(manager.getId()))
         .append("        predicates: ")
         .append(Utils.endl);

        for(i = 0;i < op.getPredicatesMap().size(); i++)
        {
          s.append(getPrefix(manager.getId()))
           .append("          ")
           .append(i)
           .append(" : ")
           .append(op.getPredicatesMap().get(i))
           .append(Utils.endl);
         }
      }

      if( op.getDisjunctsMap().size() > 0 )
      {
         s.append(getPrefix(manager.getId()))
          .append("        disjuncts: ")
          .append(Utils.endl);

         for(i = 0;i < op.getDisjunctsMap().size(); i++)
         {
           s.append(getPrefix(manager.getId()))
            .append("          ")
            .append(i)
            .append(" : ")
            .append(op.getDisjunctsMap().get(i).getFormulasValues())
            .append(Utils.endl);
         }
      }
    }

    return s;
  }

  public StringBuffer formatBranch(TestManager manager, String branch, String msg)
  {
    //if(manager.getStatus() == Test.EXECUTING)
    //{
      StringBuffer s = getPrefix(manager.getId())
                         .append(" branch: ")
                         .append(branch)
                         .append(Utils.endl);
      if(msg != null)
        s.append(getPrefix(manager.getId()))
         .append(" additional data: ")
         .append(msg)
         .append(Utils.endl);

      return s;
    /*}
    else
    {
      return getCheckingPrefix(wrid, manager.getTestId())
              .append(" branch: ")
              .append(branch)
              .append(Utils.endl);
    }*/
  }

  public StringBuffer formatDeferredBranch(TestManager manager, String branch, String msg)
  {
    //if(manager.getStatus() == Test.EXECUTING)
    //{
      StringBuffer s = getPrefix(manager.getId())
               .append(" defferred branch: ")
               .append(branch)
               .append(Utils.endl);
      if(msg != null)
          s.append(getPrefix(manager.getId()))
           .append(" additional data: ")
           .append(msg);

        return s;
    /*}
    else
    {
      return getCheckingPrefix(wrid, manager.getTestId())
              .append(" defferred branch: ")
              .append(branch)
              .append(Utils.endl);
    }*/
  }

  public StringBuffer formatMark(TestManager manager, String mark)
  {
    //if(manager.getStatus() == Test.EXECUTING)
    //{
      return getPrefix(manager.getId())
               .append(" mark: ")
               .append(mark)
               .append(Utils.endl);
    /*}
    else
    {
      return getCheckingPrefix(wrid, manager.getTestId())
              .append(" mark: ")
              .append(mark)
              .append(Utils.endl);
    }*/
  }

  public StringBuffer formatPrimeFormula(TestManager manager, int formulaId, boolean value)
  {
    //if(manager.getStatus() == Test.EXECUTING)
    //{
      return getPrefix(manager.getId())
               .append(" formula: ")
               .append(formulaId)
               .append(" is ")
               .append(value)
               .append(Utils.endl);
      /*}
    else
    {
      return getCheckingPrefix(wrid, manager.getTestId())
              .append(" formula: ")
              .append(formulaId)
              .append(" is ")
              .append(value)
              .append(Utils.endl);
    }*/
  }

  public StringBuffer formatPrimeFormula(TestManager manager, String formula, boolean value)
  {
    return getPrefix(manager.getId())
            .append(" formula: ")
            .append(formula)
            .append(" = ")
            .append(value)
            .append(Utils.endl);
  }

  public StringBuffer formatCoverageElement(TestManager manager, String coverage, int elementId)
  {
    return getPrefix(manager.getId())
             .append(" info: coverage: ")
             .append(coverage)
             .append(": ")
             .append(elementId)
             .append(Utils.endl);
  }

  public StringBuffer formatInvariantStart(TestManager manager, ModelObject o, String signature, String msg)
  {
    //if(manager.getStatus() == Test.EXECUTING)
    //{
      return getPrefix(manager.getId())
               .append(" model: invariant start ")
               .append( printObjectValue( true, o ) )
               .append('.')
               .append(signature)
               .append((msg == null)?"":" " + msg)
               .append(Utils.endl);
      /*}
    else
    {
      return getCheckingPrefix(wrid, manager.getTestId())
              .append(" model: invariant ")
              .append(signature)
              .append(Utils.endl);
    }*/
  }

  public StringBuffer formatInvariantStart(TestManager manager, String p_class, String signature, String msg)
  {
      return getPrefix(manager.getId())
               .append(" model: invariant start ")
               .append(p_class)         // possible deep string
               .append('.')
               .append(signature)
               .append((msg == null)?"":" " + msg)
               .append(Utils.endl);
  }

  public StringBuffer formatInvariantEnd(TestManager manager)
  {
    return getPrefix(manager.getId())
            .append(" model: invariant end")
            .append(Utils.endl);
  }

    public StringBuffer formatImplementationOperationStart( TestManager manager
                                                          , OperationInfo op_info
                                                          , Object obj
                                                          , Object[] args
                                                          )
    {
        StringBuffer s =  getPrefix( manager.getId() )
                         .append( " implementation: operation " )
                         .append( op_info.getPrettySignature() )
                         .append( Utils.endl )
                       ;

        if( obj != null )
        {
            s.append( getPrefix( manager.getId() ) )
             .append( " implementation: object " )
             .append( normalizeString( printObjectValue( false, obj ) ) )
             .append( Utils.endl );
        }

        s.append( getPrefix( manager.getId() ) )
         .append( " implementation: arguments: " )
         .append( Utils.endl );

        String[] argTypes = op_info.getParTypes();
        String[] argStrings = printArgsValues( false, argTypes, args );

        for( int i = 0; i < args.length; i++ )
        {
            s.append( getPrefix( manager.getId() ) )
             .append( "          " )
             .append( getRealValueType( argTypes[ i ], args[ i ] ) )
             .append( " " )
             .append( argStrings[ i ] )
             .append( Utils.endl );
        }

        return s;
    }


  public StringBuffer formatImplementationOperationEnd(TestManager manager, long stime, long etime)
  {
    return getPrefix(manager.getId())
            .append(" implementation: working time: ")
            .append(stime)
            .append(" - ")
            .append(etime)
            .append(Utils.endl)
            .append(getPrefix(manager.getId()))
            .append(" implementation: operation end")
            .append(Utils.endl);
  }

  public StringBuffer formatImplementationResult(TestManager manager, String resultType, Object result)
  {
    StringBuffer s = new StringBuffer();

    if(!Utils.isVoidType(resultType))
    {
      s.append(getPrefix(manager.getId()))
       .append(" implementation: result: ")
       .append( getRealValueType( resultType, result ) )
       .append(" ")
       .append( normalizeString( printObjectValue( Utils.isPrimitiveType( resultType )
                                                 , false
                                                 , result
                                                 )
                               )
              )
       .append(Utils.endl);
    }

    return s;
  }

  public StringBuffer formatImplementationException(TestManager manager, Throwable error)
  {
    return getPrefix(manager.getId())
            .append(" implementation: exception ")
            .append( normalizeString( printObjectValue( false, error ) ) )
            .append(Utils.endl);
  }

  public StringBuffer formatImplementationTimeInterval(TestManager manager, long startTime, long endTime)
  {
    return getPrefix(manager.getId())
             .append(" info: implementation: start - ")
             .append(startTime)
             .append(", end - ")
             .append(endTime)
             .append(Utils.endl);
  }

  public StringBuffer formatImplementationValue(TestManager manager, String kind, String type, String name, Object value)
  {
    return getPrefix(manager.getId())
             .append(" implementation: ")
             .append(kind)
             .append("".equals(type) ? "" : " " + type)
             .append("".equals(name) ? "" : " " + name)
             .append(" = ")
             .append( normalizeString( printObjectValue( false, value ) ) )
             .append(Utils.endl);
  }

  public StringBuffer formatImplementationMessage(TestManager manager, String info)
  {
    return getPrefix(manager.getId())
             .append(" info: implementation: ")
             .append(normalizeString(info))
             .append(Utils.endl);
  }

  public StringBuffer formatUserInfo(TestManager manager, String info)
  {
    //if(manager.getStatus() == Test.EXECUTING)
    //{
      return getPrefix(manager.getId())
               .append(" user_info: ")
               .append(normalizeString(info))
               .append(Utils.endl);
      /*}
    else
    {
      return getCheckingPrefix(wrid, manager.getTestId())
              .append(" user_info: ")
              .append(normalizeString(info))
              .append(Utils.endl);
    }*/
  }

  public StringBuffer formatUserInfo(TestManager manager, String[] tags, String[] values)
  {
      // TODO: Implement
      return new StringBuffer();
  }

  public StringBuffer getTraceHeader()
  {
    return new StringBuffer();
  }

  public StringBuffer getTraceFooter()
  {
    return new StringBuffer();
  }


}
//=============================================================================
// End of file FormatAsText.java
