/*
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 * FormatAsXml
 * $Id: FormatAsXml.java,v 1.39 2008/07/25 10:01:05 almo Exp $
 *
 * Author: Alexei Barantsev <barancev@ispras.ru>
 * Created on December 9, 2002
 */
//=============================================================================

package jatva.tracer.impl;

//=============================================================================

import jatva.iterators.lang.IIntegerIterator;
import jatva.lang.ModelClass;
import jatva.lang.ModelEvent;
import jatva.lang.ModelObject;
import jatva.lang.ModelReaction;
import jatva.lang.TestingModel;
import jatva.lang.coverage.DisjunctInfo;
import jatva.lang.coverage.DisjunctsMap;
import jatva.lang.coverage.SEOperationInfo;
import jatva.lang.synchronization.TestManager;
import jatva.tracer.TraceConstants;
import jatva.tracer.ValueKind;
import jatva.util.IntToStringMap;
import jatva.util.OperationInfo;
import jatva.util.Utils;

import java.io.CharArrayWriter;
import java.io.PrintWriter;
import java.util.Map;
import java.util.Properties;

//=============================================================================

/** This class formats each message as an xml tag that contains message data
 * in attributes and/or in the body.
 *
 * @author  <a href="mailto:barancev@ispras.ru">Alexei Barantsev</a>
 * @version 1.0
 */
public class FormatAsXml extends AbstractFormatter {

    public static int INDENT_SIZE = 2;
    private int indent = 1;

    /** Creates a new instance of FormatAsXml */
    public FormatAsXml() {
    }

    public void configure(Properties formatterProperties) {
    }

    public StringBuffer formatTestThreadStart( TestManager manager) {
        // TODO: Implement
        return new StringBuffer();
    }

	public StringBuffer formatScenarioStart( TestManager manager, TestingModel scenario) {
        StringBuffer res = new StringBuffer();
        addPrefix(res, 1);
        addTagStartWithTraceId(res, TraceConstants.SCENARIO_START );
        addAttr(res, TraceConstants.SCENARIO_NAME_ATTR,
                scenario == null ? TraceConstants.NO_SCENARIO : scenario.getClass().getName());
        addAttr(res, TraceConstants.SCENARIO_START_TIME_ATTR, System.currentTimeMillis());
        res.append("/>");
        res.append(Utils.endl);

        // HACK:
        res.append( "<scenario_value trace=\"1\" kind=\"test engine property\" name=\"Deferred Reactions Enabled\" value=\"true\"/>" );
        res.append( Utils.endl );

        return res;
    }

    public StringBuffer formatEnvironmentProperties( TestManager manager, Map<String, String> props )
    {
        StringBuffer res = new StringBuffer();

        for( Map.Entry<String,String> e : props.entrySet() )
        {
            addPrefix( res, 0 );
            addTagStartWithTraceId( res, TraceConstants.TRACE_ENVIRONMENT );
            addAttr( res, TraceConstants.ENV_NAME_ATTR, e.getKey() );
            addAttr( res, TraceConstants.ENV_VALUE_ATTR, e.getValue() );
            res.append( "/>" );
            res.append( Utils.endl );
        }

        return res;
    }

    public StringBuffer formatScenarioEnd( TestManager manager, String msg)
    {
        return format( -1
                     , TraceConstants.SCENARIO_END
                     , TraceConstants.SCENARIO_END_TIME_ATTR
                     , Long.toString(System.currentTimeMillis())
                     , msg.toString()
                     );
    }

    public StringBuffer formatException(TestManager manager, boolean isInterim, Throwable cause, String info, String[] tags, String[] values) {
        return formatException(manager, isInterim, cause, info, tags, values, false);
    }

    public StringBuffer formatValidException(TestManager manager, Throwable ex)
    {
      return new StringBuffer();
    }

    public StringBuffer formatInternalError(TestManager manager, Throwable cause, String info, String[] tags, String[] values) {
        return formatException(manager, false, cause, info, tags, values, true);
    }

    public StringBuffer formatState(TestManager manager, int stateId, Object state) {
        StringBuffer res = format(0, TraceConstants.STATE,
            TraceConstants.STATE_ID_ATTR, Integer.toString(stateId));
        formatValue(res, TraceConstants.SCENARIO_VALUE, ValueKind.GENSTATE.toString(), state.getClass().getName(), null, "" + state);
        return res;
    }

    public StringBuffer formatTransitionStart(TestManager manager, int transitionId) {
        return format(1, TraceConstants.TRANSITION_START,
            TraceConstants.TRANSITION_ID_ATTR, Integer.toString(transitionId));
    }

    public StringBuffer formatTransitionEnd(TestManager manager) {
        return format(-1, TraceConstants.TRANSITION_END);
    }

    public StringBuffer formatJumpStart(TestManager manager, String jumpKind) {
        return format(1, TraceConstants.JUMP_START,
            TraceConstants.JUMP_KIND_ATTR, jumpKind);
    }

    public StringBuffer formatJumpEnd(TestManager manager) {
        return format(-1, TraceConstants.JUMP_END);
    }

    public StringBuffer formatTestFrameStart(TestManager manager) {
        // TODO: Implement
        return new StringBuffer();
    }

    public StringBuffer formatTestFrameEnter(TestManager manager) {
        // TODO: Implement
        return new StringBuffer();
    }

    public StringBuffer formatTestFrameExit(TestManager manager) {
    	// TODO: Implement
    	return new StringBuffer();
    }

    public StringBuffer formatTestFrameEnd(TestManager manager) {
      // TODO: Implement test frame end
      return new StringBuffer();
    }

    public StringBuffer formatSerializationStart(TestManager manager) {
      return format(1, TraceConstants.SERIALIZATION_START);
    }

    public StringBuffer formatSerializationEnd(TestManager manager) {
      return format(-1, TraceConstants.SERIALIZATION_END);
    }

    public StringBuffer formatScenarioMethod(TestManager manager, TestingModel obj, String mth) {
      return formatValue(TraceConstants.SCENARIO_VALUE, ValueKind.SCENARIO_METHOD.toString(), null, null, obj.getClass().getName() + "." + mth);
    }

    public StringBuffer formatScenarioValue(TestManager manager, String kind, String type, String name, boolean isPrimitive, Object value)
    {
        return formatValue( TraceConstants.SCENARIO_VALUE
                          , kind
                          , type
                          , name
                          , printObjectValue( isPrimitive, true, value )
                          );
    }

    public StringBuffer formatScenarioMessage(TestManager manager, String info) {
        return formatInfo(TraceConstants.INFO, info);
    }

    public StringBuffer formatTestMessage(TestManager manager, String info) {
      return formatInfo(TraceConstants.INFO, info);
    }

    public StringBuffer formatModelEvent(TestManager manager, ModelEvent event) {
        StringBuffer res = new StringBuffer();
        addPrefix(res, 0);
        if (event.getObject() instanceof ModelClass) {
            addTagStartWithTraceId(res, TraceConstants.MODEL_CLASS );
            addAttr(res, TraceConstants.MODEL_CLASS_NAME_ATTR, event.getOperation().getDeclaringClass());
            res.append("/>");

        } else {
            addTagStartWithTraceId(res, TraceConstants.MODEL_OBJECT );
            addAttr(res, TraceConstants.MODEL_OBJECT_ID_ATTR, Integer.toString(Utils.identityHash(event.getObject()), 16));
            res.append(">");
            addText(res, printObjectValue( true, event.getObject() ) );
            addTagEnd(res, TraceConstants.MODEL_OBJECT);
        }
        res.append(Utils.endl);

        addPrefix(res, 0);
        addTagStartWithTraceId(res, TraceConstants.MODEL_CALL_START );
        addAttr(res, TraceConstants.MODEL_KIND_ATTR, event instanceof ModelReaction ? "reaction" : "stimulus");
        addAttr(res, TraceConstants.MODEL_SIGNATURE_ATTR, event.getOperation().getPrettySignature());
        addAttr(res, TraceConstants.MODEL_REFID_ATTR, event.getId());
        res.append("/>");
        res.append(Utils.endl);

        indent++;
        String[] argNames = event.getOperation().getParNames();
        String[] argTypes = event.getOperation().getParTypes();

        String[] argStrings = printArgsValues( true, argTypes, event.getArguments() );

        for(int i = 0; i < argNames.length; i++)
        {
            formatValue( res
                       , TraceConstants.MODEL_VALUE
                       , ValueKind.ARGUMENT.toString()
                       , getRealValueType( argTypes[i], event.getArguments()[i] )
                       , argNames[i]
                       , argStrings[i]
                       );
        }

        return res;
    }

    public StringBuffer formatModelEventEnd(TestManager manager, ModelEvent event) {
        StringBuffer res = new StringBuffer();

        if(event.getResult() != null) {
            String resultType = event.getOperation().getReturnType();

        if(!Utils.isVoidType(resultType))
        {
            formatValue( res
                       , TraceConstants.MODEL_VALUE
                       , ValueKind.RESULT.toString()
                       , getRealValueType( resultType, event.getResult() )
                       , null
                       , printObjectValue( Utils.isPrimitiveType( resultType )
                                         , true, event.getResult()
                                         )
                       );
            }

        } else
        if( event.getException() != null )
        {
            formatValue(res
                       , TraceConstants.MODEL_VALUE
                       , ValueKind.EXCEPTION.toString()
                       , null
                       , null
                       , printObjectValue( true, event.getException() )
                       );
        }

        format(res, -1, TraceConstants.MODEL_CALL_END);

        return res;
    }

    public StringBuffer formatModelValue(TestManager manager, String kind, String type, String name, Object value) {
        return formatValue(TraceConstants.MODEL_VALUE
                          , kind
                          , type
                          , name
                          , printObjectValue( true, value )
                          );
    }

    public StringBuffer formatModelMessage(TestManager manager, String info) {
        return formatInfo(TraceConstants.INFO, info);
    }

    public StringBuffer formatOraclePreconditionEnter(TestManager manager, String signature, int id) {
        return format(1, TraceConstants.ORACLE_START,
            TraceConstants.ORACLE_SIGNATURE_ATTR, signature,
            TraceConstants.ORACLE_REF_ATTR, "" + id);
    }

    public StringBuffer formatOraclePostconditionEnter(TestManager manager, String signature) {
      /*return format(0, TraceConstants.POSTCONDITION_START, wrid,
          TraceConstants.ORACLE_CLASS_ATTR, cname,
          TraceConstants.ORACLE_SIGNATURE_ATTR, signature);
      */
      return new StringBuffer();
  }

    public StringBuffer formatOracleExit(TestManager manager) {
        return format(-1, TraceConstants.ORACLE_END);
    }

    public StringBuffer formatPreconditionExit(TestManager manager) {
        return format(0, TraceConstants.PRECONDITION_END);
    }

    public StringBuffer formatCoverageStructure(TestManager manager, SEOperationInfo disjunctManager) {
        StringBuffer res = new StringBuffer();
        addPrefix(res, 1);
        addTagStartWithTraceId(res, TraceConstants.COVERAGE_STRUCTURE );
        res.append(">");
        res.append(Utils.endl);

        IntToStringMap formulae = disjunctManager.getFormulasMap();
        addPrefix(res, 0);
        addTagStart(res, TraceConstants.COVERAGE_FORMULAE);
        res.append(">");
        for (IIntegerIterator i = formulae.keys().iterator(); i.hasNext(); i.shift()) {
            addTagStart(res, TraceConstants.COVERAGE_FORMULA);
            addAttr(res, TraceConstants.COVERAGE_FORMULA_ID_ATTR, i.value());
            res.append(">");
            addText(res, formulae.get(i.value()));
            addTagEnd(res, TraceConstants.COVERAGE_FORMULA);
        }
        addTagEnd(res, TraceConstants.COVERAGE_FORMULAE);
        res.append(Utils.endl);

        IntToStringMap branches = disjunctManager.getBranchesMap();
        addPrefix(res, 0);
        addTagStart(res, TraceConstants.COVERAGE_BRANCHES);
        res.append(">");
        for (IIntegerIterator i = branches.keys().iterator(); i.hasNext(); i.shift()) {
            //i.next();
            addTagStart(res, TraceConstants.COVERAGE_BRANCH);
            addAttr(res, TraceConstants.COVERAGE_BRANCH_ID_ATTR, i.value());
            res.append(">");
            addText(res, branches.get(i.value()));
            addTagEnd(res, TraceConstants.COVERAGE_BRANCH);
        }
        addTagEnd(res, TraceConstants.COVERAGE_BRANCHES);
        res.append(Utils.endl);

        IntToStringMap markedPaths = disjunctManager.getMarkedPathsMap();
        if( markedPaths.size() > 0 )
        {
            addPrefix(res, 0);
            addTagStart(res, TraceConstants.COVERAGE_MPATHS);
            res.append(">");
            for (IIntegerIterator i = markedPaths.keys().iterator(); i.hasNext(); i.shift()) {
                addTagStart(res, TraceConstants.COVERAGE_MPATH);
                addAttr(res, TraceConstants.COVERAGE_MPATH_ID_ATTR, i.value());
                res.append(">");
                addText(res, markedPaths.get(i.value()));
                addTagEnd(res, TraceConstants.COVERAGE_MPATH);
            }
            addTagEnd(res, TraceConstants.COVERAGE_MPATHS);
            res.append(Utils.endl);
        }

        IntToStringMap predicates = disjunctManager.getPredicatesMap();
        if( predicates.size() > 0 )
        {
            addPrefix(res, 0);
            addTagStart(res, TraceConstants.COVERAGE_PREDICATES);
            res.append(">");
            for (IIntegerIterator i = predicates.keys().iterator(); i.hasNext(); i.shift()) {
                addTagStart(res, TraceConstants.COVERAGE_PREDICATE);
                addAttr(res, TraceConstants.COVERAGE_PREDICATE_ID_ATTR, i.value());
                res.append(">");
                addText(res, predicates.get(i.value()));
                addTagEnd(res, TraceConstants.COVERAGE_PREDICATE);
            }
            addTagEnd(res, TraceConstants.COVERAGE_PREDICATES);
            res.append(Utils.endl);
        }

        DisjunctsMap disjuncts = disjunctManager.getDisjunctsMap();
        if( disjuncts.size() > 0 )
        {
            addPrefix(res, 0);
            addTagStart(res, TraceConstants.COVERAGE_DISJUNCTS);
            res.append(">");
            for (IIntegerIterator i = disjuncts.keys().iterator(); i.hasNext(); i.shift()) {
                addTagStart(res, TraceConstants.COVERAGE_DISJUNCT);
                addAttr(res, TraceConstants.COVERAGE_DISJUNCT_ID_ATTR, i.value());
                DisjunctInfo disjunct = disjuncts.get(i.value());
                addAttr(res, TraceConstants.COVERAGE_FORMULAE, disjunct.getFormulasValues());
                addAttr(res, TraceConstants.COVERAGE_BRANCH, disjunct.getBranchId());
                addAttr(res, TraceConstants.COVERAGE_MPATH, disjunct.getMarkedPathId());
                addAttr(res, TraceConstants.COVERAGE_PREDICATE, disjunct.getPredicateId());
                res.append("/>");
            }
            addTagEnd(res, TraceConstants.COVERAGE_DISJUNCTS);
            res.append(Utils.endl);
        }
        addPrefix(res, -1);
        addTagEnd(res, TraceConstants.COVERAGE_STRUCTURE);
        res.append(Utils.endl);

        return res;
    }

    public StringBuffer formatBranch(TestManager manager, String branch, String msg) {
        return format(0, TraceConstants.BRANCH, TraceConstants.BRANCH_NAME_ATTR, branch, msg);
    }

    public StringBuffer formatDeferredBranch(TestManager manager, String branch, String msg) {
        // TODO: Implement
        return format(0, TraceConstants.DEFERRED_BRANCH, TraceConstants.BRANCH_NAME_ATTR, branch, msg);
    }

    public StringBuffer formatMark(TestManager manager, String mark) {
        return format(0, TraceConstants.MARK, TraceConstants.MARK_NAME_ATTR, mark);
    }

    public StringBuffer formatPrimeFormula(TestManager manager, int formulaId, boolean value) {
        return format(0, TraceConstants.PRIME_FORMULA,
            TraceConstants.FORMULA_ID_ATTR, Integer.toString(formulaId),
            TraceConstants.FORMULA_VALUE_ATTR, Boolean.toString(value));
    }

    public StringBuffer formatPrimeFormula( TestManager manager, String formula, boolean value) {
      return format(0, TraceConstants.PRIME_FORMULA,
          TraceConstants.FORMULA_ID_ATTR, formula,
          TraceConstants.FORMULA_VALUE_ATTR, Boolean.toString(value));
    }

    public StringBuffer formatCoverageElement( TestManager manager, String coverage, int elementId) {
        return format(0, TraceConstants.COVERAGE_ELEMENT,
            TraceConstants.COVERAGE_NAME_ATTR, coverage,
            TraceConstants.COVERAGE_EL_ID_ATTR, Integer.toString(elementId));
    }

    public StringBuffer formatInvariantStart( TestManager manager, ModelObject o, String signature, String msg) {
        return format(1, TraceConstants.INVARIANT_START, TraceConstants.INVARIANT_SIGNATURE_ATTR, signature, msg);
    }

    public StringBuffer formatInvariantStart( TestManager manager, String p_class, String signature, String msg) {
        return format(1, TraceConstants.INVARIANT_START, TraceConstants.INVARIANT_SIGNATURE_ATTR, signature, msg);
    }

    public StringBuffer formatInvariantEnd(TestManager manager) {
        return format(-1, TraceConstants.INVARIANT_END);
    }

//    public StringBuffer formatImplementationOperationStart( TestManager manager
//                                                          , String p_class
//                                                          , String signature
//                                                          , Object obj
//                                                          , Object[] args
//                                                          )
//    {
//        StringBuffer res = new StringBuffer();
//
//        if( obj != null )
//        {
//            String objAsText = printObjectValue( false, obj );
//
//            format(res, 0, TraceConstants.IMPL_OBJECT,
//                    TraceConstants.IMPL_OBJECT_ID_ATTR, Integer.toString(Utils.identityHash(obj), 16), objAsText);
//
//        } else {
//            format( res, 0, TraceConstants.IMPL_CLASS,
//                    TraceConstants.IMPL_CLASS_NAME_ATTR, p_class );
//        }
//
//        format(res, 1, TraceConstants.IMPL_CALL_START,
//                 TraceConstants.IMPL_SIGNATURE_ATTR, signature);
//
////        String[] argTypes = Utils.getParameterTypes(args.length, signature);
//        String[] argTypes = getParameterTypes(signature);
//        String[] argStrings = printArgsValues( false, argTypes, args );
//
//        for(int i = 0; i < args.length; i++)
//        {
//            formatValue( res
//                       , TraceConstants.IMPL_VALUE
//                       , ValueKind.ARGUMENT.toString()
//                       , argTypes[i]
//                       , null
//                       , argStrings[ i ]
//                      );
//        }
//
//        return res;
//    }

    public StringBuffer formatImplementationOperationStart( TestManager manager
                                                          , OperationInfo op_data
                                                          , Object obj
                                                          , Object[] args
                                                          )
    {
        StringBuffer res = new StringBuffer();

        if( obj != null )
        {
            String objAsText = printObjectValue( false, obj );
            format( res, 0
                  , TraceConstants.IMPL_OBJECT
                  , TraceConstants.IMPL_OBJECT_ID_ATTR
                  , Integer.toString( Utils.identityHash( obj ), 16 )
                  , objAsText
                  );
        }
        else
        {
            format( res, 0
                  , TraceConstants.IMPL_CLASS
                  , TraceConstants.IMPL_CLASS_NAME_ATTR
                  , op_data.getDeclaringClass()
                  );
        }

        format( res, 1
              , TraceConstants.IMPL_CALL_START
              , TraceConstants.IMPL_SIGNATURE_ATTR
              , op_data.getPrettySignature()
              );

        String[] argTypes = op_data.getParTypes();
        String[] argStrings = printArgsValues( false, argTypes, args );
        boolean has_par_names = op_data.getParNames().length > 0;

        for( int i = 0; i < args.length; i++ )
        {
            formatValue( res
                       , TraceConstants.IMPL_VALUE
                       , ValueKind.ARGUMENT.toString()
                       , getRealValueType( argTypes[ i ], args[ i ] )
                       , has_par_names ? op_data.getParNames()[ i ] : null
                       , argStrings[ i ]
                       );
        }

        return res;
    }

    public StringBuffer formatImplementationOperationEnd(TestManager manager, long startTime, long endTime) {
        StringBuffer res = format(0, TraceConstants.IMPL_WORKING_TIME,
                TraceConstants.IMPL_START_TIME_ATTR, Long.toString(startTime),
                TraceConstants.IMPL_END_TIME_ATTR, Long.toString(endTime));
        format(res, -1, TraceConstants.IMPL_CALL_END);
        return res;
    }

    public StringBuffer formatImplementationResult( TestManager manager
                                                  , String resultType
                                                  , Object result
                                                  )
    {
        if( Utils.isVoidType(resultType) )
        {
            return new StringBuffer();

        } else {
            return formatValue(TraceConstants.IMPL_VALUE
                              , ValueKind.RESULT.toString()
                              , getRealValueType( resultType, result )
                              , null
                              , printObjectValue( Utils.isPrimitiveType(resultType)
                                                , false
                                                , result
                                                )
                              );
        }
    }

    public StringBuffer formatImplementationException(TestManager manager, Throwable error) {
        CharArrayWriter buf = new CharArrayWriter();
        error.printStackTrace(new PrintWriter(buf));

        return formatValue(TraceConstants.IMPL_VALUE, ValueKind.EXCEPTION.toString(),
                error.getClass().getName(), null, buf.toString());
    }

    public StringBuffer formatImplementationTimeInterval(TestManager manager, long startTime, long endTime) {
        return format(0, TraceConstants.IMPL_WORKING_TIME,
            TraceConstants.IMPL_START_TIME_ATTR, Long.toString(startTime),
            TraceConstants.IMPL_END_TIME_ATTR, Long.toString(endTime));
    }

    public StringBuffer formatImplementationValue( TestManager manager, String kind, String type, String name, Object value)
    {
        return formatValue(TraceConstants.IMPL_VALUE
                          , kind
                          , type
                          , name
                          , printObjectValue( false, value )
                          );
    }

    public StringBuffer formatImplementationMessage(TestManager manager, String info) {
        return formatInfo(TraceConstants.INFO, info);
    }

    public StringBuffer formatUserInfo(TestManager manager, String info) {
        return formatInfo(TraceConstants.USER_INFO, info);
    }

    public StringBuffer formatUserInfo(TestManager manager, String[] tags, String[] values) {
        StringBuffer res = new StringBuffer();
        addPrefix(res, 0);
        addTagStartWithTraceId(res, TraceConstants.USER_INFO_MAP );
        res.append(">");

        if (tags != null) {
            // TODO: Check: values != null && tags.length == values.length
            for (int i = 0; i < tags.length; i++) {
                formatProperty(res, tags[i], values[i]);
            }
        }

        addTagEnd(res, TraceConstants.USER_INFO_MAP);
        res.append(Utils.endl);
        return res;
    }

    public StringBuffer getTraceHeader() {
        StringBuffer res = new StringBuffer();
        addTagStart(res, TraceConstants.TRACE);
        addAttr( res, "version", "2.3" );
        res.append(">");
        res.append(Utils.endl);
        return res;
    }

    public StringBuffer getTraceFooter() {
        StringBuffer res = new StringBuffer();
        addTagEnd(res, TraceConstants.TRACE);
        res.append(Utils.endl);
        return res;
    }

    private StringBuffer formatException( TestManager manager
                                        , boolean isInterim
                                        , Throwable cause
                                        , String info
                                        , String[] tags
                                        , String[] values
                                        , boolean internal
                                        )
    {
        StringBuffer res = new StringBuffer();
        addPrefix(res, 1);
        addTagStartWithTraceId(res, TraceConstants.EXCEPTION );

        if( isInterim )
        {
          addAttr(res, TraceConstants.EXCEPTION_INTERIM_ATTR, "true");
        }

        addAttr(res, TraceConstants.EXCEPTION_INTERNAL_ATTR, Boolean.toString(internal));
        res.append(">");
        res.append(Utils.endl);

        addPrefix(res, 0);
        addTagStart(res, TraceConstants.EXCEPTION_INFO);
        res.append(">");
        addText(res, info);
        addTagEnd(res, TraceConstants.EXCEPTION_INFO);
        res.append(Utils.endl);

        if (cause != null) {
            CharArrayWriter buf = new CharArrayWriter();
            cause.printStackTrace(new PrintWriter(buf));
            addPrefix(res, 0);
            addTagStart(res, TraceConstants.EXCEPTION_BACKTRACE);
            res.append(">");
            addText(res, buf.toString());
            addTagEnd(res, TraceConstants.EXCEPTION_BACKTRACE);
            res.append(Utils.endl);
        }

        if (tags != null) {
            // TODO: Check: values != null && tags.length == values.length
            for (int i = 0; i < tags.length; i++) {
                formatProperty(res, tags[i], values[i]);
            }
        }

        addPrefix(res, -1);
        addTagEnd(res, TraceConstants.EXCEPTION);
        res.append(Utils.endl);

        return res;
    }

    private StringBuffer formatInfo(String tag, String info) {
        StringBuffer res = new StringBuffer();
        addPrefix(res, 0);
        addTagStartWithTraceId(res, tag);
        res.append(">");
        addText(res, info);
        addTagEnd(res, tag);
        res.append(Utils.endl);
        return res;
    }

    private StringBuffer formatProperty(StringBuffer acc, String name, String value) {
        addPrefix(acc, 0);
        addTagStart(acc, TraceConstants.PROPERTY);
        addAttr(acc, TraceConstants.NAME_ATTR, name);

        if (value != null && (value.indexOf('\n') > 0 || value.indexOf('\r') > 0)) {
            acc.append(">");
            addText(acc, value);
            addTagEnd(acc, TraceConstants.PROPERTY);

        } else {
            addAttr(acc, TraceConstants.VALUE_ATTR, value);
            acc.append("/>");
        }

        acc.append(Utils.endl);

        return acc;
    }

    private StringBuffer formatValue(String tag, String kind, String type, String name, String value) {
        return formatValue(new StringBuffer(), tag, kind, type, name, value);
    }

    private StringBuffer formatValue(StringBuffer acc, String tag, String kind, String type, String name, String value)
    {
        addPrefix(acc, 0);
        // TODO :????????????????????????????
        if (testId >= 0) {
            addTagStartWithTraceId(acc, tag);
        } else {
            addTagStart(acc, tag);
        }
        addAttr(acc, TraceConstants.KIND_ATTR, kind);
        addAttr(acc, TraceConstants.TYPE_ATTR, type);
        addAttr(acc, TraceConstants.NAME_ATTR, name);

        if (value != null) {
            if (value != null && (value.indexOf('\n') > 0 || value.indexOf('\r') > 0)) {
                acc.append(">");
                addText(acc, value);
                addTagEnd(acc, tag);

            } else {
                addAttr(acc, TraceConstants.VALUE_ATTR, value);
                acc.append("/>");
            }

        } else {
            acc.append("/>");
        }

        acc.append(Utils.endl);

        return acc;
    }

    private StringBuffer format(int shift, String tag) {
        return format(new StringBuffer(), shift, tag);
    }

    private StringBuffer format(StringBuffer acc, int shift, String tag) {
        addPrefix(acc, shift);
        addTagStartWithTraceId(acc, tag);
        acc.append("/>");
        acc.append(Utils.endl);
        return acc;
    }

    private StringBuffer format(int shift, String tag, String name1, String value1) {
        return format(new StringBuffer(), shift, tag, name1, value1);
    }

    private StringBuffer format(StringBuffer acc, int shift, String tag, String name1, String value1) {
        addPrefix(acc, shift);
        addTagStartWithTraceId(acc, tag);
        if (value1 !=  null) {
            addAttr(acc, name1, value1);
        }
        acc.append("/>");
        acc.append(Utils.endl);
        return acc;
    }

    private StringBuffer format(int shift, String tag, String name1, String value1, String msg )
    {
        return format(new StringBuffer(), shift, tag, name1, value1, msg);
    }

    private StringBuffer format(StringBuffer acc, int shift, String tag, String name1, String value1, String msg) {
        addPrefix(acc, shift);
        addTagStartWithTraceId(acc, tag );
        if (value1 !=  null) {
            addAttr(acc, name1, value1);
        }
        if (msg != null) {
            acc.append(">");
            addText(acc, msg);
            addTagEnd(acc, tag);
        } else {
            acc.append("/>");
        }
        acc.append(Utils.endl);
        return acc;
    }

    private StringBuffer format(int shift, String tag, String name1, String value1, String name2, String value2) {
        StringBuffer res = new StringBuffer();
        addPrefix(res, shift);
        addTagStartWithTraceId(res, tag );
        if (value1 !=  null) {
            addAttr(res, name1, value1);
        }
        if (value2 !=  null) {
            addAttr(res, name2, value2);
        }
        res.append("/>");
        res.append(Utils.endl);
        return res;
    }

    //------------------------------------------------------------------------
    // Utility methods

    private StringBuffer addPrefix(StringBuffer acc, int shift) {
        if (shift < 0) {
            indent += shift;
        }

        for (int i = 0; i < indent; i++) {
            for (int j = 0; j < INDENT_SIZE; j++) {
                acc.append(' ');
            }
        }

        if (shift > 0) {
            indent += shift;
        }

        return acc;
    }

    private StringBuffer addAttr(StringBuffer acc, String name, String value) {
        if (value == null || "".equals(value)) {
            return acc;
        } else {
            acc.append(" ");
            acc.append(name);
            acc.append("=\"");
            addEscapeAttributeEntities(acc, value);
            acc.append("\"");
            return acc;
        }
    }

    private StringBuffer addAttr(StringBuffer acc, String name, long value) {
        acc.append(" ");
        acc.append(name);
        acc.append("=\"");
        acc.append(value);
        acc.append("\"");
        return acc;
    }

    private StringBuffer addTagStart(StringBuffer acc, String name) {
        acc.append("<");
        acc.append(name);
        return acc;
    }

    private StringBuffer addTagStartWithTraceId(StringBuffer acc, String name ) {
        acc.append("<");
        acc.append(name);
        addAttr(acc, TraceConstants.TRACE, testId );
        return acc;
    }

    private StringBuffer addTagEnd(StringBuffer acc, String name) {
        acc.append("</");
        acc.append(name);
        acc.append(">");
        return acc;
    }

    private StringBuffer addText(StringBuffer acc, String text) {
        if (text.indexOf('\n') > 0 || text.indexOf('\r') > 0) {
            acc.append("<![CDATA[");
            acc.append(text.replace('\0', ' '));
            acc.append("]]>");

        } else {
            addEscapeElementEntities(acc, text);
        }

        return acc;
    }

    /** Common buffer (for optimization). */
    //StringBuffer buffer = new StringBuffer();

    /** This will take the pre-defined entities in XML 1.0 and
     * convert their character representation to the appropriate
     * entity reference, suitable for XML elements' contents.
     */
    private StringBuffer addEscapeElementEntities(StringBuffer acc, String text) {
        char[] block = null;
        int i, last = 0, size = text.length();
        for ( i = 0; i < size; i++ ) {
            String entity = null;
            char c = text.charAt(i);
            switch( c ) {
                case '<' :
                    entity = "&lt;";
                    break;
                case '>' :
                    entity = "&gt;";
                    break;
                case '&' :
                    entity = "&amp;";
                    break;
                case '\0' :
                    entity = " ";
                    break;
                default:
                    break;
            }
            if (entity != null) {
                if ( block == null ) {
                    block = text.toCharArray();
                }
                acc.append(block, last, i - last);
                acc.append(entity);
                last = i + 1;
            }
        }
        if ( last == 0 ) {
            acc.append(text);
            return acc;
        }
        if ( last < size ) {
            if ( block == null ) {
                block = text.toCharArray();
            }
            acc.append(block, last, i - last);
        }
        return acc;
    }

    /** This will take the pre-defined entities in XML 1.0 and
     * convert their character representation to the appropriate
     * entity reference, suitable for XML attributes.
     */
    private StringBuffer addEscapeAttributeEntities(StringBuffer acc, String text) {
        char[] block = null;
        int i, last = 0, size = text.length();
        for ( i = 0; i < size; i++ ) {
            String entity = null;
            char c = text.charAt(i);
            switch( c ) {
                case '<' :
                    entity = "&lt;";
                    break;
                case '>' :
                    entity = "&gt;";
                    break;
                case '\'' :
                    entity = "&apos;";
                    break;
                case '\"' :
                    entity = "&quot;";
                    break;
                case '&' :
                    entity = "&amp;";
                    break;
                default:
                    break;
            }
            if (entity != null) {
                if ( block == null ) {
                    block = text.toCharArray();
                }
                acc.append(block, last, i - last);
                acc.append(entity);
                last = i + 1;
            }
        }
        if ( last == 0 ) {
            acc.append(text);
            return acc;
        }
        if ( last < size ) {
            if ( block == null ) {
                block = text.toCharArray();
            }
            acc.append(block, last, i - last);
        }
        return acc;
    }

}
//=============================================================================
// End of file FormatAsXml.java

