/*
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 * WriteToWriter
 * $Id: WriteToWriter.java,v 1.3 2008/05/16 13:02:36 almo Exp $
 *
 * Author: Alexei Barantsev <barancev@ispras.ru>
 * Created on March 28, 2003
 */
//=============================================================================

package jatva.tracer.impl;

//=============================================================================

import jatva.tracer.TracerException;

import java.io.IOException;
import java.io.Writer;
import java.util.Properties;

//=============================================================================

/** 
 * This class implements writing trace messages to <code>java.io.Writer</code>.
 *
 * @author  <a href="mailto:barancev@ispras.ru">Alexei Barantsev</a>
 * @version 1.0
 */
public class WriteToWriter extends AbstractWriter {

    private Writer writer;

    /** 
     * Creates a new instance of <code>WriteToWriter</code>.
     * 
     * @param writer The trace destination.
     */
    public WriteToWriter(Writer writer) {
        this.writer = writer;
    }

    public void configure(Properties writerProperties) {
        throw new TracerException("This trace writer can not be configured this way");
    }
    
    public void sendMessage(StringBuffer message) {
        try {
            synchronized (writer) {
                writer.write(message.toString());
                writer.flush(); // Needs optimization!
            }

        } catch (IOException e) {
            throw new TracerException("Could not write to writer " + writer, e);
        }
    }

}
//=============================================================================
// End of file WriteToWriter.java
