/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: AbstractHashSet.java,v 1.3 2006/02/27 16:24:53 kuliamin Exp $
 * Created on 13.05.2004
 *
 */
package jatva.util;

import jatva.iterators.lang.AbstractObjectIterator;
import jatva.iterators.lang.IObjectIterator;

/**
 * Base class for hash sets.
 * 
 * @author Victor Kuliamin
 *
 */
public abstract class AbstractHashSet
  extends    AbstractSet
  implements ISet
{
  abstract protected int     hash(Object o);
  abstract protected boolean equalObjects(Object o1, Object o2);
  abstract protected IList   makeList();
  
  abstract public    Object  clone();

  public AbstractHashSet() {}

  public AbstractHashSet(Object o)
  {
    _size = 1;
    SetNode p = new SetNode(o);
    _items[hash(o)&(_itemsNumber-1)] = p;
    p._bkw = p._frw = _last = p;   
  }
  
  public AbstractHashSet(ISet s)
  {
    super(s);
  }
  
  public void    clear()
  {
    _size  = 0;
    _last  = null;
    _items = new SetNode[_itemsNumber];
  }
  
  public int     size()
  {
    return _size;
  }
  
  public boolean add(Object o)
  {
    int h = hash(o)&(_itemsNumber-1);
    SetNode p = _items[h];
    
    while(p != null)
    {
      if(equalObjects(p._o, o)) return false;
      p = p._next;
    }
    
    p = new SetNode(o);
    p._next = _items[h];
    _items[h] = p;
    if(_size++ == 0) { _last = p._bkw = p._frw = p; } 
    else             { p._frw = _last._frw; p._bkw = _last; _last._frw = p; _last = p; } 

    return true;
  }

  public boolean contains(Object o)
  {
    int h = hash(o)&(_itemsNumber-1);
    SetNode p = _items[h];
    
    while(p != null)
    {
      if(equalObjects(p._o, o)) return true;
      p = p._next;
    }
    
    return false;
  }
  
  public boolean remove(Object o)
  {
    int h = hash(o)&(_itemsNumber-1);
    SetNode p = _items[h], q = null;
    
    while(p != null)
    {
      if(equalObjects(p._o, o))
      {
        if(q == null) _items[h] = p._next;
        else          q._next   = p._next;
        
        if(_size-- == 1)    _last = null;
        else
        { 
          p._bkw._frw = p._frw; 
          p._frw._bkw = p._bkw;
          if(p == _last) _last = p._bkw;
        }
        return true;
      }  
      
      q = p;
      p = p._next;
    }
    return false;
  }

  public boolean retain(ISet s)
  {
    boolean res = false;
    
    if(s != null && s.size() != 0)
    {
      SetNode p, q;
      for(int i = 0; i < _itemsNumber; i++)
      {
        q = null;
        if(null == (p = _items[i])) continue;
        while(p != null)
        {
          if(!s.contains(p._o))
          {
            res = true;
            if(q == null) _items[i] = p._next;
            else          q._next   = p._next;
            
            if(_size-- == 1)    _last = null;
            else
            { 
              p._bkw._frw = p._frw; 
              p._frw._bkw = p._bkw;
              if(p == _last) _last = p._bkw;
            }
          }
          else q = p;
          p = p._next;    
        }
      }
    }    
    return res;
  }
  
  public boolean replace(IMap m)
  {
    boolean res = false;
    // TODO Auto-generated method stub
    return res; 
  }
  
/*  private class Wrapper
  {
    Object _o;
    
    Wrapper(Object o) { _o = o; }
  }*/

  public IObjectIterator iterator()
  {
    return new SetIterator(this);
  }
  
  protected static class SetIterator
    extends    AbstractObjectIterator
    implements IObjectIterator
  {
    private SetNode         p = null;
    private AbstractHashSet s = null; 
        
    SetIterator(AbstractHashSet s)  
    {
      this.s = s;
      if(s._last == null) p = null;
      else                p = s._last._frw;
    }
       
    SetIterator(SetIterator r)
    {
      s = r.s;
      p = r.p;
    }
    
    public boolean stop()
    {
      if(p != null) return true;
      else          return false;
    }
  
    public void shift()
    {
      if(p == s._last) p = null;
      else             p = p._frw;
    }
      
    public Object value()
    {
      return p._o;
    }
  
    public Object clone()
    {
      return new SetIterator(this);
    }
      
    public boolean equals(Object o)
    {
      if(!(o instanceof SetIterator)) return false;
      
      SetIterator r = (SetIterator)o;
       
      return r.s == s && r.p == p;
    }

    public int hashCode()
    {
      return s.hashCode() + 17*System.identityHashCode(p);
    }
    
    public void remove()
    {
      Object o = p._o;
      p = p._frw;
      if(p == s._last._frw) p = null;
      
      s.remove(o);
    }
  }
  
  static int _itemsNumber = 16;
  
  SetNode _items[] = new SetNode[_itemsNumber];
  int     _size;//    = 0;
  SetNode _last;//    = null;
  
  protected static class SetNode
  {
    SetNode _bkw;
    SetNode _frw;
    SetNode _next;
    Object  _o;
    
    SetNode(Object o) { _o = o; }
  }
}
