/*
 * Created on 13.05.2004
 *
 */
package jatva.util;

import jatva.iterators.lang.IObjectIterator;
import jatva.iterators.lang.IObjectPairIterator;

/**
 * @author Victor Kuliamin
 *
 */
public abstract class AbstractMap
  implements IMap
{
  abstract protected int       hashKey(Object o);
  abstract protected int       hashValue(Object o);
  abstract protected boolean   equalValues(Object o1, Object o2);
  abstract protected IList     makeList();
  abstract protected ISet      makeSet();
  abstract protected IMultiSet makeMultiSet();

  abstract public Object clone();
  abstract public void   clear();
  abstract public int    size();
  
  abstract public boolean contains(Object key, Object val); 
  abstract public boolean containsKey(Object key);
  abstract public Object  get(Object key);
  abstract public Object  put(Object key, Object value);
  abstract public boolean remove(Object key, Object val);
  abstract public Object  removeKey(Object key);
  abstract public IObjectPairIterator iterator();


  public AbstractMap() {}
  
  public AbstractMap(Object key, Object val)
  {
    put(key, val);
  }
  
  public AbstractMap(IMap m)
  {
    if(m != null && m.size() != 0)
      for(IObjectPairIterator i = m.iterator(); !i.stop(); i.shift())
        put(i.value1(), i.value2());
  }
  
  public AbstractMap(ISet s)
  {
    if(s != null && s.size() != 0)
      for(IObjectIterator i = s.iterator(); !i.stop(); i.shift())
        put(i.value(), i.value());
  }

  public AbstractMap(ISet s, Object o)
  {
    if(s != null && s.size() != 0)
      for(IObjectIterator i = s.iterator(); !i.stop(); i.shift())
        put(i.value(), o);
  }

  public boolean isEmpty()
  {
    return size() == 0;
  }

  public boolean containsValue(Object value)
  {
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      if(equalValues(i.value2(), value)) return true;

    return false;
  }
  
  public boolean containsAll(IMap m)
  {
    if(m == null)          return false;
    else if(m.size() == 0) return true;
    else
      for(IObjectPairIterator i = m.iterator(); !i.stop(); i.shift())
        if(!contains(i.value1(), i.value2())) return false;
    return true;
  }

  public boolean containsRefs(Object key, Object val)
  {
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      if(i.value1() == key && i.value2() == val) return true;
      
    return false;
  }

  public boolean containsKeyRef(Object key)
  {
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      if(i.value1() == key) return true;
      
    return false;
  }

  public boolean containsValueRef(Object value)
  {
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      if(i.value2() == value) return true;
      
    return false;
  }

  public boolean putAll(IMap m)
  {
    boolean res = false;
    if(m != null && m.size() != 0)
      for(IObjectPairIterator i = m.iterator(); !i.stop(); i.shift())
        res |= !equalValues(put(i.value1(), i.value2()), i.value2());
        
    return res;
  }

  public boolean removeValue(Object val)
  {
    boolean res = false;
    IObjectPairIterator i = iterator();
    while(!i.stop())
    {
      if(equalValues(i.value2(), val)) { res = true; i.remove(); } 
      else                             i.shift();
    }
    
    return res;
  }

  public boolean removeAll(IMap m)
  {
    boolean res = false;
    if(m != null && m.size() != 0)
      for(IObjectPairIterator i = m.iterator(); !i.stop(); i.shift())
        res |= remove(i.value1(), i.value2());
        
    return res;
  }

  public boolean removeKeys(ISet s)
  {
    boolean res = false;
    if(s != null && s.size() != 0)
    {
      IObjectPairIterator i = iterator();
      while(!i.stop())
      {
        if(s.contains(i.value1())) { res = true; i.remove(); }
        else                       i.shift();
      }
    }
        
    return res;
  }

  public boolean removeValues(ISet s)
  {
    boolean res = false;
    if(s != null && s.size() != 0)
      for(IObjectIterator i = s.iterator(); !i.stop(); i.shift())
        res |= removeValue(i.value());
        
    return res;
  }

  public boolean retain(IMap m)
  {
    boolean res = false;
    if(m != null && m.size() != 0)
    {
      IObjectPairIterator i = iterator();
      while(!i.stop())
      {
        if(!m.contains(i.value1(), i.value2())) { res = true; i.remove(); }
        else                                    i.shift();
      }
    }
        
    return res;
  }

  public boolean retainKeys(ISet s)
  {
    boolean res = false;
    if(s != null && s.size() != 0)
    {
      IObjectPairIterator i = iterator();
      while(!i.stop())
      {
        if(!s.contains(i.value1())) { res = true; i.remove(); }
        else                        i.shift();
      }
    }
        
    return res;
  }

  public boolean retainValues(ISet s)
  {
    boolean res = false;
    if(s != null && s.size() != 0)
    {
      IObjectPairIterator i = iterator();
      while(!i.stop())
      {
        if(!s.contains(i.value2())) { res = true; i.remove(); }
        else                        i.shift();
      }
    }
        
    return res;
  }

  public boolean extend(IMap m)
  {
    boolean res = false;
    if(m != null && m.size() != 0)
      for(IObjectPairIterator i = m.iterator(); !i.stop(); i.shift())
        if(!containsKey(i.value1())) { res = true; put(i.value1(), i.value2()); }
        
    return res;
  }

  public boolean replace(IMap m)
  {
    boolean res = false;
    if(m != null && m.size() != 0)
      for(IObjectPairIterator i = m.iterator(); !i.stop(); i.shift())
        if(containsKey(i.value1()) && !equalValues(get(i.value1()),i.value2())) { res = true; put(i.value1(), i.value2()); }
        
    return res;
  }

  public void compose(IMap m)
  {
    if(m != null && m.size() != 0)
    {
      IObjectPairIterator i = iterator();
      while(!i.stop())
      {
        if(m.containsKey(i.value1())) { i.replaceObject2(m.get(i.value1())); i.shift(); }
        else                          i.remove();
      }
    }
  }

  public ObjectPair[] toArray()
  {
    ObjectPair[] r = new ObjectPair[size()];
    int j = 0;
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      r[j++] = new ObjectPair(i.value1(), i.value2());
    
    return r;
  }

  public IList entryList()
  {
    IList l = makeList();
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      l.addLast(new ObjectPair(i.value1(), i.value2()));
    
    return l;
  }

  public ISet entrySet()
  {
    ISet s = makeSet();
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      s.add(new ObjectPair(i.value1(), i.value2()));
    
    return s;
  }

  public ISet image(ISet s)
  {
    if(s == null) return null;
    
    ISet r = makeSet();
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      if(s.contains(i.value1())) r.add(i.value2());
    
    return r;
  }

  public ISet range()
  {
    ISet r = makeSet();
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      r.add(i.value2());
    
    return r;
  }

  public ISet preimage(ISet s)
  {
    if(s == null) return null;
    
    ISet r = makeSet();
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      if(s.contains(i.value2())) r.add(i.value1());
    
    return r;
  }

  public ISet keys()
  {
    ISet r = makeSet();
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      r.add(i.value1());
    
    return r;
  }

  public IMultiSet values(ISet s)
  {
    if(s == null) return null;
    
    IMultiSet r = makeMultiSet();
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      if(s.contains(i.value1())) r.add(i.value2());
    
    return r;
  }

  public IMultiSet values()
  {
    IMultiSet r = makeMultiSet();
    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
      r.add(i.value2());
    
    return r;
  }

  public boolean equals(Object o)
  {
    if(o == null) return false;
    else if(o == this) return true;
    else if(o instanceof IMap)
    {
      IMap m = (IMap)o;
      if(m.size() != size()) return false;
      else if(size() == 0)   return true;
      
      for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
        if(!m.contains(i.value1(), i.value2())) return false;
      
      return true;
    }
    else if(o instanceof IRelation)
    {
      IRelation c = (IRelation)o;
      if(c.size() != size()) return false;
      else if(size() == 0)   return true;
      
      for(IObjectPairIterator i = c.iterator(); !i.stop(); i.shift())
        if(!contains(i.value1(), i.value2())) return false;
        
      return true;      
    }
    else return false;
  }

  public boolean equalsByRefs(Object o)
  {
    if(o == null) return false;
    else if(o == this) return true;
    else if(o instanceof IMap)
    {
      IMap m = (IMap)o;
      if(m.size() != size()) return false;
      else if(size() == 0)   return true;
      
      for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
        if(!m.containsRefs(i.value1(), i.value2())) return false;
      
      return true;
    }
    else if(o instanceof IRelation)
    {
      IRelation c = (IRelation)o;
      if(c.size() != size()) return false;
      else if(size() == 0)   return true;
      
      for(IObjectPairIterator i = c.iterator(); !i.stop(); i.shift())
        if(!containsRefs(i.value1(), i.value2())) return false;
        
      return true;      
    }
    else return false;
  }

  public int hashCode()
  {
    int r = 0;
    if(size() != 0)
      for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
        r += hashKey(i.value1()) + hashValue(i.value2());

    return r;
  }

  public String toString()
  {
    StringBuffer s = new StringBuffer("(");
    boolean f = true;

    for(IObjectPairIterator i = iterator(); !i.stop(); i.shift())
    {
      if(!f) s.append(", ");
      else   f = false;
      s.append("" + i.value1() + " : " + i.value2());
    }
    s.append(')');
    
    return s.toString();
  }
}
