/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: AbstractSet.java,v 1.1 2005/05/12 03:53:23 kuliamin Exp $
 * Created on 20.05.2004
 *
 */
package jatva.util;

import jatva.iterators.lang.IObjectIterator;

/**
 * Abstract class implementing most of set methods
 * by means of others.
 * 
 * @author Victor Kuliamin
 */
public abstract class AbstractSet
  implements ISet
{
  abstract protected int   hash(Object o);
  abstract protected IList makeList();

  abstract public Object  clone();
  abstract public void    clear();
  abstract public int     size();
  abstract public boolean add(Object o);
  abstract public boolean contains(Object o);
  abstract public boolean remove(Object o);
  abstract public IObjectIterator iterator();
  
  public AbstractSet() {}
  
  public AbstractSet(Object o)
  {
    add(o);
  }
  
  public AbstractSet(ISet s)
  {
    if(s != null && s.size() != 0)
      for(IObjectIterator i = s.iterator(); !i.stop(); i.shift()) add(i.value());
  }

  public boolean isEmpty()
  {
    return size() == 0;
  }

  public boolean addAll(ISet s)
  {
    boolean res = false;
    if(s != null && s.size() != 0)    
      for(IObjectIterator i = s.iterator(); !i.stop(); i.shift())
        res |= add(i.value());
    
    return res;
  }
  
  public boolean containsRef(Object o)
  {
    for(IObjectIterator i = iterator(); !i.stop(); i.shift())
      if(i.value() == o) return true;
    
    return false;
  }

  public boolean containsAll(ISet s)
  {
    if(s == null)          return false;
    else if(s.size() == 0) return true;
    else
      for(IObjectIterator i = s.iterator(); !i.stop(); i.shift())
        if(!contains(i.value())) return false;
    
    return true;
  }

  public boolean isSubset(ISet s)
  {
    if(s == null) return false;
    else          return s.containsAll(this);
  }

  public boolean removeAll(ISet s)
  {
    boolean res = false;
    if(s != null && s.size() != 0)
      for(IObjectIterator i = s.iterator(); !i.stop(); i.shift())
        res |= this.remove(i.value());
    
    return res;
  }

  public boolean retain(ISet s)
  {
    boolean res = false;
    
    if(s != null && s.size() != 0)
    {
      IObjectIterator i = iterator();
      
      while(!i.stop())
      {
        if(!s.contains(i.value())) { i.remove(); res = true; } 
        else                       i.shift(); 
      }
    }
    return res;
  }
  
  public boolean replace(Object o1, Object o2)
  {
    if(remove(o1)) { add(o2); return true; }
    else           return false;
  }
  
  public boolean replace(IMap m)
  {
    // TODO Auto-generated method stub
    return false;
  }

  public IList toList()
  {
    IList l = makeList();
    for(IObjectIterator i = iterator(); !i.stop(); i.shift())
      l.addLast(i.value());
    
    return l;
  }

  public Object[] toArray()
  {
    Object res[] = new Object[size()];
    int j = 0;
    for(IObjectIterator i = iterator(); !i.stop(); i.shift())
      res[j++] = i.value();

    return res;
  }

  public ISet power()
  {
    // TODO Auto-generated method stub
    return null;
  }

  public String  toString()
  {
    StringBuffer s = new StringBuffer("{");
    boolean f = true;

    for(IObjectIterator i = iterator(); !i.stop(); i.shift())
    {
      if(!f) s.append(", ");
      else   f = false;
      s.append(i.value());
    }
    s.append('}');
    
    return s.toString();
  }

  public int hashCode()
  {
    int r = 0;
    if(size() != 0)
      for(IObjectIterator i = iterator(); !i.stop(); i.shift())
        r += hash(i.value());

    return r;
  }

  public boolean equals(Object o)
  {
    if(o == null) return false;
    else if(o == this) return true;
    else if(o instanceof ISet)
    {
      ISet s = (ISet)o;
      if(s.size() != size()) return false;
      else if(size() == 0)   return true;
      
      for(IObjectIterator i = iterator(); !i.stop(); i.shift())
        if(!s.contains(i.value())) return false;
      
      return true;
    }
    else if(o instanceof IMultiSet)
    {
      IMultiSet c = (IMultiSet)o;
      if(c.size() != size()) return false;
      else if(size() == 0)   return true;
      
      for(IObjectIterator i = c.iterator(); !i.stop(); i.shift())
        if(!contains(i.value())) return false;
        
      return true;      
    }
    else return false;
  }
  
  public boolean equalsByRefs(Object o)
  {
    if(o == null) return false;
    else if(o == this) return true;
    else if(o instanceof ISet)
    {
      ISet s = (ISet)o;
      if(s.size() != size()) return false;
      else if(size() == 0)   return true;
      
      for(IObjectIterator i = iterator(); !i.stop(); i.shift())
        if(!s.containsRef(i.value())) return false;
      
      return true;
    }
    else if(o instanceof IMultiSet)
    {
      IMultiSet c = (IMultiSet)o;
      if(c.size() != size()) return false;
      else if(size() == 0)   return true;
      
      for(IObjectIterator i = c.iterator(); !i.stop(); i.shift())
        if(!containsRef(i.value())) return false;
        
      return true;      
    }
    else return false;
  }
}
