package jatva.util;

import jatva.lang.InsufficientAccessibilityError;
import jatva.lang.TranslationError;

import java.lang.reflect.Constructor;

public class CtorInfo extends OperationInfo
{
    private Constructor<?> ctor;

    public CtorInfo( String extendedReflectSignature )
    {
        super( extendedReflectSignature );
    }

    @Override
    protected final Class<?>[] obtainExecutableMember()
    {
        ctor = Utils.getConstructorByPartialSignature( reflectSignature );

        if( ctor == null )
            throw new TranslationError( "Invalid constructor signature"
                                      , new String[]{ "constructor signature" }
                                      , new String[]{ reflectSignature }
                                      );

        try
        {
          ctor.setAccessible( true );
        }
        catch( SecurityException e )
        {
          throw new InsufficientAccessibilityError( "Insufficient rights to call implementation constructor through reflection"
                                                  , e
                                                  , new String[]{"constructor"}
                                                  , new String[]{reflectSignature}
                                                  );
        }

        this.declaringClass = Utils.prettifyClass( ctor.getDeclaringClass() );
        this.returnType = declaringClass;

        return ctor.getParameterTypes();
    }

    public Constructor<?> getCtor()
    {
        return ctor;
    }

    @Override
    protected void buildPrettySignature()
    {
        StringBuffer res = new StringBuffer();
        res.append( declaringClass );
        appendParTypes( res );
        this.prettySignature = res.toString();
    }
}
