/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: IdentityList.java,v 1.1 2005/05/12 03:53:23 kuliamin Exp $
 * Created on 16.02.2005
 *
 */
package jatva.util;

/** 
 * Imlements list of objects.
 *
 * @author Victor Kuliamin
 *
 */
public class IdentityList
{
  int _size = 0;
  IdentityListNode _head = null;

  int _lastIndex = 0;
  IdentityListNode _last = null;

  public IdentityList()
  {
  }

  public IdentityList(IdentityList other)
  {
    IdentityListNode p = other._head;
  
    if(p == null) return;
  
    _size = other._size;
    if(other._last != null) _lastIndex = other._lastIndex;
    else                    _last      = null;
  
    IdentityListNode s = new IdentityListNode(p), q = null;
    _head = s;
  
    p = p._next;
    while(p != other._head)
    {
      q = new IdentityListNode(p);
      s._next = q;
      q._prev = s;
      if(other._last == p) _last = q;
      p = p._next;
      s = q;
    }
  
    _head._prev = s;
    s._next = _head;
  }

  public int size()
  {
    return _size;
  }

  public boolean isEmpty()
  {
    return (_size == 0);
  }

  public Object first()
  {
    if(_head != null) return _head._val;
    else return null;
  }
  
  public void setFirst(Object o)
  {
    if(_head != null) _head._val = o;
    else
    {
      _head = new IdentityListNode(o);
      _head._next = _head;
      _head._prev = _head;
      _size++;
    }
  }

  public Object last()
  {
    if(_head != null) return _head._prev._val;
    else return null;
  }
  
  public void setLast(Object o)
  {
    if(_head != null) _head._prev._val = o;
    else
    {
      _head = new IdentityListNode(o);
      _head._next = _head;
      _head._prev = _head;
      _size++;
    }
  }

  public void addLast(Object val)
  {
    if(_head == null)
    {
      _head = new IdentityListNode(val);
      _head._next = _head;
      _head._prev = _head;
    }
    else
    {
      IdentityListNode p = new IdentityListNode(val);
      p._prev = _head._prev;
      p._next = _head;
      _head._prev._next = p;
      _head._prev = p;
    }

    _size++;
  }

  public void addFirst(Object val)
  {
    if(_head == null)
    {
      _head = new IdentityListNode(val);
      _head._next = _head;
      _head._prev = _head;
    }
    else
    {
      IdentityListNode p = new IdentityListNode(val);
      p._prev = _head._prev;
      p._next = _head;
      _head._prev._next = p;
      _head._prev = p;
      _head = p;
    }

    _size++;

    if(_last != null) _lastIndex++;
  }

  public void clear()
  {
    _head = null;
    _size = 0;

    _last = null;
  }

  public void removeLast()
  {
    if(_head == null)   return;
    else if(_size == 1) _head = null;
    else
    {
      _head._prev._prev._next = _head;
      _head._prev = _head._prev._prev;
    }

    _size--;

    if(_last != null && _lastIndex >= _size) _last = null;
  }

  public void removeFirst()
  {
    if(_head == null) return;
    else if(_size == 1) _head = null;
    else
    {
      _head._prev._next = _head._next;
      _head._next._prev = _head._prev;
      _head = _head._next;
    }

    _size--;

    if(_last != null && _size > 0) _lastIndex--;
    else if(_size == 0) _last = null;
  }

  protected IdentityListNode getNode(int i)
  {
    IdentityListNode s = null;
  
    if(_head != null)
    {
      if(_last != null && i - _lastIndex >= 0 && _size - i >= i - _lastIndex)
      {
        s = _last;
        while(i - _lastIndex > 0) { s = s._next; i--; }
      }
      else if(
        _last != null && i - _lastIndex >= 0 && _size - i < i - _lastIndex
        || _last == null && i >= _size - i
        )
      {
        s = _head;
        while(_size - i > 0) { s = s._prev; i++; }
      }
      else if(_last != null && i - _lastIndex < 0 && i >= _lastIndex - i)
      {
        s = _last;
        while(_lastIndex - i > 0) { s = s._prev; i++; }
      }
      else/* if(
          _last != null && i - _lastIndex < 0 && i < _lastIndex - i
       || _last == null && i < _size - i
      )*/
      {
        s = _head;
        while(i > 0) { s = s._next; i--; }
      }
    }
  
    return s;
  }

  public void add(int i, Object val)
  {
    if(i < 0 || i > _size) return;

    IdentityListNode p = new IdentityListNode(val);
    IdentityListNode s = getNode(i);

    if(s != null)
    {
      p._prev = s._prev;
      s._prev._next = p;
      p._next = s;
      s._prev = p;      
      if(i == 0) _head = p;
    }
    else
    {
      _head = p;
      p._prev = _head;
      p._next = _head;
    }

    _size++;

    _lastIndex = i;
    _last = p;
  }

  public void set(int i, Object val)
  {
    if(i < 0 || i >= _size) return;

    IdentityListNode s = getNode(i);

    s._val = val;

    _lastIndex = i;
    _last = s;
  }

  public Object get(int i)
  {
    if(i < 0 || i >= _size) return null;

    IdentityListNode s = getNode(i);

    _lastIndex = i;
    _last = s;

    return s._val;
  }

  public void remove(int i)
  {
    if(i < 0 || i >= _size) return;

    if(_size == 1)
    {
      clear();
      return;
    }

    IdentityListNode s = getNode(i);

    s._prev._next = s._next;
    s._next._prev = s._prev;
    if(s == _head) _head = s._next;

    _size--;

    _lastIndex = (i == _size)?0:i;
    _last = s._next;
  }

  public boolean equals(Object obj)
  {
    if(!(obj instanceof IdentityList)) return false;
    
    IdentityList other = (IdentityList)obj;
    if(other._size != _size) return false;
    if(other == this)       return true;

    IdentityListNode p = _head, s = other._head;

    for(int i = 0; i < _size; i++)
    {
      if(p._val != s._val) return false;

      p = p._next;
      s = s._next;
    }

    return true;
  }

  public int hashCode()
  {
    int r = 0;
    IdentityListNode p = _head;

    for(int i = 0; i < _size; i++)
    {
      r += (p._val == null)?0:p._val.hashCode();
      p = p._next;
    }

    return r;
  }

  public Object clone()
  {
    return new IdentityList(this);
  }

  public boolean contains(Object val)
  {
    IdentityListNode p = _head;

    if(p == null) return false;

    do
    {
      if(p._val == val) return true;

      p = p._next;
    }
    while(p != _head);

    return false;
  }

  public boolean remove(Object val)
  {
    IdentityListNode p = _head;

    if(p == null) return false;

    do
    {
      if(p._val == val) 
      {
        p._prev._next = p._next;
        p._next._prev = p._prev;
        if(p == _head) _head = p._next;

        _size--;

        _lastIndex = 0;
        _last = null;
        return true;
      }

      p = p._next;
    }
    while(p != _head);

    return false;
  }

  public String toString()
  {
    StringBuffer bf = new StringBuffer("[");

    IdentityListNode p = _head;

    if(p != null)
    {
      while(true)
      {
        bf.append(p._val);

        if((p = p._next) != _head) bf.append(", ");
        else break;  
      }
    }

    bf.append("]");

    return bf.toString();
  }

  public IdentityListEnumerator getEnumerator()
  {
    return new IdentityListEnumerator(this);
  }

  public class IdentityListEnumerator
  {
    IdentityList _base = null;
    IdentityListNode _current = null;

    public IdentityListEnumerator(IdentityList b)
    {
      _base = b;
    }

    public Object current()
    {
      return _current._val;
    }

    public boolean moveNext()
    {
      if(_current == null && _base._head == null) return false;
      else if(_current == null)                   { _current = _base._head; return true; }
      else if(_current._next == _base._head)      { _current = null; return false; }
      else                                        { _current = _current._next; return true; }
    }
  }

  protected class IdentityListNode
  {
    public IdentityListNode _next = null;
    public IdentityListNode _prev = null;
    public Object _val = null;
  
    public IdentityListNode(Object elem)
    {
      _val = elem;
    }

    public IdentityListNode(IdentityListNode right)
    {
      _val = right._val;
    }

    public Object clone()
    {
      return new IdentityListNode(_val);
    }
  }

}

