/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: IdentityToIntMap.java,v 1.2 2007/10/19 06:47:21 kuliamin Exp $
 * Created on 01.06.2005
 *
 */
package jatva.util;

import jatva.iterators.lang.AbstractObjectIntPairIterator;
import jatva.iterators.lang.IObjectIntPairIterator;

/** 
 *
 * @author Victor Kuliamin
 *
 */
public class IdentityToIntMap
{
  protected IdentityToIntMapNode _node = null;
  protected int _size = 0;
  
  public IdentityToIntMap()
  {
  }
  
  public synchronized String toStringX()
  {
    StringBuffer st = new StringBuffer("{");
    
    if(_node != null)
      _node.appendStringBuffer(st);
    
    st.append("}");
    
    return st.toString();
  }
  
  public synchronized String toString()
  {
    StringBuffer st = new StringBuffer("{");
    IdentityToIntMapNode n = minimum();
    
    if(n != null)
    {
      st.append("" + n._key + " : " + n._value);
      n = next(n);
    }
    
    while(n != null)
    {
      st.append(", ");
      st.append("" + n._key + " : " + n._value);
      n = next(n);
    }
      
    st.append("}");
    
    return st.toString();
  }
  
  protected void rotate(boolean left, IdentityToIntMapNode x)
  {
    IdentityToIntMapNode y;
    
    if(left)
    {
      y = x._right;
      x._right = y._left;
      if(y._left != null)
        y._left._parent = x;
      
      y._parent = x._parent;
      if(x._parent != null)
      {
        if(x == x._parent._left)
          x._parent._left = y;
        else
          x._parent._right = y;
      }
      else
        _node = y;
      
      y._left = x;
      x._parent = y;
    }
    else
    {
      y = x._left;
      x._left = y._right;
      if(y._right != null)
        y._right._parent = x;
      
      y._parent = x._parent;
      if(x._parent != null)
      {
        if(x == x._parent._left)
          x._parent._left = y;
        else
          x._parent._right = y;
      }
      else
        _node = y;
      
      y._right = x;
      x._parent = y;
    }
  }
  
  private int hashKey(Object o)
  {
    return Utils.identityHash(o);
  }
  
  public synchronized boolean containsKey(Object key)
  {
    if(_node == null)
      return false;
    
    IdentityToIntMapNode p = _node;
    
    do
    {
      if(hashKey(p._key) == hashKey(key))
        return true;
      else if(hashKey(p._key) > hashKey(key))
        p = p._left;
      else
        p = p._right;
    }
    while(p != null);
    
    return false;
  }
  
  public synchronized boolean containsValue(int value)
  {
    if(_node == null)
      return false;
    
    return _node.containsValue(value);
  }
  
  protected IdentityToIntMapNode minimum()
  {
    if(_node == null)
      return null;
    else
      return _node.minimum();
  }
  
  protected IdentityToIntMapNode maximum()
  {
    if(_node == null)
      return null;
    else
      return _node.maximum();
  }
  
  protected IdentityToIntMapNode next(IdentityToIntMapNode p)
  {
    if(p._right != null)
      return p._right.minimum();
    
    while(p._parent != null && p == p._parent._right)
      p = p._parent;

    return p._parent;
  }
  
  protected IdentityToIntMapNode prev(IdentityToIntMapNode p)
  {
    if(p._left != null)
      return p._left.maximum();
    
    while(p._parent != null && p == p._parent._left)
      p = p._parent;

    return p._parent;
  }
  
  public synchronized int get(Object key)
  {
    if(_node == null)
      return 0;
    
    IdentityToIntMapNode p = _node;
    
    do
    {
      if(hashKey(p._key) == hashKey(key))
        return p._value;
      else if(hashKey(p._key) > hashKey(key))
        p = p._left;
      else
        p = p._right;
    }
    while(p != null);
    
    return 0;
  }
  
  public synchronized int put(Object key, int value)
  {
    if(_node == null)
    {
      _node = new IdentityToIntMapNode(key, value);
      _node._red = false;
      _size++;
      
      return 0;
    }
    
    IdentityToIntMapNode p = _node, q;
    int res;
    
    do
    {
      q = p;
      
      if(hashKey(p._key) == hashKey(key))
      {
        res = p._value;
        p._value = value;
        return res;
      }
      else if(hashKey(p._key) > hashKey(key))
        p = p._left;
      else
        p = p._right;
    }
    while(p != null);
    
    if(hashKey(q._key) > hashKey(key))
      q._left = ( p =  new IdentityToIntMapNode(key, value) );
    else
      q._right = ( p = new IdentityToIntMapNode(key, value) );
    p._parent = q;
    
    _size++;
    
    IdentityToIntMapNode r;
    
//    System.out.println("After add : " + toString());
    
    while(q != null && q._red == true)
    {
      if(q == q._parent._left)
      {
//        System.out.print("q is on the left, ");
        
        r = q._parent._right;
        if(r == null || !r._red )
        {
          if(p == q._right)
          {
//            System.out.println("case 2");
                        
            q._right = p._left;
            if(p._left != null) p._left._parent = q;
            p._left = q;
            p._parent = q._parent;
            q._parent._left = p;
            q._parent = p;
            p = q;
            q = q._parent;
            
//            System.out.println("After rotation : " + toString());
          }
          
//          System.out.println("case 3");
          
          if(q._parent._parent != null)
          {
            if(q._parent._parent._left == q._parent)
              q._parent._parent._left = q;
            else
              q._parent._parent._right = q;
          }
          else
            _node = q;
          
          q._red = false;
          q._parent._red = true;
          q._parent._left = q._right;
          if(q._right != null) q._right._parent = q._parent;
          q._right = q._parent;
          q._parent = q._parent._parent;
          q._right._parent = q;
          
          return 0;
        }
        else
        {
//          System.out.println("case 1");
          
          q._red = false;
          r._red = false;
          q._parent._red = true;
          p = q._parent;
          q = p._parent;
        }
      }
      else
      {
//        System.out.print("q is on the right, ");

        r = q._parent._left;
        if(r == null || !r._red)
        {
          if(p == q._left)
          {
//            System.out.println("case 2");

            q._left = p._right;
            if(p._right != null) p._right._parent = q;
            p._right = q;
            p._parent = q._parent;
            q._parent._right = p;
            q._parent = p;
            p = q;
            q = q._parent;

//            System.out.println("After rotation : " + toString());
          }
          
//         System.out.println("case 3");
         
         if(q._parent._parent != null)
          {
            if(q._parent._parent._left == q._parent)
              q._parent._parent._left = q;
            else
              q._parent._parent._right = q;
          }
          else
            _node = q;
          
          q._red = false;
          q._parent._red = true;
          q._parent._right = q._left;
          if(q._left != null) q._left._parent = q._parent;
          q._left = q._parent;
          q._parent = q._parent._parent;
          q._left._parent = q;
          
          return 0;
        }
        else
        {
          q._red = false;
          r._red = false;
          q._parent._red = true;
          p = q._parent;
          q = p._parent;
        }
      }
    }
    
    _node._red = false;
    
    return 0;
  }
  
  public synchronized int remove(Object key)
  {
    if(_node == null)
      return 0;
    
    IdentityToIntMapNode p = _node;
    
    do
    {
      if(hashKey(p._key) > hashKey(key))
        p = p._left;
      else if(hashKey(p._key) < hashKey(key))
        p = p._right;
      else
        break;
    }
    while(p != null);
    
    if(p == null)
      return 0;
    
    boolean isLeft;
    int res = p._value;
    
    if(p._left == null)
    {
      if(p._parent == null)
      {
        _node = p._right;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._right;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._right;
          isLeft = false;
        }
      }
      
      if(p._right != null)
        p._right._parent = p._parent;
      
      if(!p._red)
        fixup(p._parent, p._right, isLeft);
    }
    else if(p._right == null)
    {
      if(p._parent == null)
      {
        _node = p._left;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._left;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._left;
          isLeft = false;
        }
      }
      
      p._left._parent = p._parent;
      if(!p._red)
        fixup(p._parent, p._left, isLeft);
    }
    else
    {
      IdentityToIntMapNode q = next(p);
      p._key = q._key;
      p._value = q._value;
      
      if(q._parent == null)
      {
        _node = q._right;
        isLeft = true;
      }
      else
      {
        if(q == q._parent._left)
        {
          q._parent._left = q._right;
          isLeft = true;
        }
        else
        {
          q._parent._right = q._right;
          isLeft = false;
        }
      }
      
      if(q._right != null)
        q._right._parent = q._parent;
      
      if(!q._red)
        fixup(q._parent, q._right, isLeft); 
    }
    
    _size--;
    return res;
  }
  
  protected void fixup(IdentityToIntMapNode p, IdentityToIntMapNode q, boolean isLeft)
  {
    while(p != null && ( q == null || q._red == false ) )
    {
      IdentityToIntMapNode w, v;
      
      if(isLeft)
      {
        w = p._right;
        if(w._red)
        {
          w._red = false;
          p._red = true;
          rotate(true, p);
          w = p._right;
        }
        
        v = w._left;
        
        if(w._right == null || !w._right._red)
        {
          if(v == null || !v._red)
          {
            w._red = true;
            
            q = p;
            p = p._parent;
            isLeft = (p == null)?true:(q == p._left);
            
            continue;
          }
          else
          {
            w._red = true;
            v._red = false;
            rotate(false, w);
            
            w = v;
            v = v._left;
          }
        }
        
        w._red = p._red;
        p._red = false;
        if(w._right != null) w._right._red = false;
        rotate(true, p);
        
        q = _node;
        p = null;
      }
      else
      {
        w = p._left;
        if(w._red)
        {
          w._red = false;
          p._red = true;
          rotate(false, p);
          w = p._left;
        }
        
        v = w._right;
        
        if(w._left == null || !w._left._red)
        {
          if(v == null || !v._red)
          {
            w._red = true;
            
            q = p;
            p = p._parent;
            isLeft = (p == null)?true:(q == p._left);
            
            continue;
          }
          else
          {
            w._red = true;
            v._red = false;
            rotate(true, w);
            
            w = v;
            v = v._right;
          }
        }
        
        w._red = p._red;
        p._red = false;
        if(w._left != null) w._left._red = false;
        rotate(false, p);
        
        q = _node;
        p = null;
      }
    }
    
    if(q != null)
      q._red = false;
  }
  
  public synchronized void clear()
  {
    _node = null;
    _size = 0;
  }
  
  public synchronized boolean isEmpty()
  {
    return (_node == null);
  }
  
  public synchronized boolean equals(Object o)
  {
    if(o == null)
      return false;
    
    if( !(o instanceof IdentityToIntMap) )
      return false;
    
    IdentityToIntMap r = (IdentityToIntMap)o;
    
    if(_size != r._size)
      return false;
    
    if(r == this)
      return true;
            
    IdentityToIntMapNode p = minimum(), q = r.minimum();
    
    while(p != null)
    {
      if(   p._key != q._key
         || p._value != q._value
        )
        return false;
      
      p = next(p);
      q = r.next(q);
    }
    
    return true;
  }

  public synchronized int hashCode()
  {
    int res = 0;
    if(_node == null) return res;

    IdentityToIntMapNode p = minimum();

    while(p != null)
    {
      res += (p._key.hashCode() << 3) + p._value;

      p = next(p);
    }

    return res;
  }
  
  public synchronized int size()
  {
    return _size;
  }
  
  public synchronized IdentitySet keySet()
  {
    IdentitySet res = new IdentitySet();
    
    if(_node != null)
      _node.appendKeySet(res);
    
    return res;
  }
  
  public IdentityToIntMap(IdentityToIntMap r)
  {
    _size = r._size;
    
    if(r._node != null)
    {
      _node = new IdentityToIntMapNode(null, 0);
      r._node.nodeClone(_node);
    }
  }
  
  public synchronized Object clone()
  {
    return new IdentityToIntMap(this);
  }
  
  public IObjectIntPairIterator iterator()
  {
    return new IdentityToIntIterator(this);
  }
  
  protected static class IdentityToIntIterator
    extends    AbstractObjectIntPairIterator
    implements IObjectIntPairIterator
  {
    private IdentityToIntMapNode p = null;
    private IdentityToIntMap     s = null; 
        
    IdentityToIntIterator(IdentityToIntMap s)  
    {
      this.s = s;
      if(s._node == null) p = null; 
      else                p = s.minimum();
    }
       
    IdentityToIntIterator(IdentityToIntIterator r)
    {
      s = r.s;
      p = r.p;
    }
    
    public boolean stop()
    {
      if(p != null) return false;
      else          return true;
    }
  
    public void shift()
    {
      if(p == s.maximum()) p = null;
      else                 p = s.next(p);
    }
      
    public Object value1()
    {
      return p._key;
    }
    
    public int value2()
    {
      return p._value;
    }
  
    public Object clone()
    {
      return new IdentityToIntIterator(this);
    }
      
    public boolean equals(Object o)
    {
      if(!(o instanceof IdentityToIntIterator)) return false;
      
      IdentityToIntIterator r = (IdentityToIntIterator)o;
       
      return r.s == s && r.p == p;
    }
  
    public int hashCode()
    {
      return s.hashCode() + 17*System.identityHashCode(p);
    }
  }  
    
  class IdentityToIntMapNode
  {
    IdentityToIntMapNode _left = null;
    IdentityToIntMapNode _right = null;
    IdentityToIntMapNode _parent = null;
    Object _key = null;
    int _value = 0;
    boolean _red = true;
    
    IdentityToIntMapNode(Object key, int value)
    {
      _key = key;
      _value = value;
    }
    
    IdentityToIntMapNode minimum()
    {
      IdentityToIntMapNode p = this;
      
      while(p._left != null)
        p = p._left;
      
      return p;
    }
    
    IdentityToIntMapNode maximum()
    {
      IdentityToIntMapNode p = this;
      
      while(p._right != null)
        p = p._right;
      
      return p;
    }
  /*  
    int fill(int array[], int ind)
    {
      int res = ind+1;
      
      array[ind] = _value;
      
      if(_left != null)
      {
        res = _left.fill(array, ind+1);
        
        if(_right != null)
          return _right.fill(array, res);
      }
      else if(_right != null)
        return _right.fill(array, ind+1);
      
      return res;
    }
  */  
    void appendStringBuffer(StringBuffer st)
    {
      st.append("" + _key + " : " + _value);
      
      if(_left != null)
      {
        st.append(", ");
        _left.appendStringBuffer(st);
        
        if(_right != null)
        {
          st.append(", ");
          _right.appendStringBuffer(st);
        }
      }
      else if(_right != null)
      {
        st.append(", ");
        _right.appendStringBuffer(st);
      }
    }
    
    void appendKeySet(IdentitySet s)
    {
      s.add(_key);
      
      if(_left != null)
        _left.appendKeySet(s);
      
      if(_right != null)
        _right.appendKeySet(s);
    }
    
    boolean containsValue(int value)
    {
      if(_value == value)
        return true;
      else if(_left != null && _left.containsValue(value))
        return true;
      else if(_right != null && _right.containsValue(value))
        return true;
      else
        return false;
    }
   
    void nodeClone(IdentityToIntMapNode n)
    {
      n._key = _key;
      n._value = _value;
      n._red = _red;
      
      if(_left != null)
      {
        n._left = new IdentityToIntMapNode(null, 0);
        n._left._parent = n;
        _left.nodeClone(n._left);
      }
      
      if(_right != null)
      {
        n._right = new IdentityToIntMapNode(null, 0);
        n._right._parent = n;
        _right.nodeClone(n._right);
      }
    }
  }
}

