/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.util;

public class IntList
{
  int _size = 0;
  IntListNode _head = null;
  
  int _lastIndex = 0;
  IntListNode _last = null;
  
  public IntList()
  {
  }
  
  public IntList(IntList other)
  {
    IntListNode p = other._head;
    
    if(p == null) return;
    
    _size = other._size;
    
    if(other._last != null) _lastIndex = other._lastIndex;
    else                    _last      = null;
    
    IntListNode s = new IntListNode(p), q = null;
    _head = s;
    
    if(other._last == p) _last = s;
    p = p._next;
       
    while(p != other._head)
    {
      q = new IntListNode(p);
      s._next = q;
      q._prev = s;
      if(other._last == p) _last = q;
      p = p._next;
      s = q;
    }
    
    _head._prev = s;
    s._next = _head;
  }
  
  public synchronized int size()
  {
    return _size;
  }
  
  public synchronized boolean isEmpty()
  {
    return (_size == 0);
  }
  
  public synchronized int first()
  {
    if(_head != null) return _head._val;
    else return 0;
  }
  
  public synchronized int last()
  {
    if(_head != null) return _head._prev._val;
    else return 0;
  }
  
  public synchronized void addLast(int val)
  {
    if(_head == null)
    {
      _head = new IntListNode(val);
      _head._next = _head;
      _head._prev = _head;
    }
    else
    {
      IntListNode p = new IntListNode(val);
      p._prev = _head._prev;
      p._next = _head;
      _head._prev._next = p;
      _head._prev = p;
    }

    _size++;
  }
  
  public synchronized void addFirst(int val)
  {
    if(_head == null)
    {
      _head = new IntListNode(val);
      _head._next = _head;
      _head._prev = _head;
    }
    else
    {
      IntListNode p = new IntListNode(val);
      p._prev = _head._prev;
      p._next = _head;
      _head._prev._next = p;
      _head._prev = p;
      _head = p;
    }
    
    _size++;
    
    if(_last != null) _lastIndex++;
  }
  
  public synchronized void clear()
  {
    _head = null;
    _size = 0;
    
    _last = null;
  }
  
  public synchronized void removeLast()
  {
    if(_head == null) return;
    else if(_size == 1)
    {
      _head = null;
    }
    else
    {
      _head._prev._prev._next = _head;
      _head._prev = _head._prev._prev;
    }

    _size--;
    
    if(_last != null && _lastIndex >= _size) _last = null;
  }
  
  public synchronized void removeFirst()
  {
    if(_head == null) return;
    else if(_size == 1)
    {
      _head = null;
    }
    else
    {
      _head._prev._next = _head._next;
      _head._next._prev = _head._prev;
      _head = _head._next;
    }

    _size--;
    
    if(_last != null && _size > 0) _lastIndex--;
    else if(_size == 0) _last = null;
  }
  
  protected IntListNode getNode(int i)
  {
    IntListNode s = null;
    
    if(_head != null)
    {
      if(_last != null && i - _lastIndex >= 0 && _size - i >= i - _lastIndex)
      {
        s = _last;
        while(i - _lastIndex > 0) { s = s._next; i--; }
      }
      else if(
            _last != null && i - _lastIndex >= 0 && _size - i < i - _lastIndex
         || _last == null && i >= _size - i
        )
      {
        s = _head;
        while(_size - i > 0) { s = s._prev; i++; }
      }
      else if(_last != null && i - _lastIndex < 0 && i >= _lastIndex - i)
      {
        s = _last;
        while(_lastIndex - i > 0) { s = s._prev; i++; }
      }
      else/* if(
            _last != null && i - _lastIndex < 0 && i < _lastIndex - i
         || _last == null && i < _size - i
        )*/
      {
        s = _head;
        while(i > 0) { s = s._next; i--; }
      }
    }
    
    return s;
  }
    
  public synchronized void add(int val, int i)
  {
    if(i < 0 || i > _size) return;
    
    IntListNode p = new IntListNode(val);
    IntListNode s = getNode(i);
    
    if(s != null)
    {
      p._prev = s._prev;
      s._prev._next = p;
      p._next = s;
      s._prev = p;      
      if(i == 0) _head = p;
    }
    else
    {
      _head = p;
      _head._next = _head;
      _head._prev = _head;
    }
    
    _size++;

    _lastIndex = i;
    _last = p;
  }

  public synchronized void set(int val, int i)
  {
    if(i < 0 || i >= _size) return;
    
    IntListNode s = getNode(i);
    s._val = val;

    _lastIndex = i;
    _last = s;
  }

  public synchronized int get(int i)
  {
    if(i < 0 || i >= _size) return 0;
    
    IntListNode s = getNode(i);

    _lastIndex = i;
    _last = s;

    return s._val;
  }

  public synchronized void remove(int i)
  {
    if(i < 0 || i >= _size) return;
    
    if(_size == 1)
    {
      clear();
      return;
    }
    
    IntListNode s = getNode(i);

    s._prev._next = s._next;
    s._next._prev = s._prev;
    if(s == _head) _head = s._next;
    
    _size--;
    
    _lastIndex = (i == _size)?0:i;
    _last = s._next;
  }
  
  public synchronized boolean equals(Object obj)
  {
    if( !(obj instanceof IntList) ) return false;
    
    IntList other = (IntList)obj;
    
    if(other.size() != _size) return false;
    
    if(_size == 0) return true;
    
    IntListNode p = _head, s = other._head;
    
    for(int i = 0; i < _size; i++)
    {
      if(p._val != s._val) return false;
      
      p = p._next;
      s = s._next;
    }
    
    return true;
  }
  
  public synchronized int hashCode()
  {
    int res = 0;
    if(_head == null) return res;

    IntListNode p = _head;

    for(int i = 0; i < _size; i++)
    {
      res += p._val;

      p = p._next;
    }

    return res;
  }

  public synchronized Object clone()
  {
    return new IntList(this);
  }
  
  public synchronized boolean contains(int val)
  {
    IntListNode p = _head;
    
    if(p == null) return false;
    
    do
    {
      if(p._val == val) return true;
      
      p = p._next;
    }
    while(p != _head);
    
    return false;
  }
  
  public synchronized String toString()
  {
    StringBuffer bf = new StringBuffer("[");
    
    IntListNode p = _head;
    
    if(p != null)
    {
      while(true)
      {
        bf.append(p._val);
        
        if((p = p._next) != _head) bf.append(", ");
        else break;  
      }
    }
    
    bf.append("]");
    
    return bf.toString();
  }
}

class IntListNode
{
  IntListNode _next = null;
  IntListNode _prev = null;
  int _val = 0;
    
  IntListNode(int elem)
  {
    _val = elem;
  }
  
  IntListNode(IntListNode right)
  {
    _val = right._val;
  }
  
  public Object clone()
  {
    return new IntListNode(_val);
  }
}