/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: IntSet.java,v 1.2 2006/02/27 16:24:53 kuliamin Exp $
 * Created on 04.04.2001
 * 
 */

package jatva.util;

import jatva.iterators.lang.AbstractIntegerIterator;
import jatva.iterators.lang.IIntegerIterator;
import jatva.iterators.lang.IObjectIterator;


/**
 * A set of int values.
 * 
 * @author Victor Kuliamin
 */
public class IntSet
{
  protected IntSetNode _node = null;
  protected int _size = 0;
  
  public IntSet() {}
  
  public IntSet(int v)
  {
    add(v);
  }
  
  public synchronized void clear()
  {
    _node = null;
    _size = 0;
  }
  
  public synchronized int size()
  {
    return _size;
  }

  public synchronized boolean isEmpty()
  {
    return (_node == null);
  }
  
  public synchronized boolean add(int value)
  {
    if(_node == null)
    {
      _node = new IntSetNode(value);
      _node._red = false;
      _size++;
      
      return true;
    }
    
    IntSetNode p = _node, q;
    
    do
    {
      q = p;
      
      if(p._value == value)
        return false;
      else if(p._value > value)
        p = p._left;
      else
        p = p._right;
    }
    while(p != null);
    
    if(q._value > value)
      q._left = ( p =  new IntSetNode(value) );
    else
      q._right = ( p = new IntSetNode(value) );
    p._parent = q;
    
    _size++;
    
    IntSetNode r;
    
    while(q != null && q._red == true)
    {
      if(q == q._parent._left)
      {
        r = q._parent._right;
        if(r == null || !r._red )
        {
          if(p == q._right)
          {
            q._right = p._left;
            if(p._left != null) p._left._parent = q;
            p._left = q;
            p._parent = q._parent;
            q._parent._left = p;
            q._parent = p;
            p = q;
            q = q._parent;
          }
          
          if(q._parent._parent != null)
          {
            if(q._parent._parent._left == q._parent)
              q._parent._parent._left = q;
            else
              q._parent._parent._right = q;
          }
          else
            _node = q;
          
          q._red = false;
          q._parent._red = true;
          q._parent._left = q._right;
          if(q._right != null) q._right._parent = q._parent;
          q._right = q._parent;
          q._parent = q._parent._parent;
          q._right._parent = q;
          
          return true;
        }
        else
        {
          q._red = false;
          r._red = false;
          q._parent._red = true;
          p = q._parent;
          q = p._parent;
        }
      }
      else
      {
        r = q._parent._left;
        if(r == null || !r._red)
        {
          if(p == q._left)
          {
            q._left = p._right;
            if(p._right != null) p._right._parent = q;
            p._right = q;
            p._parent = q._parent;
            q._parent._right = p;
            q._parent = p;
            p = q;
            q = q._parent;
          }
          
         if(q._parent._parent != null)
          {
            if(q._parent._parent._left == q._parent)
              q._parent._parent._left = q;
            else
              q._parent._parent._right = q;
          }
          else
            _node = q;
          
          q._red = false;
          q._parent._red = true;
          q._parent._right = q._left;
          if(q._left != null) q._left._parent = q._parent;
          q._left = q._parent;
          q._parent = q._parent._parent;
          q._left._parent = q;
          
          return true;
        }
        else
        {
          q._red = false;
          r._red = false;
          q._parent._red = true;
          p = q._parent;
          q = p._parent;
        }
      }
    }
    
    _node._red = false;
    
    return true;
  }
  
  public synchronized boolean addAll(IntSet s)
  {
    boolean res = false;
    if(s != null && s.size() != 0)    
      for(IIntegerIterator i = s.iterator(); !i.stop(); i.shift())
        res |= add(i.value());
    
    return res;
  }
  
  public synchronized boolean contains(int value)
  {
    if(_node == null)
      return false;
    
    IntSetNode p = _node;
    
    do
    {
      if(p._value == value)
        return true;
      else if(p._value > value)
        p = p._left;
      else
        p = p._right;
    }
    while(p != null);
    
    return false;
  }
  
  public synchronized boolean containsAll(IntSet s)
  {
    if(s == null)          return false;
    else if(s.size() == 0) return true;
    else
      for(IIntegerIterator i = s.iterator(); !i.stop(); i.shift())
        if(!contains(i.value())) return false;
    
    return true;
  }

  public synchronized boolean isSubset(IntSet s)
  {
    if(s == null) return false;
    else          return s.containsAll(this);
  }

  public synchronized boolean remove(int value)
  {
    if(_node == null)
      return false;
    
    IntSetNode p = _node;
    
    do
    {
      if(p._value > value)
        p = p._left;
      else if(p._value < value)
        p = p._right;
      else
        break;
    }
    while(p != null);
    
    if(p == null)
      return false;
    
    boolean isLeft;
    
    if(p._left == null)
    {
      if(p._parent == null)
      {
        _node = p._right;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._right;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._right;
          isLeft = false;
        }
      }
      
      if(p._right != null)
        p._right._parent = p._parent;
      
      if(!p._red)
        fixup(p._parent, p._right, isLeft);
    }
    else if(p._right == null)
    {
      if(p._parent == null)
      {
        _node = p._left;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._left;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._left;
          isLeft = false;
        }
      }
      
      p._left._parent = p._parent;
      if(!p._red)
        fixup(p._parent, p._left, isLeft);
    }
    else
    {
      IntSetNode q = next(p);
      p._value = q._value;
      
      if(q._parent == null)
      {
        _node = q._right;
        isLeft = true;
      }
      else
      {
        if(q == q._parent._left)
        {
          q._parent._left = q._right;
          isLeft = true;
        }
        else
        {
          q._parent._right = q._right;
          isLeft = false;
        }
      }
      
      if(q._right != null)
        q._right._parent = q._parent;
      
      if(!q._red)
        fixup(q._parent, q._right, isLeft); 
    }
    
    _size--;
    return true;
  }
  
  public synchronized boolean removeAll(IntSet s)
  {
    boolean res = false;
    if(s != null && s.size() != 0)
      for(IIntegerIterator i = s.iterator(); !i.stop(); i.shift())
        res |= this.remove(i.value());
    
    return res;
  }

  public synchronized boolean retain(IntSet s)
  {
    boolean res = false;
    
    if(s != null && s.size() != 0)
    {
      IIntegerIterator i = iterator();
      
      while(!i.stop())
      {
        if(!s.contains(i.value())) { i.remove(); res = true; } 
        else                       i.shift(); 
      }
    }
    return res;
  }
  
  public synchronized boolean replace(int i, int j)
  {
    if(remove(i)) { add(j); return true; }
    else           return false;
  }
  
  public synchronized boolean replace(IMap m)
  {
    // TODO Auto-generated method stub
    return false;
  }

  public synchronized IntList toList()
  {
    IntList l = new IntList();
    for(IIntegerIterator i = iterator(); !i.stop(); i.shift())
      l.addLast(i.value());
    
    return l;
  }

  public synchronized int[] toArray()
  {
    int[] res = new int[size()];
    int j = 0;
    for(IIntegerIterator i = iterator(); !i.stop(); i.shift())
      res[j++] = i.value();

    return res;
  }
  
  public synchronized IIntegerIterator iterator()
  {
    return new IntSetIterator(this);
  }
  
  public IntSet(IntSet r)
  {
    _size = r._size;
    
    if(r._node != null)
    {
      _node = new IntSetNode(0);
      r._node.nodeClone(_node);
    }
  }
  
  public synchronized Object clone()
  {
    return new IntSet(this);
  }

  public synchronized String toString()
  {
    StringBuffer st = new StringBuffer("{");
    IntSetNode n = minimum();
    
    if(n != null)
    {
      st.append(n._value);
      n = next(n);
    }
    
    while(n != null)
    {
      st.append(", ");
      st.append(n._value);
      n = next(n);
    }
      
    st.append("}");
    
    return st.toString();
  }
  
  public synchronized boolean equals(Object o)
  {
    if( !(o instanceof IntSet) ) return false;
    
    IntSet r = (IntSet)o;
    
    if(_size != r._size) return false;
    if(r == this)        return true;
            
    IntSetNode p = minimum(), q = r.minimum();
    
    while(p != null)
    {
      if(p._value != q._value) return false;
      
      p = next(p);
      q = r.next(q);
    }
    
    return true;
  }

  public synchronized int hashCode()
  {
    int res = 0;
    if(_node == null) return res;

    IntSetNode p = minimum();
    
    while(p != null)
    {
      res += p._value;
      
      p = next(p);
    }

    return res;
  }
  
  public static IntSet singleton(int i)
  {
    return new IntSet(i);
  }

  public static IntSet union(IntSet s1, IntSet s2)
  {
    IntSet res = new IntSet(s1);
    res.addAll(s2);
    
    return res;
  }
  
  public static IntSet intersection(IntSet s1, IntSet s2)
  {
    IntSet res = new IntSet(s1);
    res.retain(s2);
    
    return res;
  }
  
  public static IntSet difference(IntSet s1, IntSet s2)
  {
    IntSet res = new IntSet(s1);
    res.removeAll(s2);
    
    return res;
  }
  
  public static IntSet symDifference(IntSet s1, IntSet s2)
  {
    IntSet res = difference(s1, s2);
    res.addAll(difference(s2, s1));
    
    return res;
  }
  
  public static IntSet distrUnion(ISet s)
  {
    IObjectIterator i = s.iterator();
    IntSet res = new IntSet(), t;
    
    while(!i.stop())
    {
      t = (IntSet)i.value();
      res.addAll(t);
    }

    return res;    
  }
  
  public static IntSet distrIntersection(ISet s)
  {
    IObjectIterator i = s.iterator();
    IntSet res = null, t;
    
    while(!i.stop())
    {
      t = (IntSet)i.value();
      if(res == null) res = new IntSet(t);
      else            res.retain(t);
    }

    return res;    
  }
  
  
  public int max()
  {
    if(_node == null) return Integer.MIN_VALUE;
    else              return maximum()._value;
  }
  
  public int min()
  {
    if(_node == null) return Integer.MAX_VALUE;
    else              return minimum()._value;
  }
  
  public int numberNotGreater(int n)
  {
    IntSetNode p = minimum();
    int c = 0;
    while(p != null && p._value <= n)
    {
      c++;
      p = next(p);
    }

    return c;
  }
  
  public int numberNotLess(int n)
  {
    IntSetNode p = maximum();
    int c = 0;
    while(p != null && p._value >= n)
    {
      c++;
      p = prev(p);
    }

    return c;
  }
  
  public int maxNotGreater(int n)
  {
    if(_node == null) return Integer.MIN_VALUE;
    
    IntSetNode p = _node, q;
    
    do
    {
      if(p._value == n)
        return n;
      else if(p._value > n)
      {
        q = p;
        p = p._left;
      }
      else
      {
        q = p;
        p = p._right;
      }
    }
    while(p != null);
    
    if(q._value > n)
    {
      q = prev(q);
      return (q == null)?Integer.MIN_VALUE:q._value;
    }
    else return q._value;
  }
  
  public int minNotLess(int n)
  {
    if(_node == null) return Integer.MAX_VALUE;
    
    IntSetNode p = _node, q;
    
    do
    {
      if(p._value == n)
        return n;
      else if(p._value > n)
      {
        q = p;
        p = p._left;
      }
      else
      {
        q = p;
        p = p._right;
      }
    }
    while(p != null);
    
    if(q._value < n)
    {
      q = next(q);
      return (q == null)?Integer.MAX_VALUE:q._value;
    }
    else return q._value;
  }
  
  //-----------------------------------------------------------------
  // auxiliary methods
  //
  protected IntSetNode find(int value)
  {
    if(_node == null) return null;
    
    IntSetNode p = _node;
    
    do
    {
      if(p._value > value)
        p = p._left;
      else if(p._value < value)
        p = p._right;
      else break;
    }
    while(p != null);
    
    return p;
  }
  
  protected void remove(IntSetNode p)
  {
    boolean isLeft;
    
    if(p._left == null)
    {
      if(p._parent == null)
      {
        _node = p._right;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._right;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._right;
          isLeft = false;
        }
      }
      
      if(p._right != null)
        p._right._parent = p._parent;
      
      if(!p._red)
        fixup(p._parent, p._right, isLeft);
    }
    else if(p._right == null)
    {
      if(p._parent == null)
      {
        _node = p._left;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._left;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._left;
          isLeft = false;
        }
      }
      
      p._left._parent = p._parent;
      if(!p._red)
        fixup(p._parent, p._left, isLeft);
    }
    else
    {
      IntSetNode q = next(p);
      p._value = q._value;
      
      if(q._parent == null)
      {
        _node = q._right;
        isLeft = true;
      }
      else
      {
        if(q == q._parent._left)
        {
          q._parent._left = q._right;
          isLeft = true;
        }
        else
        {
          q._parent._right = q._right;
          isLeft = false;
        }
      }
      
      if(q._right != null)
        q._right._parent = q._parent;
      
      if(!q._red)
        fixup(q._parent, q._right, isLeft); 
    }
    
    _size--;
  }
  
  protected IntSetNode minimum()
  {
    if(_node == null)
      return null;
    else
      return _node.minimum();
  }
  
  protected IntSetNode maximum()
  {
    if(_node == null)
      return null;
    else
      return _node.maximum();
  }
  
  protected IntSetNode next(IntSetNode p)
  {
    if(p._right != null) return p._right.minimum();
    
    while(p._parent != null && p == p._parent._right)
      p = p._parent;

    return p._parent;
  }
  
  protected IntSetNode prev(IntSetNode p)
  {
    if(p._left != null) return p._left.maximum();
    
    while(p._parent != null && p == p._parent._left)
      p = p._parent;

    return p._parent;
  }
  
  protected void fixup(IntSetNode p, IntSetNode q, boolean isLeft)
  {
    while(p != null && (q == null || q._red == false))
    {
      IntSetNode w, v;
      
      if(isLeft)
      {
        w = p._right;
        if(w._red)
        {
          w._red = false;
          p._red = true;
          rotate(true, p);
          w = p._right;
        }
        
        v = w._left;
        
        if(w._right == null || !w._right._red)
        {
          if(v == null || !v._red)
          {
            w._red = true;
            
            q = p;
            p = p._parent;
            
            if(p == null)
              isLeft = true;
            else
              isLeft = (q == p._left);
            //isLeft = (p == null)?true:(q == p._left);
            
            continue;
          }
          else
          {
            w._red = true;
            v._red = false;
            rotate(false, w);
            
            w = v;
            v = v._left;
          }
        }
        
        w._red = p._red;
        p._red = false;
        
        if(w._right != null)
          w._right._red = false;
          
        rotate(true, p);
        
        q = _node;
        p = null;
      }
      else
      {
        w = p._left;
        if(w._red)
        {
          w._red = false;
          p._red = true;
          rotate(false, p);
          w = p._left;
        }
        
        v = w._right;
        
        if(w._left == null || !w._left._red)
        {
          if(v == null || !v._red)
          {
            w._red = true;
            
            q = p;
            p = p._parent;
            
            if(p == null)
              isLeft = true;
            else
              isLeft = (q == p._left);
            //isLeft = (p == null)?true:(q == p._left);
            
            continue;
          }
          else
          {
            w._red = true;
            v._red = false;
            rotate(true, w);
            
            w = v;
            v = v._right;
          }
        }
        
        w._red = p._red;
        p._red = false;
        
        if(w._left != null)
          w._left._red = false;
          
        rotate(false, p);
        
        q = _node;
        p = null;
      }
    }
    
    if(q != null)
      q._red = false;
  }
  
  protected void rotate(boolean left, IntSetNode x)
  {
    IntSetNode y;
    
    if(left)
    {
      y = x._right;
      x._right = y._left;
      if(y._left != null)
        y._left._parent = x;
      
      y._parent = x._parent;
      if(x._parent != null)
      {
        if(x == x._parent._left)
          x._parent._left = y;
        else
          x._parent._right = y;
      }
      else
        _node = y;
      
      y._left = x;
      x._parent = y;
    }
    else
    {
      y = x._left;
      x._left = y._right;
      if(y._right != null)
        y._right._parent = x;
      
      y._parent = x._parent;
      if(x._parent != null)
      {
        if(x == x._parent._left)
          x._parent._left = y;
        else
          x._parent._right = y;
      }
      else
        _node = y;
      
      y._right = x;
      x._parent = y;
    }
  }
  
  private static class IntSetNode
  {
    IntSetNode _left = null;
    IntSetNode _right = null;
    IntSetNode _parent = null;
    int _value = 0;
    boolean _red = true;
    
    IntSetNode(int value)
    {
      _value = value;
    }
    
    public int hashCode()
    {
      return _value + (_red?0:1139) + ((_parent == null)?0:67547) + ((_left == null)?0:132381) + ((_right == null)?0:541873);
    }
    
    IntSetNode minimum()
    {
      IntSetNode p = this;
      
      while(p._left != null)
        p = p._left;
      
      return p;
    }
    
    boolean isChild(IntSetNode p)
    {
      if(p == null)
        return false;
      
      if(p == this)
        return true;
        
      if(_left != null && _left.isChild(p))
        return true;
        
      if(_right != null && _right.isChild(p))
        return true;
        
      return false;
    }
      
    IntSetNode maximum()
    {
      IntSetNode p = this;
      
      while(p._right != null)
        p = p._right;
      
      return p;
    }
    
    int fill(int array[], int ind)
    {
      int res = ind+1;
      
      array[ind] = _value;
      
      if(_left != null)
      {
        res = _left.fill(array, ind+1);
        
        if(_right != null)
          return _right.fill(array, res);
      }
      else if(_right != null)
        return _right.fill(array, ind+1);
      
      return res;
    }
    
    void appendStringBuffer(StringBuffer st)
    {
      st.append(_value);
      
      if(_left != null)
      {
        st.append(", ");
        _left.appendStringBuffer(st);
        
        if(_right != null)
        {
          st.append(", ");
          _right.appendStringBuffer(st);
        }
      }
      else if(_right != null)
      {
        st.append(", ");
        _right.appendStringBuffer(st);
      }
    }
    
    void nodeClone(IntSetNode n)
    {
      n._value = _value;
      n._red = _red;
      
      if(_left != null)
      {
        n._left = new IntSetNode(0);
        n._left._parent = n;
        _left.nodeClone(n._left);
      }
      
      if(_right != null)
      {
        n._right = new IntSetNode(0);
        n._right._parent = n;
        _right.nodeClone(n._right);
      }
    }
  }
  
  private static class IntSetIterator
    extends AbstractIntegerIterator
    implements IIntegerIterator
  {
    private IntSetNode p = null;
    private IntSet     s = null; 
        
    IntSetIterator(IntSet s)  
    {
      this.s = s;
      if(s._node == null) p = null;
      else                p = s.minimum();
    }
       
    IntSetIterator(IntSetIterator r)
    {
      s = r.s;
      p = r.p;
    }
    
    public boolean stop()
    {
      if(p != null) return false;
      else          return true;
    }
  
    public void shift()
    {
      if(p == s.maximum()) p = null;
      else                 p = s.next(p);
    }
      
    public int value()
    {
      return p._value;
    }
  
    public Object clone()
    {
      return new IntSetIterator(this);
    }
      
    public boolean equals(Object o)
    {
      if(!(o instanceof IntSetIterator)) return false;
      
      IntSetIterator r = (IntSetIterator)o;
       
      return r.s == s && r.p == p;
    }
  
    public int hashCode()
    {
      return s.hashCode() + 17*System.identityHashCode(p);
    }
    
    public void remove()
    {
      if(p == s.maximum())
      {
        s.remove(p);
        p = null;
      }
      else
      {
        int x = s.next(p)._value;
        s.remove(p);
        p = s.find(x);
      }
    }
  }

  public static class Checker
  {
    public static boolean checkInvariants(IntSet s)
    {
      if(s._node == null) return s._size == 0;
      else
        return    checkForCycles(s._node) 
               && checkTreeStructure(s._node)
               && checkSize(s._node, s._size)
               && checkColors(s._node)
               && checkEqualityOfBlackHeights(s._node);
    }
    
    private static boolean checkForCycles(IntSetNode n)
    {
      java.util.IdentityHashMap m = new java.util.IdentityHashMap();
      
      return checkForCyclesAndFill(n, m);
    }
    
    private static boolean checkForCyclesAndFill(IntSetNode n, java.util.IdentityHashMap m)
    {
      boolean result = true;
      if(m.containsKey(n)) return false;
      m.put(n, null);
      
      if(n._left != null) result = checkForCyclesAndFill(n._left, m);
      if(!result) return false;

      if(n._right != null) result = checkForCyclesAndFill(n._right, m);
      return result;
    }
    
    private static boolean checkTreeStructure(IntSetNode n)
    {
      if(n._parent != null) return false;
      
      return checkForIncorrectBackReferences(n);
    }
    
    private static boolean checkForIncorrectBackReferences(IntSetNode n)
    {
      boolean result = true;
      
      if(n._left != null && n._left._parent != n) return false;
      else if(n._left != null) result = checkForIncorrectBackReferences(n._left);
      if(!result) return false;
        
      if(n._right != null && n._right._parent != n) return false;
      else if(n._right != null)  result = checkForIncorrectBackReferences(n._right);
      
      return result;
    }
    
    private static boolean checkSize(IntSetNode n, int size)
    {
      return countNumberOfNodes(n) == size;
    }
    
    private static int countNumberOfNodes(IntSetNode n)
    {
      int res = 1;
      if(n._left != null) res += countNumberOfNodes(n._left);
      if(n._right != null) res += countNumberOfNodes(n._right);
      
      return res;
    }
    
    private static boolean checkColors(IntSetNode n)
    {
      boolean result = true;
      
      if(n._red && n._left != null && n._left._red) return false;
      else if(n._left != null) result = checkColors(n._left);
      
      if(n._red && n._right != null && n._right._red) return false;
      else if(n._right != null) result = checkColors(n._right);
      
      return result;  
    }

    private static boolean checkEqualityOfBlackHeights(IntSetNode n)
    {
      java.util.IdentityHashMap m = new java.util.IdentityHashMap();
      
      countBlackHeights(n, m); 
      
      return checkBlackHeights(n, m);
    }
    
    private static int countBlackHeights(IntSetNode n, java.util.IdentityHashMap m)
    {
      int r = 1;
      if(n._left == null && n._right == null)
      {
        r += (n._red?0:1);
        m.put(n, new Integer(r));
        return r;
      }
      else if(n._left == null)
      {
        countBlackHeights(n._right, m); 
        r += (n._red?0:1);
        m.put(n, new Integer(r));
        return r;
      }
      else if(n._right == null)
      {
        countBlackHeights(n._left, m); 
        r += (n._red?0:1);
        m.put(n, new Integer(r));
        return r;
      }
      else
      {
        r = countBlackHeights(n._left, m);
        countBlackHeights(n._right, m);   
        r += (n._red?0:1);
        m.put(n, new Integer(r));
        return r;
      }
    }
    
    private static boolean checkBlackHeights(IntSetNode n, java.util.IdentityHashMap m)
    {
      if(n._left == null && n._right == null) return true;
      else if(n._left == null)
      {
        if(((Integer)(m.get(n._right))).intValue() != 1) return false;
        return checkBlackHeights(n._right, m);
      }
      else if(n._right == null)
      {
        if(((Integer)(m.get(n._left))).intValue() != 1) return false;
        return checkBlackHeights(n._left, m);
      }
      else
      {
        if(((Integer)(m.get(n._left))).intValue() != ((Integer)(m.get(n._right))).intValue()) return false;
        return checkBlackHeights(n._left, m) && checkBlackHeights(n._right, m);
      }
    }
  }
}

