/* 
 * Copyright (c) 2000 ISP RAS.
 * B.Communisticheskaya,25, Moscow, Russia.
 * All rights reserved.
 *
 */

package jatva.util;

import jatva.iterators.lang.AbstractIntPairIterator;
import jatva.iterators.lang.IIntPairIterator;

public class IntToIntMap
{
  protected IntToIntMapNode _node = null;
  protected int _size = 0;
  
  public IntToIntMap()
  {
  }
  
  public boolean checkInvariants()
  {
    if(_node != null)
      return _node.check() && _node.checkColors() && _node.checkPaths();
    
    return true;
  }
  
  public synchronized String toStringX()
  {
    StringBuffer st = new StringBuffer("{");
    
    if(_node != null)
      _node.appendStringBuffer(st);
    
    st.append("}");
    
    return st.toString();
  }
  
  public synchronized String toString()
  {
    StringBuffer st = new StringBuffer("{");
    IntToIntMapNode n = minimum();
    
    if(n != null)
    {
      st.append("" + n._key + " : " + n._value);
      n = next(n);
    }
    
    while(n != null)
    {
      st.append(", ");
      st.append("" + n._key + " : " + n._value);
      n = next(n);
    }
      
    st.append("}");
    
    return st.toString();
  }
  
  protected void rotate(boolean left, IntToIntMapNode x)
  {
    IntToIntMapNode y;
    
    if(left)
    {
      y = x._right;
      x._right = y._left;
      if(y._left != null)
        y._left._parent = x;
      
      y._parent = x._parent;
      if(x._parent != null)
      {
        if(x == x._parent._left)
          x._parent._left = y;
        else
          x._parent._right = y;
      }
      else
        _node = y;
      
      y._left = x;
      x._parent = y;
    }
    else
    {
      y = x._left;
      x._left = y._right;
      if(y._right != null)
        y._right._parent = x;
      
      y._parent = x._parent;
      if(x._parent != null)
      {
        if(x == x._parent._left)
          x._parent._left = y;
        else
          x._parent._right = y;
      }
      else
        _node = y;
      
      y._right = x;
      x._parent = y;
    }
  }
  
  public synchronized boolean containsKey(int key)
  {
    if(_node == null)
      return false;
    
    IntToIntMapNode p = _node;
    
    do
    {
      if(p._key == key)
        return true;
      else if(p._key > key)
        p = p._left;
      else
        p = p._right;
    }
    while(p != null);
    
    return false;
  }
  
  public synchronized boolean containsValue(int value)
  {
    if(_node == null)
      return false;
    
    return _node.containsValue(value);
  }
  
  protected IntToIntMapNode minimum()
  {
    if(_node == null)
      return null;
    else
      return _node.minimum();
  }
  
  protected IntToIntMapNode maximum()
  {
    if(_node == null)
      return null;
    else
      return _node.maximum();
  }
  
  protected IntToIntMapNode next(IntToIntMapNode p)
  {
    if(p._right != null)
      return p._right.minimum();
    
    while(p._parent != null && p == p._parent._right)
      p = p._parent;

    return p._parent;
  }
  
  protected IntToIntMapNode prev(IntToIntMapNode p)
  {
    if(p._left != null)
      return p._left.maximum();
    
    while(p._parent != null && p == p._parent._left)
      p = p._parent;

    return p._parent;
  }
  
  public synchronized int get(int key)
  {
    if(_node == null)
      return 0;
    
    IntToIntMapNode p = _node;
    
    do
    {
      if(p._key == key)
        return p._value;
      else if(p._key > key)
        p = p._left;
      else
        p = p._right;
    }
    while(p != null);
    
    return 0;
  }
  
  public synchronized int put(int key, int value)
  {
    if(_node == null)
    {
      _node = new IntToIntMapNode(key, value);
      _node._red = false;
      _size++;
      
      return 0;
    }
    
    IntToIntMapNode p = _node, q;
    int res;
    
    do
    {
      q = p;
      
      if(p._key == key)
      {
        res = p._value;
        p._value = value;
        return res;
      }
      else if(p._key > key)
        p = p._left;
      else
        p = p._right;
    }
    while(p != null);
    
    if(q._key > key)
      q._left = ( p =  new IntToIntMapNode(key, value) );
    else
      q._right = ( p = new IntToIntMapNode(key, value) );
    p._parent = q;
    
    _size++;
    
    IntToIntMapNode r;
    
//    System.out.println("After add : " + toString());
    
    while(q != null && q._red == true)
    {
      if(q == q._parent._left)
      {
//        System.out.print("q is on the left, ");
        
        r = q._parent._right;
        if(r == null || !r._red )
        {
          if(p == q._right)
          {
//            System.out.println("case 2");
                        
            q._right = p._left;
            if(p._left != null) p._left._parent = q;
            p._left = q;
            p._parent = q._parent;
            q._parent._left = p;
            q._parent = p;
            p = q;
            q = q._parent;
            
//            System.out.println("After rotation : " + toString());
          }
          
//          System.out.println("case 3");
          
          if(q._parent._parent != null)
          {
            if(q._parent._parent._left == q._parent)
              q._parent._parent._left = q;
            else
              q._parent._parent._right = q;
          }
          else
            _node = q;
          
          q._red = false;
          q._parent._red = true;
          q._parent._left = q._right;
          if(q._right != null) q._right._parent = q._parent;
          q._right = q._parent;
          q._parent = q._parent._parent;
          q._right._parent = q;
          
          return 0;
        }
        else
        {
//          System.out.println("case 1");
          
          q._red = false;
          r._red = false;
          q._parent._red = true;
          p = q._parent;
          q = p._parent;
        }
      }
      else
      {
//        System.out.print("q is on the right, ");

        r = q._parent._left;
        if(r == null || !r._red)
        {
          if(p == q._left)
          {
//            System.out.println("case 2");

            q._left = p._right;
            if(p._right != null) p._right._parent = q;
            p._right = q;
            p._parent = q._parent;
            q._parent._right = p;
            q._parent = p;
            p = q;
            q = q._parent;

//            System.out.println("After rotation : " + toString());
          }
          
//         System.out.println("case 3");
         
         if(q._parent._parent != null)
          {
            if(q._parent._parent._left == q._parent)
              q._parent._parent._left = q;
            else
              q._parent._parent._right = q;
          }
          else
            _node = q;
          
          q._red = false;
          q._parent._red = true;
          q._parent._right = q._left;
          if(q._left != null) q._left._parent = q._parent;
          q._left = q._parent;
          q._parent = q._parent._parent;
          q._left._parent = q;
          
          return 0;
        }
        else
        {
          q._red = false;
          r._red = false;
          q._parent._red = true;
          p = q._parent;
          q = p._parent;
        }
      }
    }
    
    _node._red = false;
    
    return 0;
  }
  
  public synchronized int remove(int key)
  {
    if(_node == null)
      return 0;
    
    IntToIntMapNode p = _node;
    
    do
    {
      if(p._key > key)
        p = p._left;
      else if(p._key < key)
        p = p._right;
      else
        break;
    }
    while(p != null);
    
    if(p == null)
      return 0;
    
    boolean isLeft;
    int res = p._value;
    
    if(p._left == null)
    {
      if(p._parent == null)
      {
        _node = p._right;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._right;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._right;
          isLeft = false;
        }
      }
      
      if(p._right != null)
        p._right._parent = p._parent;
      
      if(!p._red)
        fixup(p._parent, p._right, isLeft);
    }
    else if(p._right == null)
    {
      if(p._parent == null)
      {
        _node = p._left;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._left;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._left;
          isLeft = false;
        }
      }
      
      p._left._parent = p._parent;
      if(!p._red)
        fixup(p._parent, p._left, isLeft);
    }
    else
    {
      IntToIntMapNode q = next(p);
      p._key = q._key;
      p._value = q._value;
      
      if(q._parent == null)
      {
        _node = q._right;
        isLeft = true;
      }
      else
      {
        if(q == q._parent._left)
        {
          q._parent._left = q._right;
          isLeft = true;
        }
        else
        {
          q._parent._right = q._right;
          isLeft = false;
        }
      }
      
      if(q._right != null)
        q._right._parent = q._parent;
      
      if(!q._red)
        fixup(q._parent, q._right, isLeft); 
    }
    
    _size--;
    return res;
  }
  
  protected void fixup(IntToIntMapNode p, IntToIntMapNode q, boolean isLeft)
  {
    while(p != null && ( q == null || q._red == false ) )
    {
      IntToIntMapNode w, v;
      
      if(isLeft)
      {
        w = p._right;
        if(w._red)
        {
          w._red = false;
          p._red = true;
          rotate(true, p);
          w = p._right;
        }
        
        v = w._left;
        
        if(w._right == null || !w._right._red)
        {
          if(v == null || !v._red)
          {
            w._red = true;
            
            q = p;
            p = p._parent;
            isLeft = (p == null)?true:(q == p._left);
            
            continue;
          }
          else
          {
            w._red = true;
            v._red = false;
            rotate(false, w);
            
            w = v;
            v = v._left;
          }
        }
        
        w._red = p._red;
        p._red = false;
        if(w._right != null) w._right._red = false;
        rotate(true, p);
        
        q = _node;
        p = null;
      }
      else
      {
        w = p._left;
        if(w._red)
        {
          w._red = false;
          p._red = true;
          rotate(false, p);
          w = p._left;
        }
        
        v = w._right;
        
        if(w._left == null || !w._left._red)
        {
          if(v == null || !v._red)
          {
            w._red = true;
            
            q = p;
            p = p._parent;
            isLeft = (p == null)?true:(q == p._left);
            
            continue;
          }
          else
          {
            w._red = true;
            v._red = false;
            rotate(true, w);
            
            w = v;
            v = v._right;
          }
        }
        
        w._red = p._red;
        p._red = false;
        if(w._left != null) w._left._red = false;
        rotate(false, p);
        
        q = _node;
        p = null;
      }
    }
    
    if(q != null)
      q._red = false;
  }
  
  public synchronized void clear()
  {
    _node = null;
    _size = 0;
  }
  
  public synchronized boolean isEmpty()
  {
    return (_node == null);
  }
  
  public synchronized boolean equals(Object o)
  {
    if(o == null)
      return false;
    
    if( !(o instanceof IntToIntMap) )
      return false;
    
    IntToIntMap r = (IntToIntMap)o;
    
    if(_size != r._size)
      return false;
    
    if(r == this)
      return true;
            
    IntToIntMapNode p = minimum(), q = r.minimum();
    
    while(p != null)
    {
      if(   p._key != q._key
         || p._value != q._value
        )
        return false;
      
      p = next(p);
      q = r.next(q);
    }
    
    return true;
  }

  public synchronized int hashCode()
  {
    int res = 0;
    if(_node == null) return res;

    IntToIntMapNode p = minimum();

    while(p != null)
    {
      res += (p._key << 3) + p._value;

      p = next(p);
    }

    return res;
  }
  
  public synchronized int size()
  {
    return _size;
  }
  
  public synchronized IntSet keySet()
  {
    IntSet res = new IntSet();
    
    if(_node != null)
      _node.appendKeySet(res);
    
    return res;
  }
  
  public IntToIntMap(IntToIntMap r)
  {
    _size = r._size;
    
    if(r._node != null)
    {
      _node = new IntToIntMapNode(0, 0);
      r._node.nodeClone(_node);
    }
  }
  
  public synchronized Object clone()
  {
    return new IntToIntMap(this);
  }
  
  public IIntPairIterator iterator()
  {
    return new IntToIntMapIterator(this);
  }

  protected static class IntToIntMapIterator
    extends    AbstractIntPairIterator
    implements IIntPairIterator
  {
    private IntToIntMapNode p = null;
    private IntToIntMap     s = null; 
        
    IntToIntMapIterator(IntToIntMap s)  
    {
      this.s = s;
      if(s._node == null) p = null; 
      else                p = s.minimum();
    }
       
    IntToIntMapIterator(IntToIntMapIterator r)
    {
      s = r.s;
      p = r.p;
    }
    
    public boolean stop()
    {
      if(p != null) return false;
      else          return true;
    }
  
    public void shift()
    {
      if(p == s.maximum()) p = null;
      else                 p = s.next(p);
    }
      
    public int value1()
    {
      return p._key;
    }
    
    public int value2()
    {
      return p._value;
    }
  
    public Object clone()
    {
      return new IntToIntMapIterator(this);
    }
      
    public boolean equals(Object o)
    {
      if(!(o instanceof IntToIntMapIterator)) return false;
      
      IntToIntMapIterator r = (IntToIntMapIterator)o;
       
      return r.s == s && r.p == p;
    }
  
    public int hashCode()
    {
      return s.hashCode() + 17*System.identityHashCode(p);
    }
  }  

}

class MapCheckException
  extends Throwable
{
  MapCheckException(int value)
  {
    super();
    System.out.println(  "CHECK: not all the paths from "
                       + value
                       + " have the same black length"
                      );
  }
}

class IntToIntMapNode
{
  IntToIntMapNode _left = null;
  IntToIntMapNode _right = null;
  IntToIntMapNode _parent = null;
  int _key = 0;
  int _value = 0;
  boolean _red = true;
  
  IntToIntMapNode(int key, int value)
  {
    _key = key;
    _value = value;
  }
  
  IntToIntMapNode minimum()
  {
    IntToIntMapNode p = this;
    
    while(p._left != null)
      p = p._left;
    
    return p;
  }
  
  IntToIntMapNode maximum()
  {
    IntToIntMapNode p = this;
    
    while(p._right != null)
      p = p._right;
    
    return p;
  }
/*  
  int fill(int array[], int ind)
  {
    int res = ind+1;
    
    array[ind] = _value;
    
    if(_left != null)
    {
      res = _left.fill(array, ind+1);
      
      if(_right != null)
        return _right.fill(array, res);
    }
    else if(_right != null)
      return _right.fill(array, ind+1);
    
    return res;
  }
*/  
  void appendStringBuffer(StringBuffer st)
  {
    st.append("" + _key + " : " + _value);
    
    if(_left != null)
    {
      st.append(", ");
      _left.appendStringBuffer(st);
      
      if(_right != null)
      {
        st.append(", ");
        _right.appendStringBuffer(st);
      }
    }
    else if(_right != null)
    {
      st.append(", ");
      _right.appendStringBuffer(st);
    }
  }
  
  void appendKeySet(IntSet s)
  {
    s.add(_key);
    
    if(_left != null)
      _left.appendKeySet(s);
    
    if(_right != null)
      _right.appendKeySet(s);
  }
  
  boolean containsValue(int value)
  {
    if(_value == value)
      return true;
    else if(_left != null && _left.containsValue(value))
      return true;
    else if(_right != null && _right.containsValue(value))
      return true;
    else
      return false;
  }
 
  void nodeClone(IntToIntMapNode n)
  {
    n._key = _key;
    n._value = _value;
    n._red = _red;
    
    if(_left != null)
    {
      n._left = new IntToIntMapNode(0, 0);
      n._left._parent = n;
      _left.nodeClone(n._left);
    }
    
    if(_right != null)
    {
      n._right = new IntToIntMapNode(0, 0);
      n._right._parent = n;
      _right.nodeClone(n._right);
    }
  }
  
  boolean check()
  {
    boolean result = true;
    
    if(_left != null && _left._parent != this)
    {
      result = false;
      System.out.println(  "CHECK: Inconsistent parent-left links: left is "
                         + _left._value
                         + ", parent must be "
                         + _value
                         + ", but is "
                         + (_left._parent == null?"null":Integer.toString(_left._parent._value))
                        );
      
      _left.check();  
    }
    else if(_left != null)
      result = _left.check();
      
    if(_right != null && _right._parent != this)
    {
      result = false;
      System.out.println(  "CHECK: Inconsistent parent-right links: right is "
                         + _right._value
                         + ", parent must be "
                         + _value
                         + ", but is "
                         + (_right._parent == null?"null":Integer.toString(_right._parent._value))
                        );
      
      _right.check();
    }
    else if(_right != null)
      result = result && _right.check();
    
    return result;
  }
  
  boolean checkColors()
  {
    boolean result = true;
    
    if(_red && _left != null && _left._red)
    {
      result = false;
      System.out.println(  "CHECK: Inconsistent colors: "
                         + _value
                         + " is red, while left "
                         + _left._value
                         + " is red"
                        );

      _left.checkColors();
    }
    else if(_left != null)
      result = _left.checkColors();
    
    if(_red && _right != null && _right._red)
    {
      result = false;
      System.out.println(  "CHECK: Inconsistent colors: "
                         + _value
                         + " is red, while right "
                         + _right._value
                         + " is red"
                        );
      
      result = _right.checkColors();
    }
    else if(_right != null)
      result = result && _right.checkColors();
    
    return result;  
  }
  
  boolean checkPaths()
  {
    try
    {
      System.out.println("Black height: " + blackHeight());
    
      return true;
    }
    catch(MapCheckException err)
    {
      return false;
    }
  }
  
  int blackHeight()
    throws MapCheckException
  {
    if(_left == null && _right == null)
      return 1 + (_red?0:1);
    else if(_left == null)
    {
      if(_right.blackHeight() != 1)
        throw new MapCheckException(_value);
      else
        return 1 + (_red?0:1);
    }
    else if(_right == null)
    {
      if(_left.blackHeight() != 1)
        throw new MapCheckException(_value);
      else
        return 1 + (_red?0:1);
    }
    else
    {
      if(_left.blackHeight() != _right.blackHeight())
        throw new MapCheckException(_value);
      else
        return _left.blackHeight() + (_red?0:1);
    }
  }
}
