package jatva.util;

import jatva.lang.InsufficientAccessibilityError;
import jatva.lang.TranslationError;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

public class MethodInfo extends OperationInfo
{
    protected Method method;

    public MethodInfo( String extendedReflectSignature )
    {
        super( extendedReflectSignature );
    }

    public Method getMethod()
    {
        return method;
    }

    @Override
    protected final Class<?>[] obtainExecutableMember()
    {
        method = Utils.getMethodByStringPrefix( reflectSignature );

        if( method == null )
            throw new TranslationError( "Invalid method signature"
                                      , new String[]{ "method signature" }
                                      , new String[]{ reflectSignature }
                                      );

        try
        {
            method.setAccessible( true );
        }
        catch( SecurityException e )
        {
          throw new InsufficientAccessibilityError( "Insufficient rights to call implementation method through reflection"
                                                  , e
                                                  , new String[]{ "method"}
                                                  , new String[]{ reflectSignature }
                                                  );
        }

        this.declaringClass = Utils.prettifyClass( method.getDeclaringClass() );
        this.returnType = Utils.prettifyClass( method.getReturnType() );

        return method.getParameterTypes();
    }

    @Override
    protected final void buildPrettySignature()
    {
        boolean is_static = ( method.getModifiers() & Modifier.STATIC ) != 0;

        StringBuffer res = new StringBuffer();

        if( is_static )
            res.append( "static " );
        res.append( returnType );
        res.append( ' ' );
        res.append( declaringClass );
        res.append( '.' );
        res.append( method.getName() );
        appendParTypes( res );
        this.prettySignature = res.toString();
    }
}
