/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: StringToIntMap.java,v 1.1 2008/06/23 12:29:19 kuliamin Exp $
 * Created on 29.01.2007
 *
 */
package jatva.util;


/** 
 *
 * @author Victor Kuliamin
 *
 */
public class StringToIntMap
{
  StringToIntMapNode _node = null;
  int _size = 0;
  
  public StringToIntMap()
  {
  }
  
  public String toString()
  {
    StringBuffer st = new StringBuffer("{");
    StringToIntMapNode n = minimum();
    
    if(n != null)
    {
      st.append("" + n._key + " : " + n._value);
      n = next(n);
    }
    
    while(n != null)
    {
      st.append(", ");
      st.append("" + n._key + " : " + n._value);
      n = next(n);
    }
      
    st.append("}");
    
    return st.toString();
  }
  
  void rotate(boolean left, StringToIntMapNode x)
  {
    StringToIntMapNode y;
    
    if(left)
    {
      y = x._right;
      x._right = y._left;
      if(y._left != null) y._left._parent = x;
      
      y._parent = x._parent;
      if(x._parent != null)
      {
        if(x == x._parent._left) x._parent._left = y;
        else                     x._parent._right = y;
      }
      else _node = y;
      
      y._left = x;
      x._parent = y;
    }
    else
    {
      y = x._left;
      x._left = y._right;
      if(y._right != null) y._right._parent = x;
      
      y._parent = x._parent;
      if(x._parent != null)
      {
        if(x == x._parent._left) x._parent._left = y;
        else                     x._parent._right = y;
      }
      else _node = y;
      
      y._right = x;
      x._parent = y;
    }
  }
  
  public boolean containsKey(String key)
  {
    if(_node == null) return false;
    StringToIntMapNode p = _node;
    int c; 

    do
    {
      c = p._key.compareTo(key);
      if(c == 0)     return true;
      else if(c > 0) p = p._left;
      else           p = p._right;
    }
    while(p != null);
    
    return false;
  }
  
  public boolean containsValue(int val)
  {
    if(_node == null) return false;
    return _node.containsValue(val);
  }
  
  StringToIntMapNode minimum()
  {
    if(_node == null) return null;
    else              return _node.minimum();
  }
  
  StringToIntMapNode maximum()
  {
    if(_node == null) return null;
    else              return _node.maximum();
  }
  
  StringToIntMapNode next(StringToIntMapNode p)
  {
    if(p._right != null) return p._right.minimum();
    while(p._parent != null && p == p._parent._right) p = p._parent;
    return p._parent;
  }
  
  StringToIntMapNode prev(StringToIntMapNode p)
  {
    if(p._left != null) return p._left.maximum();
    while(p._parent != null && p == p._parent._left) p = p._parent;
    return p._parent;
  }
  
  public int get(String key)
  {
    if(_node == null) return 0;
    StringToIntMapNode p = _node;
    int c;
    
    do
    {
      c = p._key.compareTo(key);
      if(c == 0)     return p._value;
      else if(c > 0) p = p._left;
      else           p = p._right;
    }
    while(p != null);
    
    return 0;
  }

  public int put(String key, int val)
  {
    if(_node == null)
    {
      _node = new StringToIntMapNode(key, val);
      _node._red = false;
      _size++;
      
      return 0;
    }
    
    StringToIntMapNode p = _node, q;
    int res;
    int c;
    
    do
    {
      q = p;

      c = p._key.compareTo(key);
      if(c == 0)
      {
        res = p._value;
        p._value = val;
        return res;
      }
      else if(c > 0) p = p._left;
      else           p = p._right;
    }
    while(p != null);
    
    if(q._key.compareTo(key) > 0)
         q._left  = ( p = new StringToIntMapNode(key, val) );
    else q._right = ( p = new StringToIntMapNode(key, val) );
    p._parent = q;
    
    _size++;
    
    StringToIntMapNode r;
    
    //    System.out.println("After add : " + toString());
    
    while(q != null && q._red == true)
    {
      if(q == q._parent._left)
      {
        //        System.out.print("q is on the left, ");
        
        r = q._parent._right;
        if(r == null || !r._red )
        {
          if(p == q._right)
          {
            //            System.out.println("case 2");
                        
            q._right = p._left;
            if(p._left != null) p._left._parent = q;
            p._left = q;
            p._parent = q._parent;
            q._parent._left = p;
            q._parent = p;
            p = q;
            q = q._parent;
            
            //            System.out.println("After rotation : " + toString());
          }
          
          //          System.out.println("case 3");
          
          if(q._parent._parent != null)
          {
            if(q._parent._parent._left == q._parent) q._parent._parent._left  = q;
            else                                     q._parent._parent._right = q;
          }
          else _node = q;
          
          q._red = false;
          q._parent._red = true;
          q._parent._left = q._right;
          if(q._right != null) q._right._parent = q._parent;
          q._right = q._parent;
          q._parent = q._parent._parent;
          q._right._parent = q;
          
          return 0;
        }
        else
        {
          //          System.out.println("case 1");
          
          q._red = false;
          r._red = false;
          q._parent._red = true;
          p = q._parent;
          q = p._parent;
        }
      }
      else
      {
        //        System.out.print("q is on the right, ");

        r = q._parent._left;
        if(r == null || !r._red)
        {
          if(p == q._left)
          {
            //            System.out.println("case 2");

            q._left = p._right;
            if(p._right != null) p._right._parent = q;
            p._right = q;
            p._parent = q._parent;
            q._parent._right = p;
            q._parent = p;
            p = q;
            q = q._parent;

            //            System.out.println("After rotation : " + toString());
          }
          
          //         System.out.println("case 3");
         
          if(q._parent._parent != null)
          {
            if(q._parent._parent._left == q._parent) q._parent._parent._left  = q;
            else                                     q._parent._parent._right = q;
          }
          else _node = q;
          
          q._red = false;
          q._parent._red = true;
          q._parent._right = q._left;
          if(q._left != null) q._left._parent = q._parent;
          q._left = q._parent;
          q._parent = q._parent._parent;
          q._left._parent = q;
          
          return 0;
        }
        else
        {
          q._red = false;
          r._red = false;
          q._parent._red = true;
          p = q._parent;
          q = p._parent;
        }
      }
    }
    
    _node._red = false;
    
    return 0;
  }
  
  public int remove(String key)
  {
    if(_node == null) return 0;
    StringToIntMapNode p = _node;
    int c;
    
    do
    {
      c = p._key.compareTo(key);
      if(c > 0)      p = p._left;
      else if(c < 0) p = p._right;
      else           break;
    }
    while(p != null);
    
    if(p == null) return 0;
    
    boolean isLeft;
    int res = p._value;
    
    if(p._left == null)
    {
      if(p._parent == null)
      {
        _node = p._right;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._right;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._right;
          isLeft = false;
        }
      }
      
      if(p._right != null) p._right._parent = p._parent;
      if(!p._red)          fixup(p._parent, p._right, isLeft);
    }
    else if(p._right == null)
    {
      if(p._parent == null)
      {
        _node = p._left;
        isLeft = true;
      }
      else
      {
        if(p == p._parent._left)
        {
          p._parent._left = p._left;
          isLeft = true;
        }
        else
        {
          p._parent._right = p._left;
          isLeft = false;
        }
      }
      
      p._left._parent = p._parent;
      if(!p._red) fixup(p._parent, p._left, isLeft);
    }
    else
    {
      StringToIntMapNode q = next(p);
      p._key = q._key;
      p._value = q._value;
      
      if(q._parent == null)
      {
        _node = q._right;
        isLeft = true;
      }
      else
      {
        if(q == q._parent._left)
        {
          q._parent._left = q._right;
          isLeft = true;
        }
        else
        {
          q._parent._right = q._right;
          isLeft = false;
        }
      }
      
      if(q._right != null) q._right._parent = q._parent;
      if(!q._red)          fixup(q._parent, q._right, isLeft); 
    }
    
    _size--;
    return res;
  }
  
  void fixup(StringToIntMapNode p, StringToIntMapNode q, boolean isLeft)
  {
    while(p != null && ( q == null || q._red == false ) )
    {
      StringToIntMapNode w, v;
      
      if(isLeft)
      {
        w = p._right;
        if(w._red)
        {
          w._red = false;
          p._red = true;
          rotate(true, p);
          w = p._right;
        }
        
        v = w._left;
        
        if(w._right == null || !w._right._red)
        {
          if(v == null || !v._red)
          {
            w._red = true;
            
            q = p;
            p = p._parent;
            isLeft = (p == null)?true:(q == p._left);
            
            continue;
          }
          else
          {
            w._red = true;
            v._red = false;
            rotate(false, w);
            
            w = v;
            v = v._left;
          }
        }
        
        w._red = p._red;
        p._red = false;
        if(w._right != null) w._right._red = false;
        rotate(true, p);
        
        q = _node;
        p = null;
      }
      else
      {
        w = p._left;
        if(w._red)
        {
          w._red = false;
          p._red = true;
          rotate(false, p);
          w = p._left;
        }
        
        v = w._right;
        
        if(w._left == null || !w._left._red)
        {
          if(v == null || !v._red)
          {
            w._red = true;
            
            q = p;
            p = p._parent;
            isLeft = (p == null)?true:(q == p._left);
            
            continue;
          }
          else
          {
            w._red = true;
            v._red = false;
            rotate(true, w);
            
            w = v;
            v = v._right;
          }
        }
        
        w._red = p._red;
        p._red = false;
        if(w._left != null) w._left._red = false;
        rotate(false, p);
        
        q = _node;
        p = null;
      }
    }
    
    if(q != null) q._red = false;
  }
  
  public void clear()
  {
    _node = null;
    _size = 0;
  }
  
  public boolean isEmpty()
  {
    return (_node == null);
  }
  
  public boolean equals(Object o)
  {
    if(!(o instanceof StringToIntMap)) return false;
    StringToIntMap r = (StringToIntMap)o;
    if(_size != r._size) return false;
    if(r == this)        return true;
            
    StringToIntMapNode p = minimum(), q = r.minimum();
    
    while(p != null)
    {
      if(p._key != q._key || p._value != q._value) return false;
      
      p = next(p);
      q = r.next(q);
    }
    
    return true;
  }

  public int hashCode()
  {
    int res = 0;
    StringToIntMapNode p = minimum();

    while(p != null)
    {
      res += p._key.hashCode() + 7*p._value;
      p = next(p);
    }
    
    return res;
  }
  
  public int size()
  {
    return _size;
  }
  
  public StringSet keys()
  {
    StringSet res = new StringSet();
    if(_node != null) _node.appendKeySet(res);
    return res;
  }
  
  public StringToIntMap(StringToIntMap r)
  {
    _size = r._size;
    
    if(r._node != null)
    {
      _node = new StringToIntMapNode(null, 0);
      r._node.nodeClone(_node);
    }
  }
  
  public Object clone()
  {
    return new StringToIntMap(this);
  }

  public StringToIntMapEnumerator GetEnumerator()
  {
    return new StringToIntMapEnumerator(this);
  }

  public class Entry
  {
    public String key = null;
    public int value = 0;

    public Entry(String key, int val)
    {
      this.key = key;
      this.value = val;
    }
  }

  public class StringToIntMapEnumerator
  {
    StringToIntMap _base = null;
    StringToIntMapNode _current = null;

    public StringToIntMapEnumerator(StringToIntMap s)
    {
      _base = s;
    }

    public Entry current()
    {
      return new Entry(_current._key, _current._value);
    }

    public boolean moveNext()
    {
      StringToIntMapNode q = null;
      if(_current == null && _base._node == null)  return false;
      else if(_current == null)                    { _current = _base.minimum(); return true; }
      else if((q = _base.next(_current)) == null)   { _current = null; return false; }
      else                                         { _current = q; return true; }
    }
  }

  protected class StringToIntMapNode
  {
    public StringToIntMapNode _left = null;
    public StringToIntMapNode _right = null;
    public StringToIntMapNode _parent = null;
    public String _key = null;
    public int _value = 0;
    public boolean _red = true;
  
    public StringToIntMapNode(String key, int val)
    {
      _key = key;
      _value = val;
    }
  
    public StringToIntMapNode minimum()
    {
      StringToIntMapNode p = this;
      while(p._left != null) p = p._left;
      return p;
    }
  
    public StringToIntMapNode maximum()
    {
      StringToIntMapNode p = this;
      while(p._right != null) p = p._right;
      return p;
    }

    public void appendKeySet(StringSet s)
    {
      s.add(_key);
      if(_left != null) _left.appendKeySet(s);
      if(_right != null) _right.appendKeySet(s);
    }

    public boolean containsValue(int val)
    {
      if(_value == val)                                    return true;
      else if(_left != null && _left.containsValue(val))   return true;
      else if(_right != null && _right.containsValue(val)) return true;
      else                                                 return false;
    }

    public void nodeClone(StringToIntMapNode n)
    {
      n._key = _key;
      n._value = _value;
      n._red = _red;
    
      if(_left != null)
      {
        n._left = new StringToIntMapNode(null, 0);
        n._left._parent = n;
        _left.nodeClone(n._left);
      }
    
      if(_right != null)
      {
        n._right = new StringToIntMapNode(null, 0);
        n._right._parent = n;
        _right.nodeClone(n._right);
      }
    }
  }

}
