
package jatva.util;

import java.lang.reflect.Array;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

public class TraceUtils
{
  /**
   * @pre: cl != null
   */
  public static String typeToString(Class cl)
  {
    StringBuffer buf = new StringBuffer();
    
    while(cl.isArray())
    {
      buf.append("[]");
      cl = cl.getComponentType();
    }
    
    buf.insert(0, cl.getName());
    
    return buf.toString();
  }
  
  public static String objectToString(Object o)
  {
    if(o == null) return "null";
    
    Class cl = o.getClass();
    int i = 0;
    
    while(cl.isArray())
    {
      i++;
      cl = cl.getComponentType();
    }

    if(o instanceof String)          return (String)o;
    else if(o instanceof Class)      return typeToString((Class)o);
    else if(i > 0)                   return arrayToString(o, i, cl);
    else if(o instanceof List)       return listToString((List)o);
    else if(o instanceof Map)        return mapToString((Map)o);
    else if(o instanceof Collection) return collectionToString((Collection)o);
    
    else      return o.toString();
  }
  
  
  public static String objectToString(boolean o)
  {
    return "" + o;
  }
  
  public static String objectToString(char o)
  {
    return "" + o;
  }
  
  public static String objectToString(short o)
  {
    return "" + o;
  }
  
  public static String objectToString(int o)
  {
    return "" + o;
  }
  
  public static String objectToString(long o)
  {
    return "" + o;
  }
  
  public static String objectToString(float o)
  {
    return "" + o;
  }
  
  public static String objectToString(double o)
  {
    return "" + o;
  }
  
  /**
   * @pre: cl != null, o != null, i > 0
   *       o - array of i dimensions, its elements has type cl
   */
  static String arrayToString(Object o, int i, Class cl)
  {
    StringBuffer buf = new StringBuffer();
    
    putArrayInStringBuffer(buf, o, i, cl);
    
    return buf.toString();
  }
  
  static void putArrayInStringBuffer(StringBuffer buf, Object o, int i, Class cl)
  {
    int l = Array.getLength(o);
    int j = 0;

    buf.append('[');
    
    if(i == 1)
    {
      if(cl.isPrimitive())
      {
        if(cl.equals(Boolean.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(','); buf.append(Array.getBoolean(o, j)); }
        else if(cl.equals(Character.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(','); buf.append(Array.getChar(o, j));    }
        else if(cl.equals(Byte.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(','); buf.append(Array.getByte(o, j));    }
        else if(cl.equals(Short.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(','); buf.append(Array.getShort(o, j));   }
        else if(cl.equals(Integer.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(','); buf.append(Array.getInt(o, j));     }
        else if(cl.equals(Long.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(','); buf.append(Array.getLong(o, j));    }
        else if(cl.equals(Float.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(','); buf.append(Array.getFloat(o, j));   }
        else if(cl.equals(Double.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(','); buf.append(Array.getDouble(o, j));  }
        else// if(cl.equals(Void.TYPE))
          for(; j < l; j++) { if(j != 0) buf.append(',');                                     }
      }
      else
        for(; j < l; j++)   { if(j != 0) buf.append(','); buf.append(Array.get(o, j));        }
    }
    else
      for(; j < l; j++)     { if(j != 0) buf.append(','); putArrayInStringBuffer(buf, Array.get(o, j), i-1, cl); }
    
    buf.append(']');   
  }
  
  static String listToString(List l)
  {
    StringBuffer b = new StringBuffer("[");
    for(int i = 0; i < l.size(); i++)
    {
      if(i > 0) b.append(", ");
      b.append(objectToString(l.get(i)));
    }
    b.append(']');
    return b.toString();
  }
  
  static String mapToString(Map m)
  {
    StringBuffer b = new StringBuffer("{");
    boolean first = true;
    for(Iterator i = m.keySet().iterator(); i.hasNext();)
    {
      Object k = i.next();
      if(!first) b.append(", ");
      first = false;
      b.append("(");
      b.append(objectToString(k));
      b.append(" : ");
      b.append(objectToString(m.get(k)));
      b.append(")");
    }
    b.append('}');
    return b.toString();
  }
  
  static String collectionToString(Collection m)
  {
    StringBuffer b = new StringBuffer("{");
    boolean first = true;
    for(Iterator i = m.iterator(); i.hasNext();)
    {
      if(!first) b.append(", ");
      first = false;
      b.append(objectToString(i.next()));
    }
    b.append('}');
    return b.toString();
  }
}