/*
 * Created on 14.02.2005
 *
 */
package jatva.util;

/**
 * @author Victor Kuliamin
 *
 */
public class TwoBitSet
{
  private final static int CHUNCK_SIZE = 32;
  private final static byte MASK       =  3;

  private long[] _items = new long[1];
  private int    _size  = 0;

  public TwoBitSet()
  {
  }

  public TwoBitSet(int size)
  {
    _size  = size;
    _items = new long[size == 0?1:((size-1)/CHUNCK_SIZE + 1)];
  }

  public String toString()
  {
    StringBuffer st = new StringBuffer("{ ");
    int s = _size;

    l:for(int i = 0; i < _items.length; i++)
    {
      for(int j = 0; j < CHUNCK_SIZE; j++)
      {
        if(j < s)
        {
          st.append( (_items[i] >> (2*j)) & 1 );
          st.append( (_items[i] >> (2*j+1)) & 1 );
          st.append(' ');
        }
        else break l;
      }
      s -= CHUNCK_SIZE;
    }
      
    st.append("}");
    
    return st.toString();
  }
  
  public byte set(int i, byte val)
  {
    int d = i/CHUNCK_SIZE;
    int  r = (int)(i%CHUNCK_SIZE);

    if(d >= _items.length)
    {
      long[] items = new long[d + 1];
      System.arraycopy(_items, 0, items, 0, _items.length);
      _items = items;
      _items[d] = ((long)(val & MASK)) << (2*r);
      _size = i+1;

      return (byte)0;
    }
    else
    {
      byte res = (byte)((_items[d] >> (2*r)) & MASK);
      _items[d] |= ( (long)(val & MASK) ) << (2*r);
      if(_size <= i) _size = i+1; 

      return res;
    }
  }

  public boolean setLeftBit(int i, boolean value)
  {
    int d = i/CHUNCK_SIZE;
    int  r = (int)(i%CHUNCK_SIZE);

    if(d >= _items.length)
    {
      long[] items = new long[d + 1];
      System.arraycopy(_items, 0, items, 0, _items.length);
      _items = items;
      _items[d] = (value)?(2L << (2*r)):0L;
      _size = i+1;

      return false;
    }
    else
    {
      boolean res = ((_items[d] >> (2*r)) & 2L) != 0;
      if(value) _items[d] |= 2L << (2*r);
      else      _items[d] &= ~(2L << (2*r));
      if(_size <= i) _size = i+1; 

      return res;
    }
  }

  public boolean setRightBit(int i, boolean value)
  {
    int d = i/CHUNCK_SIZE;
    int  r = (int)(i%CHUNCK_SIZE);

    if(d >= _items.length)
    {
      long[] items = new long[d + 1];
      System.arraycopy(_items, 0, items, 0, _items.length);
      _items = items;
      _items[d] = (value)?(1L << (2*r)):0L;
      _size = i+1;

      return false;
    }
    else
    {
      boolean res = ((_items[d] >> (2*r)) & 1L) != 0;
      if(value) _items[d] |= 1L << (2*r);
      else      _items[d] &= ~(1L << (2*r));
      if(_size <= i) _size = i+1; 

      return res;
    }
  }

  private final static long _TS_LEFT_BITS  = 0xaaaaaaaaaaaaaaaaL;
  private final static long _TS_RIGHT_BITS = 0x5555555555555555L;

  public void clearLeftBits()
  {
    for(int i = 0; i < _items.length; i++)
      _items[i] &= _TS_RIGHT_BITS;
  }

  public void clearRightBits()
  {
    for(int i = 0; i < _items.length; i++)
      _items[i] &= _TS_LEFT_BITS;
  }

  public byte get(int i)
  {
    int d = i/CHUNCK_SIZE;
    int  r = (int)(i%CHUNCK_SIZE);

    if(d >= _items.length)
      return (byte)0;
    else
      return (byte)((_items[d] >> (2*r)) & MASK);
  }

  public boolean getLeftBit(int i)
  {
    int d = i/CHUNCK_SIZE;
    int  r = (int)(i%CHUNCK_SIZE);

    if(d >= _items.length) return false;
    else                   return ((_items[d] >> (2*r)) & 2L) != 0;
  }

  public boolean getRightBit(int i)
  {
    int d = i/CHUNCK_SIZE;
    int  r = (int)(i%CHUNCK_SIZE);

    if(d >= _items.length) return false;
    else                   return ((_items[d] >> (2*r)) & 1L) != 0;
  }

  public void clear()
  {
    for(int i = 0; i < _items.length; i++) _items[i] = 0;
    _size = 0;
  }
  
  public boolean isEmpty()
  {
    for(int i = 0; i < _items.length; i++)
      if(_items[i] != 0) return false;

    return true;
  }

  public int find(byte val)
  {
    for(int i = 0; i < _items.length; i++)
      for(int j = 0; j < CHUNCK_SIZE; j++)
        if(((_items[i] >> (2*j)) & (long)MASK) == (long)val)
          return i*CHUNCK_SIZE + j;

    return -1;
  }
  
  public boolean equals(Object o)
  {
    if(!(o instanceof TwoBitSet)) return false;
    
    TwoBitSet r = (TwoBitSet)o;
    if(_size != r._size) return false;
    if(r == this)        return true;
            
    for(int i = 0; i < _items.length; i++)
      if(_items[i] != r._items[i]) return false;

    return true;
  }

  public int hashCode()
  {
    int res = 0;

    for(int i = 0; i < _items.length; i++)
      res += ((int)(_items[i] >> CHUNCK_SIZE) + (int)(_items[i] & 0xffffffff));
    
    return res;
  }
  
  public int size()
  {
    return _size;
  }
  
  public long[] toArray()
  {
    return (long[])_items.clone();
  }
  
  public TwoBitSet(TwoBitSet r)
  {
    _size = r._size;
    _items = new long[r._items.length];
    System.arraycopy(r._items, 0, _items, 0, _items.length);
  }
  
  public Object clone()
  {
    return new TwoBitSet(this);
  }

  public TwoBitSetEnumerator GetEnumerator()
  {
    return new TwoBitSetEnumerator(this);
  }

  public class TwoBitSetEnumerator
  {
    TwoBitSet _base = null;
    int _current = -1;

    public TwoBitSetEnumerator(TwoBitSet s)
    {
      _base = s;
    }

    public boolean MoveNext()
    {
      if((int)(++_current) == _base._size) return false;
      else                          return true;
    }
  }
}

