/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: WeakIdentityMap.java,v 1.4 2008/06/23 12:29:19 kuliamin Exp $
 * Created on 30.04.2004
 */
package jatva.util;

import jatva.iterators.lang.AbstractObjectPairIterator;
import jatva.iterators.lang.IObjectPairIterator;

import java.lang.ref.WeakReference;

/**
 * @author Victo Kuliamin
 */
public class WeakIdentityMap
  extends    AbstractMap
  implements IMap
{
  public WeakIdentityMap() {}
  
  public WeakIdentityMap(Object key, Object val)
  {
    if(key != null)
    {
      _size = 1;
      MapNode p = new MapNode(key, val);
      _items[hashKey(key)&(_itemsNumber-1)] = p;  
      p._bkw = p._frw = _last = p;  
    }
  }
  
  public WeakIdentityMap(IMap m)
  {
    super(m);
  }
  
  public WeakIdentityMap(ISet s)
  {
    super(s);
  }
  
  public WeakIdentityMap(ISet s, Object o)
  {
    super(s, o);
  }

  private void strip(MapNode p, MapNode q, int i)
  {
    if(q == null) 
    {
      if(p != _items[i]) System.out.println("!!!ERORR1 in stripObject"); 
      _items[i] = p._next;
    } 
    else 
    {
      if(p != q._next) System.out.println("!!!ERORR2 in stripObject");
      q._next   = p._next; 
    } 

    if(_size-- == 1) _last = null;
    else             
    {
      p._bkw._frw = p._frw;
      p._frw._bkw = p._bkw;
      if(p == _last) _last = p._bkw;
    }
  }

  private void stripNode(MapNode p)
  {
    MapNode q = _items[p.i&(_itemsNumber-1)], r = null;
    while(q != p)
    {
      r = q;
      q = q._next;
    }
    
    if(r == null) _items[p.i&(_itemsNumber-1)] = p._next;
    else          r._next     = p._next;
    
    if(_size-- == 1) _last = null;
    else             
    {
      p._bkw._frw = p._frw;
      p._frw._bkw = p._bkw;
      if(p == _last) _last = p._bkw;
    }
  }

  protected int hashKey(Object o)
  {
    return Utils.identityHash(o);
  }

  protected int hashValue(Object o)
  {
    return Utils.hash(o);
  }

  protected boolean equalValues(Object o1, Object o2)
  {
    return Utils.equalObjects(o1, o2);
  }

  protected IList makeList()
  {
    // TODO Auto-generated method stub
    return null;
  }

  protected ISet makeSet()
  {
    return new WeakIdentitySet();
  }

  protected IMultiSet makeMultiSet()
  {
    // TODO Auto-generated method stub
    return null;
  }

  public Object clone()
  {
    return new WeakIdentityMap(this);
  }

  public boolean equalsByRefs(Object o)
  {
    return equals(o);
  }
  
  public void strip()
  {
    if(_size == 0) return;
    
    MapNode p, q;
  
    for(int i = 0; i < _itemsNumber; i++)
    {
      if(null == (p = _items[i])) continue;
      q = null;
      
      while(p != null)
      {
        if(p._key.get() == null) strip(p, q, i);
        else q = p;

        p = p._next;    
      }
    }
  }

  public int size()
  {
    return _size;
  }

  public void clear()
  {
    _size  = 0;
    _items = new MapNode[_itemsNumber];
    _last  = null;
  }

  public boolean containsKey(Object key)
  {
    if(key != null && _size != 0)
    {
      int h = hashKey(key)&(_itemsNumber-1);
      MapNode p = _items[h], q = null;
      Object x;
    
      while(p != null)
      {
        if((x = p._key.get()) != null && x == key) return true;
        else if(x == null) strip(p, q, h);
        else q = p;
        p = p._next;
      }
    }
    return false;
  }

  public boolean containsKeyRef(Object key)
  {
    return containsKey(key);
  }

  public boolean contains(Object key, Object val)
  {
    if(key != null && _size != 0)
    {
      int h = hashKey(key)&(_itemsNumber-1);
      MapNode p = _items[h], q = null;
      Object x;
    
      while(p != null)
      {
        if((x = p._key.get()) != null && x == key) return equalValues(p._val, val);
        else if(x == null) strip(p, q, h);
        else q = p;
        p = p._next;
      }
    }
    
    return false;
  }

  public boolean containsRefs(Object key, Object val)
  {
    if(key != null && _size != 0)
    {
      int h = hashKey(key)&(_itemsNumber-1);
      MapNode p = _items[h], q = null;
      Object x;
    
      while(p != null)
      {
        if((x = p._key.get()) != null && x == key) return p._val == val;
        else if(x == null) strip(p, q, h);
        else q = p;
        p = p._next;
      }
    }
    
    return false;
  }
  
  public Object get(Object key)
  {
    if(key != null && _size != 0)
    {
      int h = hashKey(key)&(_itemsNumber-1);
      MapNode p = _items[h], q = null;
      Object x;
    
      while(p != null)
      {
        if((x = p._key.get()) != null && x == key) return p._val;
        else if(x == null) strip(p, q, h);
        else q = p;
        p = p._next;
      }
    }
    
    return null;
  }

  public Object removeKey(Object key)
  {
    if(key != null && _size != 0)
    {
      int h = hashKey(key)&(_itemsNumber-1);
      MapNode p = _items[h], q = null;
      Object x;
    
      while(p != null)
      {
        if((x = p._key.get()) != null && x == key) { strip(p, q, h); return p._val; }
        else if(x == null) strip(p, q, h);
        else q = p;
        p = p._next;
      }
    }
    
    return null;
  }

  public boolean remove(Object key, Object val)
  {
    if(key != null && _size != 0)
    {
      int h = hashKey(key)&(_itemsNumber-1);
      MapNode p = _items[h], q = null;
      Object x;
    
      while(p != null)
      {
        if((x = p._key.get()) != null && x == key)
        {
          if(equalValues(p._val, val)) { strip(p, q, h); return true; }
          else                         return false; 
        }
        else if(x == null) strip(p, q, h);
        else q = p;
        p = p._next;
      }
    }

    return false;
  }

  public Object put(Object key, Object value)
  {
    if(key == null) return null;
    
    int h = hashKey(key)&(_itemsNumber-1);
    MapNode p = _items[h], q = null;

    if(_size != 0)
    {
      Object x;
      while(p != null)
      {
        if((x = p._key.get()) != null && x == key)
        {
          Object v = p._val;
          p._val = value;
          return v;
        }
        else if(x == null) strip(p, q, h);
        else q = p;
         
        p = p._next;
      }
    }
      
    p = new MapNode(key, value);
    p._next = _items[h];
    _items[h] = p;
    if(_size++ == 0) { _last = p._bkw = p._frw = p; }
    else             { p._frw = _last._frw; p._bkw = _last; _last._frw._bkw = p; _last._frw = p; _last = p; }
    return null;
  }
  
  public boolean cput(Object key, Object value)
  {
    if(key == null) return false;
    
    int h = hashKey(key)&(_itemsNumber-1);
    MapNode p = _items[h], q = null;

    if(_size != 0)
    {
      Object x;
      while(p != null)
      {
        if((x = p._key.get()) != null && x == key) return false;
        else if(x == null) strip(p, q, h);
        else q = p;
         
        p = p._next;
      }
    }
      
    p = new MapNode(key, value);
    p._next = _items[h];
    _items[h] = p;
    if(_size++ == 0) { _last = p._bkw = p._frw = p; }
    else             { p._frw = _last._frw; p._bkw = _last; _last._frw = p; _last = p; }
    return true;
  }

  public IObjectPairIterator iterator()
  {
    return new WeakIdentityMapIterator(this);
  }

  private static class WeakIdentityMapIterator
    extends     AbstractObjectPairIterator
    implements  IObjectPairIterator
  {
    private MapNode         p = null;
    private WeakIdentityMap s = null; 
    
    WeakIdentityMapIterator(WeakIdentityMap s)  
    {
      this.s = s;
      if(s._last == null) p = null;
      else                p = s._last._frw;
      
      while(p != null && p._key.get() == null)
      {
        s.stripNode(p);
        if(p == s._last) p = null;
        else             p = p._frw;        
      }
    }
       
    WeakIdentityMapIterator(WeakIdentityMapIterator r)
    {
      s = r.s;
      p = r.p;      
    }
    
    public boolean stop()
    {
      if(p == null) return true;
      else          return false;
    }
  
    public boolean equals(Object o)
    {
      if(!(o instanceof WeakIdentityMapIterator)) return false;
      
      WeakIdentityMapIterator r = (WeakIdentityMapIterator)o;
       
      return r.s == s && r.p == p;
    }

    public int hashCode()
    {
      return s.hashCode() + 17*System.identityHashCode(p);
    }
    
    public void remove()
    {
      MapNode q = p;
      if(p == s._last) p = null;
      else             p = p._frw;
      s.stripNode(q);
      
      while(p != null && p._key.get() == null)
      {
        s.stripNode(p);
        if(p == s._last) p = null;
        else             p = p._frw;        
      }
    }
  
    public void shift()
    {
      if(p == s._last) p = null;
      else             p = p._frw;
        
      while(p != null && p._key.get() == null)
      {
        s.stripNode(p);
        if(p == s._last) p = null;
        else             p = p._frw;        
      }
    }    
      
    public Object value1()
    {
      return p._key.get();
    }
  
    public Object value2()
    {
      return p._val;
    }

    public Object clone()
    {
      return new WeakIdentityMapIterator(this);
    }

    public void replaceObject1(Object o)
    {
      // TODO Auto-generated method stub
    }

    public void replaceObject2(Object o)
    {
      p._val = o;
    }
  }
  
  private static int _itemsNumber = 16;
  
  private MapNode _items[] = new MapNode[_itemsNumber];
  private int     _size    = 0;
  private MapNode _last    = null;
  
  private static class MapNode
  {
    MapNode _next;
    MapNode _frw;
    MapNode _bkw;
    WeakReference _key;
    Object _val;
    int i;
    
    MapNode(Object key, Object val) { _key = new WeakReference(key); _val = val; i = Utils.hash(key); }
  }
 
  public static IMap pair(Object key, Object value)
  {
    return new WeakIdentityMap(key, value);
  }

  public static IMap identity(ISet s)
  {
    return new WeakIdentityMap(s);
  }

  public static IMap constant(ISet s, Object value)
  {
    return new WeakIdentityMap(s, value);
  }

  public static IMap intersection(IMap s1, IMap s2)
  {
    IMap res = new WeakIdentityMap(s1);
    res.retain(s2);
    
    return res;
  }
  
  public static IMap difference(IMap s1, IMap s2)
  {
    IMap res = new WeakIdentityMap(s1);
    res.removeAll(s2);
    
    return res;
  }
  
  public static IMap symDifference(IMap s1, IMap s2)
  {
    IMap res = difference(s1, s2);
    res.putAll(difference(s2, s1));
    
    return res;
  }
  
  public static IMap composition(IMap s1, IMap s2)
  {
    IMap res = new WeakIdentityMap(s1);
    res.compose(s2);
    
    return res;
  }
  
  public static IMap extension(IMap s1, IMap s2)
  {
    IMap res = new WeakIdentityMap(s1);
    res.extend(s2);
    
    return res;
  }
}
