/*
 * Copyright (c) 2004 ISP RAS.
 * 109004, B. Communisticheskaya, 25, Moscow, Russia.
 * All rights reserved.
 *
 * $Id: WeakIdentitySet.java,v 1.4 2008/06/23 12:29:19 kuliamin Exp $
 * Created on 30.04.2004
 */
package jatva.util;

import java.lang.ref.WeakReference;

import jatva.iterators.lang.AbstractObjectIterator;
import jatva.iterators.lang.IObjectIterator;

/**
 * @author Victor Kuliamin
 *
 */
public class WeakIdentitySet
  extends AbstractSet
  implements ISet
{
  public WeakIdentitySet() {}

  public WeakIdentitySet(Object o)
  {
    if(o != null)
    {
      _size = 1;
      SetNode p = new SetNode(o);
      _items[hash(o)&(_itemsNumber-1)] = p;
      p._bkw = p._frw = _last = p;  
    }
  }
  
  /**
   * Method for debugging only
   * @return String representation of a tree
   */
  private String debug()
  {
    int cc = 0;
    StringBuffer s = new StringBuffer("" + _size + " {");
    if(_size != 0)
    {
      SetNode p;
      boolean flag = true;
      Object x;
    
      for(int i = 0; i < _itemsNumber; i++)
      {
        s.append("\n\n | (" + i + ")");
        if(null == (p = _items[i])) continue;
        while(p != null)
        {
          if((x = p._r.get()) == null)
          {
            if(!flag) s.append(", ");
            else      flag = false;
            s.append(p + "<!!!>");
            cc++;
          } 
          else
          {          
            if(!flag) s.append(", ");
            else      flag = false;
            s.append("" + p + '<' + x + '>');
            cc++;
          }
          p = p._next;    
        }
      }
    }
    s.append('}');
    if(cc != _size) s.append(" !!!ERROR, cc = " + cc + ", _size = " + _size);
    return s.toString();  
  }
  
  public WeakIdentitySet(ISet s)
  {
    super(s);
  }
  
  protected int hash(Object o)
  {
    return Utils.identityHash(o); 
  }
  
  protected IList makeList()
  {
    // TODO Auto-generated method stub
    return null;
  }

  private void stripObject(SetNode p, SetNode q, int i)
  {
    if(q == null) 
    {
      if(p != _items[i]) System.out.println("!!!ERORR1 in stripObject"); 
      _items[i] = p._next;
    } 
    else 
    {
      if(p != q._next) System.out.println("!!!ERORR2 in stripObject");
      q._next = p._next; 
    } 

    if(_size-- == 1) _last = null;
    else             
    {
      p._bkw._frw = p._frw;
      p._frw._bkw = p._bkw;
      if(p == _last) _last = p._bkw;
    }
  }

  private void stripNode(SetNode p)
  {
    SetNode q = _items[p.i&(_itemsNumber-1)], r = null;
    while(q != p)
    {
      if(q == null)
      {
        System.out.println("!!!ERORR3 in stripNode");
      }
      r = q;
      q = q._next;
    }
    
    if(r == null) _items[p.i&(_itemsNumber-1)] = p._next;
    else          r._next     = p._next;
    
    if(_size-- == 1) _last = null;
    else             
    {
      p._bkw._frw = p._frw;
      p._frw._bkw = p._bkw;
      if(p == _last) _last = p._bkw;
    }
  }
  
  public Object clone()
  {
    return new WeakIdentitySet(this);
  }  
  
  public void strip()
  {
    if(_size == 0) return;
    
    SetNode p, q;
  
    for(int i = 0; i < _itemsNumber; i++)
    {
      if(null == (p = _items[i])) continue;
      q = null;
      
      while(p != null)
      {
        if(p._r.get() == null)
        {
          stripObject(p, q, i);
        }
        else q = p;

        p = p._next;    
      }
    }
  }
  
  public int size()
  {
    return _size;
  }

  public void clear()
  {
    _size  = 0;
    _items = new SetNode[_itemsNumber]; 
    _last  = null;
  }

  public boolean add(Object o)
  {
    if(o == null) return false;
    
    
    int h = hash(o)&(_itemsNumber-1);
    SetNode p = _items[h], q = null;

    if(_size != 0)
    {
      Object x;
      while(p != null)
      {
        if((x = p._r.get()) != null && x == o)
        {
          return false; 
        } 
        else if(x == null)
        {
          stripObject(p, q, h);
        }
        else q = p;
  
        p = p._next;
      }
    }
    
    p = new SetNode(o);
    p._next = _items[h];
    _items[h] = p;
    if(_size++ == 0) { _last = p._bkw = p._frw = p; } 
    else             
    { 
      p._frw = _last._frw; 
      p._bkw = _last;
      _last._frw._bkw = p; // missed before
      
      _last._frw = p; 
      _last = p; 
    } 

    return true;
  }

  public boolean contains(Object o)
  {
    if(o != null && _size != 0)
    {
      int h = hash(o)&(_itemsNumber-1);
      SetNode p = _items[h], q = null;
      Object x;
      
      while(p != null)
      {
        if((x = p._r.get()) != null && x == o) return true;
        else if(x == null)
        {
          stripObject(p, q, h);
        }
        else q = p;
  
        p = p._next;
      }
    }
    
    return false;
  }

  public boolean remove(Object o)
  {
    if(o != null && _size != 0)
    {
      int h = hash(o)&(_itemsNumber-1);
      SetNode p = _items[h], q = null;
      Object x;
      
      while(p != null)
      {
        if((x = p._r.get()) != null && x == o)
        {
          stripObject(p, q, h);
          return true;
        }  
        else if(x == null)
        {
          stripObject(p, q, h);
        }
        else q = p;
  
        p = p._next;
      }
    }
    
    return false;
  }

  public IObjectIterator iterator()
  {
    return new WeakIdentitySetIterator(this);
  }
  
  public boolean containsRef(Object o)
  {
    return contains(o);
  }

  public boolean equalsByRefs(Object o)
  {
    return equals(o);
  }
  
  public static ISet singleton(Object o)
  {
    return new WeakIdentitySet(o);
  }

  public static ISet union(ISet s1, ISet s2)
  {
    ISet res = new WeakIdentitySet(s1);
    res.addAll(s2);
    
    return res;
  }
  
  public static ISet intersection(ISet s1, ISet s2)
  {
    ISet res = new WeakIdentitySet(s1);
    res.retain(s2);
    
    return res;
  }
  
  public static ISet difference(ISet s1, ISet s2)
  {
    ISet res = new WeakIdentitySet(s1);
    res.removeAll(s2);
    
    return res;
  }
  
  public static ISet symDifference(ISet s1, ISet s2)
  {
    ISet res = difference(s1, s2);
    res.addAll(difference(s2, s1));
    
    return res;
  }
  
  public static ISet distrUnion(ISet s)
  {
    IObjectIterator i = s.iterator();
    ISet res = new WeakIdentitySet(), t;
    
    while(!i.stop())
    {
      t = (ISet)i.value();
      res.addAll(t);
    }

    return res;    
  }
  
  public static ISet distrIntersection(ISet s)
  {
    IObjectIterator i = s.iterator();
    ISet res = null, t;
    
    while(!i.stop())
    {
      t = (ISet)i.value();
      if(res == null) res = new WeakIdentitySet(t);
      else            res.retain(t);
    }

    return res;    
  }
  
  private static class WeakIdentitySetIterator
    extends     AbstractObjectIterator
    implements  IObjectIterator
  {
    private SetNode         p = null;
    private WeakIdentitySet s = null; 
    
    WeakIdentitySetIterator(WeakIdentitySet s)  
    {
      this.s = s;
      if(s._last == null) p = null;
      else                p = s._last._frw;
      
      while(p != null && p._r.get() == null)
      {
        s.stripNode(p);
        if(p == s._last) p = null;
        else             p = p._frw;        
      }
    }
       
    WeakIdentitySetIterator(WeakIdentitySetIterator r)
    {
      s = r.s;
      p = r.p;      
    }
    
    public boolean stop()
    {
      if(p == null) return true;
      else          return false;
    }
  
    public boolean equals(Object o)
    {
      if(!(o instanceof WeakIdentitySetIterator)) return false;
      
      WeakIdentitySetIterator r = (WeakIdentitySetIterator)o;
       
      return r.s == s && r.p == p;
    }

    public int hashCode()
    {
      return s.hashCode() + 17*System.identityHashCode(p);
    }
    
    public void remove()
    {
      SetNode q = p;
      if(p == s._last) p = null;
      else             p = p._frw;
      s.stripNode(q);
      
      while(p != null && p._r.get() == null)
      {
        s.stripNode(p);
        if(p == s._last) p = null;
        else             p = p._frw;        
      }
    }
  
    public void shift()
    {
      if(p == s._last) p = null;
      else             p = p._frw;
        
      while(p != null && p._r.get() == null)
      {
        s.stripNode(p);
        if(p == s._last) p = null;
        else             p = p._frw;        
      }
    }    
      
    public Object value()
    {
      return p._r.get();
    }
  
    public Object clone()
    {
      return new WeakIdentitySetIterator(this);
    }
  }
  
  private static int _itemsNumber = 16;
  
  private SetNode _items[] = new SetNode[_itemsNumber];
  private int     _size    = 0;
  private SetNode _last    = null;
  
  private static class SetNode
  {
    SetNode _next;
    SetNode _frw;
    SetNode _bkw;
    WeakReference _r;
    int i;
    
    SetNode(Object o) { _r = new WeakReference(o); i = System.identityHashCode(o); }
  }
}
